/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/


/*
  File: Annotation.java
*/

package edu.uofminnesota.DEMAIS.core.annotation;

//Import needed packages
import java.util.*;
import edu.uofminnesota.DEMAIS.meal.*;



public class Annotation {

    // The scope of the annotation
    private DataModel scope;

    /* 
       An annotation is represented by a condition/action pair. Both are stored and parsed separately, and each will have a command stack associated with it after parsing.
     */
    private Condition condition;
    private Action action;

    // If the action has been fired during play mode
    private boolean actionFired;

    // Cached string representation
    private String stringRep;



    // Convert a String to an annotation
    // Null is returned when unsuccessful
    public static Annotation stringToAnnotation(DataModel scope, String sentence) {
	Annotation annotation = null;

	// Attempt to split apart the sentence
	int index = 0;
	String conditionS = null;
	String actionS = null;
	if ( (index = sentence.indexOf(',')) == -1) {
	    if ( (index = sentence.indexOf(';')) == -1) {
		if ( (index = sentence.indexOf(':')) == -1) {
		    return null;
		}
	    }
	}
	if (index >= 0) {
	    conditionS = sentence.substring(0, index);
	    actionS = sentence.substring(index+1);
	    annotation = new Annotation(scope, conditionS, actionS);
	}
	return annotation;
    }



    // Constructors
    private Annotation() {
    }

    public Annotation(DataModel model, String conditionS) {
	this(model, conditionS, null);
    }

    public Annotation(DataModel model, String conditionS, String actionS) {
	scope = model;
	condition = null;
	action = null;
	if (conditionS != null) {
	    condition = new Condition(conditionS);
	}
	if (actionS != null) {
	    action = new Action(actionS);
	}

	actionFired = false;
	stringRep = toString();
    }

    private Annotation(DataModel model, Condition c, Action a) {
	scope = model;
	condition = c;
	action = a;
	actionFired = false;
	stringRep = toString();
    }


    public Annotation(Annotation copy) {
	scope = copy.scope;

	condition = null;
	if (copy.condition != null) {
	    condition = (Condition) copy.condition.clone();
	}

	action = null;
	if (copy.action != null) {
	    action = (Action) copy.action.clone();
	}

	actionFired = copy.actionFired;
	stringRep = toString();
    }



    // Public methods
    public DataModel getScope() {
	return scope;
    }

    public Condition getCondition() {
	return condition;
    }

    public String getConditionString() {
	String ret = null;
	if (condition != null) {
	    ret = condition.getConditionString();
	}
	return ret;
    }

    public void setConditionString(String c) {
	if (condition == null) {
	    condition = new Condition(c);
	} else {
	    condition.setConditionString(c);
	}
	stringRep = toString();
    }


    public Action getAction() {
	return action;
    }

    public String getActionString() {
	String ret = null;
	if (action != null) {
	    ret = action.getActionString();
	} 
	return ret;
    }

    public void setActionString(String a) {
	if (action == null) {
	    action = new Action(a);
	} else {
	    action.setActionString(a);
	}
	stringRep = toString();
    }


    public int hashCode() {
	return stringRep.hashCode();
    }

    
    public Stack getConditionStack() {
	Stack cs = null;
	if (condition != null) {
	    cs = condition.getConditionStack();
	} 
	return cs;
    }


    public Stack getActionStack() {
	Stack as = null;
	if (action != null) {
	    as = action.getActionStack();
	}
	return as;
    }


    public Annotation reverse() {
	Condition reverseCondition = null;
	Action reverseAction = null;
	if (condition != null) {
	    reverseCondition = condition.reverse();
	}
	if (action != null) {
	    reverseAction = action.reverse();
	}
	return (new Annotation(scope, reverseCondition, reverseAction));
    }


    public void setConditionStack(Stack cStack) {
	if (condition == null) {
	    condition = new Condition(null, cStack);
	} else {
	    condition.setConditionStack(cStack);
	}
    }
 
    public void setActionStack(Stack aStack) {
	if (action == null) {
	    action = new Action(null, aStack);
	} else {
	    action.setActionStack(aStack);
	}
    }


    public boolean getActionFired() {
	return actionFired;
    }

    public void setActionFired(boolean fired) {
	actionFired = fired;
    }


    public boolean equals(Object obj) {
	if (obj instanceof Annotation) {
	    return (equals((Annotation) obj));

	} else if (obj instanceof String) {
	    return (equals((String) obj));
	}

	return (this == obj);
    }


    public boolean equals(Annotation a) {
	return (stringRep.equals(a.stringRep));
    }

    public boolean equals(String s) {
	return (stringRep.equals(s));
    }


    public String toString() {
	String ret = "";
	if (condition != null) {
	    ret += condition.toString();
	}
	if (action != null) {
	    ret += ", ";
	    ret += action.toString();
	}
	return ret;
    }


    public Object clone() {
	return (new Annotation(this));	
    }

}





