/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/


/* 
   File: Parser.java
*/

// Current Package
package edu.uofminnesota.DEMAIS.core.parser;

// Import needed packages
import java.lang.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.core.annotation.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;




/* 
   The parser parses a single annotation. As a result of parsing an annotation, three objects are produced:
   conditionStack - the condition of the annotation as an executable stack object
   actionStack - the action of the annotation as an executable stack object
   monitorStack - the objects (variables and media objects) that need to be monitored at runtime.
 */

public class Parser {

    // During the parsing of a variable, media object, or storyboard name, we 
    // need to know the scope in which to search the symbol table
    protected static int scope;

    // Current scope of the annotation being parsed
    protected static DataModel annotationScope;



    // static initializer
    static {
	scope = PropertyManager.LOCAL_SCOPE;
	annotationScope = null;
    }


    // Private constructor
    private Parser() {
	// The parser is not to be instantiated
    }


    // Public static methods
    public static void parse(Annotation annotation) throws ParseException {	
	if (annotation.getConditionString() == null) {
	    throw (new ParseException("Annotation condition is null"));
	}

	annotationScope = annotation.getScope();
	try {
	    Stack cStack = parseCondition(annotation.getConditionString());
	    annotation.setConditionStack(cStack);
	
	} catch (ParseException xcptn1) {
	    // Fill in more information and then re-throw a condition exception
	    xcptn1.setAnnotation(annotation);
	    throw (new ParseConditionException(xcptn1));
	}
	

	// Condition was parsed successfully
	if ((annotation.getAction() == null) || (annotation.getActionString() == null) ) {
	    annotation.setActionStack(null);
	    return;
	}

	try {
	    Stack aStack = parseAction(annotation.getActionString());
	    annotation.setActionStack(aStack);
	
	} catch (ParseException xcptn2) {
	    // Fill in more information and then re-throw an action exception
	    xcptn2.setAnnotation(annotation);
	    throw (new ParseActionException(xcptn2));
	}
    }


    // Private static methods
    private static Stack parseCondition(String condition) throws ParseException {
	return parse(condition, Nonterminal.DEMAISCondition);
    }

    private static Stack parseAction(String action) throws ParseException {
	return parse(action, Nonterminal.DEMAISAction);
    }


    // Implements Algorithm 4.3 of Aho, Sethi, and Ulman
    private static Stack parse(String parseString, Nonterminal startSymbol) throws ParseException {

	// Initialize the stream tokenizer
	Tokenizer tokenizer = new Tokenizer(parseString);

	// Create a parse stack for the LL(1) grammar
	Stack parseStack = new Stack();

	// Create an opcode stack for the commands
	Stack opCodeStack = new Stack();

	// For the opcodes, need to save the last token 
	Token prevToken = null;


	// Intialize the parse stack to have the start symbol and EOF on it,
	// with the start symbol being on top
	parseStack.push(Terminal.EOF);
	parseStack.push(startSymbol);
	
	Token token = tokenizer.getNextToken();
	prevToken = token;
	boolean topIsEOF = false;
	
	do {
	    Object X = parseStack.peek();

	    // Check if top parse stack element is an Opcode, Terminal, or Nonterminal
	    if (X instanceof Opcode) {
		Opcode opcode = (Opcode) X;
		if (opcode.equals(Opcode.REAL_NUMBER)) {
		    opcode.setRealNumber(prevToken.getValue());
		    // Push the opCode onto the opCodeStack
		    opCodeStack.push(opcode);

		} else if (opcode.equals(Opcode.MEDIA_OBJECT) ||
			   opcode.equals(Opcode.VARIABLE) ||
			   opcode.equals(Opcode.STORYBOARD)) {

		    opcode.setObject(prevToken.getObject());
		    opcode.setScope(prevToken.getScope());
		    opCodeStack.push(opcode);

		} else if (opcode.equals(Opcode.SET_SCOPE_TO_LOCAL)) {
		    Parser.scope = PropertyManager.LOCAL_SCOPE;
		  
		} else if (opcode.equals(Opcode.SET_SCOPE_TO_APPLICATION)) {
		    Parser.scope = PropertyManager.APPLICATION_SCOPE;

		} else if (isIndirectReference(opcode)) {
		    opcode = resolveIndirectReference(opcode, prevToken.getWordCount());
		    opCodeStack.push(opcode);

		} else {
		    opCodeStack.push(opcode);
		}

		// Pop this opcode from the stack
		parseStack.pop();

	    } else if (X instanceof Terminal) {

		Terminal terminal = (Terminal) X;
		if (terminal.equals(token.getTerminal()))  {

		    // Pop the top symbol from the stack
		    parseStack.pop();
		    // Advance to the next token
		    prevToken = token;
		    token = tokenizer.getNextToken();

		} else {
		    throw (new ParseException(token.getWordCount(), "Parser: parse()", "Unknown word"));
		}

	    } else if (X instanceof Nonterminal) {
		
		Nonterminal nonterminal = (Nonterminal) X;
		int production = Parsetable.Table[nonterminal.intValue()][token.getTerminal().intValue()];
		
		// Ensure we have a matching production
		if (production < 0) {
		    throw (new ParseException(token.getWordCount(), "Parser: parse()", "Unknown word"));
		}
		
		// Pop the top parse stack element
		parseStack.pop();

		/*
		  Push the production elements onto the parseStack. Recall that the production elements were stored in the parse table in reverse order so that pushing them onto the parseStack now would result in the correct ordering.
		*/
		Collection Ylist = Production.getProduction(production);
		Iterator it = Ylist.iterator();
		while (it.hasNext()) {
		    Object obj = it.next();
		    if (obj instanceof Opcode) {
			Opcode code = (Opcode) obj;
			parseStack.push(code.clone());
		    } else {
			parseStack.push(obj);
		    }
		}
		
	    } else {
		DEMAISError.showErrorMessage("Unknown object type on parse stack", "Parser", false);
	    }

	    if (! parseStack.empty()) {
		Object topStackElement = parseStack.peek();
		if (topStackElement instanceof Terminal) {
		    Terminal topTerminal = (Terminal) topStackElement;
		    if (topTerminal.equals(Terminal.EOF)) {
			topIsEOF = true;
		    }
		}

	    } else {
		topIsEOF = true;
	    }

	}
	while (! topIsEOF);


	// Return the opcode stack
	return opCodeStack;
    }


    // Locate the object to which an indirect reference refers
    private static Opcode resolveIndirectReference(Opcode code, int wordCount) throws ParseException {
	NamedModel object = SymbolTableManager.resolveIndirectReference(PropertyManager.LOCAL_SCOPE, annotationScope, code);

	Opcode newCode = null;
	if (object != null) {
	    newCode = new Opcode(Opcode.MEDIA_OBJECT);
	    newCode.setObject(object);
	    newCode.setScope(annotationScope);
	} else {
	    throw (new ParseException(wordCount, "Parser: resolveIndirectReferences()", "Could not resolve the indirect reference"));
	}
	return newCode;
    }



    private static boolean isIndirectReference(Opcode code) {
	return(code.equals(Opcode.ANIMATION) ||
	       code.equals(Opcode.AUDIO) ||
	       code.equals(Opcode.VIDEO));
    }

}


