/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/


/*
  File: SymbolTable.java

  The SymbolTable maintains a set of named objects. The table supports fast lookup of these objects using different keys.
*/

// Current package
package edu.uofminnesota.DEMAIS.core.parser;

// Import needed packages
import java.lang.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.meal.*;



public class SymbolTable extends Hashtable {

    // Constructors
    public SymbolTable() {
	this(50);
    }

    public SymbolTable(int capacity) {
	super(capacity);
    }



    // Public methods        
    public void put(NamedModel object) {
	put(object, object);
    }
    
    public void put(NamedModel key, NamedModel value) {
	super.put(key, value);
    }


    public boolean containsKey(NamedModel key) {
	return super.containsKey(key);
    }

    public boolean containsKey(Integer key) {
	return super.containsKey(key);
    }

    public boolean containsKey(String key) {
	key = key.toUpperCase();
	for (Enumeration e = keys(); e.hasMoreElements(); ) {
	    String objName = ((NamedModel) e.nextElement()).getName();
	    if ( (objName != null) && (key.equals(objName.toUpperCase()))) {
		return true;
	    }
	}
	return false;
    }


    public boolean containsValue(NamedModel value) {
	return super.containsValue(value);
    }

    public boolean containsValue(Integer value) {
	return super.containsValue(value);
    }

    public boolean containsValue(String value) {
	value = value.toUpperCase();
	for (Enumeration e = elements(); e.hasMoreElements(); ) {
	    String objName = ((NamedModel) e.nextElement()).getName();
	    if ( (objName != null) && (value.equals(objName.toUpperCase()))) {
		return true;
	    }
	}
	return false;
    }


    public NamedModel get(NamedModel key) {
	return ((NamedModel) super.get(key));
    }

    public NamedModel get(Integer key) {
	return ((NamedModel) super.get(key));
    }

    // Search the table objects for the first one with name key
    public NamedModel get(String key) {
	key = key.toUpperCase();
	for (Enumeration e = elements(); e.hasMoreElements(); ) {
	    NamedModel no = (NamedModel) e.nextElement();
	    String objName = no.getName();
	    if ( (objName != null) && (key.equals(objName.toUpperCase()))) {
		return no;
	    }
	}
	return null;
    }


    // Remove the object from the table
    public Object remove(NamedModel obj) {
	return super.remove(obj);
    }

    public Object remove(Integer key) {
	return super.remove(key);
    }

    // Remove the first entry with name key
    public Object remove(String key) {
	key = key.toUpperCase();
	NamedModel object = null;
	for (Enumeration e = keys(); e.hasMoreElements(); ) {
	    NamedModel no = (NamedModel) e.nextElement();
	    String objName = no.getName();
	    if ((objName != null) && (key.equals(objName.toUpperCase()))) {
		object = no;
		break;
	    }
	}
	if (object != null) {
	    return super.remove(object);
	}
	return null;
    }



    // Attempts to resolve an indirect reference by searching the
    // symbol table for a likely matching object
    public NamedModel resolveIndirectReference(Opcode code) {

	if (code.equals(Opcode.ANIMATION)) {
	    return resolveAnimationObject();

	} else if (code.equals(Opcode.AUDIO)){
	    return resolveAudioObject();

	} else if (code.equals(Opcode.NARRATION)) {
	    return resolveNarrationObject();

	} else if (code.equals(Opcode.VIDEO)) {
	    return resolveVideoObject();
	}

	return null;
    }


    // Private methods  
    private AnimationModel resolveAnimationObject() {
	NamedModel no = null;
	AnimationModel animation = null;
	AnimationModel best = null;

	for (Enumeration e = elements(); e.hasMoreElements(); ) {
	    no = (NamedModel) e.nextElement();
	    // Attempt to cast it as an animation object
	    if (no instanceof AnimationModel) {
		animation = (AnimationModel) no;
		// Test if this one is better than the best 
		best = animation;
	    }
	}
	return best;
    }


    private AudioModel resolveAudioObject() {
	NamedModel no = null;

	for (Enumeration e = elements(); e.hasMoreElements(); ) {
	    no = (NamedModel) e.nextElement();
	    // Is it an audio model
	    if (no instanceof AudioModel) {
		return (AudioModel) no;
	    }
	}
	return null;
    }


    private NarrationModel resolveNarrationObject() {
	NamedModel no = null;
	
	for (Enumeration e = elements(); e.hasMoreElements(); ) {
	    no = (NamedModel) e.nextElement();
	    // Is it a narration model
	    if (no instanceof NarrationModel) {
		return (NarrationModel) no;
	    }
	}
	return null;
    }


    private VideoModel resolveVideoObject() {
	NamedModel no = null;

	for (Enumeration e = elements(); e.hasMoreElements(); ) {
	    no = (NamedModel) e.nextElement();
	    // Is it a video object
	    if (no instanceof VideoModel) {
		return (VideoModel) no;
	    }
	}
	return null;
    }

}
