/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/


/*
  File: Tokenizer.java
*/

// Current package
package edu.uofminnesota.DEMAIS.core.parser;

// Import needed packages
import java.lang.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;



public class Tokenizer {

    // Annotation sentence stored as a character array
    private char[] inputStream;

    // Position in the input stream
    private int position;    
    // Length of the input stream
    private int length;

    // Number of tokens read so far
    private int tokenCount;

    // Enables a token to be pushed back into the stream
    private int pushBackPos;




    // Constructors
    public Tokenizer() {
	this("");
    }
	
    public Tokenizer(String sentence) {
	inputStream = sentence.toCharArray();
	position = 0;
	length = inputStream.length;
	tokenCount = 0;
	pushBackPos = 0;
    }


    // Public methods
    public boolean isEOF() {
	return (position >= length);
    }


    public Token getNextToken() {
	Token token = null;
	String tokenString = null;

	// Move to the next string
	moveToNextToken();

	// Read the token string
	tokenString = readToken();

	// Match the token string
	token = matchTokenString(tokenString);

	// Set the count on the token
	token.setWordCount(tokenCount);
	++tokenCount;

	return token;
    }


    // Private methods
    public void moveToNextToken() {
	skipWhiteSpace();
	pushBackPos = position;
    }


    // Read the string until the next whitespace character
    public String readToken() {
	String tokenString = "";

	// Check for EOF first
	if (isEOF()) {
	    return "EOF";
	}

	/* 
	   Check for single character tokens that may not be preceded
	   or followed by a whitespace character
	*/
	if ( (inputStream[position] == '+') ||
	     (inputStream[position] == '-') ||
	     (inputStream[position] == '*') ||
	     (inputStream[position] == '/') ||
	     (inputStream[position] == ',') ||
	     (inputStream[position] == '[') ||
	     (inputStream[position] == ']') ||
	     (inputStream[position] == '(') ||
	     (inputStream[position] == ')') ||
	     ((inputStream[position] == '!') && ((position + 1 < length) && (inputStream[position + 1] != '='))) ||
	     ((inputStream[position] == '<') && ((position + 1 < length) && (inputStream[position + 1] != '=') && (inputStream[position + 1] != '>'))) || 
	     ((inputStream[position] == '>') && ((position + 1 < length) && (inputStream[position + 1] != '='))) ||
	     ((inputStream[position] == '=') && ((position + 1 < length) && (inputStream[position + 1] != '=')))) {
   
	    tokenString += inputStream[position];
	    ++position;
	    return tokenString;
	}


	if ( ((inputStream[position] == '!') && ((position + 1 < length) && (inputStream[position + 1] == '='))) ||
	     ((inputStream[position] == '<') && ((position + 1 < length) && (inputStream[position + 1] == '='))) ||
	     ((inputStream[position] == '<') && ((position + 1 < length) && (inputStream[position + 1] == '>'))) || 
	     ((inputStream[position] == '>') && ((position + 1 < length) && (inputStream[position + 1] == '='))) ||
	     ((inputStream[position] == '=') && ((position + 1 < length) && (inputStream[position + 1] == '=')))) {
   
	    tokenString += inputStream[position];
	    ++position;
	    tokenString += inputStream[position];
	    ++position;
	    return tokenString;
	}


	// Now read in any other token	
	while( (!isEOF()) && 
	       (inputStream[position] != '+') &&
	       (inputStream[position] != '-') &&
	       (inputStream[position] != '*') &&
	       (inputStream[position] != '/') &&
	       (inputStream[position] != ',') &&
	       (inputStream[position] != '[') &&
	       (inputStream[position] != ']') &&
	       (inputStream[position] != '(') &&
	       (inputStream[position] != ')') &&
	       (inputStream[position] != '!') &&  
	       (inputStream[position] != '<') &&
	       (inputStream[position] != '>') &&
	       (inputStream[position] != '=') &&
	       (!Character.isWhitespace(inputStream[position]))) {

	    tokenString += inputStream[position];
	    ++position;
	}
	
	return tokenString;
    }


    // Find the terminal that matches this token string
    private Token matchTokenString(String tokenString) {
	// The representative token for the token string
	Token token = null;

	// Map the string to the terminal value
	Terminal terminal = Terminal.toTerminal(tokenString);
	if (terminal != null) {
	    // Found the token in the look up table
	    token = new Token(terminal, tokenString);

	} else {
	    // Need to do some more processing
	    try {
		double d = Double.parseDouble(tokenString);
		token = new Token(Terminal.REAL_NUMBER, tokenString);
		token.setValue(d);

	    } catch (NumberFormatException xcptn) {
		try {
		    int i = Integer.parseInt(tokenString); 
		    token = new Token(Terminal.REAL_NUMBER, tokenString);
		    token.setValue((double) i);

		} catch (NumberFormatException xcptn2) {
		  
		    /*
		      Must be a media object name. A media object name can be in one of two formats, either:
		      scopeId::objectId
		      or
		      objectName
		    */

		    DataModel localScope = null;
		    NamedModel no = null;
		    // Determine which format the string is in
		    int scopeIndex = tokenString.indexOf("::");
		    if (scopeIndex >= 0) {
			// scopeId::objectId
			String scopeStringId = tokenString.substring(0, scopeIndex);
			// Remove the "__" prepended to the front
			scopeStringId = scopeStringId.substring(2);

			String noStringId = tokenString.substring(scopeIndex + 2);
			// Remove the "__" prepended to the front
			noStringId = noStringId.substring(2);

			try {
			    Integer scopeId = new Integer(Integer.parseInt(scopeStringId));
			    Integer noId = new Integer(Integer.parseInt(noStringId));

			    localScope = (DataModel) SymbolTableManager.getFromApplicationScope(scopeId);
			    no = SymbolTableManager.get(PropertyManager.LOCAL_SCOPE, localScope, noId);

			} catch (NumberFormatException xcptn3) {
			}

		    } else {
			// Just the object name in the string
			// Can be either an ID string or an actual object name
			if (tokenString.length() > 2) {
			    try {
				String scopeStringId = tokenString.substring(2);
				Integer scopeId = new Integer(Integer.parseInt(scopeStringId));
				no = (NamedModel) SymbolTableManager.get(PropertyManager.GLOBAL_SCOPE, null, scopeId);
			    } catch (NumberFormatException xcptn4) {
			    }
			}
			if (no == null) {
			    no = (NamedModel) SymbolTableManager.get(PropertyManager.GLOBAL_SCOPE, null, tokenString);
			}
		    }

		    if (no != null) {
			if (no instanceof Variable) {
			    token = new Token(Terminal.VARIABLE, tokenString, 0.0, no, localScope, 0);

			} else if ((no instanceof MediaModel) || 
				   (no instanceof TextEntryModel) ||
				   (no instanceof NarrationModel) ) {
			    token = new Token(Terminal.MEDIA_OBJECT, tokenString, 0.0, no, localScope, 0);
			    
			    // If its an audio object, realize it now
			    if (no instanceof AudioModel) {
				AudioModel model = (AudioModel) no;
				model.realize();
			    }

			} else if (no instanceof StoryboardModel) {
			    // Storyboards are not scoped, so use tokenString
			    token = new Token(Terminal.STORYBOARD, tokenString, 0.0, no, null, 0);

			} else {
			    token = new Token(Terminal.UNKNOWN, tokenString);
			}

		    } else {
			token = new Token(Terminal.UNKNOWN, tokenString);
		    }
		}
	    }
	}
	return token;
    }
    

    // Move to the next non-whitespace input character 
    private void skipWhiteSpace() {
	while((!isEOF()) && Character.isWhitespace(inputStream[position])) {
	    ++position;
	}
    }
}


