/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File: CAudioCapture.java
*/
package edu.uofminnesota.DEMAIS.meal;


// Import necessary packages
import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.media.*;
import javax.media.format.*;
import javax.media.protocol.*;
import javax.media.control.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;




public class CAudioCapture implements javax.media.ControllerListener {

    // Base name of the output file name
    private String baseFileName;

    // Reference to the mic channel
    private CaptureDeviceInfo microphone;

    // Processor for the mic data
    private Processor processor;

    // If the processor is configured
    private boolean configured;

    // If the processor is realized
    private boolean realized;

    // If the process is closed
    private boolean closed;

    // Location of the output media file
    private MediaLocator mediaLocation;

    // Data sink for storing the audio
    private DataSink fileWriter;





    public static boolean isMicrophoneAvailable() {
	CaptureDeviceInfo mic = null;
	Vector deviceList = CaptureDeviceManager.getDeviceList(new AudioFormat(AudioFormat.LINEAR));
	if (deviceList.size() > 0) {
	    mic = (CaptureDeviceInfo) deviceList.firstElement();

	} else {
	    deviceList = CaptureDeviceManager.getDeviceList(new AudioFormat(AudioFormat.ALAW));
	    if (deviceList.size() > 0) {
		mic = (CaptureDeviceInfo) deviceList.firstElement();

	    } else {	
		deviceList = CaptureDeviceManager.getDeviceList(new AudioFormat(AudioFormat.ULAW));
		if (deviceList.size() > 0) {
		    mic = (CaptureDeviceInfo) deviceList.firstElement();
		}
	    }
	}

	// If mic is not null, then microphone is available
	return (mic != null);
    }


    
    // Constructor
    private CAudioCapture() {
	this(null);
    }

    // baseFileName is a file name without a path or extension
    public CAudioCapture(String baseFileName) {
	this.baseFileName = baseFileName;
	microphone = null;
	processor = null;
	configured = false;
	realized = false;
	closed = false;
	mediaLocation = null;
	fileWriter = null;
    }


    public String getAudioFileName() {
	return baseFileName + ".mp3";
    }


    public Component getControlPanelComponent() {
	Component c = null;
	if (processor != null) {
	    c = processor.getControlPanelComponent();
	}
	return c;
    }

    public boolean beginCapture() {

	// Get the microphone
	if (! getMicrophoneChannel()) {
	    DEMAISError.showErrorMessage("Audio capture through a microphone is unavailable", "CAudioCapture", false);
	    return false;
	}

	if (! getProcessor()) {
	    return false;
	}

	if (! configureProcessor()) {
	    DEMAISError.showErrorMessage("Could not configure processor", "CAudioCapture", false);
	    return false;
	}

	if (! setEncoding()) {
	    return false;
	}

	if (! realizeProcessor()) {
	    DEMAISError.showErrorMessage("Could not realize processor", "CAudioCapture", false);
	    return false;
	}
	
	if (! getMediaLocation()) {
	    DEMAISError.showErrorMessage("Could not determine media location", "CAudioCapture", false);
	    return false;
	}
	
	// get the output of the processor
	DataSource source = processor.getDataOutput();
	
	// create a datasink to do the file writing & open the sink to
	// make sure we can write to it.
	fileWriter = null;
	try {
	    fileWriter = Manager.createDataSink(source, mediaLocation);
	    fileWriter.open();
	    fileWriter.start();
	    processor.start();

	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "CAudioCapture", false);
	    DEMAISError.logException(xcptn);
	    return false;
	}

	return true;
    }

  
    public void endCapture() {
	try {

	    closeProcessor();
	    if (fileWriter != null) {
		fileWriter.close();
	    }
	    if (processor != null) {
		processor.deallocate();
	    }
	
	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "CAudioCapture", false);
	    DEMAISError.logException(xcptn);
	}
    }
    


    // Private Methods
    
    private boolean getMicrophoneChannel() {
	// Get a mic channel
	microphone = null;
	Vector deviceList = CaptureDeviceManager.getDeviceList(new AudioFormat(AudioFormat.LINEAR));
	if (deviceList.size() > 0) {
	    microphone = (CaptureDeviceInfo) deviceList.firstElement();
	    
	} else {
	    deviceList = CaptureDeviceManager.getDeviceList(new AudioFormat(AudioFormat.ALAW));
	    if (deviceList.size() > 0) {
		microphone = (CaptureDeviceInfo) deviceList.firstElement();
		
	    } else {	
		deviceList = CaptureDeviceManager.getDeviceList(new AudioFormat(AudioFormat.ULAW));
		if (deviceList.size() > 0) {
		    microphone = (CaptureDeviceInfo) deviceList.firstElement();		    
		}
	    }
	}
	return (microphone != null);
    }
    
    
    private boolean getProcessor() {
	if (microphone != null) {
	    try {
		processor = Manager.createProcessor(microphone.getLocator());
		processor.addControllerListener(this);
		
	    } catch (Exception xcptn) {
		processor = null;
		DEMAISError.showErrorMessage(xcptn.toString(), "CAudioCapture", false);
		DEMAISError.logException(xcptn);
	    } 
	}
	return (processor != null);
    }
    
    
    private boolean configureProcessor() {
	if (processor != null) {
	    int timeOut = 5000;
	    long startTime = System.currentTimeMillis();
	    
	    // Configure the processor (asynchronous)
	    synchronized(this) {
		processor.configure();
		while (! configured) {
		    try {
			wait(timeOut);
		    } catch (InterruptedException ie) {
		    }
		    if ((System.currentTimeMillis() - startTime) > timeOut) {
			break;
		    }
		}
	    }
	}
	return configured;
    }
    
    
    private boolean setEncoding() {
	boolean encodingPossible = false;
	
	if (processor != null) {
	    // Set the output encoding type
	    processor.setContentDescriptor(new FileTypeDescriptor(FileTypeDescriptor.MPEG_AUDIO));
	    TrackControl track[] = processor.getTrackControls();		
	    for (int i=0; i<track.length; ++i) {
		try {
		    track[i].setFormat(new AudioFormat(AudioFormat.MPEGLAYER3));
		    encodingPossible = true;
		} catch (Exception xcptn) {
		    DEMAISError.showErrorMessage(xcptn.toString(), "CAudioCapture", false);
		    DEMAISError.logException(xcptn);
		}
	    }
	}
	return encodingPossible;
    }
    
    
    private boolean realizeProcessor() {
	if (processor != null) {
	    int timeOut = 5000;
	    long startTime = System.currentTimeMillis();
	    
	    // Realize the processor (asynchronous)
	    synchronized(this) {
		processor.realize();
		while (! realized) {
		    try {
			wait(timeOut);
		    } catch (InterruptedException ie) {
		    }
		    if ((System.currentTimeMillis() - startTime) > timeOut) {
			break;
		    }
		}
	    }
	}

	return realized;
    }
    


    
    private boolean getMediaLocation() {
	// Determine location of the output audio file
	mediaLocation = null;

	File audioDir = null;
	File audioDir1 = new File(PropertyManager.APPLICATION_AUDIO_DIRECTORY);
	File appDir = new File(PropertyManager.APPLICATION_DIRECTORY);
	if (audioDir1.exists()) {
	    audioDir = audioDir1;
	} else {
	    audioDir = appDir;
	}
	
	String directory = "";
	if (audioDir != null) {
	    directory = audioDir.getAbsolutePath() + PropertyManager.FILE_SEPARATOR;
	}
	String audioFileName = directory + baseFileName + ".mp3";
	
	// create a File protocol MediaLocator with the location of the
	// file to which the data is to be written
	URL mediaURL = null;
	try {
	    mediaURL = new URL("file:///" + audioFileName);
	    mediaLocation = new MediaLocator(mediaURL);

	} catch (Exception xcptn) {
	    mediaLocation = null;
	    DEMAISError.showErrorMessage(xcptn.toString(), "CAudioCapture", false);
	    DEMAISError.logException(xcptn);
	}

	return (mediaLocation != null);
    }
    
    

    private boolean closeProcessor() {
	if (processor != null) {
	    synchronized (this) {
		processor.close();
		while (! closed) {
		    try {
			wait(5000);
		    } catch (InterruptedException ie) {
		    }
		}
	    }
	    processor.removeControllerListener(this);
	}
	return closed;
    }


    public synchronized void controllerUpdate(ControllerEvent event) {
	if (event instanceof ConfigureCompleteEvent) {
	    configured = true;	
	    
	} else if (event instanceof RealizeCompleteEvent) {
	    realized = true;
	    
	} else if (event instanceof ControllerClosedEvent) {
	    closed = true;
	    return;
	   
	} else {
	    return;
	}
	
	notifyAll();
    }
}

