/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: ContinuousMediaModel.java
*/

// Current package
package edu.uofminnesota.DEMAIS.meal;

// Import needed packages
import java.lang.*;
import java.util.*;
import java.io.*;




public abstract class ContinuousMediaModel extends MediaModel
    implements ContinuousMediaModelInterface {
    
    // Listeners to notify when the continuous media model begins or ends
    private Vector beginListeners;
    private Vector endListeners;
    
    // Event status variables
    private boolean wasBeginEvent;
    private boolean wasEndEvent;



    // Constructors
    private ContinuousMediaModel() {
	this(0, null, null, null);
    }
    
    public ContinuousMediaModel(int objScope, DataModel localScope, String modelName, String mediaFileName) {
	super(objScope, localScope, modelName, mediaFileName);
	beginListeners = new Vector();
	endListeners = new Vector();
	wasBeginEvent = false;
	wasEndEvent = false;
    }
    
    
    public ContinuousMediaModel(ContinuousMediaModel copy) {
	this(copy, copy.localScope);
    }
    
    public ContinuousMediaModel(ContinuousMediaModel copy, DataModel toModel) {
	super(copy, toModel);
	beginListeners = new Vector();
	endListeners = new Vector();
	wasBeginEvent = false;
	wasEndEvent = false;
    }

    
    public ContinuousMediaModel(DataInputStream stream, DataModel localScope) throws IOException {
	super(stream, localScope);
	beginListeners = new Vector();
	endListeners = new Vector();
	wasBeginEvent = false;
	wasEndEvent = false;
    }

    

    // ContinuousMediaModelInterface - implement the interface
    // here, but rely on subclasses to override
    public void reset() {
    }
    public void begin() {
    }
    public void end() {
    }
    public void pause() {
    }
    public void resume() {
    }
    public void setMediaTime(double value) {
    }
    public double getMediaTime() {
	return 0.0;
    }
    public void setRate(double value) {
    }
    public double getRate() {
	return 0.0;
    }

    public void mediaBeginEvent() {
	setWasBeginEvent(true, true);
    }

    public void setWasBeginEvent(boolean b, boolean notify) {
	wasBeginEvent = b;
	if (notify) {
	    notifyBeginListeners();
	}
    }

    public boolean wasBeginEvent() {
	return wasBeginEvent;
    }


    public void mediaEndEvent() {
	setWasEndEvent(true, true);
    }

    public void setWasEndEvent(boolean b, boolean notify) {
	wasEndEvent = b;
	if (notify) {
	    notifyEndListeners();
	}
    }

    public boolean wasEndEvent() {
	return wasEndEvent;
    }




    public void addBeginListener(ContinuousMediaModelListenerInterface cli) {
	if (! beginListeners.contains(cli)) {
	    beginListeners.add(cli);
	}
    }

    public void removeBeginListener(ContinuousMediaModelListenerInterface cli) {
	beginListeners.remove(cli);
    }

    public void removeBeginListeners() {
	beginListeners.clear();
    }


    public void addEndListener(ContinuousMediaModelListenerInterface cli) {
	if (! endListeners.contains(cli)) {
	    endListeners.add(cli);
	}
    }

    public void removeEndListener(ContinuousMediaModelListenerInterface cli) {
	endListeners.remove(cli);
    }

    public void removeEndListeners() {
	endListeners.clear();
    }


    public void removeContinuousMediaModelListeners() {
	removeBeginListeners();
	removeEndListeners();
    }

    
    public void removeListeners() {
	super.removeListeners();
	removeContinuousMediaModelListeners();
    }


    public void operationalize() {
	super.operationalize();
	wasBeginEvent = false;
	wasEndEvent = false;
    }

    public void unOperationalize() {
	super.unOperationalize();
	removeContinuousMediaModelListeners();
	wasBeginEvent = false;
	wasEndEvent = false;
    }


    public void deleteModel() {
	super.deleteModel();
	removeContinuousMediaModelListeners();
    }


    
    protected void notifyBeginListeners() {
	Iterator it = beginListeners.iterator();
	while (it.hasNext()) {
	    ContinuousMediaModelListenerInterface listener = (ContinuousMediaModelListenerInterface) it.next();
	    listener.CMediaModelBegun(this);
	}
    }

    protected void notifyEndListeners() {
	Iterator it = endListeners.iterator();
	while (it.hasNext()) {
	    ContinuousMediaModelListenerInterface listener = (ContinuousMediaModelListenerInterface) it.next();
	    listener.CMediaModelEnded(this);
	}
    }

}
