/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: JMFContinuousMediaObject.java
*/

// Current package
package edu.uofminnesota.DEMAIS.meal;

// Import needed packages
import java.lang.*;
import java.util.*;
import java.io.*;
import java.net.*;
import javax.media.*;
import java.awt.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;





public abstract class JMFContinuousMediaModel extends ContinuousMediaModel 
    implements ControllerListener {


    // JMF player object 
    private Player player;

    // Location of media file
    private URL mURL;

    // The current rate of the media object
    private double rate;

    // Status variables
    private int state;
    private boolean realized;
    private boolean prefetched;

    /*
      A list of objects to notify when the continuous media object
      changes rate or value
    */
    private Vector rateListeners;
    private Vector valueListeners;




    // Constructors
    private JMFContinuousMediaModel() {
	this(0, null, null, null);
    }

    public JMFContinuousMediaModel(int objScope, DataModel localScope, String modelName, String mediaFileName) {
	super(objScope, localScope, modelName, mediaFileName);
	player = null;
	mURL = null;
	rate = 0.0;
	state = ContinuousMediaModelInterface.INITIAL;
	realized = false;
	prefetched = false;
	rateListeners = new Vector();
	valueListeners = new Vector();

	try {
	    // Check absolute path name first
	    mURL = null;
	    File file = new File(getMediaFilePathName());
	    if (file.exists()) {
		mURL = file.toURL();

	    } else {
		file = new File(getRelativeMediaFilePathName());
		if (file.exists()) {
		    mURL = file.toURL();
		}
	    }

	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "JMFContinuousMediaModel", false);
	    DEMAISError.logException(xcptn);
	}
    }
    
    
    public JMFContinuousMediaModel(JMFContinuousMediaModel copy) {
	this(copy, copy.localScope);
    }
    

    public JMFContinuousMediaModel(JMFContinuousMediaModel copy, DataModel toModel) {
	super(copy, toModel);

	player = null;
	mURL = copy.mURL;
	rate = 0.0;
	state = ContinuousMediaModel.INITIAL;
	realized = false;
	prefetched = false;

	rateListeners = null;
	if (copy.rateListeners != null) {
	    rateListeners = (Vector) copy.rateListeners.clone();
	}

	valueListeners = null;
	if (copy.valueListeners != null) {
	    valueListeners = (Vector) copy.valueListeners.clone();  
	}
    }
    

    public JMFContinuousMediaModel(DataInputStream stream, DataModel localScope) throws IOException {
	super(stream, localScope);
	player = null;
	mURL = null;
	rate = 0.0;
	state = ContinuousMediaModelInterface.INITIAL;
	realized = false;
	prefetched = false;

	rateListeners = new Vector();
	valueListeners = new Vector();
	
	try {
	    // Check absolute path name first
	    File file = new File(getMediaFilePathName());
	    if (file.exists()) {
		mURL = file.toURL();
		
	    } else {
		file = new File(getRelativeMediaFilePathName());
		if (file.exists()) {
		    mURL = file.toURL();
		}
	    }

	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "JMFContinuousMediaModel", false);
	    DEMAISError.logException(xcptn);
	}
    }



    // Public methods
    public void realizePlayer() {
	if ( (player == null) && (mURL != null) ) {
	    try {
		Manager.setHint(Manager.LIGHTWEIGHT_RENDERER, Boolean.TRUE); 
		player = Manager.createPlayer(mURL);

	    } catch (Exception xcptn) {
		System.out.println(xcptn.toString());
		player = null;
		return;
	    }
	    player.addControllerListener(this);

	    // Realize the player (asynchronous)
	    int timeOut = 5000;
	    long startTime = System.currentTimeMillis();

	    synchronized(this) {
		player.realize();
		while (! realized) {
		    try {
			wait(timeOut);
		    } catch (InterruptedException ie) {
		    }
		    if ((System.currentTimeMillis() - startTime) > timeOut) {
			realized = false;
			break;
		    }
		}
	    }

	    // Prefetch the media (asynchronous)
	    startTime = System.currentTimeMillis();
	    synchronized(this) {
		player.prefetch();
		while (! prefetched) {
		    try {
			wait(timeOut);
		    } catch (InterruptedException ie) {
		    }
		    if ((System.currentTimeMillis() - startTime) > timeOut) {
			prefetched = true;
			break;
		    }
		}
	    }
	    if ((player != null) &&
		(player.getState() == Controller.Realized)) {
		GainControl control = player.getGainControl();
		if (control != null) {
		    control.setLevel(ContinuousMediaModelInterface.GAIN);
		}
	    }
	}
    }
    


    public void deleteModel() {
	super.deleteModel();
	removeJMFContinuousMediaModelListeners();
	if (isRealized()) {
	    // Releases all resources
	    player.close();
	}
    }


    public boolean isRealized() {
	return realized;
    }


    public void realize() {
	super.realize();
	if (! isRealized()) {
	    realizePlayer();
	    state = ContinuousMediaModelInterface.INITIAL;
	} else {
	    //reset();
	}
    }

    public void unRealize() {
	super.unRealize();
	reset();
    }


    public void operationalize() {
	super.operationalize();
	if (! isRealized()) {
	    realize();
	} else {
	    reset();
	}
    }


    public void unOperationalize() {
	super.unOperationalize();
	removeJMFContinuousMediaModelListeners();
	reset();
    }
    

    public void reset() {
	state = ContinuousMediaModelInterface.RESET;
	rate = 0.0;
	if (isRealized()) {
	    // Generates a stop event
	    player.stop();
	}
    }

    
    public void begin() {
	if (isRealized()) {
	    if (state != ContinuousMediaModelInterface.PAUSED) {		
		// Not resuming from a pause - start from the beginning
		setMediaTime(0.0);
		state = ContinuousMediaModelInterface.BEGINNING;
	    }

	    // Generates a start event
	    player.start();
	}
    }


    private void finishBegin() {
	rate = 1.0;
	if ( (state == ContinuousMediaModelInterface.INITIAL) || 
	     (state == ContinuousMediaModelInterface.BEGINNING) ) {
	    state = ContinuousMediaModelInterface.BEGUN;
	    // notify begin and rate listeners
	    setWasBeginEvent(true, true);
	    notifyRateListeners();

	} else if (state == ContinuousMediaModelInterface.PAUSED) {
	    // Notify rate listeners
	    state = ContinuousMediaModelInterface.BEGUN;
	    notifyRateListeners();
	}
    }


    public boolean hasBegun() {
	return ( (state == ContinuousMediaModelInterface.BEGUN) ||
		 (state == ContinuousMediaModelInterface.PAUSED));
    }

    
    public void end() {
	if (isRealized()) {
	    // Prevent from being called more than once on a stopped object
	    if ((state > ContinuousMediaModelInterface.INITIAL) &&
		(state < ContinuousMediaModelInterface.ENDING)) {
		state = ContinuousMediaModelInterface.ENDING;

		// Stop the media object - generates a StopEvent
		player.stop();

		// Set to the last frame
		setMediaTime(getDuration());
	    }
	}
    }

    private void finishEnd() {
	rate = 0.0;
	if (state == ContinuousMediaModelInterface.ENDING) {
	    state = ContinuousMediaModelInterface.ENDED;
	    // Notify rate and end listeners
	    notifyRateListeners();
	    setWasEndEvent(true, true);
	}
    }


    public boolean hasEnded() {
	return (state == ContinuousMediaModelInterface.ENDED);
    }


    public void pause() {
	if (isRealized()) {
	    state = ContinuousMediaModelInterface.PAUSING;
	    // Pause the playing object - generates a StopEvent
	    player.stop();
	}
    }

    private void finishPause() {
	rate = 0.0;
	if (state == ContinuousMediaModelInterface.PAUSING) {
	    state = ContinuousMediaModelInterface.PAUSED;
	    notifyRateListeners();
	}
    }


    public void resume() {
	begin();
    }


    public double getDuration() {
	double duration = Double.MAX_VALUE;
	if (isRealized()) {
	    Time time = player.getDuration();
	    if (time != Time.TIME_UNKNOWN)
		duration = time.getSeconds();
	}
	return duration;
    }


    public void setMediaTime(double value) {
	if (isRealized()) {
	    try {
		// Generates a MediaTimeSetEvent
		if (state == ContinuousMediaModelInterface.BEGUN) {
		    player.stop();
		    player.setMediaTime(new Time(value));
		    player.start();

		} else {
		    player.setMediaTime(new Time(value));
		}

	    } catch (Exception xcptn) {
		System.out.println(xcptn.toString());
	    }
	}
    }


    public double getMediaTime() {
	double v = 0.0;
	if (isRealized()) {
	    v = player.getMediaTime().getSeconds();
	}
	return v;
    }


    public void setRate(double r) {
	if (isRealized()) {
	    rate = r;
	    // Generates a RateChangedEvent
	    player.setRate((float)rate);
	}
    }

    public double getRate() {
	// Return the cached value as its more accurate
	// than the rate that the JMF keeps. For example,
	// JMF maintains a rate value of 1, even when a 
	// media object is paused!
	return rate;
    }


    public int getState() {
	return state;
    }

    // Specialized function 
    public void setState(int s, double mediaTime) {
	if (isRealized()) {
	    state = s;
	    if (state == ContinuousMediaModelInterface.INITIAL) {
		// Do nothing
	    
	    } else if ((state == ContinuousMediaModelInterface.BEGINNING) ||
		       (state == ContinuousMediaModelInterface.BEGUN)) {
		state = ContinuousMediaModelInterface.BEGUN;
		rate = 1.0;
		setMediaTime(mediaTime);
		player.start();
		
	    } else if ((state == ContinuousMediaModelInterface.PAUSING) ||
		       (state == ContinuousMediaModelInterface.PAUSED)) {
		state = ContinuousMediaModelInterface.PAUSING;
		setMediaTime(mediaTime);
		player.stop();
		
	    } else if ((state == ContinuousMediaModelInterface.ENDING) ||
		       (state == ContinuousMediaModelInterface.ENDED)) {
		state = ContinuousMediaModelInterface.ENDED;
		rate = 0.0;
		setMediaTime(mediaTime);
		player.stop();
	    } 
	} 
    }


    public Component getVisualComponent() {
	Component c = null;
	if (isRealized()) {
	    c = player.getVisualComponent();
	}
	return c;
    }


    public Component getControlPanelComponent() {
	Component c = null;
	if (isRealized()) {
	    c = player.getControlPanelComponent();
	} 
	return c;
    }


    public void addRateListener(ContinuousMediaModelListenerInterface cli) {
	if (! rateListeners.contains(cli)) {
	    rateListeners.add(cli);
	}
    }

    public void removeRateListener(ContinuousMediaModelListenerInterface cli) {
	rateListeners.remove(cli);
    }

    public void removeRateListeners() {
	rateListeners.clear();
    }


    public void addValueListener(ContinuousMediaModelListenerInterface cli) {
	if (! valueListeners.contains(cli)) {
	    valueListeners.add(cli);
	}
    }

    public void removeValueListener(ContinuousMediaModelListenerInterface cli) {
	valueListeners.remove(cli);
    }

    public void removeValueListeners() {
	valueListeners.clear();
    }


    public void removeJMFContinuousMediaModelListeners() {
	removeRateListeners();
	removeValueListeners();
    }
    

    public void removeListeners() {
	super.removeListeners();
	removeJMFContinuousMediaModelListeners();
    }



    
    protected void notifyRateListeners() {
	Iterator it = rateListeners.iterator();
	while (it.hasNext()) {
	    ContinuousMediaModelListenerInterface listener = (ContinuousMediaModelListenerInterface) it.next();
	    listener.CMediaModelRateChanged(this);
	}
    }
    
    protected void notifyValueListeners() {
	Iterator it = valueListeners.iterator();
	while (it.hasNext()) {
	    ContinuousMediaModelListenerInterface listener = (ContinuousMediaModelListenerInterface) it.next();
	    listener.CMediaModelValueChanged(this);
	}
    }



    /*
      ControllerListener interface - Listen for events that are generated
      from system updates (playout without interaction)
    */
    public synchronized void controllerUpdate(ControllerEvent event) {

	if (event instanceof EndOfMediaEvent) {
	    // Result of natural playout
	    int mode = ((Integer) PropertyManager.getProperty(PropertyManager.APPLICATION_MODE)).intValue();
	    if (mode == PropertyManager.APPLICATION_PLAY) {
		state = ContinuousMediaModelInterface.ENDING;
		finishEnd();
	    }

	} else if (event instanceof StartEvent) {
	    if ((state == ContinuousMediaModelInterface.INITIAL) ||
		(state == ContinuousMediaModelInterface.BEGINNING) ||
		(state == ContinuousMediaModelInterface.PAUSED)) {		
		finishBegin();
	    }

	} else if (event instanceof StopEvent) {
	    if (state == ContinuousMediaModelInterface.ENDING) {
		finishEnd();

	    } else if (state == ContinuousMediaModelInterface.PAUSING) {
		finishPause();
	    
	    } else if (state == ContinuousMediaModelInterface.RESET) {
		setMediaTime(0.0);
		state = ContinuousMediaModelInterface.INITIAL;
	    }

	} else if (event instanceof RateChangeEvent) {
	    rate = player.getRate();
	    if (rate == 0.0) {
		state = ContinuousMediaModelInterface.PAUSED;
	    } else {
		state = ContinuousMediaModelInterface.BEGUN;
	    }
	    notifyRateListeners();

	} else if (event instanceof MediaTimeSetEvent) {
	    notifyValueListeners();
	    
	} else if (event instanceof RealizeCompleteEvent) {
	    realized = true;
	    notifyAll();

	} else if (event instanceof PrefetchCompleteEvent) {
	    prefetched = true;
	    notifyAll();
	}
    }

}


