/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: MediaModel.java
*/

// Current package
package edu.uofminnesota.DEMAIS.meal;

// Import needed packages
import java.lang.*;
import java.util.*;
import java.io.*;
import edu.uofminnesota.DEMAIS.property.*;
import javax.swing.JFileChooser;
import edu.uofminnesota.DEMAIS.usrinterface.util.*;




public abstract class MediaModel extends NamedModel 
    implements MediaModelInterface {


    // File path name for this media object's data
    protected String mediaFilePathName; // absolute path name
    protected String relativeMediaFilePathName; // Relative to directory where DEMAIS was started

    // status variables
    protected boolean wasSingleClicked;
    protected boolean wasDoubleClicked;
    protected boolean isOver;
    protected boolean isHighlighted;

    // Event listeners
    private Vector singleClickListeners;
    private Vector doubleClickListeners;
    private Vector rolloverListeners;




    // Constructors
    private MediaModel() {
	this(0, null, null, null);
    }

    public MediaModel(int scope, DataModel localScope, String modelName, String mediaFileName) {
	super(scope, localScope, modelName);
	mediaFilePathName = mediaFileName;
	relativeMediaFilePathName = mediaFilePathName;

	String appDirectory = PropertyManager.APPLICATION_DIRECTORY;
	if ((mediaFilePathName != null) &&
	    (mediaFilePathName.startsWith(appDirectory))) {
	    // Media path name is relative to directory where DEMAIS was started
	    // Extract the relative path name
	    relativeMediaFilePathName = mediaFilePathName.substring(appDirectory.length() + 1);
	}

	wasSingleClicked = false;
	wasDoubleClicked = false;
	isOver = false;
	isHighlighted = false;

	singleClickListeners = new Vector();
	doubleClickListeners = new Vector();
	rolloverListeners = new Vector();

	if (mediaFilePathName != null) {
	    // Set the object name based on the mediaFilePathName
	    // Example: G:\Video1.mpg -> Video1
	    int index = mediaFilePathName.lastIndexOf(PropertyManager.FILE_SEPARATOR);
	    String shortName = mediaFilePathName.substring(index+1);
	    index = shortName.lastIndexOf(".");
	    if (index >= 0) {
		shortName = shortName.substring(0, index);
	    }
	    setName(shortName);
	}
    }


    public MediaModel(MediaModel copy) {
	this(copy, copy.localScope);
    }


    public MediaModel(MediaModel copy, DataModel toModel) {
	super(copy, toModel);
	mediaFilePathName = null;
	if (copy.mediaFilePathName != null) {
	    mediaFilePathName = new String(copy.mediaFilePathName);
	}

	relativeMediaFilePathName = null;
	if (copy.relativeMediaFilePathName != null) {
	    relativeMediaFilePathName = new String(copy.relativeMediaFilePathName);
	}

	wasSingleClicked = copy.wasSingleClicked;
	wasDoubleClicked = copy.wasDoubleClicked;
	isOver = copy.isOver;
	isHighlighted = copy.isHighlighted;

	singleClickListeners = new Vector();
	doubleClickListeners = new Vector();
	rolloverListeners = new Vector();
    }


    public MediaModel(DataInputStream stream, DataModel localScope) throws IOException {
	super(stream, localScope);

	// Read the relative file name first
	relativeMediaFilePathName = null;
	boolean hasRelativeName = stream.readBoolean();
	if (hasRelativeName) {
	    int length = stream.readInt();
	    relativeMediaFilePathName = "";
	    for (int i=0; i<length; ++i) {
		relativeMediaFilePathName += stream.readChar();
	    }
	}

	// Read the absolute file name
	mediaFilePathName = null;
	boolean hasMediaName = stream.readBoolean();
	if (hasMediaName) {
	    int length = stream.readInt();
	    mediaFilePathName = "";
	    for (int i=0; i<length; ++i) {
		mediaFilePathName += stream.readChar();
	    }
	}

	wasSingleClicked = false;
	wasDoubleClicked = false;
	isOver = false;
	isHighlighted = false;
	
	singleClickListeners = new Vector();
	doubleClickListeners = new Vector();
	rolloverListeners = new Vector();

	// Ensure that the file exists
	if ((mediaFilePathName != null) && 
	    (relativeMediaFilePathName != null)) {
	    File file = new File(mediaFilePathName);
	    if (! file.exists()) {
		// Absolute name no longer exists
		file = new File(relativeMediaFilePathName);
		if (! file.exists()) {
		    // Display a dialog so the user can locate the file
		    String lastDir = (String) PropertyManager.getProperty(PropertyManager.APPLICATION_LAST_MEDIA_DIRECTORY);
		    int index = mediaFilePathName.lastIndexOf(PropertyManager.FILE_SEPARATOR);
		    String fileName = mediaFilePathName.substring(index+1);
		    // Create chooser starting at this directory
		    DFileChooser chooser = new DFileChooser(lastDir);
		    chooser.setCursor(PropertyManager.APPLICATION_CURSOR);
		    chooser.setDialogTitle("Where is " + fileName);
		    int returnVal = chooser.showOpenDialog(PropertyManager.APPLICATION_FRAME_WIDGET);
		    if (returnVal == JFileChooser.APPROVE_OPTION) {
			String selectedDir = chooser.getCurrentDirectory().getAbsolutePath();
			// Store the directory for the next time
			PropertyManager.setProperty(PropertyManager.APPLICATION_LAST_MEDIA_DIRECTORY, selectedDir);	    
			mediaFilePathName = chooser.getSelectedFile().getAbsolutePath();
			relativeMediaFilePathName = mediaFilePathName;			
			if (mediaFilePathName.startsWith(PropertyManager.APPLICATION_DIRECTORY)) {
			    // Media path name is relative to directory where DEMAIS was started
			    // Extract the relative path name
			    relativeMediaFilePathName = mediaFilePathName.substring(PropertyManager.APPLICATION_DIRECTORY.length() + 1);
			}
		    } else {
			throw (new IOException("MediaModel: Could not locate media file " + mediaFilePathName));
		    }
		    
		} else {
		    mediaFilePathName = file.getAbsolutePath();
		}
	    }
	}
    }



    // Public interface methods
    public String getMediaFilePathName() {
	return mediaFilePathName;
    }

    public void setMediaFilePathName(String fn) {
	mediaFilePathName = fn;
    }


    public String getRelativeMediaFilePathName() {
	return relativeMediaFilePathName;
    }

    public void setRelativeMediaFilePathName(String rfn) {
	relativeMediaFilePathName = rfn;
    }


    public String getBaseFileName() {
	String baseName = null;
	if (mediaFilePathName != null) {
	    int index = mediaFilePathName.lastIndexOf(PropertyManager.FILE_SEPARATOR);
	    if (index >= 0) {
		baseName = mediaFilePathName.substring(index + 1);
	    }
	}
	return baseName;
    }


    public boolean isHighlighted() {
	return isHighlighted;
    }

    public void setIsHighlighted(boolean b) {
	isHighlighted = b;
    }


    public void mouseEnter() {
	isOver = true;
	notifyRolloverListeners();
    }

    public void mouseLeave() {
	wasSingleClicked = false;
	wasDoubleClicked = false;
	isOver = false;
	notifyRolloverListeners();
    }


    public boolean isOver() {
	return isOver;
    }

    public void setIsOver(boolean b) {
	isOver = b;
    }


    public void mouseSingleClicked() {
	setWasSingleClicked(true, true);
    }

    public void setWasSingleClicked(boolean b, boolean notify) {
	wasSingleClicked = b;
	if (notify) {
	    notifySingleClickListeners();
	}
    }

    public boolean wasSingleClicked() {
	return wasSingleClicked;
    }


    public void mouseDoubleClicked() {
	setWasDoubleClicked(true, true);
    }

    public void setWasDoubleClicked(boolean b, boolean notify) {
	wasDoubleClicked = b;
	if (notify) {
	    notifyDoubleClickListeners();
	}
    }

    public boolean wasDoubleClicked() {
	return wasDoubleClicked;
    }


    public void display() {
    }
    public void unDisplay() {
    }
    public void highlight() {
    }
    public void unHighlight() {
    }


    public void realize() {
    }
    public void unRealize() {
    }


    public void addSingleClickListener(MediaModelListenerInterface obj) {
	if (! singleClickListeners.contains(obj)) {
	    singleClickListeners.add(obj);
	}
    }

    public void removeSingleClickListener(MediaModelListenerInterface obj) {
	singleClickListeners.remove(obj);
    }

    public void removeSingleClickListeners() {
	singleClickListeners.clear();
    }



    public void addDoubleClickListener(MediaModelListenerInterface obj) {
	if (! doubleClickListeners.contains(obj)) {
	    doubleClickListeners.add(obj);
	}
    }

    public void removeDoubleClickListener(MediaModelListenerInterface obj) {
	doubleClickListeners.remove(obj);
    }
    
    public void removeDoubleClickListeners() {
	doubleClickListeners.clear();
    }



    public void addRolloverListener(MediaModelListenerInterface obj) {
	if (! rolloverListeners.contains(obj)) {
	    rolloverListeners.add(obj);
	}
    }
 
    public void removeRolloverListener(MediaModelListenerInterface obj) {
	rolloverListeners.remove(obj);
    }

    public void removeRolloverListeners() {
	rolloverListeners.clear();
    }


    public void removeMediaModelListeners() {
	removeSingleClickListeners();
	removeDoubleClickListeners(); 
	removeRolloverListeners();
    }


    public void removeListeners() {
	super.removeListeners();
	removeMediaModelListeners();
    }


    public void unOperationalize() {
	super.unOperationalize();
	removeMediaModelListeners();
	wasSingleClicked = false;
	wasDoubleClicked = false;
	isOver = false;
	isHighlighted = false;
    }

    public void deleteModel() {
	super.deleteModel();
	removeMediaModelListeners();
    }


    // Private methods
    private void notifySingleClickListeners() {
	Iterator it = singleClickListeners.iterator();
	while (it.hasNext()) {
	    MediaModelListenerInterface listener = (MediaModelListenerInterface) it.next();
	    listener.mediaModelSingleClicked(this);
	}
    }

    private void notifyDoubleClickListeners() {	
	Iterator it = doubleClickListeners.iterator();
	while (it.hasNext()) {
	    MediaModelListenerInterface listener = (MediaModelListenerInterface) it.next();
	    listener.mediaModelDoubleClicked(this);
	}
    }


    private void notifyRolloverListeners() {
	Iterator it = rolloverListeners.iterator();
	while (it.hasNext()) {
	    MediaModelListenerInterface listener = (MediaModelListenerInterface) it.next();
	    listener.mediaModelRolledOver(this);
	}
    }



    public void write(DataOutputStream stream) throws IOException {
	// Write the parent information
	super.write(stream);

	// Write the relative media file path name
	stream.writeBoolean(relativeMediaFilePathName != null);
	if (relativeMediaFilePathName != null) {
	    // Write file name length
	    stream.writeInt(relativeMediaFilePathName.length());
	    // Write the file name
	    stream.writeChars(relativeMediaFilePathName);
	} 
	
	// Write the absolute media file path name
	stream.writeBoolean(mediaFilePathName != null);
	if (mediaFilePathName != null) {
	    // Write file name length
	    stream.writeInt(mediaFilePathName.length());
	    // Write the file name
	    stream.writeChars(mediaFilePathName);
	}
    }
}



