/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: MultiViewModel.java
*/

// Current package
package edu.uofminnesota.DEMAIS.meal;

// Import needed packages
import java.lang.*;
import java.util.*;
import java.io.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.meal.layer.*;
import edu.uofminnesota.DEMAIS.usrinterface.contentmanager.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.multiview.*;





public class MultiViewModel extends DataModel {

    // Used to create subsequent names
    private static int MV_IDENTIFIER;

    // A global list of multiview models in the system
    private static Vector multiViewModels;


    // The set of layers associated with this multiview
    private LayerList layers;

    // The multiview panel
    private MultiViewPanel multiViewPanel;





    // Static initializer
    static {
	MV_IDENTIFIER = 0;
	multiViewModels = new Vector();
    }


    private static int getMVIdentifier() {
	++MV_IDENTIFIER;
	return MV_IDENTIFIER;
    }

    public static void setMVIdentifier(int id) {
	MV_IDENTIFIER = id;
    }


    public static void addMultiViewModel(MultiViewModel model) {
	multiViewModels.add(model);
	PropertyManager.setProperty(PropertyManager.APPLICATION_MULTIVIEW_COUNT, new Integer(multiViewModels.size()));

    }


    public static void removeMultiViewModel(MultiViewModel model) {
	multiViewModels.remove(model);
	PropertyManager.setProperty(PropertyManager.APPLICATION_MULTIVIEW_COUNT, new Integer(multiViewModels.size()));

	
    }


    public static MultiViewModel getFirstMultiViewModel() {
	if (multiViewModels.size() > 0) {
	    return (MultiViewModel) multiViewModels.firstElement();
	}
	return null;
    }
    
    // Returns the multiview that comes before model in the list
    // If model is the 0th element, then the 1st element is returned
    public static MultiViewModel getPreviousMultiViewModel(MultiViewModel model) {
	return getPreviousMultiViewModel(model, true);
    }
    
    public static MultiViewModel getPreviousMultiViewModel(MultiViewModel model, boolean flag) {
	MultiViewModel prev = null;
	
	boolean first = true;
	Iterator it = multiViewModels.iterator();
	while (it.hasNext()) {
	    MultiViewModel current = (MultiViewModel) it.next();
	    if (current.equals(model)) {
		break;
	    }
	    first = false;
	    prev = current;
	}
     
	if (first && flag) {
	    if (multiViewModels.size() > 1) {
		// First model was deleted, so return the following one
		prev = (MultiViewModel) multiViewModels.elementAt(1);
	    }
	}

	return prev;
    }


    // Returns the multiView model that comes after model in the list
    public static MultiViewModel getNextMultiViewModel(MultiViewModel model) {
	MultiViewModel next = null;
	Iterator it = multiViewModels.iterator();
	while (it.hasNext()) {
	    MultiViewModel current = (MultiViewModel) it.next();
	    if (current.equals(model)) {
		if (it.hasNext()) {
		    next = (MultiViewModel) it.next();
		}
	    }
	}
	return next;
    }



    public static MultiViewModel getMultiViewModelById(Integer id) {
	Iterator it = multiViewModels.iterator();
	while (it.hasNext()) {
	    MultiViewModel current = (MultiViewModel) it.next();
	    if (current.equals(id)) {
		return current;
	    }
	}
	return null;
    }



    public static Vector getMultiViewModels() {
	return multiViewModels;
    }
    


    // Constructors
    private MultiViewModel() {
	this(0, null, null);
    }

    public MultiViewModel(int scope, DataModel localScope, String modelName) {
	super(scope, localScope, modelName);
	if (modelName == null) {
	    setName("MultiView" + Integer.toString(MultiViewModel.getMVIdentifier()), false);
	}

	// Create initial set of layers
	layers = new LayerList();
	layers.addLayer(new ContentLayer("Content", 0, true, true));

	// Create the multiview panel
	multiViewPanel = new MultiViewPanel(this);

	MultiViewModel.addMultiViewModel(this);
	ContentManagerFrame.getFrame().addDataModel(this);
    }


    public MultiViewModel(MultiViewModel copy) {
	super(copy, true);
    }
    
    public MultiViewModel(MultiViewModel copy, boolean copyOfModels) {
	super(copy, copyOfModels);
	setName("MultiView" + Integer.toString(MultiViewModel.getMVIdentifier()), false);

	// Copy the layer vector
	layers = null;
	if (copy.layers != null) { 
	    // Makes a deep copy of the layer list
	    layers = (LayerList) copy.layers.clone();
	    
	} else {
	    layers = new LayerList();
	    layers.addLayer(new ContentLayer("Content", 0, true, true));
	}

	// Clone the multiview panel
	multiViewPanel = null;
	if (copy.multiViewPanel != null) {
	    multiViewPanel = (MultiViewPanel) copy.multiViewPanel.clone(this, copyOfModels);
	}

	// Add to the global list of multiview models
	MultiViewModel.addMultiViewModel(this);
	ContentManagerFrame.getFrame().addDataModel(this);
	
    }


    public MultiViewModel(DataInputStream stream) throws IOException {
	super(stream);

	// Read and set the class identifier value
	MV_IDENTIFIER = stream.readInt();

	// Read the layer information
	layers = new LayerList(stream);

	// Read the multiview panel
	multiViewPanel = new MultiViewPanel(stream, this);

	// Add to the global list of multiview models
	MultiViewModel.addMultiViewModel(this);
	ContentManagerFrame.getFrame().addDataModel(this);
    }


    public void completeIO() {
	multiViewPanel.completeIO();
    }


    public Storyboard getStoryboard() {
	return getMultiViewPanel();
    }

    public MultiViewPanel getMultiViewPanel() {
	return multiViewPanel;
    }


    public void operationalize() {
	layers.operationalize();
	multiViewPanel.operationalize();
	
	// Operationalize the multiview first, which sets the annotations
	// Then operationalize the parent, which parses those annotations
	super.operationalize();
    }

    public void unOperationalize() {
	super.unOperationalize();
	layers.unOperationalize();
	multiViewPanel.unOperationalize();
    }

    public void realize(boolean showControls, boolean recreate) {
	super.realize(showControls, recreate);
	multiViewPanel.realize(showControls, recreate);
    }
    
    public void unRealize(boolean keepVisuals) {
	super.unRealize(keepVisuals);
	multiViewPanel.unRealize(keepVisuals);
    }


    // Layer methods
    public boolean isLayerVisible(int layer) {
	return layers.isLayerVisible(layer);
    }

    public boolean isLayerSelected(int layer) {
	return layers.isLayerSelected(layer);
    }

    public int getSelectedLayer() {
	return layers.getSelectedLayer();
    }

    public int getNumLayers() {
	return layers.getNumLayers();
    }

    public LayerModel getSelectedLayerModel() {
	return layers.getSelectedLayerModel();
    }

    public String getLayerName(int layer) {
	return layers.getLayerName(layer);
    }

    public LayerModel getLayerModel(int layer) {
	return layers.getLayerModel(layer);
    }

    public void setLayerName(int layer, String name) {
	layers.setLayerName(layer, name);
    }

    public void setLayerVisible(int layer, boolean visible) {
	layers.setLayerVisible(layer, visible);
    }

    public void setLayerSelected(int layer, boolean selected) {
	layers.setLayerSelected(layer, selected);
    }

    public LayerList getLayers() {
	return layers;
    }



    // Delete the model from the system - close out all views on this model
    public void deleteModel() {
	// Destroy the parent model first
	super.deleteModel();

	// Allow the multiview to clean up
	if (multiViewPanel != null) {
	    multiViewPanel.deleteMultiView();
	}

	// Set to the next available multiview, if we are deleting the current one
	DataModel cModel = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	if (cModel != null) {
	    if (this.equals(cModel)) {
		DataModel nextModel = MultiViewModel.getPreviousMultiViewModel(this);
		if (nextModel == null) {
		    nextModel = StoryboardModel.getFirstStoryboardModel();
		}
		PropertyManager.setProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL, nextModel);
	    }	    
	}

	// Remove the multiview from the global list
	MultiViewModel.removeMultiViewModel(this);
    }



    // View operations
    public ContentManagerCellView getContentManagerCellView() {
	MultiViewContentManagerCellView view = new MultiViewContentManagerCellView(this);
	addView(view);
	return view;
    }


    public Object clone() {
	return (new MultiViewModel(this));
    }


    public Object clone(boolean copyOfModels) {
	return (new MultiViewModel(this, copyOfModels));
    }



    public void write(DataOutputStream stream) throws IOException {
	super.write(stream);

	// Write the static identifier
	stream.writeInt(MV_IDENTIFIER);

	// Write the layer vector
	layers.write(stream);


	// Write the multiview panel information
	multiViewPanel.write(stream);
    }

}
