/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: NamedModel.java
*/

// Current package
package edu.uofminnesota.DEMAIS.meal;

// Import needed packages
import java.lang.*;
import java.util.*;
import java.io.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.core.parser.*;



// Represents an object with a name 
public abstract class NamedModel extends Object {

    // Static variable for assigning ids
    private static int NO_IDENTIFIER;


    // The identifier of the object - globally unique
    private Integer identifier;

    // The scope into which this object was placed
    // LOCAL_SCOPE, GLOBAL_SCOPE, or APPLICATION_SCOPE
    protected int scope;

    // If placed into LOCAL_SCOPE, the reference to that scope
    protected DataModel localScope;

    // The name of the object - need not be unique
    private String name;

    // A list of objects to notify when this object's name changes
    private Vector nameListeners;





    // Static initializer
    static {
	NO_IDENTIFIER = -1;
    }


    // Static functions
    public static Integer getNextIdentifier() {
	++NO_IDENTIFIER;
	return (new Integer(NO_IDENTIFIER));
    }




    // Constructors
    private NamedModel() {
    }

    public NamedModel(int scope, DataModel localScope, String modelName) {
	identifier = NamedModel.getNextIdentifier();
	this.scope = scope;
	this.localScope = localScope;
	name = modelName;
	if (name == null) {
	    // Set a default unique name
	    name = "Model" + identifier.toString();
	}

	nameListeners = new Vector();

	// Place this object into the symbol table
	SymbolTableManager.put(scope, localScope, this);

	// A new system object has been created, a save is needed
	PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
    }


    public NamedModel(NamedModel copy) {
	this(copy, copy.localScope);
    }


    public NamedModel(NamedModel copy, DataModel toModel) {
	identifier = NamedModel.getNextIdentifier();
	scope = copy.scope;
	localScope = toModel;

	name = null;
	if (copy.name != null) {
	    name = new String(copy.name);
	} else {
	    name = "Model" + identifier.toString();
	}

	// Do not copy the name listeners. Listening objects 
	// will need to add themselves to the copy
	nameListeners = new Vector();

	// Place this new object into the symbol table
	SymbolTableManager.put(scope, localScope, this);

	PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
    }


    public NamedModel(DataInputStream stream, DataModel localScope) throws IOException {
	// Read the class identifier
	NO_IDENTIFIER = stream.readInt();

	// Read the object identifier
	identifier = new Integer(stream.readInt());

	// Read the scope
	scope = stream.readInt(); 
	this.localScope = localScope;
	
	// Read the model name
	name = "";
	int length = stream.readInt();
	for (int i=0; i<length; ++i) {
	    name += stream.readChar();
	}
	
	// Create name listener structure
	nameListeners = new Vector();
	
	// Place this object into the symbol table
	SymbolTableManager.put(scope, localScope, this);
    }




    // Public methods
    public Integer getIdentifier() {
	return identifier;
    }

    // Prepend '__' to prevent the parser from detecting it as a raw integer
    public String getIdentifierId() {
	return "__" + identifier.toString();
    }

    public String getName() {
	return name;
    }

 
    public void setName(String newName) {
	setName(newName, true);
    }

    public void setName(String newName, boolean notify) {
	String oldName = name;
	name = newName;
	if (notify) {
	    notifyNameListeners(oldName, newName);
	}
	PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
    }



    // Force concrete subclasses to implement this method
    public void addNameListener(NamedModelNameListenerInterface nli) {
	if (! nameListeners.contains(nli)) {
	    nameListeners.add(nli);
	}
    }
    
    // Remove the specified name listener
    public void removeNameListener(NamedModelNameListenerInterface nli) {
	nameListeners.remove(nli);
    }

    // Remove all name listeners
    public void removeNameListeners() {
	nameListeners.clear();
    }

    public void removeListeners() {
	removeNameListeners();
    }

    public void operationalize() {
    }

    public void unOperationalize() {
	removeNameListeners();
    }

    public void deleteModel() {
	// Remove listeners
	removeNameListeners();

	// Remove from the symbol table
	SymbolTableManager.remove(scope, localScope, this);
    }

    
    public int hashCode() {
	return identifier.hashCode();
    }


    public boolean equals(Object obj) {
	if (obj instanceof NamedModel) {
	    return (equals((NamedModel) obj));
	
	} else if (obj instanceof Integer) {
	    return (equals((Integer) obj));
	
	} else if (obj instanceof String) {
	    return (equals((String) obj));
	} 
	
	return (this == obj);
    }

    public boolean equals(NamedModel no) {
	return (identifier.intValue() == no.identifier.intValue());
    }

    public boolean equals(Integer i) {
	return (identifier.intValue() == i.intValue());
    }

    public boolean equals(String s) {
	return (name.equals(s));
    }


    // Notify objects listening for a name change
    private void notifyNameListeners(String oldName, String newName) {
	Iterator it = nameListeners.iterator();
	while (it.hasNext()) {
	    NamedModelNameListenerInterface nli = (NamedModelNameListenerInterface) it.next();
	    nli.nameChanged(this, oldName, newName);
	}
    }


    public void write(DataOutputStream stream) throws IOException {
	
	// Write the value of the class identifier
	stream.writeInt(NO_IDENTIFIER);
	
	// Write the object identifier
	stream.writeInt(identifier.intValue());
	
	// Write the scope
	stream.writeInt(scope);
	
	// Write name length
	if (name != null) {
	    stream.writeInt(name.length());
	    // Write name
	    stream.writeChars(name);
	} else {
	    stream.writeInt(0);
	}
    }

}

