/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: NarrationModel.java
  
  Same as a voice script
*/

// Current package
package edu.uofminnesota.DEMAIS.meal;

// Import needed packages
import java.lang.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.print.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.speech.synthesis.*;
import edu.uofminnesota.DEMAIS.core.parser.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.usrinterface.narration.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.contentmanager.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;





public class NarrationModel extends DataModel 
    implements ContinuousMediaModelInterface,
	       Printable {
    

    // Used to create subsequent narration names
    private static int NM_IDENTIFIER;

    // A global list of narration models in the system
    private static Vector narrationModels;

    private static final Font PRINT_FONT;




    // The document model for the narration
    private DefaultStyledDocument docModel;

    // Indexes into the narration string to
    // pinpoint the synchronization markers
    private Vector syncMarkers;

    // Maps each marker to a variable at runtime
    private Vector variableList;



    /*
      Listeners to notify when the narration begins or ends
    */
    private Vector beginListeners;
    private Vector endListeners;

    private boolean wasBeginEvent;
    private boolean wasEndEvent;

    // Speakable Listener
    private Listener listener;

    // Current state of playback
    private int state;

    // change update listener
    private ChangeListener changeListener;




    // Static initializer
    static {
	NM_IDENTIFIER = 0;
	narrationModels = new Vector();
	PRINT_FONT = new Font("Times", Font.PLAIN, 10);
    }

    private static int getNMIdentifier() {
	++NM_IDENTIFIER;
	return NM_IDENTIFIER;
    }

    public static void setNMIdentifier(int id) {
	NM_IDENTIFIER = id;
    }

    
    public static void addNarrationModel(NarrationModel model) {
	narrationModels.add(model);
    }
    
    
    public static void removeNarrationModel(NarrationModel model) {
	narrationModels.remove(model);
    }

    
    public static NarrationModel getNarrationModelById(Integer id) {
	Iterator it = narrationModels.iterator();
	while (it.hasNext()) {
	    NarrationModel current = (NarrationModel) it.next();
	    if (current.equals(id)) {
		return current;
	    }
	}
	return null;
    }


    public static Vector getNarrationModels() {
	return narrationModels;
    }
    





    // Constructors
    private NarrationModel() {
	this(0, null, null, null);
    }

    public NarrationModel(int scope, DataModel localScope, String modelName, String narrationText) {
	super(scope, localScope, modelName);
	setName("Narration" + Integer.toString(NarrationModel.getNMIdentifier()), false);

	docModel = new DefaultStyledDocument();
	if (narrationText != null) {
	    try {
		docModel.insertString(0, narrationText, null);
	    } catch (Exception xcptn) {
		DEMAISError.showErrorMessage(xcptn.toString(), "NarrationModel", false);
		DEMAISError.logException(xcptn);
	    }
	}

	syncMarkers = new Vector();
	variableList = new Vector();

	beginListeners = new Vector();
	endListeners = new Vector();
	wasBeginEvent = false;
	wasEndEvent = false;

	listener = new Listener();
	state = ContinuousMediaModelInterface.INITIAL;

	changeListener = new ChangeListener();
	docModel.addDocumentListener(changeListener);

	NarrationModel.addNarrationModel(this);
	ContentManagerFrame.getFrame().addDataModel(this);
    }


    public NarrationModel(NarrationModel copy) {
	this(copy, true);
    }

    public NarrationModel(NarrationModel copy, boolean copyOfModels) {
	super(copy, copyOfModels);
	setName("Narration" + Integer.toString(NarrationModel.getNMIdentifier()), false);

	// Copy the document model
	docModel = new DefaultStyledDocument();
	String copyNarrationText = copy.getNarrationText(true);
	if (copyNarrationText != null) {
	    try {
		docModel.insertString(0, copyNarrationText, null);
		
	    } catch (Exception xcptn1) {
		DEMAISError.showErrorMessage(xcptn1.toString(), "NarrationModel", false);
		DEMAISError.logException(xcptn1);
	    }
	}

	// Copy the sync marker locations
	syncMarkers = null;
	if (copy.syncMarkers != null) {
	    syncMarkers = new Vector();
	    Iterator it = copy.syncMarkers.iterator();
	    while (it.hasNext()) {
		Integer index = (Integer) it.next();
		syncMarkers.add(new Integer(index.intValue()));
	    }
	} else {
	    syncMarkers = new Vector();
	}

	// Position the sync markers
	positionSyncMarkers();


	// Copy the variables
	variableList = null;
	if (copy.variableList != null) {
	    variableList = new Vector();
	    Iterator it = copy.variableList.iterator();
	    while (it.hasNext()) {
		Variable var = (Variable) it.next();
		variableList.add(var.clone(this));
	    }
	} else {
	    variableList = new Vector();
	}

	beginListeners = new Vector();
	endListeners = new Vector();
	wasBeginEvent = false;
	wasEndEvent = false;

	listener = new Listener();
	state = ContinuousMediaModelInterface.INITIAL;

	changeListener = new ChangeListener();
	docModel.addDocumentListener(changeListener);

	NarrationModel.addNarrationModel(this);
	ContentManagerFrame.getFrame().addDataModel(this);
    }


    
    public NarrationModel(DataInputStream stream) throws IOException {
	super(stream);

	// Read the static identifier value
	NM_IDENTIFIER = stream.readInt();

	// Read the number of chars to follow
	String narrationText = "";
	int numChars = stream.readInt();
	// Read the text
	for (int c=0; c<numChars; ++c) {
	    narrationText += stream.readChar();
	}

	docModel = new DefaultStyledDocument();
	try {
	    docModel.insertString(0, narrationText, null);
	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "NarrationModel", false);
	    DEMAISError.logException(xcptn);
	}

	// Read the syncMarker list
	boolean hasSyncMarkers = stream.readBoolean();
	if (hasSyncMarkers) {
	    int numSyncMarkers = stream.readInt();
	    syncMarkers = new Vector();
	    for (int i=0; i<numSyncMarkers; ++i) {
		int smPosition = stream.readInt();
		syncMarkers.add(new Integer(smPosition));
	    }	 
	} else {
	    syncMarkers = new Vector();
	}

	positionSyncMarkers();

	// Read the variable information
	boolean hasVariables = stream.readBoolean();
	if (hasVariables) {
	    int numVariables = stream.readInt();
	    variableList = new Vector();
	    for (int i=0; i<numVariables; ++i) {
		Variable v = new Variable(stream, this);
		variableList.add(v);
	    }    
	} else {
	    variableList = new Vector();
	}
	

	beginListeners = new Vector();
	endListeners = new Vector();
	wasBeginEvent = false;
	wasEndEvent = false;

	listener = new Listener();
	state = ContinuousMediaModelInterface.INITIAL;

	changeListener = new ChangeListener();
	docModel.addDocumentListener(changeListener);

	NarrationModel.addNarrationModel(this);
	ContentManagerFrame.getFrame().addDataModel(this);
    }
    

    private void positionSyncMarkers() {
	if (syncMarkers != null) {
	    // Insert the sync icons into the new document.
	    SimpleAttributeSet inputAttributes = new SimpleAttributeSet();
	    StyleConstants.setIcon(inputAttributes, NarrationTextPane.SYNC_ICON);
	    Iterator it = syncMarkers.iterator();
	    while (it.hasNext()) {
		int index = ( (Integer) it.next()).intValue();
		try {
		    docModel.insertString(index, " ", inputAttributes);
		} catch (Exception xcptn) {
		    DEMAISError.showErrorMessage(xcptn.toString(), "PositionSyncMarkers", false);
		    DEMAISError.logException(xcptn);
		}
	    }    
	} 
    }
    

    public String getNarrationText(boolean build) {
	return getNarrationText(0, docModel.getLength(), build);
    }

    public String getNarrationText(int end, boolean build) {
	if (end > docModel.getLength()) {
	    end = docModel.getLength();
	}
	return getNarrationText(0, end, build);
    }

    public String getNarrationText(int begin, int end, boolean build) {
	if (build) {
	    buildSyncMarkersAndVariables();
	}
	String ret = null;
	try {
	    // Get the entire text string
	    StringBuffer buffer = new StringBuffer(docModel.getText(0, docModel.getLength()));

	    // The sync points show up as spaces in this text. Remove locations
	    // matching sync point locations before returning the text
	    Vector descending = sortDescending(syncMarkers);
	    Iterator it = descending.iterator();
	    while (it.hasNext()) {
		int index = ((Integer) it.next()).intValue();
		buffer.deleteCharAt(index);
	    }

	    if (begin > buffer.length()) {
		begin = buffer.length() - 1;
	    }

	    if (end > buffer.length()) {
		end = buffer.length();
	    }

	    ret = buffer.substring(begin, end);

	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "NarrationModel::getNarrationText()", false);
	    DEMAISError.logException(xcptn);
	}

	return ret;
    }


    public DefaultStyledDocument getDocument() {
	return docModel;
    }

    
    public Collection getSyncMarkers() {
	return syncMarkers;
    }


    public Variable getVariableAtIndex(int index) {
	if ( (variableList == null) ||
	     (index < 0) || 
	     (index > variableList.size())) {
	    return null;
	}
	return (Variable) variableList.elementAt(index);
    }


    public NarrationEditView getNarrationEditView() {
	// Before creating a new edit view, determine if an edit view
	// already exists for this model.
	NarrationEditView v = null;
	Iterator it = openModelViews.iterator();
	while (it.hasNext()) {
	    DataModelViewInterface dmvi = (DataModelViewInterface) it.next();
	    if (dmvi instanceof NarrationEditView) {
		v = (NarrationEditView) dmvi;
		break;
	    }
	}

	if (v == null) {
	    v = new NarrationEditView(this);
	    addView(v);
	}

	return v;
    }


    public ContentManagerCellView getContentManagerCellView() {
	NarrationContentManagerCellView view = new NarrationContentManagerCellView(this);
	addView(view);
	return view;
    }



    // Called when the editor window is closed
    public void buildSyncMarkersAndVariables() {
	// Clear the out-of-date sets
	if (syncMarkers == null) {
	    syncMarkers = new Vector();
	} else {
	    syncMarkers.clear();
	}

	if (variableList == null) {
	    variableList = new Vector();
	} else {
	    Iterator it = variableList.iterator();
	    while (it.hasNext()){
		Variable variable = (Variable) it.next();
		SymbolTableManager.remove(PropertyManager.LOCAL_SCOPE, this, variable);
	    }
	    variableList.clear();
	}
	
	// Look at each of the elements
	for (int i=0; i<docModel.getLength(); ++i) {
	    Element e = docModel.getCharacterElement(i);
	    if (e.getName().equals("icon")) {
		// Found a sync marker
		syncMarkers.add(new Integer(i));
		variableList.add(new Variable(PropertyManager.LOCAL_SCOPE, this, null, 0.0));
	    }
	}
    }


    public void addBeginListener(ContinuousMediaModelListenerInterface cli) {
	if (! beginListeners.contains(cli)) {
	    beginListeners.add(cli);
	}
    }

    public void removeBeginListener(ContinuousMediaModelListenerInterface cli) {
	beginListeners.remove(cli);
    }

    public void removeBeginListeners() {
	beginListeners.clear();
    }


    public void addEndListener(ContinuousMediaModelListenerInterface cli) {
	if (! endListeners.contains(cli)) {
	    endListeners.add(cli);
	}
    }

    public void removeEndListener(ContinuousMediaModelListenerInterface cli) {
	endListeners.remove(cli);
    }

    public void removeEndListeners() {
	endListeners.clear();
    }



    public void removeNarrationModelListeners() {
	removeBeginListeners();
	removeEndListeners();
    }


    public void removeListeners() {
	super.removeListeners();
	removeNarrationModelListeners();
    }



    public void deleteModel() {
	super.deleteModel();
	removeNarrationModelListeners();
	NarrationModel.removeNarrationModel(this);
	end();
    }


    public void realize(boolean showControls, boolean recreate) {
	super.realize(showControls, recreate);
	state = ContinuousMediaModelInterface.INITIAL;
    }

    public void unRealize(boolean keepVisuals) {
	super.unRealize(keepVisuals);
	reset();
    }


    public void operationalize() {
	super.operationalize();
	reset();

	// Make sure variable values are 0
	Iterator it = variableList.iterator();
	while (it.hasNext()) {
	    Variable variable = (Variable) it.next();
	    variable.setValue(0.0, false);
	}
    }

    public void unOperationalize() {
	super.unOperationalize();
	removeNarrationModelListeners();
	reset();	
    }
    

    // ContinuousMediaModelInterface
    public void reset() {
	state = ContinuousMediaModelInterface.RESET;
	TextToSpeechSynthesizer.getTTS().stopSpeaking();
    }


    public void begin() {
	if (state != ContinuousMediaModelInterface.PAUSED) {
	    // Start/Restart the synthesizer
	    state = ContinuousMediaModelInterface.BEGINNING;
	    TextToSpeechSynthesizer.getTTS().speakJSMLText(getJSMLText(), listener);
	
	} else {
	    TextToSpeechSynthesizer.getTTS().resumeSpeaking();
	}
    }

    private void finishBegin() {
	if ( (state == ContinuousMediaModelInterface.INITIAL) || 
	     (state == ContinuousMediaModelInterface.BEGINNING) ) {
	    state = ContinuousMediaModelInterface.BEGUN;

	    // Reset variables
	    Iterator it = variableList.iterator();
	    while (it.hasNext()) {
		Variable variable = (Variable) it.next();
		variable.setValue(0.0);
	    }
	    setWasBeginEvent(true, true);

	} else {
	    state = ContinuousMediaModelInterface.BEGUN;
	}

    }

    public boolean hasBegun() {
	return ( (state == ContinuousMediaModelInterface.BEGUN) ||
		 (state == ContinuousMediaModelInterface.PAUSED));
    }


    public void end() {
	// Prevent from being called more than once on a stopped object
	if (state < ContinuousMediaModelInterface.ENDING) {
	    state = ContinuousMediaModelInterface.ENDING;
	    TextToSpeechSynthesizer.getTTS().stopSpeaking();
	}
    }
    
    private void finishEnd() {
	state = ContinuousMediaModelInterface.ENDED;
	setWasEndEvent(true, true);
    }
	    
    public boolean hasEnded() {
	return (state == ContinuousMediaModelInterface.ENDED);
    }


    public void pause() {
	state = ContinuousMediaModelInterface.PAUSING;
	TextToSpeechSynthesizer.getTTS().pauseSpeaking();
    }

    private void finishPause() {
	state = ContinuousMediaModelInterface.PAUSED;
    }


    public void resume() {
	begin();
    }

    public void setMediaTime(double value) {
    }
    public double getMediaTime() {
	return 0.0;
    }
    public void setRate(double rate) {
    }
    public double getRate() {
	return 0.0;
    }


    public void mediaBeginEvent() {
	setWasBeginEvent(true, true);
    }

    public boolean wasBeginEvent() {
	return wasBeginEvent;
    }

    public void setWasBeginEvent(boolean b, boolean notify) {
	wasBeginEvent = b;
	if (notify) {
	    notifyBeginListeners();
	}
    }


    public void mediaEndEvent() {
	setWasEndEvent(true, true);
    }

    public boolean wasEndEvent() {
	return wasEndEvent();
    }

    public void setWasEndEvent(boolean b, boolean notify) {
	wasEndEvent = b;
	if (notify) {
	    notifyEndListeners();
	}
	if (!wasEndEvent) {
	    // Make sure variable values are 0
	    Iterator it = variableList.iterator();
	    while (it.hasNext()) {
		Variable variable = (Variable) it.next();
		variable.setValue(0.0, false);
	    }
	}
    }



    // For edit mode
    public void beginEditNarration(int fromIndex) {
	beginEditNarration(fromIndex, null);
    }
    
    public void beginEditNarration(int fromIndex, SpeakableListener listener) {
	int length = docModel.getLength();
	if (fromIndex >= length) {
	    fromIndex = 0;
	}
	TextToSpeechSynthesizer.getTTS().speakPlainText(getNarrationText(fromIndex, length, true), listener);
    }

    public void pauseEditNarration() {
       	TextToSpeechSynthesizer.getTTS().pauseSpeaking();
    }

    public void resumeEditNarration() {
       	TextToSpeechSynthesizer.getTTS().resumeSpeaking();
    }

    public void endEditNarration() {
       	TextToSpeechSynthesizer.getTTS().stopSpeaking();
    }



    public Object clone() {
	return (new NarrationModel(this));
    }

    public Object clone(boolean copyOfModels) {
	return (new NarrationModel(this, copyOfModels));
    }



    public int print(Graphics g, PageFormat pf, int pi)
	throws PrinterException {

	Graphics2D g2d = (Graphics2D)g;
	g2d.translate(pf.getImageableX(), pf.getImageableY());
	g2d.setFont(PRINT_FONT);
	g2d.setColor(Color.black);	
	FontMetrics fm = g2d.getFontMetrics();

	int maxWidth = (int)Math.round(pf.getImageableWidth());
	int maxHeight = (int)Math.round(pf.getImageableHeight());
	int cX = 5;
	int cY = fm.getHeight();

	String printS = getNarrationText(true);
	for (int i=0; i<printS.length(); ++i) {
	    String s = "";
	    char ch = printS.charAt(i);
	    s += ch;
	    g2d.drawString(s, cX, cY);

	    if ((cX + fm.charWidth(ch)) > maxWidth) {
		cX = 5;
		cY += fm.getHeight();
	    } else {
		cX += fm.charWidth(ch);
	    }
	}
	
	return Printable.PAGE_EXISTS;
    }
    



    private String getJSMLText() {
	StringBuffer jsmlText = new StringBuffer(getNarrationText(false));

	// Sort in descending order
	Vector descending = sortDescending(syncMarkers);
	Iterator it = descending.iterator();
	int position = descending.size() - 1;
	int offset = position;

	while (it.hasNext()) {
	    Integer index = (Integer) it.next();
	    jsmlText.insert(index.intValue() - offset, "<MARKER MARK=\"" + Integer.toString(position) + "\"/>");
	    --position;
	    --offset;
	}
	jsmlText.insert(0, "<JSML>");
	jsmlText.append("</JSML>");

	return jsmlText.toString();
    }



    protected void notifyBeginListeners() {
	Iterator it = beginListeners.iterator();
	while (it.hasNext()) {
	    ContinuousMediaModelListenerInterface listener = (ContinuousMediaModelListenerInterface) it.next();
	    listener.CMediaModelBegun(this);
	}
    }

    protected void notifyEndListeners() {
	Iterator it = endListeners.iterator();
	while (it.hasNext()) {
	    ContinuousMediaModelListenerInterface listener = (ContinuousMediaModelListenerInterface) it.next();
	    listener.CMediaModelEnded(this);
	}
    }

    public Storyboard getStoryboard() {
	return null;
    }


    // Sort DESCENDING
    private Vector sortDescending(Vector ascending) {
	Vector descending = new Vector();
	Iterator it = ascending.iterator();
	while (it.hasNext()) {
	    int i = ((Integer) it.next()).intValue();
	    descending.insertElementAt(new Integer(i), 0);
	}

	return descending;
    }

    


    public class Listener extends SpeakableAdapter {

	public Listener() {
	}
	

	public void markerReached(SpeakableEvent evt) {
	    // Map the marker to the variable that it affects
	    String markerIndex = evt.getText();
	    int index = -1;
	    try {
		index = Integer.parseInt(markerIndex);
	    } catch (Exception xcptn) {
	    }

	    if (index >= 0) {
		Variable variable = (Variable) getVariableAtIndex(index);
		if (variable != null) {
		    // Change Monitor will detect the change
		    variable.setValue(1.0);
		} 
	    }
	}
	
	public void speakableStarted(SpeakableEvent evt) {
	    finishBegin();
	}
	
	public void speakablePaused(SpeakableEvent evt) {
	    finishPause();	
	}
	
	public void speakableResumed(SpeakableEvent evt) {
	    state = ContinuousMediaModelInterface.BEGUN;
	}
	
	public void speakableEnded(SpeakableEvent evt) {
	    // Could end from normal playout
	    state = ContinuousMediaModelInterface.ENDING;
	    finishEnd();
	}	
    }


 
    
    public void write(DataOutputStream stream) throws IOException {
	// Write the parent information
	super.write(stream);

	// Write the static class identifier
	stream.writeInt(NM_IDENTIFIER);

	// Write the text
	String text = getNarrationText(true);
	// Write length of text
	if (text != null) {
	    stream.writeInt(text.length());
	    // Write text
	    stream.writeChars(text);

	} else {
	    stream.writeInt(0);
	}

	// Write number of sync markers
	stream.writeBoolean(syncMarkers != null);
	if (syncMarkers != null) {
	    stream.writeInt(syncMarkers.size());
	    // Write the markers
	    Iterator it = syncMarkers.iterator();
	    while (it.hasNext()) {
		Integer location = (Integer) it.next();
		stream.writeInt(location.intValue());
	    }
	}

	// Write number of variables
	stream.writeBoolean(variableList != null);
	if (variableList != null) {
	    stream.writeInt(variableList.size());
	    // Write the variable information
	    Iterator it = variableList.iterator();
	    while (it.hasNext()) {
		Variable variable = (Variable) it.next();
		// Make sure the variable value is 0!
		variable.setValue(0.0, false);
		variable.write(stream);
	    }
	}
    }


    class ChangeListener implements DocumentListener {
	public ChangeListener() {
	}

	public void changedUpdate(DocumentEvent e) {
	}

	public void insertUpdate(DocumentEvent e) {
	    PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
	}

	public void removeUpdate(DocumentEvent e) {
	    PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
	}

    }

}

