/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: TextModel.java
*/

// Current package
package edu.uofminnesota.DEMAIS.meal;

// Import needed packages
import java.lang.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import javax.speech.synthesis.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.*;




public class TextModel extends StaticMediaModel 
    implements ContinuousMediaModelInterface {


    // Corresponding graphical object
    private TextGraphicalObject textGraphicalObject;

    // Sequence of characters typed by the user
    private Vector charSequence;


    // Listeners to notify when the narration begins or ends
    private Vector beginListeners;
    private Vector endListeners;
    
    private boolean wasBeginEvent;
    private boolean wasEndEvent;

    // Speakable Listener
    private Listener listener;

    // Current state of playback
    private int state;




    // Constructors
    private TextModel() {
	this(0, null, null, null);
    }

    public TextModel(int scope, DataModel localScope, String modelName, TextGraphicalObject tgo) {
	super(scope, localScope, modelName, null);
	textGraphicalObject = tgo;
	charSequence = new Vector();
	beginListeners = new Vector();
	endListeners = new Vector();
	listener = new Listener();
	state = ContinuousMediaModelInterface.INITIAL;
    }


    public TextModel(TextModel copy, TextGraphicalObject tgo) {
	this(copy, tgo, tgo.getStoryboard().getModel());
    }


    public TextModel(TextModel copy, TextGraphicalObject tgo, DataModel toModel) {
	super(copy, toModel);
	textGraphicalObject = tgo;
	
	charSequence = null;
	if (copy.charSequence != null) {
	    charSequence = new Vector(copy.charSequence.size());
	    Iterator it = copy.charSequence.iterator();
	    while (it.hasNext()) {
		char c = ((Character) it.next()).charValue();
		charSequence.add(new Character(c));
	    }
	} else {
	    charSequence = new Vector();
	}

	beginListeners = new Vector();
	endListeners = new Vector();
	wasBeginEvent = false;
	wasEndEvent = false;

	listener = new Listener();
	state = ContinuousMediaModelInterface.INITIAL;
    }


    public TextModel(DataInputStream stream, TextGraphicalObject tgo, DataModel localScope) throws IOException {
	super(stream, localScope);
	textGraphicalObject = tgo;

	boolean hasChars = stream.readBoolean();
	if (hasChars) {
	    // Read number of characters
	    int numChars = stream.readInt();
	    charSequence = new Vector();
	    for (int i=0; i<numChars; ++i) {
		char ch = stream.readChar();
		charSequence.add(new Character(ch));
	    }
	} else {
	    charSequence = new Vector();
	}

	beginListeners = new Vector();
	endListeners = new Vector();
	wasBeginEvent = false;
	wasEndEvent = false;

	listener = new Listener();
	state = ContinuousMediaModelInterface.INITIAL;
    }




    // Public methods
    public void addChar(char c) {
	addChar(new Character(c));
    }

    public void addChar(Character c) {
	charSequence.add(c);
    }

    public void addChar(int index, char c) {
	addChar(index, new Character(c));
    }

    public void addChar(int index, Character c) {
	charSequence.insertElementAt(c, index);
    }


    public Vector getCharSequence() {
	return charSequence;
    }


    public char charAt(int index) {
	if (index >= charSequence.size()) {
	    index = charSequence.size() - 1;
	} else if (index < 0) {
	    index = 0;
	}
	return ((Character)charSequence.elementAt(index)).charValue();
    }


    public void removeCharAt(int index) {
	if ( (index < 0) || (index >= charSequence.size()) ) {
	    return;
	}
	charSequence.removeElementAt(index);
    }


    public int size() {
	return charSequence.size();
    }


    public String getText() {
	String s = "";
	Iterator it = charSequence.iterator();
	while (it.hasNext()) {
	    s += ((Character) it.next()).charValue();
	}
	return s;
    }


    public String getString() {
	return getText();
    }


    public void display() {
	textGraphicalObject.display();
    }

    public void unDisplay() {
	textGraphicalObject.unDisplay();
    }

    public void highlight() {
	textGraphicalObject.highlight();
    }

    public void unHighlight() {
	textGraphicalObject.unHighlight();
    }


    public Object clone(TextGraphicalObject tgo) {
	return (new TextModel(this, tgo));
    }



    public void addBeginListener(ContinuousMediaModelListenerInterface cli) {
	if (! beginListeners.contains(cli)) {
	    beginListeners.add(cli);
	}
    }

    public void removeBeginListener(ContinuousMediaModelListenerInterface cli) {
	beginListeners.remove(cli);
    }

    public void removeBeginListeners() {
	beginListeners.clear();
    }


    public void addEndListener(ContinuousMediaModelListenerInterface cli) {
	if (! endListeners.contains(cli)) {
	    endListeners.add(cli);
	}
    }

    public void removeEndListener(ContinuousMediaModelListenerInterface cli) {
	endListeners.remove(cli);
    }

    public void removeEndListeners() {
	endListeners.clear();
    }


    public void removeTextModelListeners() {
	removeBeginListeners();
	removeEndListeners();
    }

    public void removeListeners() {
	super.removeListeners();
	removeTextModelListeners();
    }

    
    public void deleteModel() {
	super.deleteModel();
	removeTextModelListeners();
	reset();
    }




    public void realize() {
	super.realize();
	state = ContinuousMediaModelInterface.INITIAL;
    }
    
    
    public void unRealize() {
	super.unRealize();
	reset();
    }


    public void operationalize() {
	super.operationalize();
	reset();
    }
    

    public void unOperationalize() {
	super.unOperationalize();
	removeTextModelListeners();
	reset();
    }


    // ContinuousMediaModelInterface
    public void reset() {
	state = ContinuousMediaModelInterface.INITIAL;
	TextToSpeechSynthesizer.getTTS().stopSpeaking();
    }


    public void begin() {
	if (state != ContinuousMediaModelInterface.PAUSED) {
	    // Start/Restart the synthesizer
	    state = ContinuousMediaModelInterface.BEGINNING;
	    TextToSpeechSynthesizer.getTTS().speakPlainText(getText(), listener);
	
	} else {
	    TextToSpeechSynthesizer.getTTS().resumeSpeaking();
	}
    }

    private void finishBegin() {
	if ( (state == ContinuousMediaModelInterface.INITIAL) || 
	     (state == ContinuousMediaModelInterface.BEGINNING) ) {
	    state = ContinuousMediaModelInterface.BEGUN;
	    setWasBeginEvent(true, true);

	} else {
	    state = ContinuousMediaModelInterface.BEGUN;
	}
    }

    public boolean hasBegun() {
	return ( (state == ContinuousMediaModelInterface.BEGUN) ||
		 (state == ContinuousMediaModelInterface.PAUSED));
    }


    public void end() {
	// Prevent from being called more than once on a stopped object
	if (state < ContinuousMediaModelInterface.ENDING) {
	    state = ContinuousMediaModelInterface.ENDING;
	    TextToSpeechSynthesizer.getTTS().stopSpeaking();
	}
    }
    
    private void finishEnd() {
	state = ContinuousMediaModelInterface.ENDED;
	setWasEndEvent(true, true);
    }
	    
    public boolean hasEnded() {
	return (state == ContinuousMediaModelInterface.ENDED);
    }


    public void pause() {
	state = ContinuousMediaModelInterface.PAUSING;
	TextToSpeechSynthesizer.getTTS().pauseSpeaking();
    }

    private void finishPause() {
	state = ContinuousMediaModelInterface.PAUSED;
    }


    public void resume() {
	begin();
    }



    public void setMediaTime(double value) {
    }
    public double getMediaTime() {
	return 0.0;
    }
    public void setRate(double rate) {
    }
    public double getRate() {
	return 0.0;
    }
    

    public void mediaBeginEvent() {
	setWasBeginEvent(true, true);
    }

    public boolean wasBeginEvent() {
	return wasBeginEvent;
    }

    public void setWasBeginEvent(boolean b, boolean notify) {
	wasBeginEvent = b;
	if (notify) {
	    notifyBeginListeners();
	}
    }


    public void mediaEndEvent() {
	setWasEndEvent(true, true);
    }

    public boolean wasEndEvent() {
	return wasEndEvent;
    }

    public void setWasEndEvent(boolean b, boolean notify) {
	wasEndEvent = b;
	if (notify) {
	    notifyEndListeners();
	}
    }



    protected void notifyBeginListeners() {
	Iterator it = beginListeners.iterator();
	while (it.hasNext()) {
	    ContinuousMediaModelListenerInterface listener = (ContinuousMediaModelListenerInterface) it.next();
	    listener.CMediaModelBegun(this);
	}
    }

    protected void notifyEndListeners() {
	Iterator it = endListeners.iterator();
	while (it.hasNext()) {
	    ContinuousMediaModelListenerInterface listener = (ContinuousMediaModelListenerInterface) it.next();
	    listener.CMediaModelEnded(this);
	}
    }



    public class Listener extends SpeakableAdapter {

	public Listener() {
	}
	
	public void speakableStarted(SpeakableEvent evt) {
	    finishBegin();
	}
	
	public void speakablePaused(SpeakableEvent evt) {
	    finishPause();	
	}
	
	public void speakableResumed(SpeakableEvent evt) {
	    state = ContinuousMediaModelInterface.BEGUN;
	}
	
	public void speakableEnded(SpeakableEvent evt) {
	    state = ContinuousMediaModelInterface.ENDING;
	    finishEnd();
	}
	
    }

    
    public void write(DataOutputStream stream) throws IOException {
	super.write(stream);
	
	stream.writeBoolean(charSequence != null);
	if (charSequence != null) {
	    // Write number of characters
	    stream.writeInt(charSequence.size());
	    // Write the characters
	    Iterator it = charSequence.iterator();
	    while (it.hasNext()) {
		Character c = (Character) it.next();
		stream.writeChar(c.charValue());
	    }
	}
    }
	
}


