/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: LayerList.java
*/

// Current package
package edu.uofminnesota.DEMAIS.meal.layer;

// Import needed packages
import java.lang.*;
import java.util.*;
import java.io.*;
import edu.uofminnesota.DEMAIS.property.*;


/* 
   Layers are stored with the lowest numbered layer being first in the list. Thus, a layer model's number equals its position in the list.
*/


public class LayerList extends LinkedList {


    public LayerList() {
	super();
    }

    public LayerList(LayerList copy) {
	super();

	Iterator it = copy.iterator();
	while (it.hasNext()) {
	    LayerModel lm = (LayerModel) it.next();
	    add((LayerModel)lm.clone());
	}
    }

    public LayerList(DataInputStream stream) throws IOException {
	super();

	// Read the number of layer models
	int numLayers = stream.readInt();

	// Read the layers
	for (int i=0; i<numLayers; ++i) {
	    LayerModel lm = null;

	    int classid = stream.readInt();
	    switch (classid) {
	    case ContentLayer.CLASSID:
	    case InteractionLayer.CLASSID:
		lm = new ContentLayer(stream);
		break;

	    case AnnotationLayer.CLASSID:
		lm = new AnnotationLayer(stream);
		break;

	    default:
		break;
	    }

	    if (lm != null) {
		add(lm);
	    }
	}
    }


    // Public methods
    public boolean isLayerVisible(int layer) {
	LayerModel lm = getLayerModelByNumber(layer);
	if (lm != null) {
	    return lm.isVisible();
	
	} else if (layer == -1) {
	    Integer mode = (Integer) PropertyManager.getProperty(PropertyManager.APPLICATION_MODE);
	    if (mode.intValue() == PropertyManager.APPLICATION_EDIT) {
		return true;
	    } else {
		return false;
	    }
	}
	return false;
    }

    public boolean isLayerSelected(int layer) {
	LayerModel lm = getLayerModelByNumber(layer);
	if (lm != null) {
	    return lm.isSelected();
	}
	return false;
    }


    public int getSelectedLayer() {
	Iterator it = iterator();
	while (it.hasNext()) {
	    LayerModel lm = (LayerModel) it.next();
	    if (lm.isSelected()) {
		// There can only be 1 selected layer
		return lm.getLayer();
	    }
	}
	return 0;
    }
    

    public int getNumLayers() {
	return size();
    }


    public LayerModel getSelectedLayerModel() {
	Iterator it = iterator();
	while (it.hasNext()) {
	    LayerModel lm = (LayerModel) it.next();
	    if (lm.isSelected()) {
		// There can only be 1 selected layer
		return lm;
	    }
	}
	return null;
    }
    
    
    public String getLayerName(int layer) {
	String ret = "";
	LayerModel lm = getLayerModelByNumber(layer);
	if (lm != null) {
	    ret = lm.getName();
	}
	return ret;
    }

    
    public LayerModel getLayerModel(int layer) {
	return getLayerModelByNumber(layer);
    }


    public void setLayerName(int layer, String name) {
	LayerModel lm = getLayerModelByNumber(layer);
	if (lm != null) {
	    lm.setName(name);
	}
    }
    

    public void setLayerVisible(int layer, boolean visible) {
	LayerModel lm = getLayerModelByNumber(layer);
	if (lm != null) {
	    lm.setVisible(visible);
	}
    }
    

    public void setLayerSelected(int layer, boolean selected) {
	LayerModel lm = getLayerModelByNumber(layer);
	if (lm != null) {
	    lm.setSelected(selected);
	}
    }


    public void newLayer(LayerModel lm) {
	addLayer(lm);
    }


    public void addLayer(LayerModel lm) {
	add(lm);
    }


    public void copyLayer(int layerNum, String newLayerName, int newLayerNum, boolean visible, boolean selected) {
	LayerModel lm = (LayerModel) getLayerModelByNumber(layerNum);
	if (lm != null) {
	    LayerModel lm_clone = (LayerModel) lm.clone();
	    lm_clone.setName(newLayerName);
	    lm_clone.setLayer(newLayerNum);
	    lm_clone.setVisible(visible);
	    lm_clone.setSelected(selected);
	    addLayer(lm_clone);
	}
    }


    public void deleteLayer(int layerNum) {
	LayerModel lm = getLayerModelByNumber(layerNum);
	if (lm != null) {
	    remove(lm);
	    renumberLayers();
	}
    }


    public void repositionLayer(int layerNum, int toPosition) {
	if (layerNum == toPosition) {
	    return;
	}

	LayerModel lm = getLayerModelByNumber(layerNum);
	if (lm != null) {
	    add(toPosition, lm);
	    
	    if (layerNum > toPosition) {
		++layerNum;
	    }
	    remove(layerNum);   
	    renumberLayers();
	}
    }


    public void flattenLayers() {
	LayerModel first = (LayerModel) getFirst();
	clear();
	first.setSelected(true);
	first.setVisible(true);
	addLayer(first);
    }


    public void operationalize() {
	Iterator it = iterator();
	while (it.hasNext()) {
	    LayerModel lm = (LayerModel) it.next();
	    lm.operationalize();
	}
    }

    public void unOperationalize() {
	Iterator it = iterator();
	while (it.hasNext()) {
	    LayerModel lm = (LayerModel) it.next();
	    lm.unOperationalize();
	}
    }


    public Object clone() {
	return (new LayerList(this));
    }


    // Private methods
    private void renumberLayers() {
	int count = 0;
	Iterator it = iterator();
	while (it.hasNext()) {
	    LayerModel lm = (LayerModel) it.next();
	    lm.setLayer(count);
	    ++count;
	}
    }


    private LayerModel getLayerModelByNumber(int layerNum) {
	// Use the fact that the layer number represents its location in the list
	LayerModel lm = null;
	if ((layerNum >= 0) && (layerNum < size())) {
	    lm = (LayerModel) get(layerNum);
	}
	return lm;
    }


    public void write(DataOutputStream stream) throws IOException {
	// Write length of list
	stream.writeInt(size());
	
	// write the layers
	Iterator it = iterator();
	while (it.hasNext()) {
	    LayerModel lm = (LayerModel) it.next();
	    lm.write(stream);
	}
    }

}
