/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: PropertyManagerQueue.java
*/

 
// Current package
package edu.uofminnesota.DEMAIS.property;

// Import needed packages
import java.lang.*;
import java.util.*;
import javax.swing.*;



public class PropertyManagerQueue extends Vector 
    implements Runnable {

    // Holds the properties from firing until this object is signaled
    private boolean signaled;




    // Constructors
    public PropertyManagerQueue() {
	this(25);
    }

    public PropertyManagerQueue(int capacity) {
	super(capacity);
	signaled = false;
    }


    // Public methods
    public void signal() {
	signaled = true;
	if (size() > 0) {
	    SwingUtilities.invokeLater(this);
	}
    }


    public void insert(Integer property, Object oldV, Object newV) {
	QueueItem item = new QueueItem(property, oldV, newV);

	/*
	  If Q already contains the property, then remove it 
	  and add the newest one - collapses duplicate properties
	*/
	if (contains(item)) {
	    remove(item);
	}
	add(item);

	/* 
	   If this is the first item inserted into the Q, and it has been
	   signaled (meaning the program is past the initialization code),
	   then schedule the notifications.
	*/
	if ((size() == 1) && (signaled)) {
	    SwingUtilities.invokeLater(this);
	}
    }


    protected void flush() {
	// Propagate the properties now, as opposed to later
	propagate();
    }


    // Runnable interface
    public void run() {
	propagate();
    }


    private void propagate() {
	/*
	  Propagate the entire queue of properties - Note that this could be a potential infinite loop. But the property dependencies are minimal, so it should not be an issue
	*/
	while (size() > 0) {
	    // Remove the head of the queue
	    QueueItem item = (QueueItem) remove(0);
	    PropertyManager.notifyPropertyListeners(item.property, item.oldValue, item.newValue);    
	}
    }
 


    // An element placed on the queue - wraps the property
    private class QueueItem {
	public Integer property;
	public Object oldValue;
	public Object newValue;

	QueueItem(Integer p, Object oldV, Object newV) {
	    property = p;
	    oldValue = oldV;
	    newValue = newV;
	}

	public boolean equals(Object obj) {
	    if (obj instanceof QueueItem) {
		return (equals((QueueItem) obj));
	    }
	    return (this == obj);
	}

	public boolean equals(QueueItem item) {
	    return (property.equals(item.property));
	}
    }
}

