/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: ChangeMonitor.java

Once an annotation is parsed, it is given to the ChangeMonitor. The ChangeMonitor looks through the condition stack for variables, continuous media objects, and behaviors which may change during runtime as a result of user interaction or system updates. When a change occurs, the associated annotation conditions need to be re-evaluated. The ChangeMonitor places itself on the listener queues for each of the changeable objects. When an object changes at runtime, the ChangeMonitor callback is invoked and the necesssary annotations are re-evaluated.
*/

// Current package
package edu.uofminnesota.DEMAIS.runtime;

// Import needed packages
import java.lang.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.core.annotation.*;
import edu.uofminnesota.DEMAIS.core.parser.*;




public class ChangeMonitor implements VariableListenerInterface, MediaModelListenerInterface, StaticMediaModelListenerInterface, ContinuousMediaModelListenerInterface, AnimationModelListenerInterface, TextEntryModelListenerInterface {

    // Follow a singleton design pattern
    private static ChangeMonitor cm;


    // Queue of changed objects to process
    private ChangeMonitorQueue cmQ;

    // The table maps a NamedModel object to a vector of annotations
    // which need to be notified when a change to that object occurs
    private Hashtable objectToAnnotation;



    // static initializer
    static {
	cm = null;
    }


    // Retrieve the singleton instance
    public static ChangeMonitor getCM() {
	if (cm == null) {
	    // Create a single instance of the change monitor
	    cm = new ChangeMonitor();
	}
	return cm;
    }


    // Private constructor to prevent additional instantiation
    private ChangeMonitor() {
	cmQ = new ChangeMonitorQueue(25);
	objectToAnnotation = new Hashtable(50);	
    }



    // Public methods
    public void monitorAnnotation(Annotation annotation) throws EvaluationException {
	buildBehaviorsAndTraceables(annotation);
    }


    // Removes this annotation from the CM
    public void removeAnnotation(Annotation annotation) {
	Vector traceables = annotation.getCondition().getTraceables();
	if (traceables != null) {
	    Iterator it = traceables.iterator();
	    while (it.hasNext()) {
		NamedModel key = (NamedModel) it.next();
		if (objectToAnnotation.containsKey(key)) {
		    // Assured that the key exists
		    Vector v = (Vector) objectToAnnotation.get(key);
		    // Remove the annotation from the vector
		    v.remove(annotation);
		    // Was this the last element in the vector?
		    if (v.size() == 0) {
			// Remove all listeners for this traceable as there are
			// no more annotations that need to be notified
			key.removeListeners();
			objectToAnnotation.remove(key);
		    }
		}
	    }
	}
    }


    // Clear out the ChangeMonitor
    public void removeAll() {
	// Loop through the keys (NamedModels) in the hash table
	for (Enumeration e = objectToAnnotation.keys(); e.hasMoreElements(); ) {   
	    NamedModel obj = (NamedModel) e.nextElement();
	    // Remove all listeners for this object
	    obj.removeListeners();
	}

	// Clear out the hashtable
	objectToAnnotation.clear();

	// Clear out the queue
	cmQ.removeAllElements();
    }



    /*
      Go through the condition opcode stack looking for variables, static media objects, and continous media objects. In some cases, the behavior represented by the objects will be created. 
    */
    private void buildBehaviorsAndTraceables(Annotation annotation) throws EvaluationException {
	// Make a local shallow copy of the condition stack
	Stack lStack = (Stack) annotation.getConditionStack().clone();

	while (! lStack.empty()) {
	
	    // Get the opcode on top of the stack
	    Opcode code = (Opcode) lStack.peek();

	    try {
		// Look at only the necessary opcodes
		if (code.equals(Opcode.BEGINS)) {
		    // Pop this opcode from the stack
		    lStack.pop();
		    // Next is a continuous media object
		    Opcode op = (Opcode) lStack.pop();
		    NamedModel nm = (NamedModel) op.getObject();
		    ContinuousMediaModelInterface cmi = (ContinuousMediaModelInterface) nm;
		    // Add the ChangeMonitor as a begin listener
		    cmi.addBeginListener(ChangeMonitor.cm);
		    // Also add as a EndListener
		    cmi.addEndListener(ChangeMonitor.cm);
		    
		    // Add this object to the traceables table
		    addTraceable(nm, annotation);
		    
		} else if (code.equals(Opcode.ENDS)) {
		    // Pop current opcode
		    lStack.pop();
		    // Next is a continuous media object
		    Opcode op = (Opcode) lStack.pop();
		    NamedModel nm = (NamedModel) op.getObject();
		    ContinuousMediaModelInterface cmi = (ContinuousMediaModelInterface) nm;
		    // Add the ChangeMonitor as a listener
		    cmi.addEndListener(ChangeMonitor.cm);
		    
		    // Also add as a BeginListener
		    cmi.addBeginListener(ChangeMonitor.cm);
		    
		    // Add this object to the traceables table
		    addTraceable(nm, annotation);
		    
		} else if (code.equals(Opcode.SELECTS) || 
			   code.equals(Opcode.CLICKS) ||
			   code.equals(Opcode.PRESSES)) {
		    // Pop this opcode
		    lStack.pop();
		    // Next object is a media object
		    Opcode op = (Opcode) lStack.pop();
		    MediaModel mm = (MediaModel) op.getObject();
		    mm.addSingleClickListener(ChangeMonitor.cm);
		    // Add this object to the traceables table
		    addTraceable(mm, annotation);
		    
		} else if (code.equals(Opcode.DOUBLECLICKS)) {
		    // Pop this opcode
		    lStack.pop();
		    // Next object is a media object
		    Opcode op = (Opcode) lStack.pop();
		    MediaModel mm = (MediaModel) op.getObject();
		    mm.addDoubleClickListener(ChangeMonitor.cm);
		    // Add this object to the traceables table
		    addTraceable(mm, annotation);
		    
		} else if (code.equals(Opcode.ROLLOVER)) {
		    lStack.pop();
		    Opcode op = (Opcode) lStack.pop();
		    MediaModel mm = (MediaModel) op.getObject();
		    mm.addRolloverListener(ChangeMonitor.cm);
		    
		    // Add this object to the traceables table
		    addTraceable(mm, annotation);
		    
		} else if (code.equals(Opcode.REACHES)) {
		    // Pop reaches
		    lStack.pop();
		    // Pop point 2
		    Opcode r1 = (Opcode) lStack.pop();
		    // Pop point 1
		    Opcode r2 = (Opcode) lStack.pop();
		    // Pop the animation object
		    Opcode obj = (Opcode) lStack.pop();
		    AnimationModel animation = (AnimationModel) obj.getObject();
		    // Add the ChangeMonitor as a listener
		    animation.addLocationListener(ChangeMonitor.cm, (int)Math.round(r1.getRealNumber()), (int)Math.round(r2.getRealNumber()));
		    addTraceable(animation, annotation);
		    
		} else if (code.equals(Opcode.DRAG_AND_DROP)) {
		    // Pop drags
		    lStack.pop();
		    // Pop target
		    Opcode op2 = (Opcode) lStack.pop();
		    // Pop the source
		    Opcode op1 = (Opcode) lStack.pop();
		    
		    MediaModel target = (MediaModel) op2.getObject();
		    StaticMediaModel source = (StaticMediaModel) op1.getObject();
		    source.addDragAndDropListener(ChangeMonitor.cm);
		    
		    addTraceable(source, annotation);
		    
		} else if (code.equals(Opcode.ENTERS_TEXT)) {
		    // Pop enters_text
		    lStack.pop();
		    // Next is the media object
		    Opcode op = (Opcode) lStack.pop();
		    TextEntryModel entry = (TextEntryModel) op.getObject();
		    entry.addTextEntryListener(ChangeMonitor.cm);
		    addTraceable(entry, annotation);
		    
		} else if (code.equals(Opcode.MEDIA_OBJECT)) {
		    Object object = code.getObject();
		    lStack.pop();
		    
		    if (object instanceof JMFContinuousMediaModel) {
			JMFContinuousMediaModel jmf = (JMFContinuousMediaModel) object;
			jmf.addValueListener(ChangeMonitor.cm);
			jmf.addRateListener(ChangeMonitor.cm);
			
			addTraceable(jmf, annotation);
		    }
		    
		} else if (code.equals(Opcode.VARIABLE)) {
		    Variable v = (Variable) code.getObject();
		    v.addValueListener(ChangeMonitor.cm);
		    lStack.pop();
		    
		    addTraceable(v, annotation);
		    
		} else {
		    lStack.pop();
		}
	    
	    } catch (Exception xcptn) {
		System.out.println(xcptn.toString());
	    }
	}
    }


    private void addTraceable(NamedModel obj, Annotation annotation) {
	Vector v = null;

	// Map the traceable object to the annotation
	if (objectToAnnotation.containsKey(obj)) {
	    v = (Vector) objectToAnnotation.get(obj);
	} else {
	    v = new Vector();
	    objectToAnnotation.put(obj, v);
	}
	v.add(annotation);

	// Also add the traceable to the annotation
	annotation.getCondition().addTraceable(obj);
    }



    /*
      For the listener interfaces, when an object changes:
	  (1) Ask the Scheduler to remove pending actions for the annotations
	  (2) Ask the Evaluator to calculate new predictions for the annotations
    */

    // VariableListenerInterface
    public void valueChanged(Variable v) {
	handleObjectChanged(v);
    }

   
    // MediaModelListenerInterface
    public void mediaModelSingleClicked(MediaModel mm) {
	handleObjectChanged(mm);
    }

    public void mediaModelDoubleClicked(MediaModel mm) {
	handleObjectChanged(mm);
    }

    public void mediaModelRolledOver(MediaModel mm) {
	handleObjectChanged(mm);
    }
    
    
    // StaticMediaModelListenerInterface
    public void wasDraggedAndDropped(StaticMediaModel source, MediaModel target) {
	handleObjectChanged(source);
    }
    

    // ContinuousMediaModelListenerInterface
    public void CMediaModelBegun(ContinuousMediaModelInterface cmi) {
	handleObjectChanged((NamedModel)cmi);
    }

    public void CMediaModelEnded(ContinuousMediaModelInterface cmi) {
	handleObjectChanged((NamedModel)cmi);
    }

    public void CMediaModelRateChanged(ContinuousMediaModelInterface cmi) {
	handleObjectChanged((NamedModel)cmi);
    }

    public void CMediaModelValueChanged(ContinuousMediaModelInterface cmi) {
	handleObjectChanged((NamedModel)cmi);
    }


    // AnimationListenerInterface
    public void AnimationModelLocationReached(AnimationModel animModel) {
	handleObjectChanged(animModel);
    }


    // TextEntryModelListenerInterface
    public void textEntered(TextEntryModel tem) {
	handleObjectChanged(tem);
    }


    public boolean equals(Object obj) {
	return (this == obj);
    }



    // Private methods
    private void handleObjectChanged(NamedModel obj) {
	// Inserts the changed object into the queue and
	// schedules it for handling
	cmQ.insert(obj);
    }



    // Callback from the ChangeMonitorQueue
    // Determine all annotations that need to be re-evaluated as a result of the change
    protected void doObjectChanged(NamedModel obj) {
	Vector vector = (Vector) objectToAnnotation.get(obj);
	if (vector == null) {
	    return;
	}

	boolean wasSingleClick = false;
	boolean wasDoubleClick = false;
	boolean wasDragAndDrop = false;
	boolean wasBeginEvent = false;
	boolean wasEndEvent = false;
	boolean wasTextEvent = false;

	// Mark the type of change
	if (obj instanceof ContinuousMediaModelInterface) {
	    ContinuousMediaModelInterface cmm = (ContinuousMediaModelInterface) obj;
	    if (cmm.wasBeginEvent()) {
		wasBeginEvent = true;

	    } else if (cmm.wasEndEvent()) {
		wasEndEvent = true;
	    }
	} 

	if (obj instanceof StaticMediaModelInterface) {
	    StaticMediaModelInterface smm = (StaticMediaModelInterface) obj;
	    if (smm.wasDraggedAndDropped()) {
		wasDragAndDrop = true;
	    }
	} 

	if (obj instanceof MediaModelInterface) {
	    MediaModelInterface mm = (MediaModelInterface) obj;
	    if (mm.wasSingleClicked()) {
		wasSingleClick = true;

	    } else if (mm.wasDoubleClicked()) {
		wasDoubleClick = true;
	    }
	} 

	if (obj instanceof TextEntryModelInterface) {
	    TextEntryModelInterface tem = (TextEntryModelInterface) obj;
	    if (tem.wasTextEntered()) {
		wasTextEvent = true;
	    }
	}

	//System.out.println("Changed Object: " + obj.getIdentifierId());

	// Do not use an iterator here, due to co-modification exceptions
	for (int i=0; i<vector.size(); ++i) {
	    Annotation annotation = (Annotation) vector.elementAt(i);

	    // Remove pending actions for this annotation
	    Scheduler.getScheduler().removeAnnotation(annotation);
	    
	    // Calculate new predictions
	    try {
		//System.out.println(annotation.toString());
		//System.out.println("Current LV: " + Evaluator.getEvaluator().getLogicValue(annotation).toString());
		LogicValue lv = Evaluator.getEvaluator().reEvaluateAnnotation(annotation);
		
		//System.out.println("New LV: " + lv.toString());
		//System.out.println();

	    } catch (EvaluationException xcptn) {
		System.out.println("ChangeMonitor:doObjectChanged() - " + xcptn.toString());
	    }
	}
	//System.out.println();
	//System.out.println();
	
	// Reset the scheduler's timer
	Scheduler.getScheduler().setTimer();

	// Reset the annotations
	if (wasSingleClick) {
	    for (int i=0; i<vector.size(); ++i) {
		Annotation annotation = (Annotation) vector.elementAt(i);
		Evaluator.getEvaluator().resetLogicValue(annotation);
	    }
	    MediaModelInterface mm = (MediaModelInterface) obj;
	    mm.setWasSingleClicked(false, false);
		
	} 

	if (wasDoubleClick) {
	    for (int i=0; i<vector.size(); ++i) {
		Annotation annotation = (Annotation) vector.elementAt(i);
		Evaluator.getEvaluator().resetLogicValue(annotation);
	    }
	    MediaModelInterface mm = (MediaModelInterface) obj;
	    mm.setWasDoubleClicked(false, false);

	} 

	if (wasDragAndDrop) {
	    for (int i=0; i<vector.size(); ++i) {
		Annotation annotation = (Annotation) vector.elementAt(i);
		Evaluator.getEvaluator().resetLogicValue(annotation);
	    }
	    StaticMediaModelInterface smm = (StaticMediaModelInterface) obj;
	    smm.setWasDraggedAndDroppedOn(false, null, false);
	    
	} 
	
	if (wasBeginEvent) {
	    for (int i=0; i<vector.size(); ++i) {
		Annotation annotation = (Annotation) vector.elementAt(i);
		Evaluator.getEvaluator().resetLogicValue(annotation);
	    }
	    ContinuousMediaModelInterface cmm = (ContinuousMediaModelInterface) obj;
	    cmm.setWasBeginEvent(false, false);
	    
	} 
	
	if (wasEndEvent) {
	    for (int i=0; i<vector.size(); ++i) {
		Annotation annotation = (Annotation) vector.elementAt(i);
		Evaluator.getEvaluator().resetLogicValue(annotation);
	    }
	    ContinuousMediaModelInterface cmm = (ContinuousMediaModelInterface) obj;
	    cmm.setWasEndEvent(false, false);

	} 

	if (wasTextEvent) {
	    for (int i=0; i<vector.size(); ++i) {
		Annotation annotation = (Annotation) vector.elementAt(i);
		Evaluator.getEvaluator().resetLogicValue(annotation);
	    }
	    TextEntryModelInterface tem = (TextEntryModelInterface) obj;
	    tem.setWasTextEntered(false, false);
	    
	}
    } 
}


