/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: Evaluator.java

  The Evaluator evaluates both an annotation's condition and action. To evaluate an annotation, the Evaluator uses our predictive logic developed as part of the Nsync project. The predictive logic uses two temporal values, WBT and WBF. See the ACM Multimedia 1998 paper for further details.
*/


// Current package
package edu.uofminnesota.DEMAIS.runtime;

// Import needed packages
import java.lang.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.core.annotation.*;
import edu.uofminnesota.DEMAIS.core.parser.*;
import edu.uofminnesota.DEMAIS.usrinterface.transition.*;




public class Evaluator {

    // Implement a singleton design pattern
    private static Evaluator evaluator;


    // Current LogicValues for each condition in the annotation
    private Hashtable lValues;



    // Static initializer
    static {
	evaluator = null;
    }

    // Return the singleton instance
    public static Evaluator getEvaluator() {
	if (evaluator == null) {
	    evaluator = new Evaluator();
	}
	return evaluator;
    }



    // Private constructor to prevent additional instantiation
    private Evaluator() {
	lValues = new Hashtable(50);
    }


    // Public methods

    // This method should be called ONLY for the first evaluation of an annotation
    public LogicValue evaluateAnnotation(Annotation annotation) throws EvaluationException {
	// Initialize the logic value in the table
	lValues.put(annotation, LogicValue.FALSE);

	// Evaluate the annotation's condition
	return reEvaluateAnnotation(annotation);
    }


    public void removeAnnotation(Annotation annotation) {
	lValues.remove(annotation);
    }


    // Set the Evaluator back to its initial state
    public void removeAll() {
	lValues.clear();
    }



    // Protected static methods
    protected LogicValue getLogicValue(Annotation annotation) {
	return (LogicValue) lValues.get(annotation);
    }


    // Re-evaluate an annotation's condition
    protected LogicValue reEvaluateAnnotation(Annotation annotation) throws EvaluationException {
	Stack cStack = (Stack) annotation.getConditionStack().clone();
	LogicValue lv = LogicValue.FALSE;

	// Only perform re-evaluation for When expressions
	// Or for the first time of an After/Once expression
	Opcode code = (Opcode) cStack.peek();
	if (code.equals(Opcode.WHEN)) {
	    lv = doEvaluation(cStack);
	    setLogicValue(annotation, lv);
	    
	} else if ((code.equals(Opcode.AFTER)) ||
		   (code.equals(Opcode.ONCE))) {
	    if (! annotation.getActionFired()) {
		// Can only be fired once
		lv = doEvaluation(cStack);
		setLogicValue(annotation, lv);
	    }

	} else {
	    System.out.println("Evaluator: Unknown opcode on top of stack");
	}

	return lv;
    }


    protected void setLogicValue(Annotation annotation, LogicValue newValue) throws EvaluationException {
	LogicValue currentValue = (LogicValue) lValues.get(annotation);
	if ((newValue == null) || (currentValue == null)) {
	    return;
	}

	//System.out.println("SET: " + annotation.toString() + " " + newValue.toString());
	lValues.put(annotation, newValue);

	/*
	  Register with the scheduler first, so that the scheduler's timer can tick while the action is potentially running. Should help reduce latency effects.
	*/

	// Check if we need to set a scheduler action; e.g., false->WBF(10)
	if ((newValue instanceof WBT) && (! newValue.isFalseAlways()) ) {
	    SchedulerAction action = new SchedulerActionWBT(annotation);
	    Scheduler.getScheduler().scheduleAction(newValue, action);
	} else if ((newValue instanceof WBF) && (! newValue.isTrueAlways())) {
	    SchedulerAction action = new SchedulerActionWBF(annotation);
	    Scheduler.getScheduler().scheduleAction(newValue, action);
	}

	if (currentValue == null) {
	    System.exit(0);
	}

	// Check for a false->true transition
	if ((currentValue.isFalseNow()) && (newValue.isTrueNow())) {
	    // A false->true transition occurred
	    // Invoke the annotation's action if it is not null
	    annotation.setActionFired(true);
	    Stack actionStack = annotation.getActionStack();
	    if (actionStack != null) {
		Stack acStack = (Stack) actionStack.clone();
		doEvaluation(acStack);
	    }
	} 
    }


    protected void resetLogicValue(Annotation annotation) {
	//System.out.println("RESET: " + annotation.toString() + " " + "WBT(-1)");
	lValues.put(annotation, LogicValue.FALSE);
    }



    // Private static methods

    // The stack evaluation routine evaluates both a condition and action stack!
    private LogicValue doEvaluation(Stack stack) throws EvaluationException {

	// The stack should never be empty!
	if (stack.empty()) {
	    throw (new EvaluationException("Evaluator: doEvaluation() - Stack was empty"));
	}
	
	// Pop the opcode on top of the stack
	Opcode code = (Opcode) stack.pop();

	// Evaluate the opcode
	if (code.equals(Opcode.AFTER)) {
	    LogicValue lv = doEvaluation(stack);
	    // For the After command, a constant must be transformed into a WBT(constant)
	    // This supports a rule such as "After 2 seconds, do ..."
	    if (lv instanceof LogicConstant) {
		// Transform the logic constant into a WBT(value)
		lv = new WBT(lv.getValue());
	    } 
	    return lv;

	} else if (code.equals(Opcode.AND)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (op1.and(op2));

	} else if (code.equals(Opcode.BEGIN)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    Object obj = op.getObject();
	    if (obj instanceof ContinuousMediaModelInterface) {
		navigateToScope(op);
		ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) obj;
		cm.begin();
	    } 
	} else if (code.equals(Opcode.BEGINS) ||
		   code.equals(Opcode.BEGUN)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) op.getObject();
	    return (LogicValue.Value(cm.wasBeginEvent()));

	} else if (code.equals(Opcode.CLICKS) ||
		   code.equals(Opcode.PRESSES) ||
		   code.equals(Opcode.SELECTS)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    MediaModelInterface mm = (MediaModelInterface) op.getObject();
	    return (LogicValue.Value(mm.wasSingleClicked()));

	} else if (code.equals(Opcode.DOUBLECLICKS)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    MediaModelInterface mm = (MediaModelInterface) op.getObject();
	    return (LogicValue.Value(mm.wasDoubleClicked()));

	} else if (code.equals(Opcode.DISPLAY)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    navigateToScope(op);
	    MediaModelInterface mm = (MediaModelInterface) op.getObject();
	    mm.display();

	} else if (code.equals(Opcode.DIVIDE)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (new LogicConstant(op1.getValue() / op2.getValue()));

	} else if (code.equals(Opcode.DRAG_AND_DROP)) {
	    Opcode op1 = (Opcode) stack.pop();
	    Opcode op2 = (Opcode) stack.pop();

	    MediaModel target = (MediaModel) op1.getObject();
	    StaticMediaModelInterface source = (StaticMediaModelInterface) op2.getObject();
	    if (source.wasDraggedAndDroppedOn(target)) {
		return LogicValue.TRUE;
	    } else {
		return LogicValue.FALSE;
	    }

	} else if (code.equals(Opcode.END)) {
	    Opcode op = (Opcode) stack.pop();
	    Object obj = op.getObject();
	    if (obj instanceof ContinuousMediaModelInterface) {
		navigateToScope(op);
		ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) obj;
		cm.end();
	    } 

	} else if (code.equals(Opcode.ENDED) || code.equals(Opcode.ENDS)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) op.getObject();
	    return (LogicValue.Value(cm.wasEndEvent()));

	} else if (code.equals(Opcode.ENTERS_TEXT)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    TextEntryModelInterface tem = (TextEntryModelInterface) op.getObject();
	    return (LogicValue.Value(tem.wasTextEntered()));

	} else if (code.equals(Opcode.EQUAL)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (handleEqual(op1, op2));

	} else if (code.equals(Opcode.GOTO)) {
	    Opcode op = (Opcode) stack.pop();
	    StoryboardModel sm = (StoryboardModel) op.getObject();
	    navigateTo(sm);

	} else if (code.equals(Opcode.GREATERTHAN)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (handleGreaterThan(op1, op2));
	    
	} else if (code.equals(Opcode.GREATERTHAN_EQUAL)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (handleGreaterThanEqual(op1, op2));

	} else if (code.equals(Opcode.HIGHLIGHT)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    navigateToScope(op);
	    MediaModelInterface mm = (MediaModelInterface) op.getObject();
	    mm.highlight();

	} else if (code.equals(Opcode.LESSTHAN)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (handleLessThan(op1, op2));

	} else if (code.equals(Opcode.LESSTHAN_EQUAL)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (handleLessThanEqual(op1, op2));

	} else if (code.equals(Opcode.MEDIA_OBJECT)) {
	    ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) code.getObject();
	    return (new LogicCMedia(cm.getMediaTime(), cm.getRate(), 0.0));

	} else if (code.equals(Opcode.MINUS)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (new LogicConstant(op1.getValue() - op2.getValue()));

	} else if (code.equals(Opcode.MULTIPLY)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (new LogicConstant(op1.getValue() * op2.getValue()));

	} else if (code.equals(Opcode.NARRATE)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) op.getObject();
	    cm.begin();

	} else if (code.equals(Opcode.NOT)) {
	    LogicValue op = doEvaluation(stack);
	    return (op.not());

	} else if (code.equals(Opcode.ONCE)) {
	    return (doEvaluation(stack));

	} else if (code.equals(Opcode.OR)) {
	    LogicValue leftOperand = doEvaluation(stack);
	    LogicValue rightOperand = doEvaluation(stack);
	    return (leftOperand.or(rightOperand));

	} else if (code.equals(Opcode.PAUSE)) {
	    Opcode op = (Opcode) stack.pop();
	    Object obj = op.getObject();
	    if (obj instanceof ContinuousMediaModelInterface) {
		navigateToScope(op);
		ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) obj;
		cm.pause();
	    }

	} else if (code.equals(Opcode.PLUS)) {
	    LogicValue op2 = doEvaluation(stack);
	    LogicValue op1 = doEvaluation(stack);
	    return (new LogicConstant(op1.getValue() + op2.getValue()));

	} else if (code.equals(Opcode.REAL_NUMBER)) {
	    return (LogicValue.Constant(code.getRealNumber()));

	} else if (code.equals(Opcode.REACHES)) {


	} else if (code.equals(Opcode.ROLLOVER)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    MediaModelInterface mm = (MediaModelInterface) op.getObject();
	    return (LogicValue.Value(mm.isOver()));

	} else if (code.equals(Opcode.SET)) {
	    // Get the real number
	    Opcode op1 = (Opcode) stack.pop();
	    double value = op1.getRealNumber();
	    // Get the variable
	    Opcode op2 = (Opcode) stack.pop();

	    // Can be a Variable or ContinuousMediaObject
	    Object obj2 = op2.getObject();
	    if (obj2 instanceof Variable) {
		Variable var = (Variable) obj2;
		var.setValue(value);
	    } else if (obj2 instanceof ContinuousMediaModelInterface) {
		navigateToScope(op2);
		ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) obj2;
		cm.setMediaTime(value);
	    }


	} else if (code.equals(Opcode.SPEAK)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    ContinuousMediaModelInterface cm = (ContinuousMediaModelInterface) op.getObject();
	    cm.begin();
	    
	}  else if (code.equals(Opcode.TRANSITION)) {
	    Opcode op = (Opcode) stack.pop();
	    StoryboardModel sm = (StoryboardModel) op.getObject();
	    Opcode speed = (Opcode) stack.pop();
	    Opcode type = (Opcode) stack.pop();
	    transitionTo(sm, type, speed);

	} else if (code.equals(Opcode.UNDISPLAY)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    navigateToScope(op);
	    MediaModelInterface mm = (MediaModelInterface) op.getObject();
	    mm.unDisplay();

	} else if (code.equals(Opcode.UNHIGHLIGHT)) {
	    // Get the next operand
	    Opcode op = (Opcode) stack.pop();
	    navigateToScope(op);
	    MediaModel mm = (MediaModel) op.getObject();
	    mm.unHighlight();

	} else if (code.equals(Opcode.VARIABLE)) {
	    Variable v = (Variable) code.getObject();
	    return (LogicValue.Constant(v.getValue()));

	} else if (code.equals(Opcode.WAIT) || 
		   code.equals(Opcode.DELAY) ||
		   code.equals(Opcode.HOLD)) {
	    // Get the next operand - a real number
	    Opcode op = (Opcode) stack.pop();
	    double time = op.getRealNumber();
	    // Convert to milliseconds
	    long ms = Math.round(1000.0 * time);
	    // Cause the delay to occur - This method restricts interactivity!
	    try {
		Thread.sleep(ms);
	    } catch (InterruptedException xcptn) {
		System.out.println(xcptn.toString());
	    }

	} else if (code.equals(Opcode.WHEN)) {
	    return (doEvaluation(stack));

	} else {
	    throw (new EvaluationException("Unknown opcode: " + code.toString()));
	}

	return null;

    }


    private void navigateToScope(Opcode code) {
	DataModel toModel = code.getScope();
	navigateTo((StoryboardModel) toModel);
    }

    
    private void navigateTo(StoryboardModel toModel) {
	DataModel cModel = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	if ((toModel != null) && 
	    (! toModel.equals(cModel))) {
	    PropertyManager.setProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL, toModel);
	    // Fixes an unusual bug when two or more navigate actions are attached to the same object
	    PropertyManager.flushQ();
	}
    }


    private void transitionTo(StoryboardModel sModel, Opcode opType, Opcode opSpeed) {
	int type = PropertyManager.DISSOLVE_TRANSITION;
	if (opType.equals(Opcode.DISSOLVE)) {
	    type = PropertyManager.DISSOLVE_TRANSITION;
	} else if (opType.equals(Opcode.WIPE)) {
	    type = PropertyManager.WIPE_TRANSITION;
	}
	
	int speed = PropertyManager.TRANSITION_SLOW;
	if (opSpeed.equals(Opcode.SLOW)) {
	    speed = PropertyManager.TRANSITION_SLOW;
	} else if (opSpeed.equals(Opcode.MEDIUM)) {
	    speed = PropertyManager.TRANSITION_MEDIUM;
	} else if (opSpeed.equals(Opcode.FAST)) {
	    speed = PropertyManager.TRANSITION_FAST;
	}
	
	transitionTo(sModel, type, speed);
    }


    private void transitionTo(StoryboardModel toModel, int type, int speed) {
	Transition transition = new Transition(toModel, type, speed);
	navigateTo(toModel);
    }



    private LogicValue handleLessThan(LogicValue op1, LogicValue op2) {
	double v1 = op1.getValue();
	double v2 = op2.getValue();
	double r1 = op1.getRate();
	double r2 = op2.getRate();

	if (v1 < v2) {
	    if (r1 <= r2) {
		// op1 is less than op2 and will stay that way
		return LogicValue.TRUE;
	    } else {
		// op1 is less than op2, but will not stay that way
		double p = (v2 - v1) / (r1 - r2);
		return (new WBF(p));
	    }

	} else if (r1 >= r2) {
	    // It is impossible for them to cross
	    return (LogicValue.FALSE);

	} else {
	    // Will cross, compute prediction
	    double p = (v1 - v2) / (r2 - r1);
	    return (new WBT(p));
	}
    }


    private LogicValue handleLessThanEqual(LogicValue op1, LogicValue op2) {
	double v1 = op1.getValue();
	double v2 = op2.getValue();
	double r1 = op1.getRate();
	double r2 = op2.getRate();

	if (v1 <= v2) {
	    if (r1 <= r2) {
		// op1 is less than op2 and will stay that way
		return LogicValue.TRUE;
	    } else {
		// op1 is less than op2, but will not stay that way
		double p = (v2 - v1) / (r1 - r2);
		return (new WBF(p));
	    }

	} else if (r1 >= r2) {
	    // It is impossible for them to cross
	    return (LogicValue.FALSE);

	} else {
	    // Will cross, compute prediction
	    double p = (v1 - v2) / (r2 - r1);
	    return (new WBT(p));
	}
    }


    private LogicValue handleGreaterThan(LogicValue op1, LogicValue op2) {
	double v1 = op1.getValue();
	double v2 = op2.getValue();
	double r1 = op1.getRate();
	double r2 = op2.getRate();

	if (v1 > v2) {
	    if (r1 >= r2) {
		// op1 is greater than op2 and will stay that way
		return LogicValue.TRUE;
	    } else {
		// op1 is greater than op2, but will not stay that way
		double p = (v1 - v2) / (r2 - r1);
		return (new WBF(p));
	    }

	} else if (r1 <= r2) {
	    // It is impossible for them to cross
	    return (LogicValue.FALSE);

	} else {
	    // Will cross, compute prediction
	    double p = (v2 - v1) / (r1 - r2);
	    return (new WBT(p));
	}
    }

   private LogicValue handleGreaterThanEqual(LogicValue op1, LogicValue op2) {
	double v1 = op1.getValue();
	double v2 = op2.getValue();
	double r1 = op1.getRate();
	double r2 = op2.getRate();

	if (v1 >= v2) {
	    if (r1 >= r2) {
		// op1 is greater than op2 and will stay that way
		return LogicValue.TRUE;
	    } else {
		// op1 is greater than op2, but will not stay that way
		double p = (v1 - v2) / (r2 - r1);
		return (new WBF(p));
	    }

	} else if (r1 <= r2) {
	    // It is impossible for them to cross
	    return (LogicValue.FALSE);

	} else {
	    // Will cross, compute prediction
	    double p = (v2 - v1) / (r1 - r2);
	    return (new WBT(p));
	}
    }


    private LogicValue handleEqual(LogicValue op1, LogicValue op2) {
	double v1 = op1.getValue();
	double v2 = op2.getValue();
	double r1 = op1.getRate();
	double r2 = op2.getRate();

	if ( (r1 == 0.0) && (r2 == 0.0) ) {
	    // Rates are both zero
	    if (v1 == v2) {
		return LogicValue.TRUE;
	    } else {
		return LogicValue.FALSE;
	    }
	}
	// For media time equality we need to check if either will
        // cross the other, and then give that time        
	LogicValue lv1 = handleGreaterThanEqual(op1, op2);
	LogicValue lv2 = handleLessThanEqual(op1, op2);

	if (lv1 instanceof WBT) {
	    return lv1;
	} else if (lv2 instanceof WBT) {
	    return lv2;
	} else {
	    return LogicValue.FALSE;
	}
    }  
}

