/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: LogicValue.java
*/

// Current package
package edu.uofminnesota.DEMAIS.runtime;

// Import needed packages
import java.lang.*;

/*
  For details of the predictive logic, see
  Bailey, B., et al. Nsync - A Toolkit for Building Interactive Multimedia Presentations. Proceedings ACM Multimedia 1998.
*/

public abstract class LogicValue {

    // True value
    public static final LogicValue TRUE;
    // False value
    public static final LogicValue FALSE;

    // A value of -1 represents infinity
    public static final int INFINITY;



    // The current logical time of a continuous media object
    protected double value;
    // The current playback rate of a continuous media object
    protected double rate;
    // The predicted time at which a condition will switch logic value
    protected double prediction;
    // All of the value are in seconds



    static {
	INFINITY = -1;
	TRUE = new WBF(INFINITY);
	FALSE = new WBT(INFINITY);
    }
    

    // Static functions
    public static LogicValue Constant(double c) {
	return (new LogicConstant(c));
    }

    public static LogicValue Value(boolean b) {
	if (b) {
	    return LogicValue.TRUE;
	} else {
	    return LogicValue.FALSE;
	}
    }



    // Constructors
    private LogicValue() {
	this(0.0, 0.0, LogicValue.INFINITY);
    }

    public LogicValue(double v, double r, double p) {
	value = v;
	rate = r;
	prediction = p;
    }


    // Public methods
    public void setValue(double v) {
	value = v;
    }

    public double getValue() {
	return value;
    }

    public void setRate(double r) {
	rate = r;
    }

    public double getRate() {
	return rate;
    }

    public void setPrediction(double p) {
	prediction = p;
    }

    public double getPrediction() {
	return prediction;
    }


    // Equality functions
    public boolean equals(LogicValue lv) {
	return ( (value == lv.value) && 
		 (rate == lv.rate) &&
		 (prediction == lv.prediction) );
    }

  
    // Force the subclasses to implement these methods
    public abstract LogicValue and(WBT wbt);
    public abstract LogicValue and(WBF wbf);
    public abstract LogicValue and(LogicConstant lc);
    public abstract LogicValue and(LogicCMedia cm);
    public abstract LogicValue and(LogicValue lv);

    public abstract LogicValue or(WBT wbt);
    public abstract LogicValue or(WBF wbf);
    public abstract LogicValue or(LogicConstant lc);
    public abstract LogicValue or(LogicCMedia cm);
    public abstract LogicValue or(LogicValue lv);

    public abstract LogicValue not();

    public abstract boolean isFalseNow();
    public abstract boolean isFalseAlways();
    public abstract boolean isTrueNow();
    public abstract boolean isTrueAlways();

}



