/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: Scheduler.java

  The Scheduler schedules a WBT(x) or WBF(x) action to be invoked at the predicted time x on behalf of the Evaluator. The Evaluator invokes an annotation's action, not the Scheduler. Only the Evaluator knows if an annotation's condition made a logic transition from false->true.

*/


// Current package
package edu.uofminnesota.DEMAIS.runtime;

// Import needed packages
import java.lang.*;
import java.util.*;
import javax.swing.Timer;
import java.awt.event.*;
import edu.uofminnesota.DEMAIS.core.annotation.*;



public class Scheduler implements ActionListener {

    // Implement a singleton design pattern
    private static Scheduler scheduler;


    // Sorted queue of scheduled actions
    private SchedulerQueue schedulerQ;

    // A timer which pops for the first action item in the queue
    private javax.swing.Timer timer;
 


    // Static initializer
    static {
	scheduler = null;
    }

    public static Scheduler getScheduler() {
	if (scheduler == null) {
	    scheduler = new Scheduler();
	}
	return scheduler;
    }



    // Private constructor to prevent additional instantiation
    private Scheduler() {
	schedulerQ = new SchedulerQueue(10);
	
	// Set up the timer
	timer = new Timer(10000, this);
	timer.setRepeats(false);
	timer.setCoalesce(true);
	timer.stop();
    }


    // Public methods 
    public void scheduleAction(LogicValue lv, SchedulerAction sa) {
	// Calculate the time at which the action is to be invoked
	long scheduledTime = Now() + (long) (Math.round(1000 * lv.getPrediction()));

	// Build a scheduler action object
	SchedulerActionItem scheduledAction = new SchedulerActionItem(scheduledTime, sa);

	// Insert the object into the scheduler queue based on the scheduled time
	if (schedulerQ.insertActionItem(scheduledTime, scheduledAction) == 0) {   
	    // If the new scheduledAction was inserted at the front of the queue, reset the timer to that one
	    setTimer();
	}
    }


    // Clear out the Scheduler - access to the Timer must be synchronized
    public void removeAll() {
	timer.stop();
	schedulerQ.removeAllElements();
    }


    // Clear actions for a particular Annotation - Depending on how the predictive logic is implemented
    // it may be possible for a single annotation to have more than one action scheduled.
    public void removeAnnotation(Annotation annotation) {
	Vector removeList = new Vector();
	
	Iterator it = schedulerQ.iterator();
	while (it.hasNext()) {   
	    SchedulerActionItem actionItem = (SchedulerActionItem) it.next();
	    SchedulerAction action = actionItem.getAction(); 
	    if (action.contains(annotation)) {
		removeList.add(actionItem);
	    }
	}

	// Remove all of these actions from the scheduler's list
	schedulerQ.removeAll(removeList);
    }


    // Set the Timer for the scheduled time
    protected void setTimer() {
	timer.stop();
	
	SchedulerActionItem item = null;
	if (schedulerQ.size() > 0) {
	    item = (SchedulerActionItem) schedulerQ.elementAt(0);
	    int delay = (int) (item.getTime() - Now());
	    if (delay < 0) {
		//System.out.println("Delay is: " + Integer.toString(delay));
		delay = 1;
	    }
	    timer.setInitialDelay(delay);
	    timer.setDelay(delay);
	    timer.start();
	}
    }
    

    // Return the current number of milliseconds since Jan. 1, 1970
    private long Now() {
	return System.currentTimeMillis();
    }


    // Callback for a timer pop
    public void actionPerformed(ActionEvent event) {

	// Stop the current timer
	timer.stop();

	/* 
	   Execute all actions whose scheduled times are <= Now()
	*/
	long target = Now();	
	boolean finished = false;
	while ((! finished) && (schedulerQ.size() > 0)) {

	    // Retrieve the head element and its scheduled time 
	    SchedulerActionItem actionItem = (SchedulerActionItem) schedulerQ.elementAt(0);
	    if (actionItem.getTime() <= target) {
		schedulerQ.removeElementAt(0);
		/*
		  The action requests that the Evaluator perform a re-evaluation of the annotation. If the Evaluator detects a logic transition from false to true, then it will invoke the annotations's action.
		*/ 
		SchedulerAction action = actionItem.getAction();
		try {
		    action.performAction();
		} catch (EvaluationException xcptn) {
		    System.out.println("Scheduler:actionPerformed() - " + xcptn.toString());
		}

	    } else {
		finished = true;
	    }
	}

	// Set the scheduler's timer for the first action on the possibly modified list
	setTimer();
    }

}

