/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: MultiViewMenu.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.base;

// Necessary imports
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;




public class MultiViewMenu extends JMenu {

    // The listener for all menu items (an inner class below)
    private Listener listener;

    // The permanent menu items
    private JMenuItem mvNew, mvDelete, mvProperties;

    // Maps a multiview model to a JCheckBoxMenuItem
    private Hashtable modelToMenu;
    private Hashtable menuToModel;



    public MultiViewMenu() {
	super("MultiView");

	listener = new Listener();

	mvNew = new JMenuItem("New");
	mvNew.setEnabled(true);
	mvNew.addActionListener(listener);
	add(mvNew);
	
	mvDelete = new JMenuItem("Delete");
	mvDelete.setEnabled(false);
	mvDelete.addActionListener(listener);
	add(mvDelete);

	addSeparator();

	mvProperties = new JMenuItem("Properties");
	mvProperties.setEnabled(false);
	mvProperties.addActionListener(listener);
	add(mvProperties);

	addSeparator();

	// Create the menu table and the reverse mapping
	modelToMenu = new Hashtable(5);
	menuToModel = new Hashtable(5);
	
	synchronizeMenuItems();
    }


    public void addMenuItem(MultiViewModel model) {
	// Default to selected
	addMenuItem(model, true);
    }

    public void addMenuItem(MultiViewModel model, boolean selected) {
	if (model != null) {
	    JCheckBoxMenuItem menu = new JCheckBoxMenuItem(model.getName(), selected);
	    menu.addActionListener(listener);
	    add(menu);
	    modelToMenu.put(model, menu);
	    menuToModel.put(menu, model);
	    model.addNameListener(listener);
	}
    }

    public void removeMenuItem(MultiViewModel model) {
	JCheckBoxMenuItem menu = (JCheckBoxMenuItem) modelToMenu.get(model);
	remove(menu);
	modelToMenu.remove(model);
	menuToModel.remove(menu);
	model.removeNameListener(listener);
    }


    // Keeps the menu items synchronized with what's in the
    // content manager view
    public void synchronizeMenuItems() {
	removeAll();

	add(mvNew);
	add(mvDelete);
	addSeparator();
	add(mvProperties);
	addSeparator();

	modelToMenu.clear();
	menuToModel.clear();

	Collection list = MultiViewModel.getMultiViewModels();
	DataModel cModel = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	Iterator it = list.iterator();
	while (it.hasNext()) {
	    MultiViewModel model = (MultiViewModel) it.next();
	    boolean selected = false;
	    if ((cModel != null) && (cModel.equals(model))) {
		selected = true;
	    }
	    addMenuItem((MultiViewModel)model, selected);
	}
    }
 

    public void selectOnlyItem(JCheckBoxMenuItem item) {
	for (Enumeration e = modelToMenu.elements(); e.hasMoreElements(); ) {
	    JCheckBoxMenuItem menu = (JCheckBoxMenuItem) e.nextElement();
		menu.setSelected(menu == item);
	}
    }

    public void selectOnlyItem(MultiViewModel model) {
	JCheckBoxMenuItem menuItem = (JCheckBoxMenuItem) modelToMenu.get(model);
	if (menuItem != null) {
	    selectOnlyItem(menuItem);
	} else {
	    addMenuItem(model);
	    selectOnlyItem(model);
	}
    }



    // An inner class to handle all menu item events
    class Listener implements ActionListener, NamedModelNameListenerInterface, PropertyListenerInterface {

	// Constructor
	public Listener() {
	    PropertyManager.addPropertyListener(PropertyManager.APPLICATION_CURRENT_DATA_MODEL, this);
	    PropertyManager.addPropertyListener(PropertyManager.APPLICATION_MULTIVIEW_COUNT, this);
	    PropertyManager.addPropertyListener(PropertyManager.APPLICATION_MODE, this);
	}


	// Menu item callback
	public void actionPerformed(ActionEvent e) {
	    Object selection = e.getSource();

	    if (selection == mvNew) {
		handleMultiViewNew();
		
	    } else if (selection == mvDelete) {
		handleMultiViewDelete();

	    } else if (selection == mvProperties) {
		handleMultiViewProperties();

	    } else {

		// Look at the multiview menu items
		JCheckBoxMenuItem menuSelection = null;
		for (Enumeration items = modelToMenu.elements(); items.hasMoreElements(); ) {
		    JCheckBoxMenuItem menu = (JCheckBoxMenuItem) items.nextElement();
		    if (selection == menu) {
			menuSelection = menu;
			MultiViewModel model = (MultiViewModel) menuToModel.get(menu);
			handleGoToMultiView(model);
		    }
		}
		selectOnlyItem(menuSelection);
	    }
	}
	

	// PropertyListener interface
	public void propertyChanged(Integer property, Object oldValue, Object newValue) {
	   
	    if (property.equals(PropertyManager.APPLICATION_MULTIVIEW_COUNT)) {
		int count = ((Integer) newValue).intValue();
		mvDelete.setEnabled(count > 0);
		mvProperties.setEnabled(count > 0);
		synchronizeMenuItems();
	    
	    } else if (property.equals(PropertyManager.APPLICATION_CURRENT_DATA_MODEL)) {
		DataModel model = (DataModel) newValue;
		if ( (model != null) && (model instanceof MultiViewModel) ) {
		    selectOnlyItem((MultiViewModel)model);
		}

	    } else if (property.equals(PropertyManager.APPLICATION_MODE)) {
		int mode = ((Integer) newValue).intValue();
		switch(mode) {

		case PropertyManager.APPLICATION_EDIT:
		    mvNew.setEnabled(true);

		    int count = ((Integer) PropertyManager.getProperty(PropertyManager.APPLICATION_MULTIVIEW_COUNT)).intValue();
		    mvDelete.setEnabled(count > 0);
		    mvProperties.setEnabled(count > 0);

		    for (Enumeration e = menuToModel.keys(); e.hasMoreElements(); ) {
			JCheckBoxMenuItem menuItem = (JCheckBoxMenuItem) e.nextElement();
			menuItem.setEnabled(true);
		    }

		    break;

		case PropertyManager.APPLICATION_PLAY:
		    mvNew.setEnabled(false);
		    mvDelete.setEnabled(false);
		    mvProperties.setEnabled(false);

		    for (Enumeration e = menuToModel.keys(); e.hasMoreElements(); ) {
			JCheckBoxMenuItem menuItem = (JCheckBoxMenuItem) e.nextElement();
			menuItem.setEnabled(false);
		    }

		    break;

		default:
		    break;
		}
	    } 
	}

	// NameListenerInterface
	public void nameChanged(NamedModel obj, String oldName, String newName) {
	    MultiViewModel model = (MultiViewModel) obj;
	    JCheckBoxMenuItem menuItem = (JCheckBoxMenuItem) modelToMenu.get(model);
	    if (menuItem != null) {
		menuItem.setText(newName);
	    }
	}


	private void handleMultiViewNew() {
	    MultiViewModel model = new MultiViewModel(PropertyManager.APPLICATION_SCOPE, null, null);
	    handleGoToMultiView(model);
	}

	private void handleMultiViewDelete() {
	    DataModel dataModel = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	    if (dataModel != null) {
		// Deleting the model will reset the current model
		dataModel.deleteModel();
	    }
	}
	
	private void handleMultiViewProperties() {
	    DataModel model = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	    if (model != null) {
		ContentPropertyFrame frame = new ContentPropertyFrame("MultiView Properties");
	    }
	}

	private void handleGoToMultiView(MultiViewModel model) {
	    PropertyManager.setProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL, model);
	}
    }
}



