/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: StoryboardMenu.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.base;

// Necessary imports
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;




public class StoryboardMenu extends JMenu {

    // The listener for all menu items (an inner class below)
    private Listener listener;

    // The permanent menu items
    private JMenuItem sNew, sCopy, sDelete, sProperties;

    // Maps a storyboard model to a JCheckBoxMenuItem
    private Hashtable modelToMenu;
    private Hashtable menuToModel;



    public StoryboardMenu() {
	super("Storyboard");

	listener = new Listener();

	sNew = new JMenuItem("New");
	sNew.setEnabled(true);
	sNew.addActionListener(listener);
	add(sNew);
	
	sCopy = new JMenuItem("Copy");
	sCopy.setEnabled(true);
	sCopy.addActionListener(listener);
	add(sCopy);
	
	sDelete = new JMenuItem("Delete");
	sDelete.setEnabled(false);
	sDelete.addActionListener(listener);
	add(sDelete);
	
	addSeparator();

	sProperties = new JMenuItem("Properties");
	sProperties.setEnabled(true);
	sProperties.addActionListener(listener);
	add(sProperties);

	addSeparator();

	// Create the menu table and the reverse mapping
	modelToMenu = new Hashtable(5);
	menuToModel = new Hashtable(5);

	synchronizeMenuItems();
    }


    public void addMenuItem(StoryboardModel model) {
	// Default to selected
	addMenuItem(model, true);
    }

    public void addMenuItem(StoryboardModel model, boolean selected) {
	if (model != null) {
	    JCheckBoxMenuItem menu = new JCheckBoxMenuItem(model.getName(), selected);
	    menu.addActionListener(listener);
	    add(menu);
	    modelToMenu.put(model, menu);
	    menuToModel.put(menu, model);
	    model.addNameListener(listener);
	}
    }


    public void removeMenuItem(StoryboardModel model) {
	JCheckBoxMenuItem menu = (JCheckBoxMenuItem) modelToMenu.get(model);
	remove(menu);
	modelToMenu.remove(model);
	menuToModel.remove(menu);
	model.removeNameListener(listener);
    }


    // Keeps the menu items synchronized with what's in the
    // content manager view
    public void synchronizeMenuItems() {
	removeAll();

	add(sNew);
	add(sCopy);
	add(sDelete);
	addSeparator();
	add(sProperties);
	addSeparator();

	modelToMenu.clear();
	menuToModel.clear();

	Collection list = StoryboardModel.getStoryboardModels();
	DataModel cModel = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	Iterator it = list.iterator();
	while (it.hasNext()) {
	    StoryboardModel model = (StoryboardModel) it.next();
	    boolean selected = false;
	    if ((cModel != null) && (cModel.equals(model))) {
		selected = true;
	    }
	    addMenuItem((StoryboardModel)model, selected);
	}
    }

    public void selectOnlyItem(JCheckBoxMenuItem item) {
	for (Enumeration e = modelToMenu.elements(); e.hasMoreElements(); ) {
	    JCheckBoxMenuItem menu = (JCheckBoxMenuItem) e.nextElement();
	    menu.setSelected(menu == item);
	}
    }

    public void selectOnlyItem(StoryboardModel model) {
	JCheckBoxMenuItem menuItem = (JCheckBoxMenuItem) modelToMenu.get(model);
	if (menuItem != null) {
	    selectOnlyItem(menuItem);
	} else {
	    addMenuItem(model);
	    selectOnlyItem(model);
	}
    }



    // An inner class to handle all menu item events
    class Listener implements ActionListener, NamedModelNameListenerInterface, PropertyListenerInterface {

	// Constructor
	public Listener() {
	    PropertyManager.addPropertyListener(PropertyManager.APPLICATION_CURRENT_DATA_MODEL, this);
	    PropertyManager.addPropertyListener(PropertyManager.APPLICATION_STORYBOARD_COUNT, this);
	    PropertyManager.addPropertyListener(PropertyManager.APPLICATION_MODE, this);
	}


	// Menu item callback
	public void actionPerformed(ActionEvent e) {
	    Object selection = e.getSource();

	    if (selection == sNew) {
		handleStoryboardNew();
		
	    } else if (selection == sCopy) {
		handleStoryboardCopy();
		
	    } else if (selection == sDelete) {
		handleStoryboardDelete();

	    } else if (selection == sProperties) {
		handleStoryboardProperties();

	    } else {

		// Look at the storyboard menu items
		JCheckBoxMenuItem menuSelection = null;
		for (Enumeration items = modelToMenu.elements(); items.hasMoreElements(); ) {
		    JCheckBoxMenuItem menu = (JCheckBoxMenuItem) items.nextElement();
		    if (selection == menu) {
			menuSelection = menu;
			StoryboardModel model = (StoryboardModel) menuToModel.get(menu);
			handleGoToStoryboard(model);
		    }
		}
		selectOnlyItem(menuSelection);
	    }
	}
	

	// PropertyListener interface
	public void propertyChanged(Integer property, Object oldValue, Object newValue) {
	    
	    if (property.equals(PropertyManager.APPLICATION_STORYBOARD_COUNT)) {
		int count = ((Integer) newValue).intValue();
		sDelete.setEnabled(count > 0);
		sCopy.setEnabled(count > 0);
		sProperties.setEnabled(count > 0);
		synchronizeMenuItems();

	    } else if (property.equals(PropertyManager.APPLICATION_CURRENT_DATA_MODEL)) {
		DataModel model = (DataModel) newValue;
		if ( (model != null) && (model instanceof StoryboardModel) ) {
		    selectOnlyItem((StoryboardModel)model);
		}

	    } else if (property.equals(PropertyManager.APPLICATION_MODE)) {
		int mode = ((Integer) newValue).intValue();
		switch (mode) {

		case PropertyManager.APPLICATION_EDIT:
		    sNew.setEnabled(true);

		    int count = ((Integer)PropertyManager.getProperty(PropertyManager.APPLICATION_STORYBOARD_COUNT)).intValue();
		    sCopy.setEnabled(count > 0);
		    sDelete.setEnabled(count > 0);
		    sProperties.setEnabled(count > 0);

		    for (Enumeration e = menuToModel.keys(); e.hasMoreElements(); ) {
			JCheckBoxMenuItem menuItem = (JCheckBoxMenuItem) e.nextElement();
			menuItem.setEnabled(true);
		    }

		    break;

		case PropertyManager.APPLICATION_PLAY:
		    sNew.setEnabled(false);
		    sCopy.setEnabled(false);
		    sDelete.setEnabled(false);
		    sProperties.setEnabled(false);

		    for (Enumeration e = menuToModel.keys(); e.hasMoreElements(); ) {
			JCheckBoxMenuItem menuItem = (JCheckBoxMenuItem) e.nextElement();
			menuItem.setEnabled(false);
		    }

		    break;

		default:
		    break;
		}
	    }
	}

	// NameListenerInterface
	public void nameChanged(NamedModel obj, String oldName, String newName) {
	    StoryboardModel model = (StoryboardModel) obj;
	    JCheckBoxMenuItem menuItem = (JCheckBoxMenuItem) modelToMenu.get(model);
	    if (menuItem != null) {
		menuItem.setText(newName);
	    }
	}


	private void handleStoryboardNew() {
	    StoryboardModel model = new StoryboardModel(PropertyManager.APPLICATION_SCOPE, null, null);
	    handleGoToStoryboard(model);
	}


	private void handleStoryboardCopy() {
	    DataModel model = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	    if (model != null) {
		StoryboardModel newModel = (StoryboardModel) model.clone();
		handleGoToStoryboard(newModel);
	    }
	}

	private void handleStoryboardDelete() {
	    DataModel dataModel = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	    if (dataModel != null) {		
		// Deleting the model will reset the current storyboard model
		dataModel.deleteModel();
	    }
	}

	private void handleStoryboardProperties() {
	    DataModel dataModel = (DataModel) PropertyManager.getProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL);
	    if (dataModel != null) {		
		ContentPropertyFrame frame = new ContentPropertyFrame("Storyboard Properties");
	    }
	}


	private void handleGoToStoryboard(StoryboardModel model) {
	    PropertyManager.setProperty(PropertyManager.APPLICATION_CURRENT_DATA_MODEL, model);
	}
    }
}

