/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: ContentManagerCellView.java
 
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.contentmanager;

// Necessary imports
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;




public abstract class ContentManagerCellView 
    extends JPanel 
    implements DataModelViewInterface {
    

    // Instance variables
    protected JPanel titlePanel, contentPanel;
    protected JLabel titleLabel;

    private DataModel model;
    private boolean selected;
    private Listener listener;
    private ContentManagerCell parentCell;



    protected ContentManagerCellView() {
	// Empty
    }

    public ContentManagerCellView(DataModel m) {
	super(new GridBagLayout());
	model = m;
	selected = false;
	setBackground(Color.white);

	listener = new Listener();
	model.addNameListener(listener);

	GridBagConstraints c = new GridBagConstraints();
	c.gridx = 0;
	c.gridy = 0;
	c.fill = GridBagConstraints.BOTH;
	c.weightx = 1.0;
	c.weighty = 1.0;
	add(createContentPanel(), c);
	
	c.gridy = 1;
	c.weighty = 0.0;
	add(createTitlePanel(), c);

	// Set the appropripate colors
	unSelectCell();
	parentCell = null;
    }



    // Derived classes must override these methods!
    public abstract void handleOpenCell();



    public void deleteView(DataModel model, boolean remove) {
	parentCell.deleteView();
	setCellSelected(false);
	model.removeNameListener(listener);
    }

    public void refreshView(DataModel model) {
	repaint();
    }


    public DataModel getModel() {
	return model;
    }
 

    public void setParentCell(ContentManagerCell cell) {
	parentCell = cell;
    }

    public boolean isCellSelected() {
	return selected;
    }

 
    public void setCellSelected(boolean value) {
	if (selected == value) {
	    // No sense in setting it to the same value
	    // Also necessary for the counts to remain correct
	    return;
	}
	
	selected = value;
	if (selected) {
	    selectCell();
	    incrementSelectCount();
	} else {
	    unSelectCell();
	    decrementSelectCount();
	}
    }
    

    protected void selectCell() {
	// Set the highlight to this object
	contentPanel.setBorder(BorderFactory.createLineBorder(PropertyManager.CONTENT_MANAGER_SELECT_COLOR, 2));
	titlePanel.setBackground(PropertyManager.CONTENT_MANAGER_SELECT_COLOR);
	titleLabel.setBackground(PropertyManager.CONTENT_MANAGER_SELECT_COLOR);
    }

    protected void unSelectCell() {
	// Unhighlight this object
	contentPanel.setBorder(BorderFactory.createLineBorder(Color.black, 1));
	titlePanel.setBackground(Color.white);
	titleLabel.setBackground(Color.white);
    }

    protected void incrementSelectCount() {
	int count = ( (Integer)PropertyManager.getProperty(PropertyManager.CONTENT_MANAGER_OBJECT_SELECTED_COUNT)).intValue();
	PropertyManager.setProperty(PropertyManager.CONTENT_MANAGER_OBJECT_SELECTED_COUNT, new Integer(count + 1));
    }
    
    protected void decrementSelectCount() {
	int count = ( (Integer)PropertyManager.getProperty(PropertyManager.CONTENT_MANAGER_OBJECT_SELECTED_COUNT)).intValue();
	PropertyManager.setProperty(PropertyManager.CONTENT_MANAGER_OBJECT_SELECTED_COUNT, new Integer(count - 1));
    }
    
    protected void selectOnlyMe() {
	ContentManagerFrame.getFrame().getPane().unSelectAll();
	setCellSelected(true);
    }
    
    
    protected void handleMouseClicked(MouseEvent e) {
	if ((SwingUtilities.isRightMouseButton(e)) || (e.isMetaDown())) {
	    // right mouse button
	    selectOnlyMe();
	    ContentManagerPopupMenu.getPopupMenu().show(e.getComponent(), e.getX(), e.getY());
	
	} else {
	    // left mouse button
	    if (e.getClickCount() == 1) {
		// left single click
		if (! e.isShiftDown()) {
		    if (PropertyManager.APPLICATION_INPUT_DEVICE == PropertyManager.STYLUS_INPUT_DEVICE) {
			// With a stylus, a single click opens the selected cell
			handleOpenCell();
			
		    } else {
			// With a mouse, a single clicks selects a different item
			selectOnlyMe();
		    }

		} else {
		    // Single click, shift is down
		    setCellSelected(!isCellSelected());		    
		}
		
	    } else if (e.getClickCount() == 2) {
		// left double click
		handleOpenCell();
	    }
	}
    }


    private JPanel createTitlePanel() {
	titlePanel = new JPanel(new GridBagLayout());
	titlePanel.setBorder(BorderFactory.createLoweredBevelBorder());
	titlePanel.addMouseListener(listener);
	
	titleLabel = new JLabel(model.getName(), JLabel.CENTER);
	titleLabel.addMouseListener(listener);
	titleLabel.setFont(PropertyManager.CONTENT_MANAGER_TITLE_FONT);

	GridBagConstraints c = new GridBagConstraints();	
	c.gridx = 0;
	c.gridy = 0;
	c.fill = GridBagConstraints.HORIZONTAL;
	c.weightx = 1.0;
	titlePanel.add(titleLabel, c);

	return titlePanel;
    }
	

    private JPanel createContentPanel() {
	contentPanel = new JPanel(new GridBagLayout());
	contentPanel.setBorder(BorderFactory.createLineBorder(Color.black, 1));
	contentPanel.setBackground(Color.white);
	contentPanel.addMouseListener(listener);

	return contentPanel;
    }


    class Listener extends MouseAdapter implements NamedModelNameListenerInterface {

	public Listener() {
	}

	// The model's name changed
	public void nameChanged(NamedModel obj, String oldName, String newName) {
	    titleLabel.setText(newName);
	}
	
	public void mouseClicked(MouseEvent e) {
	    handleMouseClicked(e);
	}

    }
}
