/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: ContentManagerFrame.java

  Implemented as a singleton pattern
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.contentmanager;

// Necessary imports
import java.util.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.border.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;




public class ContentManagerFrame extends JFrame {

    // Reference to the singleton frame instance
    private static ContentManagerFrame cmFrame;

    // Reference to the internal pane
    private ContentManagerPane cmPane;

    // Property listener
    private Listener listener;




    // Static initializer
    static {
	cmFrame = null;
    }

    public static ContentManagerFrame getFrame() {
	if (cmFrame == null) {
	    cmFrame = new ContentManagerFrame();

	    // Force the popup menu to load as well
	    ContentManagerPopupMenu.getPopupMenu();
	}
	return cmFrame;
    }



    public ContentManagerFrame() {
	super("Content Manager");

	setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);

	listener = new Listener();

	addWindowListener(listener);

	setLookAndFeel();
	createMenus();
	setupContentPane();
	
	setResizable(false);
	pack();
	setRootLocation();
	setVisible(false);
    }


    public void makeVisible(boolean visible) {
	if (visible) {
	    setState(Frame.NORMAL);
	    setVisible(true);
	    PropertyManager.setProperty(PropertyManager.CONTENT_MANAGER_VISIBLE, Boolean.TRUE);

	} else {
	    setState(Frame.ICONIFIED);
	    PropertyManager.setProperty(PropertyManager.CONTENT_MANAGER_VISIBLE, Boolean.FALSE);

	}
	setCursor(PropertyManager.APPLICATION_CURSOR);
    }



    public void closeWindow() {
	makeVisible(false);
    }


    public void addDataModel(DataModel model) {
	cmPane.addDataModel(model);
    }

    public void addDataModel(DataModel model, int position) {
	cmPane.addDataModel(model, position);
    }


    public Collection getSelectedCells() {
	return cmPane.getSelectedCells();
    }

    public void deleteSelectedCells() {
	cmPane.deleteSelectedCells();
    }

    public ContentManagerCellView getFirstSelectedCellView() {
	return cmPane.getFirstSelectedCellView();
    }


    // Private methods
    private void setLookAndFeel() {
	try {
	    UIManager.setLookAndFeel(DEMAISFrame.getLookAndFeelClassName());
	
	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "Content Manager", false);
	    DEMAISError.logException(xcptn);
	}
    }


    private void createMenus() {
	this.setJMenuBar(new ContentManagerFrameMenuBar(this));
    }



    protected ContentManagerPane getPane() {
	return cmPane;
    }



    private void setupContentPane() {
	JPanel contentPane = new JPanel(new GridBagLayout());
	contentPane.setBorder(BorderFactory.createLoweredBevelBorder());

	GridBagConstraints c = new GridBagConstraints();

	int cmWidth = PropertyManager.CONTENT_MANAGER_WIDTH;
	int cmHeight = PropertyManager.CONTENT_MANAGER_HEIGHT;
	contentPane.setPreferredSize(new Dimension(cmWidth, cmHeight));


	this.setContentPane(contentPane);

	// Add the main content manager pane
	c.gridx = 0;
	c.gridy = 0;
	c.gridwidth = GridBagConstraints.REMAINDER;
	c.gridheight = GridBagConstraints.REMAINDER;
	c.fill = GridBagConstraints.BOTH;
	c.weightx = 1.0;
	c.weighty = 1.0;

	cmPane = new ContentManagerPane();
	contentPane.add(cmPane, c);
    }


    private void setRootLocation() {
	Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
	Dimension me = this.getSize();

	Point p = new Point( (int)(screen.getWidth() - me.getWidth()) / 2, 
			     (int)(screen.getHeight() - me.getHeight()) / 2);
	setLocation(p);
    }



    class Listener extends WindowAdapter implements PropertyListenerInterface {
	public Listener() {
	    PropertyManager.addPropertyListener(PropertyManager.APPLICATION_MODE, this);
	}


	// Do not destroy the window, just iconify it (due to a weird bug in the JVM)
	public void windowClosing(WindowEvent e) {
	    makeVisible(false);
	}
	
	public void windowIconified(WindowEvent e) {
	    PropertyManager.setProperty(PropertyManager.CONTENT_MANAGER_VISIBLE, Boolean.FALSE);
	}

	public void propertyChanged(Integer property, Object oldValue, Object newValue) {
	    if (property.equals(PropertyManager.APPLICATION_MODE)) {
		int mode = ((Integer) newValue).intValue();
		if (mode == PropertyManager.APPLICATION_PLAY) {
		    makeVisible(false);
		}
	    }
	}
    }

}




