/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File MultiViewPanel.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.multiview;
    

// Necessary imports
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import java.io.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;
import edu.uofminnesota.DEMAIS.usrinterface.visuallanguage.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture.commands.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.stroke.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.behavior.*;
import edu.uofminnesota.DEMAIS.usrinterface.util.*;





// Storyboard class
public class MultiViewPanel extends Storyboard 
    implements DataModelViewInterface {


    // Number of compartment items
    private static final int NUM_COMPARTMENTS = 6;

    private MultiViewCompartment compartments[];




    private MultiViewPanel() {
	this((MultiViewModel)null);
    }

    public MultiViewPanel(MultiViewModel mvm) {
	super(mvm);

	compartments = new MultiViewCompartment[NUM_COMPARTMENTS];
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    compartments[i] = null;  
	}

	MultiViewDropTargetListener mvDropTargetListner = new MultiViewDropTargetListener(this);
    }


    public MultiViewPanel(MultiViewPanel copy) {
	this(copy, (MultiViewModel)copy.model);
    }

    public MultiViewPanel(MultiViewPanel copy, MultiViewModel toModel) {
	super(copy, toModel, false);

	compartments = new MultiViewCompartment[NUM_COMPARTMENTS];
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    compartments[i] = null;
	}
    }


    public MultiViewPanel(DataInputStream stream, MultiViewModel toModel) throws IOException {
	super(stream, toModel);

	compartments = new MultiViewCompartment[NUM_COMPARTMENTS];
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    compartments[i] = null;  
	
	    if (stream.readBoolean()) {
		// Compartment contains a data model
		// Read the class id to determine which one
		
		int classid = stream.readInt();
		switch (classid) {
		case MultiViewStoryboardCompartment.CLASSID:
		    compartments[i] = new MultiViewStoryboardCompartment(stream, this);
		    break;
		case MultiViewNarrationCompartment.CLASSID:
		    compartments[i] = new MultiViewNarrationCompartment(stream, this);
		    break;
		default:
		    break;
		}
		
	    }
	}
	MultiViewDropTargetListener mvDropTargetListner = new MultiViewDropTargetListener(this);
    }


    public void completeIO() {
	super.completeIO();
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    if (compartments[i] != null) {
		compartments[i].completeIO();
		DataModel dmodel = compartments[i].getModel();
		if (dmodel != null) {
		    dmodel.addView(this);
		}
	    }
	}
    }




    public void handleStoryboardModelDrop(StoryboardModel sModel, int compartment) {
	Rectangle r = mapCompartmentToRectangle(compartment);
	compartments[compartment] = new MultiViewStoryboardCompartment(r, this, sModel);
	sModel.addView(this);
	repaint();
	PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
    }


    public void handleNarrationModelDrop(NarrationModel nModel, int compartment) {
	Rectangle r = mapCompartmentToRectangle(compartment);
	compartments[compartment] = new MultiViewNarrationCompartment(r, this, nModel);
	nModel.addView(this);
	repaint();
	PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
    }




    // Multiview has only one layer!
    public int getCurrentLayer() {
	return 0;
    }


    public MultiViewCompartment getCompartment(int i) {
	return compartments[i];
    }

    public MultiViewCompartment[] getCompartments() {
	return compartments;
    }


    public boolean isCompartmentAvailable() {
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    if (compartments[i] == null) {
		return true;
	    }
	}
	return false;
    }


    public boolean isCompartmentAvailable(int i) {
	if ( (i < 0) || (i >= NUM_COMPARTMENTS) || (compartments[i] != null) ) {
	    return false;
	}
	return true;
    }


    public boolean isModelPresent(DataModel model) {
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    if ( (compartments[i] != null) &&
		 (model.equals(compartments[i].getModel()))) {
		return true;
	    }
	}
	return false;
    }


    public int getCompartmentForModel(DataModel model) {
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    if ( (compartments[i] != null) &&
		 (model.equals(compartments[i].getModel()))) {
		return i;
	    }
	}
	return -1;
    }


    public DataModel getModelInCompartment(int i) {
	DataModel model = null;
	if (compartments[i] != null) {
	    model = compartments[i].getModel();
	}
	return model;
    }
    

    public Rectangle mapCompartmentToRectangle(int compartment) {
	int x = 0;
	int y = 0;	
	int w = Storyboard.WIDTH / 3;
	int h = Storyboard.HEIGHT / 2;

	if (compartment == 0) {
	    x = 0;
	    y = 0;
	} else if (compartment == 1) {
	    x = w;
	    y = 0;
	} else if (compartment == 2) {
	    x = 2 * w;
	    y = 0;
	} else if (compartment == 3) {
	    x = 0;
	    y = h;
	} else if (compartment == 4) {
	    x = w;
	    y = h;
	} else if (compartment == 5) {
	    x = 2 * w;
	    y = h;
	}

	return (new Rectangle(x, y, w, h));
    }


    public int mapXYToCompartment(int x, int y) {
	int c = 0;
	int w = Storyboard.WIDTH;
	int h = Storyboard.HEIGHT;
	int horizontal = h / 2;

	if (y <= horizontal) {
	    if (x < (w / 3)) {
		c = 0;
	    } else if (x < ((2*w) / 3)) {
		c = 1;
	    } else {
		c = 2;
	    }

	} else {
	    if (x < (w / 3)) {
		c = 3;
	    } else if (x < ((2*w) / 3)) {
		c = 4;
	    } else {
		c = 5;
	    }
	}
	return c;
    }
    
    
    public void onInstall() {
	// Install the compartments
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    if (compartments[i] != null) {
		compartments[i].onInstall();
	    }
	}
    }


   public void onUnInstall() {
       unSelectAllObjects(false);
       
       // UnInstall the compartments
       for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	   if (compartments[i] != null) {
	       compartments[i].onUnInstall();
	   }
       }
   }


    
    public boolean contains (JComponent c) {
	Component components[] = getComponents();
	for (int i=0; i<components.length; ++i) {
	    if (components[i] == c) {
		return true;
	    }
	}
	return false;
    }

    

    // Override storybord paste functions
    public Collection pasteGraphicalObjects(Collection c) {
	return c;
    }
    
    public Collection pasteGraphicalObjects(Collection c, int deltaX, int deltaY, Storyboard parentS) {
	return c;
    }
    


    // Delete the entire view
    public void deleteMultiView() {
	super.deleteStoryboard();

	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    if (compartments[i] != null) {
		DataModel dmodel = compartments[i].getModel();
		if (dmodel != null) {
		    model.removeView(this);
		}
	    }
	}
    }


    public void deleteView(DataModel model, boolean remove) {
	int compartment = getCompartmentForModel(model);
	if (compartment >= 0) {
	    deleteView(compartment, remove);
	}
	repaint();
    }

    
    public void deleteView(int compartment, boolean remove) {
	// Determine which cell this data model is in
	DataModel removeModel = compartments[compartment].getModel();
	if (removeModel != null) {
	    if (remove) {
		removeModel.removeView(this);
	    }

	    // Delete the compartment's collection
	    compartments[compartment].deleteView();

	    // Remove the compartment object
	    compartments[compartment] = null;

	    Vector toRemove = new Vector();
	    // Remove all ink strokes that begin/end in that compartment
	    Iterator it = graphicalObjects.iterator();
	    while (it.hasNext()) {
		InkStroke inkStroke = (InkStroke) it.next();
		Point p = inkStroke.getFirstPoint();
		int c = mapXYToCompartment(p.x, p.y);
		if (c == compartment) {
		    // First point is in deleted compartment
		    toRemove.add(inkStroke);
		} else {
		    // If not, check last point
		    p = inkStroke.getLastPoint();
		    c = mapXYToCompartment(p.x, p.y);
		    if (c == compartment) {
			toRemove.add(inkStroke);
		    }
		}
	    }
	    graphicalObjects.removeAll(toRemove);
	}
    }




    public void refreshView(DataModel model) {
	int compartment = getCompartmentForModel(model);
	if (compartment >= 0) {
	    compartments[compartment].refreshView();

	    Graphics2D graphics = (Graphics2D) this.getGraphics();
	    if (graphics != null) {
		Rectangle r = compartments[compartment].getBounds();
		graphics.setColor(PropertyManager.STORYBOARD_BACKGROUND_COLOR);
		graphics.fillRect(r.x + 1, r.y + 1, r.width - 1, r.height - 1);
		compartments[compartment].render(graphics);
	    } else {
		repaint();
	    }
	}
    }



    
    // Paint this component
   public void paintComponent() {
	Graphics2D g = (Graphics2D) getGraphics();
	if (g != null) {
	    paintComponent(g);
	} else {
	    repaint();
	}
    }

    public void paintStoryboardComponent(Graphics2D g) {
	// Paint the separators
	int width = Storyboard.WIDTH;
	int height = Storyboard.HEIGHT;
	if ( (scaleFactorX < 1.0) || (scaleFactorY < 1.0)) {
	    width = getWidth();
	    height = getHeight();
	}
	
	g.setColor(Color.black);
	g.setStroke(GraphicalObject.LIGHT_STROKE);
	g.drawLine(width / 3, 0, width / 3, height);
	g.drawLine((2*width) / 3, 0, (2*width) / 3, height);
	g.drawLine(0, height / 2, width, height / 2);
	
	// Paint the compartments
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    if (compartments[i] != null) {
		compartments[i].render(g);
	    }
	}

	// Paint the objects
	graphicalObjects.render(g);
    }
    

    public void paintChildren(Graphics g) {
	super.superPaintChildren(g);
    }

    
    public void onMouseReleased(MouseEvent evt, TimedStroke ts) {
	if (mouseDragged) {
	    mouseDragged = false;
	    return;
	}

	//Check for a pen tap
	if (ts.npoints < PropertyManager.APPLICATION_MAX_TAP_POINTS) {
	    // Pen tap
	    if (ts.isLeftMouseButton()) {
		// Repaint called later
		unSelectAllObjects(false);
	    }

	} else if (ts.isLeftMouseButton()) {
	    InkStroke iStroke = new InkStroke(ts, getCurrentLayer(), true, false, this);

	    // See if this object intersects others		
	    Point sourcePoint = iStroke.getFirstPoint();
	    int sourceCompartment = mapXYToCompartment(sourcePoint.x, sourcePoint.y);
	    DataModel sourceModel = getModelInCompartment(sourceCompartment);

	    Point destPoint = iStroke.getLastPoint();
	    int destCompartment = mapXYToCompartment(destPoint.x, destPoint.y);
	    DataModel destModel = getModelInCompartment(destCompartment);
	    
	    if ((sourceModel != null) &&
		(destModel != null) &&
		(sourceCompartment != destCompartment)) {

		iStroke.setPerformAnalysis(true);
		performAnalysis(iStroke);
	    }
	    // Repaint called later
	    unSelectAllObjects(false);

	} else {
	    // Is a gesture command
	    Gesture gesture = new Gesture(ts);
	    GestureCommand cmd = new GestureCommand(gesture, this);
	    cmd.performAction();
	}

	repaint();
    }

    
    public void onVisualLanguageIconDropped(MouseEvent evt, int code, BehavioralInkStroke bStroke) {
	int x = evt.getX();
	int y = evt.getY();
	
	if (VLButton.isVLIconASourceEvent(code)) {
	    handleSourceEventDrop(bStroke, code, x, y);
	    
	} else {
	    handleActionDrop(bStroke, code, x, y);
	}
	
	PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
	
	// Repaint the mvp
	repaint();
    }


    public void onHoldAndPress(MouseEvent evt) {
	// Do nothing
    }


    public void performAnalysis(GraphicalObject graphicalObject) {
	// Did the user want this object analyzed?
	if (! graphicalObject.performAnalysis()) {
	    return;
	}

	// Determine what the object is
	if (graphicalObject instanceof BehavioralInkStroke) {
	} else if (graphicalObject instanceof InkStroke) {
	    // Analyze all ink strokes
	    performAnalysis((InkStroke) graphicalObject);
	}
    }

  
    public void performAnalysis(InkStroke iStroke) {
	Point sourcePoint = iStroke.getFirstPoint();
	int sourceCompartment = mapXYToCompartment(sourcePoint.x, sourcePoint.y);
	DataModel sourceModel = getModelInCompartment(sourceCompartment);
	Rectangle sourceRectangle = compartments[sourceCompartment].getBounds();

	Point destPoint = iStroke.getLastPoint();
	int destCompartment = mapXYToCompartment(destPoint.x, destPoint.y);
	DataModel destModel = getModelInCompartment(destCompartment);
	Rectangle destRectangle = compartments[destCompartment].getBounds();

	
	if ((sourceModel == null) || (destModel == null)) {
	    return;
	}

	GraphicalObject sourceObj = null;
	SourceEvent sourceEvent = null;
	if (sourceModel instanceof StoryboardModel) {
	    int x = (int)Math.round(((double)Storyboard.WIDTH / (double)sourceRectangle.width) * (sourcePoint.x - sourceRectangle.x));
	    int y = (int)Math.round(((double)Storyboard.HEIGHT / (double)sourceRectangle.height) * (sourcePoint.y - sourceRectangle.y));
	    sourceObj = sourceModel.getStoryboard().getRectangleOrTextContaining(x, y, sourceModel.getNumLayers(), true);
	    if (sourceObj == null) {
		iStroke.setPerformAnalysis(false);
		return;
	    }
	}

	GraphicalObject actionObj = null;
	DestinationAction action = null;
	if (destModel instanceof StoryboardModel) {
	    int x = (int)Math.round(((double)Storyboard.WIDTH / (double) destRectangle.width) * (destPoint.x - destRectangle.x));
	    int y = (int)Math.round(((double)Storyboard.HEIGHT / (double) destRectangle.height) * (destPoint.y - destRectangle.y));
	    actionObj = destModel.getStoryboard().getRectangleOrTextContaining(x, y, destModel.getNumLayers(), true);
	}


	// Determine default source event 
	if (sourceModel instanceof NarrationModel) {
	    NarrationModel narrationSourceModel = (NarrationModel) sourceModel;
	    int syncIndex = ((MultiViewNarrationCompartment)compartments[sourceCompartment]).getSyncMarkerIndexAt(sourcePoint);
	    if (syncIndex >= 0) {
		// Found a sync marker at start point, create a sync marker event
		sourceEvent = new NarrationSyncMarkerEvent(narrationSourceModel, syncIndex);
	    }
	    
	} else {
	    // Determine default source event
	    if (sourceObj == null) {
		//System.out.println("Source is null");
	    }

	    // Its a storyboard model - careful, actionObj may be null
	    if (sourceObj instanceof ContinuousMediaRectangleInkStroke) {
		ContinuousMediaRectangleInkStroke cmris = (ContinuousMediaRectangleInkStroke) sourceObj;
		double mediaTime = cmris.getMediaTime();
		double duration = cmris.getDuration();
		
		if (mediaTime <= 1.0) {
		    sourceEvent = new BeginMediaEvent(cmris);
		} else if (mediaTime <= (duration - 1.0)) {
		    sourceEvent = new SyncMarkerEvent(cmris, mediaTime);
		} else {
		    sourceEvent = new EndMediaEvent(cmris);
		}

	    } else if (sourceObj instanceof RectangleInkStroke) {
		// Handle ImageRectangleInkStroke as well
		
		if (actionObj == null) {
		    sourceEvent = new SingleMouseClickEvent((RectangleInkStroke)sourceObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    sourceEvent = new RolloverEvent((RectangleInkStroke)sourceObj);
		} else if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		    sourceEvent = new SingleMouseClickEvent((RectangleInkStroke)sourceObj);
		} else if (actionObj instanceof RectangleInkStroke) {
		    sourceEvent = new SingleMouseClickEvent((RectangleInkStroke)sourceObj);
		}
		
	    } else if (sourceObj instanceof TextGraphicalObject) {
		
		if (actionObj == null) {
		    sourceEvent = new SingleMouseClickEvent((TextGraphicalObject) sourceObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    sourceEvent = new RolloverEvent((TextGraphicalObject)sourceObj);
		} else if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		    sourceEvent = new RolloverEvent((TextGraphicalObject)sourceObj);
		} else if (actionObj instanceof RectangleInkStroke) {
		    sourceEvent = new SingleMouseClickEvent((TextGraphicalObject)sourceObj);
		}
		
	    } else if (sourceObj instanceof BeginAnchor) {
		sourceEvent = new TemporalEvent(sourceObj);
	    }	    
	}
	

	// Determine default destination action	

	if (destModel instanceof NarrationModel) {
	    action = new BeginMediaAction((NarrationModel) destModel);

	} else {

	    // Destination is a storyboard model

	    if (sourceObj instanceof ContinuousMediaRectangleInkStroke) {

		if (actionObj == null) {
		    action = new NavigationAction((StoryboardModel)destModel);
		    
		} else if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		    ContinuousMediaRectangleInkStroke cmris = (ContinuousMediaRectangleInkStroke) actionObj;
		    double mediaTime = cmris.getMediaTime();
		    double duration = cmris.getDuration();
		    
		    if (mediaTime <= 1.0) {
			action = new BeginMediaAction(cmris);
		    } else if (mediaTime <= (duration - 1.0)) {
			action = new SetMediaTimeAction(cmris, mediaTime);
		    } else {
			action = new EndMediaAction(cmris);
		    }

		} else if (actionObj instanceof RectangleInkStroke) {
		    action = new HighlightAction((RectangleInkStroke) actionObj);
		    
		} else if (actionObj instanceof TextGraphicalObject) {
		    action = new DisplayAction((TextGraphicalObject) actionObj);
		}
		
	    } else {
		// Matches all other source object types
		
		if (actionObj == null) {
		    action = new NavigationAction((StoryboardModel)destModel);
		   
		} else if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		    ContinuousMediaRectangleInkStroke cmris = (ContinuousMediaRectangleInkStroke) actionObj;
		    double mediaTime = cmris.getMediaTime();
		    double duration = cmris.getDuration();
		    
		    if (mediaTime <= 1.0) {
			action = new BeginMediaAction(cmris);
		    } else if (mediaTime <= (duration - 1.0)) {
			action = new SetMediaTimeAction(cmris, mediaTime);
		    } else {
			action = new EndMediaAction(cmris);
		    }
		    
		} else if (actionObj instanceof TextGraphicalObject) {		   
		    action = new DisplayAction((TextGraphicalObject) actionObj);
		    
		} else if (actionObj instanceof RectangleInkStroke) {
		    // Handle images as well
		    action = new DisplayAction((RectangleInkStroke) actionObj);
		}
	    }
	}
	
	// Determine the location for the source event object
	TimedStroke stroke = iStroke.getTimedStroke();
	int locX = -1;
	int locY = -1;
	int i = -1;

	// Compute the transformed source event object
	if (sourceObj != null) {	    
	    
	    GraphicalObject sourceEventObj = (GraphicalObject) sourceObj.clone(this, false);
	    sourceEventObj.scale( (double)sourceRectangle.width / (double) Storyboard.WIDTH, 
				  (double)sourceRectangle.height / (double) Storyboard.HEIGHT);
	    sourceEventObj.doTranslate(sourceRectangle.x, sourceRectangle.y);
	    
	    // Locate the first point along the stroke that is outside the source object
	    i=1;
	    while ( (i < stroke.npoints) && 
		    (sourceEventObj.contains(stroke.xpoints[i], stroke.ypoints[i])) ) {
		++i;
	    }
	    if (i >= stroke.npoints) {
		i = 1;
	    }
	    
	    locX = stroke.xpoints[i];
	    locY = stroke.ypoints[i];
	    
	    // Adjust for direction
	    if ( (stroke.xpoints[i] - stroke.xpoints[i-1]) < 0) {
		// Moving to the left
		locX -= BehaviorEvent.VLI_WIDTH;
	    }
	    if ( (stroke.ypoints[i] - stroke.ypoints[i-1]) < 0) {
		// Moving towards the top of the screen
		locY -= BehaviorEvent.VLI_HEIGHT;
	    }
	    sourceEvent.setLocation(locX, locY);

	}

	// Compute the transformed action object
	if (actionObj != null) {
	    GraphicalObject actionEventObj = (GraphicalObject) actionObj.clone(this, false);
	    actionEventObj.scale( (double)destRectangle.width / (double) Storyboard.WIDTH, 
				  (double)destRectangle.height / (double) Storyboard.HEIGHT);
	    	    actionEventObj.doTranslate(destRectangle.x, destRectangle.y);
	    
	    i = stroke.npoints - 2;
	    while ((i >= 0) && (actionEventObj.contains(stroke.xpoints[i], stroke.ypoints[i]))) {
		--i;
	    }
	    if (i <= 0) {
		i = stroke.npoints - 2;
	    }
	
	    locX = stroke.xpoints[i];
	    locY = stroke.ypoints[i];
	    
	    // Adjust for direction
	    if ( (stroke.xpoints[i] - stroke.xpoints[i-1]) > 0) {
		// Moving to the right
		locX -= BehaviorEvent.VLI_WIDTH;
	    }
	    if ( (stroke.ypoints[i] - stroke.ypoints[i-1]) > 0) {
		// Moving towards the bottom of the screen
		locY -= BehaviorEvent.VLI_HEIGHT;
	    }	    

	} else {
	    i = stroke.npoints - 2;
		
	    locX = stroke.xpoints[i];
	    locY = stroke.ypoints[i];
	}
	action.setLocation(locX, locY);

	if ((sourceEvent != null) && (action != null)) {
	    BehavioralInkStroke bStroke = new BehavioralInkStroke(iStroke, sourceEvent, action);
	    addGraphicalObject(bStroke);
	    PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
	} 
    }


    public void handleSourceEventDrop(BehavioralInkStroke bStroke, int code, int x, int y) {
	DataModel sourceModel = null;
	GraphicalObject sourceObj = bStroke.getSourceEventObject();
	if (sourceObj != null) {
	    // Scope of the source object takes precedent - fixes bug of 
	    // a source icon straddling a compartment boundary
	    sourceModel = sourceObj.getScope();
	
	} else {
	    Toolkit.getDefaultToolkit().beep();
	    return;
	}
	
	if ((sourceModel == null) || (sourceModel instanceof NarrationModel)) {
	    Toolkit.getDefaultToolkit().beep();
	    return;
	}

	// Map the code to the source event
	SourceEvent sEvent = null;
	switch (code) {
	    
	case VLButton.SINGLE_MOUSE_CLICK_EVENT:
	    if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new SingleMouseClickEvent((RectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new SingleMouseClickEvent((TextGraphicalObject)sourceObj);
	    }
	    break;
	    
	case VLButton.DOUBLE_MOUSE_CLICK_EVENT:
	    if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new DoubleMouseClickEvent((RectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new DoubleMouseClickEvent((TextGraphicalObject)sourceObj);
	    }     
	    break;
	    
	case VLButton.ROLLOVER_EVENT:
	    if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new RolloverEvent((RectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new RolloverEvent((TextGraphicalObject)sourceObj);
	    }
	    break;
	    
	case VLButton.DRAG_AND_DROP_EVENT:
	    break;
	    
	case VLButton.BEGIN_MEDIA_EVENT:
	    if (sourceObj instanceof ContinuousMediaRectangleInkStroke) {
		sEvent = new BeginMediaEvent((ContinuousMediaRectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new BeginMediaEvent((TextGraphicalObject)sourceObj);
	    }
	    break;
	    
	case VLButton.END_MEDIA_EVENT:
	    if (sourceObj instanceof ContinuousMediaRectangleInkStroke) {
		sEvent = new EndMediaEvent((ContinuousMediaRectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new EndMediaEvent((TextGraphicalObject)sourceObj);
	    }
	    break;
	    
	case VLButton.ENTER_TEXT_EVENT:
	    if  ((sourceObj instanceof ContinuousMediaRectangleInkStroke) ||
		 (sourceObj instanceof ImageRectangleInkStroke)) {
		break;
	    } else if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new EnterTextEvent((RectangleInkStroke)sourceObj);
	    }
	    break;
	    
	case VLButton.TEMPORAL_EVENT:
	    if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new TemporalEvent((RectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new TemporalEvent((TextGraphicalObject)sourceObj);
	    } else if (sourceObj instanceof BeginAnchor) {
		sEvent = new TemporalEvent((BeginAnchor)sourceObj);
	    }
	    break;
	    
	case VLButton.SYNC_MARKER_EVENT:
	    if (sourceObj instanceof ContinuousMediaRectangleInkStroke) {
		sEvent = new SyncMarkerEvent((ContinuousMediaRectangleInkStroke)sourceObj);
	    } 
	    break;

	default:
	    break;
	}
	
	if (sEvent != null) {
	    bStroke.setSourceEvent(sEvent);
	} else {
	    Toolkit.getDefaultToolkit().beep();
	}
    }


    private void handleActionDrop(BehavioralInkStroke bStroke, int code, int x, int y) {
	DataModel actionModel = null;
	GraphicalObject actionObj = bStroke.getActionObject();
	if (actionObj != null) {
	    // Scope of the action object takes precedent - fixes bug of 
	    // an action icon straddling a compartment boundary
	    actionModel = actionObj.getScope();
	} else {
	    int compartment = mapXYToCompartment(x, y);
	    if (compartments[compartment] != null) {
		actionModel = compartments[compartment].getModel();
	    }
	}
	if (actionModel == null) {
	    Toolkit.getDefaultToolkit().beep();
	    return;
	}

	// When placing a Display action on a currently Navigation/Transition object
	// need to relocate the actionObj
	if (actionObj == null) { 
	    if (actionModel instanceof StoryboardModel) {
		Point destPoint = bStroke.getLastPoint();
		int compartment = getCompartmentForModel(actionModel);
		Rectangle bounds = compartments[compartment].getBounds();
		int dX = (int)Math.round(((double)Storyboard.WIDTH / (double) bounds.width) * (destPoint.x - bounds.x));
		int dY = (int)Math.round(((double)Storyboard.HEIGHT / (double) bounds.height) * (destPoint.y - bounds.y));
		actionObj = actionModel.getStoryboard().getRectangleOrTextContaining(dX, dY, actionModel.getNumLayers(), true);
	    }
	}

	// Map the code to the event or action
	DestinationAction action = null;

	if (actionModel instanceof NarrationModel) {
	    NarrationModel nModel = (NarrationModel) actionModel;
	    
	    switch(code) {
	    case VLButton.BEGIN_MEDIA_ACTION:
		action = new BeginMediaAction(nModel);
		break;
		
	    case VLButton.END_MEDIA_ACTION:
		action = new EndMediaAction(nModel);
		break;
		
	    case VLButton.PAUSE_MEDIA_ACTION:
		action = new PauseMediaAction(nModel);
		break;
	    default:
		break;
	    }

	} else if (actionObj == null) {

	    switch(code) {
	    case VLButton.NAVIGATE_ACTION:
		action = new NavigationAction((StoryboardModel) actionModel);
		if (! action.hasDestModel()) {
		    return;
		}
		break;
		
	    case VLButton.TRANSITION_ACTION:
		action = new TransitionAction((StoryboardModel) actionModel);
		if (! action.hasDestModel()) {
		    return;
		}
		break;
	    default:
		break;
	    }

	} else {

	    switch (code) {
		
	    case VLButton.NAVIGATE_ACTION:
		action = new NavigationAction((StoryboardModel)actionModel);
		if (! action.hasDestModel()) {
		    return;
		}
		break;
		
	    case VLButton.BEGIN_MEDIA_ACTION:
		if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		    action = new BeginMediaAction((ContinuousMediaRectangleInkStroke)actionObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    action = new BeginMediaAction((TextGraphicalObject)actionObj);
		}
		break;
		
	    case VLButton.END_MEDIA_ACTION:
		if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		    action = new EndMediaAction((ContinuousMediaRectangleInkStroke)actionObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    action = new EndMediaAction((TextGraphicalObject)actionObj);
		}
		break;
		
	    case VLButton.PAUSE_MEDIA_ACTION:
		if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		    action = new PauseMediaAction((ContinuousMediaRectangleInkStroke)actionObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    action = new PauseMediaAction((TextGraphicalObject)actionObj);
		}
		break;
		
	    case VLButton.SET_MEDIA_TIME_ACTION:
		if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		    action = new SetMediaTimeAction((ContinuousMediaRectangleInkStroke)actionObj);
		}
		break;
		
	    case VLButton.DELAY_ACTION:
		action = new DelayAction(actionObj);
		break;

	    case VLButton.TRANSITION_ACTION:
		action = new TransitionAction((StoryboardModel) actionModel);
		if (! action.hasDestModel()) {
		    return;
		}
		break;
		
	    case VLButton.DISPLAY_ACTION:
		if (actionObj instanceof RectangleInkStroke) {
		    action = new DisplayAction((RectangleInkStroke)actionObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    action = new DisplayAction((TextGraphicalObject)actionObj);
		}
		break;
		
	    case VLButton.UNDISPLAY_ACTION:
		if (actionObj instanceof RectangleInkStroke) {
		    action = new UnDisplayAction((RectangleInkStroke)actionObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    action = new UnDisplayAction((TextGraphicalObject)actionObj);
		}
		break; 
		
	    case VLButton.HIGHLIGHT_ACTION:
		if (actionObj instanceof RectangleInkStroke) {
		    action = new HighlightAction((RectangleInkStroke)actionObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    action = new HighlightAction((TextGraphicalObject)actionObj);
		}
		break;
		
	    case VLButton.UNHIGHLIGHT_ACTION:
		if (actionObj instanceof RectangleInkStroke) {
		    action = new UnHighlightAction((RectangleInkStroke)actionObj);
		} else if (actionObj instanceof TextGraphicalObject) {
		    action = new UnHighlightAction((TextGraphicalObject)actionObj);
		}
		break;
		
	    case VLButton.PLAY_AUDIO_ACTION:
		break;
		
	    default:
		break;
	    }
	}
	
	if (action != null) {
	    bStroke.setAction(action);
	} else {
	    Toolkit.getDefaultToolkit().beep();
	}
    }



    public void operationalize() {
	Iterator it = graphicalObjects.iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go instanceof BehavioralInkStroke) {
		BehavioralInkStroke bis = (BehavioralInkStroke) go;
		Point p = bis.getFirstPoint();
		int compartment = mapXYToCompartment(p.x, p.y);
		DataModel dModel = getModelInCompartment(compartment);
		if (dModel != null) {
		    bis.operationalize(dModel);
		}
	    } 
	}
	setCursor(Storyboard.OP_DEFAULT_CURSOR);
    }


    public void unOperationalize() {
	graphicalObjects.unOperationalize();
	setCursor(PropertyManager.APPLICATION_CURSOR);
    }



    public Object clone() {
	return (new MultiViewPanel(this));
    }

    public Object clone(MultiViewModel toModel) {
	return (new MultiViewPanel(this, toModel));
    }



    public void write(DataOutputStream stream) throws IOException {
	super.write(stream);

	// Write the compartment information
	for (int i=0; i<NUM_COMPARTMENTS; ++i) {
	    stream.writeBoolean(compartments[i] != null);
	    if (compartments[i] != null) {
		//Write the compartment info
		compartments[i].write(stream);
	    }
	}
    }
}


