/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File MultiViewStoryboardCompartment.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.multiview;
    

// Necessary imports
import java.awt.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.util.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.behavior.*;


	
	

public class MultiViewStoryboardCompartment extends MultiViewCompartment {

    // Class id 
    public static final int CLASSID = 2010;

    // Dashed stroke for navigational links
    private final static float _dash[] = {10.0f};
    private static final Stroke DASH = new BasicStroke(1.0f, 
						       BasicStroke.CAP_BUTT, 
						       BasicStroke.JOIN_MITER, 
						       10.0f, _dash, 0.0f);
    // Arrow for navigational link
    private static Point DEFAULT_ARROW[];




    // Underlying storyboard model
    private StoryboardModel model;

    // For file I/O only
    private int ioStoryboard;

    // Current list of graphical objects for this compartment
    private GraphicalObjectCollection currentCollection;

    // List of intra-storyboard navigation/transition links to drawn  
    private Vector navigationActions;



    // Static initializer
    static {
	DEFAULT_ARROW = new Point[3];
	// Define arrow in Java's coordinate system (Y increases downward)
	DEFAULT_ARROW[0] = new Point(-4, -4);
	DEFAULT_ARROW[1] = new Point(3, 0);
	DEFAULT_ARROW[2] = new Point(-4, 4);
    }



    private MultiViewStoryboardCompartment() {
	this(null, null, null);
    }
    
    public MultiViewStoryboardCompartment(Rectangle r, MultiViewPanel mp, StoryboardModel sModel) {
	super(r, mp);
	model = sModel;
	navigationActions = new Vector();
	onInstall();
    }

    public MultiViewStoryboardCompartment(DataInputStream stream, MultiViewPanel mp) throws IOException {
	super(stream, mp);
	ioStoryboard = stream.readInt();
	currentCollection = null;
	navigationActions = new Vector();
    }

    public void completeIO() {
	super.completeIO();
	model = StoryboardModel.getStoryboardModelById(new Integer(ioStoryboard));
    }


    public void onInstall() {
	super.onInstall();

  	// Retrieve a new copy of the underlying objects - as the referenced storyboard may
	// have been updated
	GraphicalObjectCollection goc = (GraphicalObjectCollection) getCollection();

	// Look for intra-storyboard navigational actions
	Iterator it = goc.iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go instanceof BehavioralInkStroke) {
		BehavioralInkStroke bis = (BehavioralInkStroke) go;
		DestinationAction action = bis.getAction();
		if (action != null) {
		    StoryboardModel destModel = action.getDestModel();
		    if (destModel != null) {
			navigationActions.add(action);
		    }
		}
	    }
	}

	// Copy over the collection
	currentCollection = (GraphicalObjectCollection) goc.clone(mvp, false);
	currentCollection.scaleTo(rectangle.width, rectangle.height);
	currentCollection.translate(rectangle.x, rectangle.y, 10000, 10000);
	currentCollection.realize(true, true);
    }


    public void onUnInstall() {
	super.onUnInstall();
	deleteCollection();
	if (navigationActions != null) {
	    navigationActions.clear();
	}
    }



    public DataModel getModel() {
	return model;
    }


    public GraphicalObjectCollection getCollection() {
	return model.getStoryboard().getGraphicalObjectCollection();
    }


    public void render(Graphics2D g) {
	integrityCheck();

	// Render the title
	g.setColor(Color.black);
	g.setFont(PropertyManager.MULTIVIEW_TITLE_FONT);
	FontMetrics fm = g.getFontMetrics();
	g.drawString(model.getName(), rectangle.x + 4, rectangle.y + rectangle.height - fm.getMaxDescent());

	if (currentCollection != null) {
	    currentCollection.render(g);
	}
	
	// Draw navigation links
	if ((navigationActions != null) && (navigationActions.size() > 0)) {
	    
	    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			       RenderingHints.VALUE_ANTIALIAS_ON);
	    drawNavigationActions(g);    
	    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			       RenderingHints.VALUE_ANTIALIAS_OFF);
	}
    }




    public void drawNavigationActions(Graphics2D g) {
	g.setStroke(DASH);
	
	double sx = ((double) rectangle.width  / (double) mvp.WIDTH);
	double sy = ((double) rectangle.height / (double) mvp.HEIGHT);
	
	Iterator it = navigationActions.iterator();
	while (it.hasNext()) {    
	    DestinationAction action = (DestinationAction) it.next();
	    if (action != null) {
		Point sourceP = new Point(rectangle.x + (int)Math.round((action.getX() * sx)), 
					  rectangle.y + (int)Math.round((action.getY() * sy)));
		
		StoryboardModel destModel = action.getDestModel();
		if (destModel != null) {
		    int c = mvp.getCompartmentForModel(destModel);
		    if (c >= 0) {
			MultiViewCompartment mvc = mvp.getCompartment(c);
			if (mvc != null) {
			    Rectangle r = mvc.getBounds();
			    Point attachPoints[] = mvc.getAttachPoints();
			    Point destP = getClosestAttachPoint(sourceP, attachPoints);

			    // Draw a circle in the source object
			    g.fillArc(sourceP.x - 3, sourceP.y - 3, 6, 6, 0, 360);

			    // Draw the dashed line
			    g.drawLine(sourceP.x, sourceP.y, destP.x, destP.y);

			    Polygon arrow = computeArrowLocation(sourceP, destP);
			    if (arrow != null) {
				g.fillPolygon(arrow);
			    }
			}
		    }
		}
	    }
	}
    }

    
    protected void deleteCollection() {
	if (currentCollection != null) {
	    currentCollection.deleteGraphicalObjects(false);
	}
    }


    private Point getClosestAttachPoint(Point sourceP, Point points[]) {
	double bestD = Double.MAX_VALUE;
	Point bestP = null;
	for (int i=0; i<points.length; ++i) {
	    double d = DMath.distance(sourceP.x, sourceP.y, points[i]);
	    if (d < bestD) {
		bestD = d;
		bestP = points[i];
	    }
	}
	return bestP;
    }




    private Polygon computeArrowLocation(Point sourceP, Point destP) {
	boolean up = false;
	int diff = destP.y - sourceP.y;
	if (diff < 0) {
	    // Moving towards the top of the screen
	    up = true;
	} 

	double distance = DMath.distance(destP.x, destP.y, sourceP.x, sourceP.y);
	double dot_product = (destP.x - sourceP.x);
	double alpha = Math.acos(dot_product / distance); // in radians
	if (! up) {
	    alpha = -1.0 * alpha;
	}
	    
	Polygon arrow = new Polygon();
	double x, y;
	int x1, y1;
	for (int i=0; i<3; ++i) {
	    x = DEFAULT_ARROW[i].getX();
	    y = DEFAULT_ARROW[i].getY();

	    // Rotate in Java's coordinate system (Y increases downward)
	    x1 = ((int)Math.round(Math.cos(alpha)*x + Math.sin(alpha)*y)) + destP.x;
	    y1 = ((int)Math.round(Math.cos(alpha)*y - Math.sin(alpha)*x)) + destP.y;
	    arrow.addPoint(x1, y1);
	}

	return arrow;
    }



    private void integrityCheck() {
	Vector toRemove = new Vector();
	Iterator it = mvp.getGraphicalObjectCollection().iterator();
	while (it.hasNext()) {
	    BehavioralInkStroke bStroke = (BehavioralInkStroke) it.next();
	    Point sourcePoint = bStroke.getFirstPoint();
	    
	    if (contains(sourcePoint)) {
		GraphicalObject assignedSourceObject = bStroke.getSourceEventObject();
		if (assignedSourceObject != null) {
		    int x = (int)Math.round(((double)Storyboard.WIDTH / (double)rectangle.width) * (sourcePoint.x - rectangle.x));
		    int y = (int)Math.round(((double)Storyboard.HEIGHT / (double)rectangle.height) * (sourcePoint.y - rectangle.y));
		    GraphicalObject cSourceObj = model.getStoryboard().getRectangleOrTextContaining(x, y, model.getNumLayers(), true);
		    if (cSourceObj != assignedSourceObject) { 		    
			toRemove.add(bStroke);
		    }
		}
	    }
	}

	// do NOT perform a repaint
	mvp.deleteGraphicalObjects(toRemove, false);
    }



    public void write(DataOutputStream stream) throws IOException {
	stream.writeInt(CLASSID);

	super.write(stream);
	stream.writeInt(model.getIdentifier().intValue());
    }
}
