/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: NarrationEditControlPanel.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.narration;

// Necessary imports
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.speech.synthesis.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;




public class NarrationEditControlPanel extends JPanel {

    private static final int STOP = 0;
    private static final int PLAY = 1;
    private static final int PLAY_BUT_PAUSED = 2;
    private static final int RECORDING = 3;
    

    private NarrationEditView view;
    private JButton record, play, pause, resume, stop;
    private Listener listener;
    private int mode;
    private javax.swing.Timer timer;
    private boolean blinkToggle;
    private Color normalBackground;
    private AudioTimerLabel timerLabel;




    // Constructor
    private NarrationEditControlPanel() {
	this(null);
    }

    public NarrationEditControlPanel(NarrationEditView v) {
	view = v;
	setLayout(new GridBagLayout());
	listener = new Listener();
	mode = STOP;

	// Create the icons and buttons
	record = createIconButton("microphone", false);
	record.setEnabled(CAudioCapture.isMicrophoneAvailable());
	record.setToolTipText("Record your voice");

	play = createIconButton("play", true);
	play.setEnabled(TextToSpeechSynthesizer.getTTS().isTextToSpeechAvailable());
	play.setToolTipText("Begin narration");

	pause = createIconButton("pause", false);
	pause.setVisible(false);
	pause.setEnabled(false);
	pause.setToolTipText("Pause narration");
	
	resume = createIconButton("resume", false);
	resume.setVisible(false);
	resume.setEnabled(false);
	resume.setToolTipText("Resume narration");

	stop = createIconButton("stop", false);
	stop.setVisible(false);
	stop.setEnabled(false);
	stop.setToolTipText("Stop narration");

	GridBagConstraints c = new GridBagConstraints();
	c.gridx = 0;
	c.gridy = 0;
	c.insets = new Insets(0, 3, 0, 0);
	c.anchor = GridBagConstraints.WEST;
	add(record, c);

	++c.gridx;
	add(play, c);

	++c.gridx;
	add(stop, c);

	++c.gridx;
	add(pause, c);

	++c.gridx;
	add(resume, c);

	++c.gridx;
	c.insets = new Insets(0, 20, 0, 0);
	c.weightx = 0.0;
	c.weighty = 0.0;
	c.fill = GridBagConstraints.NONE;
	timerLabel = new AudioTimerLabel();
	add(timerLabel, c);


	// Set up the timer
	timer = new Timer(10000, new RecordListener());
	timer.setRepeats(true);
	timer.stop();
	blinkToggle = false;
	normalBackground = record.getBackground();
    }

  
    private JButton createIconButton(String name, boolean needsDisabled) {
	ImageIcon enabled = 
	    new ImageIcon(PropertyManager.APPLICATION_IMAGE_DIRECTORY +
			  PropertyManager.FILE_SEPARATOR +
			  name + 
			  "-enabled.gif");

	JButton button = new JButton(enabled);
	if (needsDisabled) {
	    ImageIcon disabled = 
		new ImageIcon(PropertyManager.APPLICATION_IMAGE_DIRECTORY +
			      PropertyManager.FILE_SEPARATOR +
			      name + 
			      "-disabled.gif");
	    button.setDisabledIcon(disabled);
	}

	button.setActionCommand(name);
	button.addActionListener(listener);
	Dimension d = new Dimension(28, 22);
	button.setPreferredSize(d);
	button.setMinimumSize(d);
	button.setMaximumSize(d);
	return button;
    }



    private void handleRecord() {
	if (mode == STOP) {
	    mode = RECORDING;
	    record.setToolTipText("Stop Recording");
	    play.setVisible(false);
	    play.setEnabled(false);

	    if (view.beginAudioRecording()) {
		timer.setInitialDelay(300);
		timer.setDelay(600);
		timer.start();
		
		timerLabel.start();
	    }

	} else if (mode == RECORDING) {
	    mode = STOP;
	    timer.stop();
	    timerLabel.stop();

	    view.endAudioRecording();

	    record.setBackground(normalBackground);
	    play.setVisible(true);
	    play.setEnabled(true);
	    record.setToolTipText("Record your voice");
	    blinkToggle = false;
	}
    }


    private void handleRecordBlink() {
	if (blinkToggle) {
	    record.setBackground(Color.red);
	} else {
	    record.setBackground(normalBackground);
	}
	blinkToggle = (! blinkToggle);
    }


    
    private void handlePlay() {
	mode = PLAY;
	record.setVisible(false);
	record.setEnabled(false);

	play.setVisible(false);
	play.setEnabled(false);

	resume.setVisible(false);
	resume.setEnabled(false);

	stop.setVisible(true);
	stop.setEnabled(true);

	pause.setVisible(true);
	pause.setEnabled(true);

	view.getModel().beginEditNarration(view.getCaretPosition(), listener);
	timerLabel.start();
    }


    private void handlePause() {
	// Pause button pressed
	mode = PLAY_BUT_PAUSED;
	pause.setVisible(false);
	pause.setEnabled(false);

	resume.setVisible(true);
	resume.setEnabled(true);

	view.getModel().pauseEditNarration();
	timerLabel.pause();
    }
    
    
    private void handleResume() {
	mode = PLAY;
	timerLabel.resume();
	pause.setVisible(true);
	pause.setEnabled(true);

	resume.setVisible(false);
	resume.setEnabled(false);
	view.getModel().resumeEditNarration();
    }


    private void handleStopAndEdit() {
	// Stop the narration first 
	view.getModel().endEditNarration();

	finishStopAndEdit();
    }


    private void finishStopAndEdit() {
	// Going from PLAY to STOP
	mode = STOP;

	// Stop the counter
	timerLabel.stop();

	record.setVisible(true);
	record.setEnabled(true);

	play.setVisible(true);
	play.setEnabled(true);

	stop.setVisible(false);
	stop.setEnabled(false);

	pause.setVisible(false);
	pause.setEnabled(false);

	resume.setVisible(false);
	resume.setEnabled(false);
    }


  

    class Listener extends SpeakableAdapter implements ActionListener {

	public Listener() {
	    // Empty for now
	}


	public void actionPerformed(ActionEvent e) {
	    String cmd = e.getActionCommand();
	    if (cmd.equals("record")) {
		handleRecord();
	    
	    } else if (cmd.equals("play")) {
		handlePlay();

	    } else if (cmd.equals("pause")) {
		handlePause();

	    } else if (cmd.equals("resume")) {
		handleResume();
		
	    } else if (cmd.equals("stop")) {
		handleStopAndEdit();

	    }
	}


	// SpeakableListener interface
	public void speakableCancelled(SpeakableEvent e) {
	    finishStopAndEdit();
	}

	public void speakableEnded(SpeakableEvent e) {
	    finishStopAndEdit();
	}
    } 


    class RecordListener implements ActionListener {
	public RecordListener() {
	}
	
	public void actionPerformed(ActionEvent event) {
	    handleRecordBlink();
	}
    }
    
}

