/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: NarrationEditMenu.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.narration;

// Necessary imports
import java.lang.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.undo.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;



public class NarrationEditMenu extends JMenu {

    // The listener for all menu items (an inner class below)
    private Listener listener;

    // The menu items
    private JMenuItem editUndo, editCut, editCopy, editPaste;

    // Top level view window of the editor
    private NarrationEditView view;

    // The text pane to perform the cut, copy, and paste actions
    private NarrationTextPane textPane;

    // Hastable of text editor actions
    private Hashtable actions;

    // The undo manager
    private UndoManager undo;

    // The undo listener
    private UndoListener undoListener;




    public NarrationEditMenu(NarrationEditView view, NarrationTextPane tPane) {
	super("Edit");
	this.view = view;
	textPane = tPane;
	listener = new Listener();
	undoListener = new UndoListener();

	createActionTable();

	undo = new UndoManager();
	Document doc = view.getModelDocument();
	doc.addUndoableEditListener(undoListener);

	editUndo = new JMenuItem("Undo");
	editUndo.setEnabled(false);
	editUndo.addActionListener(listener);
	add(editUndo);

	addSeparator();

	editCut = new JMenuItem("Cut");
	editCut.setEnabled(false);
	editCut.addActionListener(listener);
	add(editCut);

	editCopy = new JMenuItem("Copy");
	editCopy.setEnabled(false);
	editCopy.addActionListener(listener);
	add(editCopy);

	editPaste = new JMenuItem("Paste");
	editPaste.setEnabled(false);
	editPaste.addActionListener(listener);
	add(editPaste);

    }



    private void createActionTable() {
	actions = new Hashtable();
	Action[] actionsArray = textPane.getActions();
	for (int i = 0; i < actionsArray.length; ++i) {
	    Action a = actionsArray[i];
	    actions.put(a.getValue(Action.NAME), a);
	}
    }   


    private Action getActionByName(String name) {
	return (Action)(actions.get(name));
    }
    

    

    // An inner class to handle undo events from the document
    class UndoListener implements UndoableEditListener {

	public UndoListener() {
	}

	public void undoableEditHappened(UndoableEditEvent e) {
	    //Remember the edit and update the menus
	    undo.addEdit(e.getEdit());
	    editUndo.setEnabled(undo.canUndo());

	}
    }


    
    // An inner class to handle all menu item events
    class Listener implements ActionListener, PropertyListenerInterface {
	
	// Constructor
	public Listener() {	
	    PropertyManager.addPropertyListener(PropertyManager.VS_TEXT_SELECTED, this);
	    PropertyManager.addPropertyListener(PropertyManager.VS_TEXT_IN_CLIPBOARD, this);   
	    //PropertyManager.addPropertyListener(PropertyManager.VS_UNDO_LIST_EMPTY, this);   
	}


	// Menu item callback
	public void actionPerformed(ActionEvent e) {
	    Object selection = e.getSource();
	 
	    if (selection == editUndo) {
		handleUndo();

	    } else if (selection == editCut) {
		handleCut(e);

	    } else if (selection == editCopy) {
		handleCopy(e);
 
	    } else if (selection == editPaste) {
		handlePaste(e);

	    } else {
		System.out.println("ERROR: Unknown edit menu command");
	    }
	}


	// PropertyListener interface
	public void propertyChanged(Integer property, Object oldValue, Object newValue) {
	    boolean b = ((Boolean) newValue).booleanValue();
	    if (property.equals(PropertyManager.VS_TEXT_SELECTED)) {
		editCut.setEnabled(b);
		editCopy.setEnabled(b);

	    } else if (property.equals(PropertyManager.VS_TEXT_IN_CLIPBOARD)) {
		editPaste.setEnabled(b);

	    } 
	}


	private void handleUndo() {
	    try {
		undo.undo();
	    } catch (CannotUndoException xcptn) {
		DEMAISError.showErrorMessage(xcptn.toString(), "Narration Undo", false);
		DEMAISError.logException(xcptn);
	    }
	    editUndo.setEnabled(undo.canUndo());
	}

	private void handleCut(ActionEvent e) {
	    Action action = getActionByName(DefaultEditorKit.cutAction);
	    action.actionPerformed(e);
	    PropertyManager.setProperty(PropertyManager.VS_TEXT_IN_CLIPBOARD, Boolean.TRUE);

	}

	private void handleCopy(ActionEvent e) {
	    Action action = getActionByName(DefaultEditorKit.copyAction);
	    action.actionPerformed(e);
	    PropertyManager.setProperty(PropertyManager.VS_TEXT_IN_CLIPBOARD, Boolean.TRUE);

	}
	
	private void handlePaste(ActionEvent e) {
	    Action action = getActionByName(DefaultEditorKit.pasteAction);
	    action.actionPerformed(e);
	}

    }
}
