/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/*
  File: NarrationEditView.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.narration;

// Necessary imports
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.border.*;
import javax.swing.text.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;
import edu.uofminnesota.DEMAIS.usrinterface.contentmanager.*;




public class NarrationEditView extends JFrame 
    implements DataModelViewInterface {


    private NarrationModel model;
    private Listener listener;
    private CAudioCapture microphone;
    private NarrationEditPanel nePanel;
    private NarrationTextPane textPane;



    private NarrationEditView() {
    }


    public NarrationEditView(NarrationModel nm) {
	super("Voice Script - " + nm.getName());

	setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);

	// Store the model
	model = nm;
	// Create the listener
	listener = new Listener();

	microphone = null;

	try {

	    // Listen for window events
	    addWindowListener(listener);

	    // Listen for name changes
	    model.addNameListener(listener);

	    // Set the look and feel
	    setLookAndFeel();

	    // Create the actual text editor now (so that the
	    // trimmings (menus, buttons, etc) can have access to it)
	    textPane = new NarrationTextPane(this);

	    createMenus(textPane);
	    setupContentPane(textPane);
	    makeVisible();

	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "Narration Editor", false);
	    DEMAISError.logException(xcptn);
	}
    }
    


    protected int getCaretPosition() {
	return textPane.getCaretPosition();
    }

    
    // Private methods
    private void setLookAndFeel() {
	try {
	    UIManager.setLookAndFeel(DEMAISFrame.getLookAndFeelClassName());
	} catch (Exception xcptn) {
	    DEMAISError.showErrorMessage("Could not set look and feel", "Narration Editor", false);
	    DEMAISError.logException(xcptn);
	}
    }


    private void createMenus(NarrationTextPane textPane) {
	this.setJMenuBar(new NarrationFrameMenuBar(this, textPane));
    }

    
    private void setupContentPane(NarrationTextPane textPane) {
	JPanel contentPane = new JPanel(new GridBagLayout());
	contentPane.setBorder(BorderFactory.createLoweredBevelBorder());
	setContentPane(contentPane);

	// Add the main content pane - which switches views as the user directs
	nePanel = new NarrationEditPanel(this, textPane);
	GridBagConstraints c = new GridBagConstraints();
	c.gridx = 0;
	c.gridy = 0;
	c.fill = GridBagConstraints.BOTH;
	c.weightx = 1.0;
	c.weighty = 1.0;
	contentPane.add(nePanel, c);
    }



    // Make the interface visible to the user
    public void makeVisible() {
	pack();
	setRootLocation();
	setVisible(true);
    }



    // Protected methods

    protected NarrationModel getModel() {
	return model;
    }

    protected String getModelName() {
	return model.getName();
    }

    protected String getModelText(boolean build) {
	return model.getNarrationText(build);
    }

    protected DefaultStyledDocument getModelDocument() {
	return model.getDocument();
    }


    protected boolean beginAudioRecording() {
	nePanel.setEditable(false);
	microphone = new CAudioCapture(getModelName());
	return (microphone.beginCapture());
    }

    protected void endAudioRecording() {
	nePanel.setEditable(true);
	microphone.endCapture();
	// Inform the user where this file
	JOptionPane.showMessageDialog(this, "Audio file: " + microphone.getAudioFileName(), "Location of Audio Recording", JOptionPane.INFORMATION_MESSAGE);
	microphone = null;
    }


    private void setRootLocation() {
	Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
	Dimension me = this.getSize();

	Point p = new Point( (int)(screen.getWidth() - me.getWidth()) / 2, 
			     (int)(screen.getHeight() - me.getHeight()) / 2);
	setLocation(p);
    }


    public void deleteView(DataModel model, boolean remove) {
	closeWindow(false, remove);
    }

    public void refreshView(DataModel model) {
	// Do nothing
    }


    public void closeWindow() {
	closeWindow(true, true);
    }

    public void closeWindow(boolean refresh, boolean remove) {
	// Stop narration if active
	model.endEditNarration();

	// Stop recording if active
	if (microphone != null) {
	    endAudioRecording();
	}
	
	if (remove) {
	    // Remove this as a view 
	    model.removeView(this);
	}

	// Remove as a name listener on the model
	model.removeNameListener(listener);

	if (refresh) {
	    model.buildSyncMarkersAndVariables();
	    model.refreshViews();
	}

       	this.dispose();
    }



    // Listener class
    class Listener extends WindowAdapter
	implements NamedModelNameListenerInterface {
	
	public Listener() {
	}


	// NameListener interface
	public void nameChanged(NamedModel obj, String oldName, String newName) {
	    setTitle("Voice Script - " + newName);
	}


	// Window Listener interface
	public void windowClosing(WindowEvent e) {
	    // The window is closing, store the model
	    closeWindow(true, true);
	}
     
    }
}




