/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File Storyboard.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard;
    

// Necessary imports
import java.awt.*;
import java.awt.event.*;
import java.awt.print.*;
import javax.swing.*;
import java.util.*;
import java.io.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.core.annotation.*;
import edu.uofminnesota.DEMAIS.core.parser.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.stroke.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.behavior.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture.commands.*;
import edu.uofminnesota.DEMAIS.usrinterface.util.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.mediation.*;






// Storyboard class
public class Storyboard extends JLayeredPane
    implements Printable {

    // The Rubine gesture recognizer is the default!
    public static final Recognizer defaultGestureRecognizer;
    // The default Ink Recognizer
    public static final Recognizer defaultInkRecognizer;

    // Cursors for the runtime storyboard
    public static final Cursor OP_DEFAULT_CURSOR;
    public static final Cursor OP_HAND_CURSOR;
    public static final Cursor OP_MOVE_CURSOR;

    // Width and height of the storyboard
    public static int WIDTH;
    public static int HEIGHT;




    // The model to which this storyboard belongs
    protected DataModel model;

    // Assembles the current stroke
    protected StrokeAssembler assembler;

    // Current gesture recognizer - initialized to the default
    protected Recognizer gestureRecognizer;

    // Current ink recognizer - initialized to the default
    protected Recognizer inkRecognizer;

    // Drop target listener for the storyboard
    protected StoryboardDropTargetListener dropTargetListener;

    // The key listener for the storyboard
    protected StoryboardKeyListener keyListener;


    // A list of objects to render
    protected GraphicalObjectCollection graphicalObjects;

    // The active text object, if any
    protected TextGraphicalObject activeTextObject;

    // If the storyboard is editable
    protected boolean editable;

    // Damaged area that will be repainted - if null, entire storyboard will be repainted
    protected Rectangle damagedArea;

    // Current scale factor applied to this storyboard
    protected double scaleFactorX, scaleFactorY;

    // The last location of the mouse for moving selected objects
    protected int lastMouseX, lastMouseY;

    protected boolean mouseDragged;



    // OPERATIONAL section
    private StoryboardOpMouseListener opMouseListener;
    private LinkedList opSourceEventObjs;
    private GraphicalObject opCurrentObject;
    private Collection opCurrentCollection;
    private int opLastMouseX;
    private int opLastMouseY;
    private boolean opDraggingObject;
    private GraphicalObject opPressedObject;




    // Static initializer
    static {
	RubineInterpreter rubineInterpreter = RubineInterpreter.getRubineInterpreter();
	// Get the gesture recognizer
	defaultGestureRecognizer = rubineInterpreter.getRecognizer(RubineInterpreter.DEMAIS_GESTURE_FILE);
	
	// Now get the ink recognizer
	defaultInkRecognizer = rubineInterpreter.getRecognizer(RubineInterpreter.STANDARD_RECTANGLE_FILE);


	
	OP_DEFAULT_CURSOR = Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR);
	OP_HAND_CURSOR = Cursor.getPredefinedCursor(Cursor.HAND_CURSOR);
	OP_MOVE_CURSOR = Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR);
    }

    

    // Constructors
    protected Storyboard(MultiViewModel parentModel) {
	setLayout(null);
	setBackground(PropertyManager.STORYBOARD_BACKGROUND_COLOR);

	model = parentModel;
	assembler = new StrokeAssembler(this, false);

	gestureRecognizer = Storyboard.defaultGestureRecognizer;
	inkRecognizer = Storyboard.defaultInkRecognizer;
	dropTargetListener = null;
	keyListener = new StoryboardKeyListener(this);
	addKeyListener(keyListener);

	graphicalObjects = new GraphicalObjectCollection(this, 20);
	activeTextObject = null;
	editable = true;
	damagedArea = null;
	scaleFactorX = 1.0;
	scaleFactorY = 1.0;
	lastMouseX = 0;
	lastMouseY = 0;
	mouseDragged = false;

	// Operational portion
	opMouseListener = null;
	opSourceEventObjs = null;
	opCurrentObject = null;
	opCurrentCollection = null;
	opLastMouseX = 0;
	opLastMouseY = 0;
	opDraggingObject = false;
	opPressedObject = null;
	setCursor(PropertyManager.APPLICATION_CURSOR);
	PieMenuBar.getPieMenuBar().addPieMenuTo(this);
    }



    protected Storyboard(DataInputStream stream, MultiViewModel parentModel) throws IOException {
	setLayout(null);
	setBackground(PropertyManager.STORYBOARD_BACKGROUND_COLOR);

	model = parentModel;
	assembler = new StrokeAssembler(this, false);

	gestureRecognizer = Storyboard.defaultGestureRecognizer;
	inkRecognizer = Storyboard.defaultInkRecognizer;
	dropTargetListener = null;

	keyListener = new StoryboardKeyListener(this);
	addKeyListener(keyListener);

	graphicalObjects = new GraphicalObjectCollection(stream, this);

	activeTextObject = null;
	editable = true;
	damagedArea = null;
	scaleFactorX = 1.0;
	scaleFactorY = 1.0;
	lastMouseX = 0;
	lastMouseY = 0;
	mouseDragged = false;

	// Operational portion
	opMouseListener = null;
	opSourceEventObjs = null;
	opCurrentObject = null;
	opCurrentCollection = null;
	opLastMouseX = 0;
	opLastMouseY = 0;
	opDraggingObject = false;
	opPressedObject = null;
	setCursor(PropertyManager.APPLICATION_CURSOR);
    }



    public Storyboard(DataModel parentModel) {
	setLayout(null);
	setBackground(PropertyManager.STORYBOARD_BACKGROUND_COLOR);

	model = parentModel;
	assembler = new StrokeAssembler(this);
	gestureRecognizer = Storyboard.defaultGestureRecognizer;
	inkRecognizer = Storyboard.defaultInkRecognizer;

	// Install the listeners
	dropTargetListener = new StoryboardDropTargetListener(this);
	keyListener = new StoryboardKeyListener(this);
	addKeyListener(keyListener);

	graphicalObjects = new GraphicalObjectCollection(this, 100);
	addGraphicalObject(new BeginAnchor(-1, this));

	activeTextObject = null;
	editable = true;
	damagedArea = null;
	scaleFactorX = 1.0;
	scaleFactorY = 1.0;
	lastMouseX = 0;
	lastMouseY = 0;
	mouseDragged = false;
	
	opMouseListener = null;
	opSourceEventObjs = new LinkedList();
	opCurrentObject = null;
	opCurrentCollection = null;
	opLastMouseX = 0;
	opLastMouseY = 0;
	opDraggingObject = false;
	opPressedObject = null;
	setCursor(PropertyManager.APPLICATION_CURSOR);

	PieMenuBar.getPieMenuBar().addPieMenuTo(this);
    }


    public Storyboard(Storyboard copy) {
	this(copy, copy.model);
    }

    public Storyboard(Storyboard copy, DataModel parentModel) {
	this(copy, parentModel, true);
    }

    public Storyboard(Storyboard copy, DataModel parentModel, boolean copyOfModels) {
	setLayout(null);
	setBackground(PropertyManager.STORYBOARD_BACKGROUND_COLOR);

	model = parentModel;

	// If the copy had an assembler, then create one for this storyboard
	assembler = null;
	if (copy.assembler != null) {
	    assembler = new StrokeAssembler(this);
	}

	gestureRecognizer = copy.gestureRecognizer;
	inkRecognizer = copy.inkRecognizer;

	dropTargetListener = null;
	if (copy.dropTargetListener != null) {
	    dropTargetListener = new StoryboardDropTargetListener(this);
	}

	keyListener = null;
	if (copy.keyListener != null) {
	    keyListener = new StoryboardKeyListener(this);
	    addKeyListener(keyListener);
	}

	// Copy the graphical collection objects
	graphicalObjects = (GraphicalObjectCollection) copy.graphicalObjects.clone(this, copyOfModels);
	activeTextObject = null;
	editable = copy.editable;
	damagedArea = null;
	scaleFactorX = copy.scaleFactorX;
	scaleFactorY = copy.scaleFactorY;
	lastMouseX = 0;
	lastMouseY = 0;
	mouseDragged = false;

	// Operational section
	opMouseListener = null;
	opSourceEventObjs = new LinkedList();
	opCurrentObject = null;
	opCurrentCollection = null;
	opLastMouseX = 0;
	opLastMouseY = 0;
	opDraggingObject = false;
	opPressedObject = null;
	setCursor(PropertyManager.APPLICATION_CURSOR);

	PieMenuBar.getPieMenuBar().addPieMenuTo(this);
    }


    
    public Storyboard(DataInputStream stream, DataModel parentModel) throws IOException {
	setLayout(null);
	setBackground(PropertyManager.STORYBOARD_BACKGROUND_COLOR);

	model = parentModel;
	assembler = new StrokeAssembler(this);
	gestureRecognizer = Storyboard.defaultGestureRecognizer;
	inkRecognizer = Storyboard.defaultInkRecognizer;

	// Install the listeners
	dropTargetListener = new StoryboardDropTargetListener(this);
	keyListener = new StoryboardKeyListener(this);
	addKeyListener(keyListener);

	// Read the graphical objects from the stream
	graphicalObjects = new GraphicalObjectCollection(stream, this);
	activeTextObject = null;
	editable = true;
	damagedArea = null;
	scaleFactorX = 1.0;
	scaleFactorY = 1.0;
	lastMouseX = 0;
	lastMouseY = 0;
	mouseDragged = false;
	
	opMouseListener = null;
	opSourceEventObjs = new LinkedList();
	opCurrentObject = null;
	opCurrentCollection = null;
	opLastMouseX = 0;
	opLastMouseY = 0;
	opDraggingObject = false;
	opPressedObject = null;
	setCursor(PropertyManager.APPLICATION_CURSOR);
    }


    public void completeIO() {
	// Maps ids to objects 
	graphicalObjects.completeIO();
	PieMenuBar.getPieMenuBar().addPieMenuTo(this);
	
    }




    // Public Methods

    // Get/Set methods

    public DataModel getModel() {
	return model;
    }


    public Recognizer getGestureRecognizer() {
	return gestureRecognizer;
    }


    public Recognizer getInkRecognizer() {
	return inkRecognizer;
    }


    public StoryboardKeyListener getStoryboardKeyListener() {
	return keyListener;
    }


    public boolean isEditable() {
	return editable;
    }

    public void setEditable(boolean value) {
	editable = value;
	if (! editable) {
	    endTextEntry();

	    if (assembler != null) {
		assembler.removeStrokeAssembler();
		assembler = null;
	    }
	  
	    if (dropTargetListener != null) {
		dropTargetListener = null;
	    }

	    if (keyListener != null) {
		removeKeyListener(keyListener);
		keyListener = null;
	    }

	} else {
	    if (assembler == null) {
		assembler = new StrokeAssembler(this);
	    }

	    if (dropTargetListener == null) {
		dropTargetListener = new StoryboardDropTargetListener(this);
	    }

	    if (keyListener == null) {
		keyListener = new StoryboardKeyListener(this);
		addKeyListener(keyListener);
	    }
	}
    }


    public int getCurrentLayer() {
	int layer = 0;
	if (model != null) {
	    layer = model.getSelectedLayer();
	}
	return layer;
    }


    public boolean isLayerVisible(int layer) {
	return model.isLayerVisible(layer);
    }

    public boolean isLayerSelected(int layer) {
	return model.isLayerSelected(layer);
    }



    // Layer routines
    public void newLayer(int newLayerNum) {
	endTextEntry();
	unSelectAllObjects();
    }

    public void copyLayer(int layerNum, int newLayerNum) {
	endTextEntry();
	graphicalObjects.copyLayer(layerNum, newLayerNum);
	ObjectPropertyFrame.getObjectPropertyFrame().updateProperties();
	repaint();
    }
    
    public void deleteLayer(int layerNum, int newSelectedLayer) {
	endTextEntry();
	graphicalObjects.deleteLayer(layerNum);
	unSelectAllObjects();
	repaint();
    }

    public void repositionLayer(int layerNum, int toPosition) {
	endTextEntry();
	graphicalObjects.repositionLayer(layerNum, toPosition);
	unSelectAllObjects();
	repaint();
    }

    public void flattenLayers(int nextSelectedLayer) {
	endTextEntry();
	graphicalObjects.flattenLayers();
	unSelectAllObjects();
	repaint();
    }

    public void setLayerVisible(int layerNum, boolean visible) {
	if (layerNum == model.getSelectedLayer()) {
	    endTextEntry();
	}
	graphicalObjects.setLayerVisible(layerNum, visible);
	repaint();
    }

    public void setLayerSelected(int layerNum, boolean selected) {
	endTextEntry();
	graphicalObjects.setLayerSelected(layerNum, selected);
	// Unselect calls repaint()
	unSelectAllObjects();
    }
    

    // Painting and related routines

    public void setDamagedArea(Rectangle r) {
	damagedArea = r;
    }


    public boolean isOptimizedDrawingEnabled() {
	return false;
    }
    
    // Repaint entire storyboard by default
    public void paintComponent(Graphics g) {
	// Paint the parent
	super.paintComponent(g);

	// Paint the storyboard
	g.setColor(PropertyManager.STORYBOARD_BACKGROUND_COLOR);
	g.fillRect(0, 0, getWidth(), getHeight());
	paintStoryboardComponent((Graphics2D) g);
    }


    public void paintComponent() {
	paintImmediately(0, 0, getWidth(), getHeight());
    }


    public void paintStoryboardComponent(Graphics2D g) {
	if (damagedArea == null) {
	    // Render all graphical objects
	    graphicalObjects.render(g);
	    
	} else {
	    // Clear the damaged area
	    g.setColor(PropertyManager.STORYBOARD_BACKGROUND_COLOR);
	    g.fillRect(damagedArea.x, damagedArea.y,
		       damagedArea.width, damagedArea.height);
	    
	    // Now paint only the objects that overlap (intersect) this area
	    // Paint objects in all layers that intersect
	    graphicalObjects.render(getBoundsIntersectingObjects(damagedArea, model.getNumLayers(), true), g);
	    damagedArea = null;
	}
    }


    /*
      Very simple heuristic to force ink strokes to paint on top of widegts. In some cases, this procedure will produce visual artifacts (incorrect drawing), however, it is appropriate for demonstrational purposes. A more integrated solution would be a valuable future project
    */
    
    public void paintChildren(Graphics g) {
	super.paintChildren(g);

	Graphics2D graphics = (Graphics2D) g;
	Component children[] = getComponents();
	for (int i=0; i<children.length; ++i) {
	    if (children[i] instanceof JComponent) {
		JComponent jChildComponent = (JComponent) children[i];
		GraphicalObject childObject = (GraphicalObject) jChildComponent.getClientProperty("DEMAIS");
		if (childObject != null) {
		    Rectangle childBounds = childObject.getBounds();
		    int childLayer = childObject.getLayer();
		    ListIterator it = graphicalObjects.listIterator(graphicalObjects.indexOf(childObject));
		    while (it.hasNext()) {
			GraphicalObject go = (GraphicalObject) it.next();
			if (childObject.strokeIntersects(go.getBounds())) {
			    go.render(graphics);
			}   
		    }
		} 
	    }
	}
    }


    public void superPaintChildren(Graphics g) {
	super.paintChildren(g);
    }


    // Paint the contents of the storyboard to the printer
    public int print(Graphics g, PageFormat pf, int pi) throws PrinterException {
	Graphics2D g2d = (Graphics2D)g;
	g2d.translate(pf.getImageableX(), pf.getImageableY());
	RepaintManager currentManager = RepaintManager.currentManager(this);
	currentManager.setDoubleBufferingEnabled(false);	
	g2d.scale(0.65, 0.7);
	paintStoryboardComponent(g2d);
	paintChildren(g2d);
	currentManager.setDoubleBufferingEnabled(true);
	return Printable.PAGE_EXISTS;
    }



    // Mouse callback routines
    public void onMousePressed(MouseEvent evt, TimedStroke ts) {
	lastMouseX = evt.getX();
	lastMouseY = evt.getY();
	endTextEntry();
	mouseDragged = false;
    }


    public void onMouseDragged(MouseEvent evt) {
	int x = evt.getX();
	int y = evt.getY();
	translateAndUpdate(getSelectedGraphicalObjects(getCurrentLayer()), x - lastMouseX, y - lastMouseY);
	lastMouseX = x;
	lastMouseY = y;
	mouseDragged = true;
    }

    public void onMouseDragged(TimedStroke ts) {
	drawCurrentStroke(ts);	
    }


    public void onMouseReleased(MouseEvent evt, TimedStroke ts) {
	if (mouseDragged) {
	    mouseDragged = false;
	    return;
	}

	// Check for a pen tap
	if (ts.npoints < PropertyManager.APPLICATION_MAX_TAP_POINTS) {
	    // Pen tap
	    if (ts.isLeftMouseButton()) {
		RectangleInkStroke ris = (RectangleInkStroke) getRectangleContaining(evt.getX(), evt.getY(), getCurrentLayer(), false);
		if (ris != null) {
		    loadMediaIntoRectangle(ris);
		    PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);

		} else {
		    // Repaint called later
		    unSelectAllObjects(false);
		}

	    } else {
		// Right button
		unSelectAllObjects(false);
	    }

	} else if (ts.isLeftMouseButton()) {
	    InkStroke go = null;
	    if ((! model.getSelectedLayerModel().isAnnotationLayer()) &&
		((!(Toolkit.getDefaultToolkit().getLockingKeyState(KeyEvent.VK_CAPS_LOCK))) &&
		 (!evt.isControlDown())) &&
		(inkRecognizer.isRectangle(new Gesture(ts)))) {

		// Sketched object was a rectangle!
		go = new RectangleInkStroke(ts, getCurrentLayer(), isLayerVisible(getCurrentLayer()), false, this);
		// Flash the rectangle to inform the user of the classification
		flashInkStroke(go);
		
	    } else {
		// Just a normal ink stroke
		go = new InkStroke(ts, getCurrentLayer(), isLayerVisible(getCurrentLayer()), false, this);
	    }

	    // Add the graphical object to the list
	    addGraphicalObject(go);

	    // Perform an analysis on the object, if we are not in an annotation layer and caps lock is not on
	    // and CTRL is not being held down
	    if ((!model.getSelectedLayerModel().isAnnotationLayer()) && 
		(!(Toolkit.getDefaultToolkit().getLockingKeyState(KeyEvent.VK_CAPS_LOCK))) &&
		(!evt.isControlDown())) {
		
		go.setPerformAnalysis(true);
		performAnalysis(go);
	   
	    } else {
		go.setPerformAnalysis(false);
	    }
	   
	    PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);

	} else {
	    // Must be a gesture command
	    Gesture gesture = new Gesture(ts);
	    GestureCommand cmd = new GestureCommand(gesture, this);
	    cmd.performAction();
	}

	repaint();
    }


    public void cancelStroke() {
	if (assembler != null) {
	    assembler.cancelStroke();
	    repaint();
	}
    }


    public void forceRectangleRecognition(GraphicalObject go, boolean value) {
	if (go instanceof RectangleInkStroke) {
	    if (! value) {
		// Force current rectangle back to a normal ink stroke
		InkStroke inkStroke = new InkStroke((InkStroke)go);
		int index = graphicalObjects.indexOf(go);
		deleteGraphicalObject(go);

		if (index >= 0) {
		    addGraphicalObject(index, inkStroke);
		} else {
		    addGraphicalObject(inkStroke);
		}

		flashInkStroke(inkStroke);
		repaint();
	    }

	} else if (go instanceof InkStroke) {
	    if (value) {
		InkStroke inkStroke = (InkStroke) go;
		// Force current ink stroke to be a rectangle
		RectangleInkStroke ris = new RectangleInkStroke(inkStroke.getTimedStroke(), inkStroke.getLayer(), inkStroke.isVisible(), inkStroke.isSelected(), this);

		int index = graphicalObjects.indexOf(inkStroke);
		deleteGraphicalObject(inkStroke);

		if (index >= 0) {
		    addGraphicalObject(index, ris);
		} else {
		    addGraphicalObject(ris);
		}

		flashInkStroke(ris);
		repaint();
	    }
	}
    }


    public void onVisualLanguageIconDropped(MouseEvent evt, int code, BehavioralInkStroke bStroke) {
	endTextEntry();
	if (dropTargetListener != null) {
	    dropTargetListener.handleDrop(code, evt.getX(), evt.getY(), bStroke);
	}
    }


    public void onHoldAndPress(MouseEvent evt) {
	endTextEntry();
	int x = evt.getX();
	int y = evt.getY();

	if (SwingUtilities.isLeftMouseButton(evt) && (!evt.isMetaDown())) {
	    GraphicalObject go = getTextContaining(x, y, getCurrentLayer(), false);
	    if (go != null) {
		// Unselect all objects except for this one, if its selected
		boolean selected = go.isSelected();
		unSelectAllObjects(false);
		go.setSelected(selected);

		// Asynchronous repaint causes problem with text
		paintImmediately(0, 0, getWidth(), getHeight());

		// Edit this text object
		activeTextObject = (TextGraphicalObject) go;
		activeTextObject.editText(x, y);

	    } else {
		unSelectAllObjects(false);
		// Asynchronous repaint causes problem with text
		paintImmediately(0, 0, getWidth(), getHeight());

		// Create a new Text object
		activeTextObject = new TextGraphicalObject(x, y, getCurrentLayer(), true, false, this);
		activeTextObject.setLayer(getCurrentLayer());
		addGraphicalObject(activeTextObject);
		activeTextObject.setPerformAnalysis(true);
	    }

	}
    }


    // Popup a file browser to enter an image or video object
    private void loadMediaIntoRectangle(RectangleInkStroke ris) {	
	String file = showAudioImageVideoFileBrowser();
	if (file != null) {
	    int index = graphicalObjects.indexOf(ris);
	    GraphicalObject newGO = null;
	    
	    if (FileExtension.isImageFile(file)) {
		newGO = new ImageRectangleInkStroke(ris, file);
		
	    } else if (FileExtension.isVideoFile(file)) {
		newGO = new VideoRectangleInkStroke(ris, file);
		
	    } else if (FileExtension.isAudioFile(file)) {
		newGO = new AudioRectangleInkStroke(ris, file);
		
	    } else {
		DEMAISError.showErrorMessage("Unknown media type", "Media file error", false);
	    }

	    if (newGO != null) {
		deleteGraphicalObject(ris);
		if (index >= 0) {
		    addGraphicalObject(index, newGO);
		} else {
		    addGraphicalObject(newGO);
		}
	    }
	}
    }




    // Graphical object manipulation routines
    public GraphicalObjectCollection getGraphicalObjectCollection() {
	return graphicalObjects;
    }

    public void setGraphicalObjectCollection(GraphicalObjectCollection goc) {
	graphicalObjects = goc;
    }




    // The storyboard is about to be installed into the ContentPanel
    // (switched in)
    public void onInstall() {
	realize(true, true);
    }


    // The storyboard is about to be removed from the ContentPanel
    // (switched out)
    public void onUnInstall() {
	endTextEntry();

	// Stop any media that is playing
	unRealize(false);
	unSelectAllObjects(false);
    }


    public void onRuntimeUnInstall(DataModel toModel) {
	if ( (toModel != null) && (toModel instanceof StoryboardModel)) {
	    onRuntimeUnInstall((StoryboardModel) toModel);
	}
    }

    public void onRuntimeUnInstall(StoryboardModel toModel) {
	opCurrentObject = null;
	opCurrentCollection = null;
	opDraggingObject = false;
	opPressedObject = null;
	setCursor(OP_DEFAULT_CURSOR);

	if (model.equals(toModel)) {
	    return;
	}

	Iterator it = graphicalObjects.iterator(); 
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.setMouseIsOver(false);
	    go.setIsHighlighted(false);
	    go.resetToOriginalLocation();
	    if (go.isDraggable()) {
		addSourceEventObject(go);
	    }
	    if (go instanceof ContinuousMediaRectangleInkStroke) {
		ContinuousMediaRectangleInkStroke from_cmris = (ContinuousMediaRectangleInkStroke) go;
		String fromFile = from_cmris.getMediaFilePathName();

		Iterator it2 = toModel.getStoryboard().getGraphicalObjectCollection().iterator();
		while (it2.hasNext()) {
		    GraphicalObject go2 = (GraphicalObject) it2.next();
		    if (go2 instanceof ContinuousMediaRectangleInkStroke) {

			ContinuousMediaRectangleInkStroke to_cmris = (ContinuousMediaRectangleInkStroke) go2;
			String toFile = to_cmris.getMediaFilePathName();
			if (fromFile.equals(toFile)) {
			    // Same continuous media object - copy the state
			    to_cmris.setState(from_cmris.getState(), from_cmris.getMediaTime());
			}
		    }
		}
		// Unrealize (stop) the one being switched out
		from_cmris.unRealize(true);
	    
	    }
	}
    }


    public void onRuntimeInstall() {
	setCursor(OP_DEFAULT_CURSOR);
	realize(false, true);
    }



    // Realize any continuous media content
    public void realize(boolean showControls, boolean recreate) {
	graphicalObjects.realize(showControls, recreate);
    }

    // Unrealize any continuous media content
    public void unRealize(boolean keepVisuals) {
	graphicalObjects.unRealize(keepVisuals);
    }




    // Geometry  methods
    public Collection getBoundsIntersectingObjects(Rectangle r, int layer, boolean searchLowerLayers) {
	// Will potentially return the begin anchor as part of the list
	return graphicalObjects.getBoundsIntersectingObjects(r, layer, searchLowerLayers);
    }

    public Collection getRectangleOrTextBoundsIntersectingObjects(Rectangle r, int layer, boolean searchLowerLayers) {
	return graphicalObjects.getRectangleOrTextBoundsIntersectingObjects(r, layer, searchLowerLayers);
    }

    public Collection getStrokeIntersectingOrContainedObjects(Rectangle r, int layer, boolean searchLowerLayers) {
	// For cut gesture operation
	return graphicalObjects.getStrokeIntersectingOrContainedObjects(r, layer, searchLowerLayers);
    }

    
    public BehavioralInkStroke getBehavioralInkStrokeContaining(int x, int y, boolean bySource, int layer, boolean searchLowerLayers) {
	return graphicalObjects.getBehavioralInkStrokeContaining(x, y, bySource, layer, searchLowerLayers);
    }
    
    public BehavioralInkStroke getBehavioralInkStrokeContaining(Point p, boolean bySource, int layer, boolean searchLowerLayers) {
	return getBehavioralInkStrokeContaining(p.x, p.y, bySource, layer, searchLowerLayers);
    }



    // Containment methods
    public GraphicalObject getRectangleContaining(int x, int y, int layer, boolean searchLowerLayers) {
	return graphicalObjects.getRectangleContaining(x, y, layer, searchLowerLayers);
    }


    public GraphicalObject getRectangleOrTextContaining(int x, int y, int layer, boolean searchLowerLayers) {
	return graphicalObjects.getRectangleOrTextContaining(x, y, layer, searchLowerLayers);
    }

    public Vector getRectangleOrTextContainingBIS(int x, int y, int layer, boolean searchLowerLayers) {
	return graphicalObjects.getRectangleOrTextContainingBIS(x, y, layer, searchLowerLayers);
    }


    public GraphicalObject getTextContaining(int x, int y, int layer, boolean searchLowerLayers) {
	return graphicalObjects.getTextContaining(x, y, layer, searchLowerLayers);
    }



    public Vector getObjectsContainedBy(Rectangle r, int layer, boolean searchLowerLayers) {
	return graphicalObjects.getObjectsContainedBy(r, layer, searchLowerLayers);
    }



    // Graphical object methods
    public GraphicalObject getGraphicalObjectById(int id) {
	return graphicalObjects.getGraphicalObjectById(id);
    }


    public void addGraphicalObject(GraphicalObject go) {
	graphicalObjects.add(go);
    }

    public void addGraphicalObject(int index, GraphicalObject go) {
	graphicalObjects.add(index, go);
    }

    public void deleteGraphicalObject(GraphicalObject go) {
	graphicalObjects.delete(go);
    }


    public void deleteGraphicalObjects(Collection c) {
	deleteGraphicalObjects(c, true);
    }
    
    public void deleteGraphicalObjects(Collection c, boolean rePaint) {
	graphicalObjects.deleteAll(c);
	if (rePaint) {
	    repaint();
	}
    }


    public void unSelectAllObjects() {
	unSelectAllObjects(true);
    }

    public void unSelectAllObjects(boolean paint) {
	graphicalObjects.unSelectAllObjects();
	ObjectPropertyFrame.getObjectPropertyFrame().updateProperties(null);
	PropertyManager.setProperty(PropertyManager.STORYBOARD_OBJECT_SELECTED, Boolean.FALSE);
	if (paint) {
	    repaint();
	}
    }


    public void selectGraphicalObjects(Collection c) {
	// First, unselect any selected objects
	unSelectAllObjects();

	Iterator it = c.iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.setSelected(true);
	}
	if (c.size() > 0) {
	    PropertyManager.setProperty(PropertyManager.STORYBOARD_OBJECT_SELECTED, Boolean.TRUE);
	}
	ObjectPropertyFrame.getObjectPropertyFrame().updateProperties(c);
	repaint();
    }


    public Collection getSelectedGraphicalObjects() {
	return getSelectedGraphicalObjects(getCurrentLayer());
    }

    public Collection getSelectedGraphicalObjects(int layer) {
	return (graphicalObjects.getSelectedGraphicalObjects(layer));
    }



    public boolean isGraphicalObjectSelected() {
	return isGraphicalObjectSelected(getCurrentLayer());
    }

    public boolean isGraphicalObjectSelected(int layer) {
	return graphicalObjects.isGraphicalObjectSelected(layer);
    }


    public void setVisibleOnObjectsContainedBy(Rectangle r, boolean visible, int layer, boolean searchLowerLayers, GraphicalObject exempt) {
	graphicalObjects.setVisibleOnObjectsContainedBy(r, visible, layer, searchLowerLayers, exempt);
    }

    public void setMouseIsOverOnObjectsContainedBy(Rectangle r, boolean isOver, int layer, boolean searchLowerLayers, GraphicalObject exempt) {
	graphicalObjects.setMouseIsOverOnObjectsContainedBy(r, isOver, layer, searchLowerLayers, exempt);
    }




    // CUT and PASTE operations
    public Collection cutSelectedGraphicalObjects() {
	Collection c = getSelectedGraphicalObjects(getCurrentLayer());
	cutGraphicalObjects(c);
	return c;
    }


    public Collection cutGraphicalObjects(Collection c) {
	graphicalObjects.deleteAll(c);
	ObjectPropertyFrame.getObjectPropertyFrame().updateProperties();
	PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
	PropertyManager.setProperty(PropertyManager.STORYBOARD_OBJECT_SELECTED, Boolean.FALSE);
	repaint();
	return c;
    }


    
    // Returns a reference to the newly pasted objects
    public Collection pasteGraphicalObjects(Collection c) {
	return pasteGraphicalObjects(c, 0, 0, this);
    }
    
    public Collection pasteGraphicalObjects(Collection c, int deltaX, int deltaY, Storyboard parentS) {
	Collection pasted = graphicalObjects.pasteObjects(c, deltaX, deltaY, getCurrentLayer(), true, parentS);
	if (pasted.size() > 0) {
	    ObjectPropertyFrame.getObjectPropertyFrame().updateProperties(pasted);
	    PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
	    PropertyManager.setProperty(PropertyManager.STORYBOARD_OBJECT_SELECTED, Boolean.TRUE);
	    repaint();
	}
	return pasted;
    }
    


    public void changeFontSize(Collection c, int delta) {
	Iterator it = c.iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go instanceof TextGraphicalObject) {
		TextGraphicalObject tgo = (TextGraphicalObject) go;
		tgo.changeFontSize(delta);
	    }
	}
    }


    // Scaling and translation methods

    public void resetToOriginalScale() {
	graphicalObjects.resetToOriginalScale();
	repaint();
    }


    public void scaleTo(int width, int height) {
	scaleTo(width, height, 0);
    }

    public void scaleTo(int width, int height, int vAdjustment) {
	double sx = ( (double) width  / (double) WIDTH);
	double sy = ( (double) (height - vAdjustment) / (double) HEIGHT);
	scale(sx, sy);
    }

    
    public void scale(int percent) {
	scale((double)percent);
    }

    public void scale(double percent) {
	scale(percent / 100.0, percent / 100.0);
    }

    public void scale(double sx, double sy) {
	graphicalObjects.scale(sx, sy);
	scaleFactorX = sx;
	scaleFactorY = sy;
    }


    public void translateAndUpdate(Collection c, int deltaX, int deltaY) {
	if (c != null) {
	    // Use screen's width and height!
	    int mWidth = getWidth();
	    int mHeight = getHeight();
	    
	    // Translate as a group!!
	    boolean canTranslateGroup = true;
	    Iterator it = c.iterator();
	    while(it.hasNext()) {
		GraphicalObject go = (GraphicalObject) it.next();
		if (! go.canTranslate(deltaX, deltaY, mWidth, mHeight)) {
		    canTranslateGroup = false;
		    break;
		}
	    }
	    
	    if (canTranslateGroup) {
		Rectangle r = null;
		it = c.iterator();
		while(it.hasNext()) {
		    GraphicalObject go = (GraphicalObject) it.next();
		    if (r != null) {
			r = r.union(go.getBounds());
		    } else {
			r = go.getBounds();
		    }
		    go.translate(deltaX, deltaY, mWidth, mHeight);
		}
		setDamagedArea(r);
		if (r != null) {
		    paintImmediately(r.x, r.y, r.width, r.height);
		} else {
		    paintImmediately(0, 0, mWidth, mHeight);
		}
		PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);
	    }
	}
    }



    public void deleteStoryboard() {
	graphicalObjects.deleteAll();
    }

    public void deleteView() {
	deleteStoryboard();
    }


    public Object clone(DataModel parentModel, boolean copyOfModels) {
	endTextEntry();
	return (new Storyboard(this, parentModel, copyOfModels));
    }
    


    // Protected methods
    protected void drawCurrentStroke(TimedStroke ts) {
	drawCurrentStroke((Graphics2D)this.getGraphics(), ts);
    }

    protected void drawCurrentStroke(Graphics2D g, TimedStroke ts) {
	if (ts != null) {
	    ts.setColor(g);
	    ts.setStroke(g);

	    if (ts.getNumPoints() > 1) {
		Point p1 = ts.getNextToLastPoint();
		Point p2 = ts.getLastPoint();
		if ((p1 != null) && 
		    (p2 != null)) {
		    g.drawLine(p1.x, p1.y, p2.x, p2.y);
		}
	    }
	}
    }



    // Private methods
    private String showAudioImageVideoFileBrowser() {
	String lastDir = (String) PropertyManager.getProperty(PropertyManager.APPLICATION_LAST_MEDIA_DIRECTORY);
	// Create chooser starting at last directory used
	DFileChooser chooser = new DFileChooser(lastDir);
	chooser.setCursor(PropertyManager.APPLICATION_CURSOR);

	int returnVal = chooser.showOpenDialog(this);	
	if (returnVal == JFileChooser.APPROVE_OPTION) {
	    String selectedDir = chooser.getCurrentDirectory().getAbsolutePath();
	    // Store the directory for the next time
	    PropertyManager.setProperty(PropertyManager.APPLICATION_LAST_MEDIA_DIRECTORY, selectedDir);	    
	    return chooser.getSelectedFile().getAbsolutePath();
	}
	return null;
    }



    private void flashInkStroke(InkStroke inkStroke) {
	try {
	    Rectangle r = inkStroke.getBasicBounds();
	    Graphics2D g = (Graphics2D) this.getGraphics();

	    // Allow user to see ink stroke for a moment
	    Thread.sleep(50);

	    // Erase ink stroke
	    g.setColor(PropertyManager.STORYBOARD_BACKGROUND_COLOR);
	    g.fillRect(r.x, r.y, r.width, r.height);

	    // Leave ink stroke erased for a moment
	    Thread.sleep(100);

	    // Redraw the ink stroke
	    inkStroke.render(g);

	} catch(Exception e) {
	}
    }


    private void endTextEntry() {
	if (activeTextObject != null) {
	    activeTextObject.endTextEntry();
	    activeTextObject = null;
	}
    }


    public boolean equals(Object obj) {
	return (this == obj);
    }



    
    protected GraphicalObject mediateSelection(Vector cand) {
	if (cand.size() == 0) {
	    return null;
	    // stop mediation for now
	} else if (cand.size() >= 1) {
	    return (GraphicalObject) cand.firstElement();
	    
	} else {
	    // Ask the user
	    StoryboardMediatorFrame frame = new StoryboardMediatorFrame(cand);
	    return frame.getSelectedObject();
	}
    }





    /* 
       --------------------------------------------------------------
       OPERATIONAL SECTION
       --------------------------------------------------------------
    */
    
   
    public void installRuntimeMouseListeners() {
	opMouseListener = new StoryboardOpMouseListener(this);
    }


    public void unInstallRuntimeMouseListeners() {
	if (opMouseListener != null) {
	    opMouseListener.removeListener();
	}
	opMouseListener = null;
    }


    public void addSourceEventObject(GraphicalObject obj) {
	if (! opSourceEventObjs.contains(obj)) {
	    opSourceEventObjs.add(obj);
	}
    }

    public void removeSourceEventObject(GraphicalObject obj) {
	opSourceEventObjs.remove(obj);
    }

    public LinkedList getSourceEventObjects() {
	return opSourceEventObjs;
    }

  
    public GraphicalObject getSourceEventObjectContaining(int x, int y) {
	ListIterator it = opSourceEventObjs.listIterator(opSourceEventObjs.size());
	while (it.hasPrevious()) {
	    GraphicalObject go = (GraphicalObject) it.previous();
	    if ((go.isVisible()) &&
		(go.contains(x, y))) {
		return go;
	    }
	}
	return null;
    }



 
    /*
      Analyze the graphical objects for behavior. Then, convert the behaviors and left over text statements into Annotations and add them to the owning storyboard model.
    */
    public void operationalize() {
	endTextEntry();
	setEditable(false);

	graphicalObjects.operationalize();

	installRuntimeMouseListeners();
	damagedArea = null;
	setCursor(OP_DEFAULT_CURSOR);
   }


    public void unOperationalize() {
	// Remove the source objects
	opSourceEventObjs.clear();

	opCurrentObject = null;
	opCurrentCollection = null;
	unInstallRuntimeMouseListeners();
	graphicalObjects.unOperationalize();
	setEditable(true);
	setCursor(PropertyManager.APPLICATION_CURSOR);
    }

  
    public void performAnalysis(InkStroke iStroke) {
	// See if this object intersects others		
	Point sourcePoint = iStroke.getFirstPoint();
	GraphicalObject sourceObject = mediateSelection(getRectangleOrTextContainingBIS(sourcePoint.x, sourcePoint.y, iStroke.getLayer(), true));
	
	Point destPoint = iStroke.getLastPoint();
	GraphicalObject destObject = mediateSelection(getRectangleOrTextContainingBIS(destPoint.x, destPoint.y, iStroke.getLayer(), true));
	
	Rectangle rectangle = iStroke.getBasicBounds();
	if ( (sourceObject != null) && 
	     (destObject != null) &&
	     (!(destObject instanceof BeginAnchor)) &&
	     ((! sourceObject.equals(destObject)) ||
	      (sourceObject instanceof ContinuousMediaRectangleInkStroke)) &&
	     (! sourceObject.contains(rectangle)) &&
	     (! destObject.contains(rectangle))) {
	    
	    SourceEvent sourceEvent = getDefaultSourceEvent(sourceObject, destObject);
	    DestinationAction action = getDefaultAction(sourceObject, destObject);

	    BehavioralInkStroke bStroke = new BehavioralInkStroke(iStroke, sourceEvent, action);
	    deleteGraphicalObject(iStroke);
	    addGraphicalObject(bStroke);
	}
    }



    public SourceEvent getDefaultSourceEvent(GraphicalObject sourceEventObj, GraphicalObject actionObj) {
	SourceEvent sourceEvent = null;
	if (sourceEventObj instanceof ContinuousMediaRectangleInkStroke) {
	    ContinuousMediaRectangleInkStroke cmris = (ContinuousMediaRectangleInkStroke) sourceEventObj;
	    double mediaTime = cmris.getMediaTime();
	    double duration = cmris.getDuration();

	    if (mediaTime <= 1.0) {
		sourceEvent = new BeginMediaEvent(cmris);
	    } else if (mediaTime <= (duration - 1.0)) {
		sourceEvent = new SyncMarkerEvent(cmris, mediaTime);
	    } else {
		sourceEvent = new EndMediaEvent(cmris);
	    }

	} else if (sourceEventObj instanceof RectangleInkStroke) {
	    // Handle ImageRectangleInkStroke as well

	    if (actionObj instanceof TextGraphicalObject) {
		sourceEvent = new RolloverEvent((RectangleInkStroke)sourceEventObj);
	    } else if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		sourceEvent = new SingleMouseClickEvent((RectangleInkStroke)sourceEventObj);
	    } else if (actionObj instanceof RectangleInkStroke) {
		sourceEvent = new DragAndDropEvent((RectangleInkStroke)sourceEventObj, (RectangleInkStroke)actionObj);
	    }
	
	} else if (sourceEventObj instanceof TextGraphicalObject) {

	    if (actionObj instanceof TextGraphicalObject) {
		sourceEvent = new RolloverEvent((TextGraphicalObject)sourceEventObj);
	    } else if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		sourceEvent = new RolloverEvent((TextGraphicalObject)sourceEventObj);
	    } else if (actionObj instanceof RectangleInkStroke) {
		sourceEvent = new DragAndDropEvent((TextGraphicalObject)sourceEventObj, (RectangleInkStroke)actionObj);
	    }
	
	} else if (sourceEventObj instanceof BeginAnchor) {
	    sourceEvent = new TemporalEvent(sourceEventObj);
	}

	return sourceEvent;
    }



    public DestinationAction getDefaultAction(GraphicalObject sourceEventObj, GraphicalObject actionObj) {
	DestinationAction action = null;
	if (sourceEventObj instanceof ContinuousMediaRectangleInkStroke) {

	    if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		ContinuousMediaRectangleInkStroke cmris = (ContinuousMediaRectangleInkStroke) actionObj;
		double mediaTime = cmris.getMediaTime();
		double duration = cmris.getDuration();
		
		if (mediaTime <= 1.0) {
		    action = new BeginMediaAction(cmris);
		} else if (mediaTime <= (duration - 1.0)) {
		    action = new SetMediaTimeAction(cmris, mediaTime);
		} else {
		    action = new EndMediaAction(cmris);
		}

	    } else if (actionObj instanceof RectangleInkStroke) {
		action = new HighlightAction((RectangleInkStroke) actionObj);
	    
	    } else if (actionObj instanceof TextGraphicalObject) {
		
		TextGraphicalObject tgo = (TextGraphicalObject) actionObj;
		// What type of statement is it?
		Terminal terminal = null;
		if ((terminal = tgo.isNavigationAction()) != null) {
		    action = new NavigationAction(tgo);
		    
		} else if ((terminal = tgo.isCMAction()) != null) {
		    if (terminal.equals(Terminal.BEGIN)) {
			action = new BeginMediaAction(tgo);
		    } else if (terminal.equals(Terminal.END)) {
			action = new EndMediaAction(tgo);
		    } else if (terminal.equals(Terminal.PAUSE)) {
			action = new PauseMediaAction(tgo);
		    } 
		} else {
		    action = new DisplayAction(tgo);
		}
	    }
	    
	} else {
	    // Matches all other source object types

	    if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		ContinuousMediaRectangleInkStroke cmris = (ContinuousMediaRectangleInkStroke) actionObj;
		double mediaTime = cmris.getMediaTime();
		double duration = cmris.getDuration();
		
		if (mediaTime <= 1.0) {
		    action = new BeginMediaAction(cmris);
		} else if (mediaTime <= (duration - 1.0)) {
		    action = new SetMediaTimeAction(cmris, mediaTime);
		} else {
		    action = new EndMediaAction(cmris);
		}

	    } else if (actionObj instanceof TextGraphicalObject) {
		TextGraphicalObject tgo = (TextGraphicalObject) actionObj;
		// What type of statement is it?
		Terminal terminal = null;
		if ((terminal = tgo.isNavigationAction()) != null) {
		    action = new NavigationAction(tgo);
		    
		} else if ((terminal = tgo.isCMAction()) != null) {
		    if (terminal.equals(Terminal.BEGIN)) {
			action = new BeginMediaAction(tgo);
		    } else if (terminal.equals(Terminal.END)) {
			action = new EndMediaAction(tgo);
		    } else if (terminal.equals(Terminal.PAUSE)) {
			action = new PauseMediaAction(tgo);
		    }
		    
		}  else {
		    action = new DisplayAction(tgo);
		}
		

	    } else if (actionObj instanceof RectangleInkStroke) {
		// Handles Images as well
		action = new PlayAudioAction((RectangleInkStroke) actionObj, "Exclaim");
	    } 
	}
	
	return action;
    }
    


    public void handleMousePressed(MouseEvent evt) {
	if (opCurrentObject != null) {
	    opPressedObject = opCurrentObject;
	    if (opCurrentObject.isDraggable()) {
		setCursor(Storyboard.OP_MOVE_CURSOR);
	    }
	}
	opLastMouseX = evt.getX();
	opLastMouseY = evt.getY();
	opCurrentCollection = null;
    }


    public GraphicalObject getOpPressedObject() {
	return opPressedObject;
    }

    public GraphicalObject getOpCurrentObject() {
	return opCurrentObject;
    }

    public void handleMouseReleased(MouseEvent evt) {
	int x = evt.getX();
	int y = evt.getY();

	if (opCurrentObject != null) {
	    if (opCurrentObject.isDraggable()) {
		setCursor(OP_DEFAULT_CURSOR);
	    }
	    
	    if (opDraggingObject) {
		opCurrentObject.wasDraggedAndDropped();
		// Reset the current object the mouse is over
		handleMouseMoved(evt);
	    }
	}
	opDraggingObject = false;
    }


    public void handleMouseSingleClicked(MouseEvent evt) {
	if (opCurrentObject != null) {
	    opCurrentObject.mouseWasSingleClicked();
	}
    }


    public void handleMouseDoubleClicked(MouseEvent evt) {
	if (opCurrentObject != null) {
	    opCurrentObject.mouseWasDoubleClicked();
	}
    }


    public void handleMouseMoved(MouseEvent evt) {
	int x = evt.getX();
	int y = evt.getY();

	if (opCurrentObject != null) {
	    // Currently within an object

 	    if (! opCurrentObject.contains(x, y)) { 
		// Object no longer contains cursor - just moved out
		setCursor(OP_DEFAULT_CURSOR);
		opCurrentObject.mouseHasLeft();
		
		// Check if cursor is now within a different operational object
		opCurrentObject = getSourceEventObjectContaining(x, y);
		if (opCurrentObject != null) {
		    if (! opCurrentObject.isDraggable()) {
			setCursor(Storyboard.OP_HAND_CURSOR);
		    }
		    opCurrentObject.mouseHasEntered();
		}
	
	    } else {
		// Check if cursor is within a different operational object
		GraphicalObject go = getSourceEventObjectContaining(x, y);
		if ((go != null) && (! opCurrentObject.equals(go))) {
		    
		    // Mouse actually may not have left - the other object may be on top of it.
		    opCurrentObject.mouseHasLeft();
		    opCurrentObject = go;
		    if (! opCurrentObject.isDraggable()) {
			setCursor(Storyboard.OP_HAND_CURSOR);
		    }
		    opCurrentObject.mouseHasEntered();
		}	
	    }
	    // Still contains object - do nothing
	    
	} else {
	    // Check if cursor is within a new operational object
	    opCurrentObject = getSourceEventObjectContaining(x, y);
	    if (opCurrentObject != null) {
		if (! opCurrentObject.isDraggable()) {
		    setCursor(Storyboard.OP_HAND_CURSOR);
		}
		opCurrentObject.mouseHasEntered();
	    }	
	}
    }
    
    
    
    public void handleMouseDragged(MouseEvent evt) {
	int x = evt.getX();
	int y = evt.getY();
	opPressedObject = null;

	if (opCurrentObject != null) {
	    if (opCurrentObject.isDraggable()) {		    
		int deltaX = x - opLastMouseX;
		int deltaY = y - opLastMouseY;
		int width = getWidth();
		int height = getHeight();
		
		Rectangle bounds = opCurrentObject.getBounds();		
		if (opCurrentCollection == null) {
		    opCurrentCollection = getObjectsContainedBy(bounds, opCurrentObject.getLayer(), true);
		    if (opCurrentCollection != null) {
			opCurrentCollection.remove(opCurrentObject);
		    }
		}
		
		boolean canTranslate = opCurrentObject.canTranslate(deltaX, deltaY, width, height);
		if (canTranslate) {
		    if (opCurrentCollection != null) {
			Iterator it = opCurrentCollection.iterator();
			while (it.hasNext()) {
			    GraphicalObject go = (GraphicalObject) it.next();
			    go.opTranslate(deltaX, deltaY, width, height);
			}
		    }
		    opCurrentObject.opTranslate(deltaX, deltaY, width, height);
		    setDamagedArea(bounds);
		    paintImmediately(bounds.x, bounds.y, bounds.width, bounds.height);
		}
		
		opDraggingObject = true;
	    }
	}
	opLastMouseX = x;
	opLastMouseY = y;
    }

    
    public void opResetDraggableToOriginalLocation() {
	if (opCurrentObject != null) {
	    opCurrentObject.resetToOriginalLocation();
	}
	if (opCurrentCollection != null) {
	    Iterator it = opCurrentCollection.iterator();
	    while (it.hasNext()) {
		GraphicalObject go = (GraphicalObject) it.next();
		go.resetToOriginalLocation();
	    }
	}
    }


    public void opSnapDraggableToLocation(Rectangle targetLocation) {
	if (opCurrentObject != null) {
	    Rectangle opCurrentLoc = opCurrentObject.getBasicBounds();
	    
	    if (opCurrentCollection != null) {
		int offsetx = 0;
		int offsety = 0;
		Iterator it = opCurrentCollection.iterator();
		while (it.hasNext()) {
		    GraphicalObject go = (GraphicalObject) it.next();
		    Rectangle goLoc = go.getBasicBounds();
		    
		    offsetx = goLoc.x - opCurrentLoc.x;
		    offsety = goLoc.y - opCurrentLoc.y;
		    
		    go.doTranslate(targetLocation.x - goLoc.x + offsetx,
				   targetLocation.y - goLoc.y + offsety);
		}
	    }
	    
	    opCurrentObject.doTranslate(targetLocation.x - opCurrentLoc.x,
					targetLocation.y - opCurrentLoc.y);
	    
	}
    }



    public void opResetCMObjects(GraphicalObject exempt) {
	graphicalObjects.opResetCMObjects(exempt);
    }


    public void write(DataOutputStream stream) throws IOException {
	// Write the graphical objects
	graphicalObjects.write(stream);
    }


}
