/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File StoryboardDropTargetListener.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard;
    

// Necessary imports
import java.awt.*;
import javax.swing.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.core.parser.*;
import edu.uofminnesota.DEMAIS.usrinterface.visuallanguage.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.behavior.*;




// Storyboard class
public class StoryboardDropTargetListener {
    
    // The storyboard affected by the events
    private Storyboard storyboard;



    private StoryboardDropTargetListener() {
	this(null);
    }

    public StoryboardDropTargetListener(Storyboard pStoryboard) {
	storyboard = pStoryboard;
    }


    // Dropped a source event or action icon
    public void handleDrop(int code, int x, int y, BehavioralInkStroke bStroke) {
	if (VLButton.isVLIconASourceEvent(code)) {
	    handleSourceEventDrop(bStroke, code, x, y);

	} else {
	    handleActionDrop(bStroke, code, x, y);
	}

	PropertyManager.setProperty(PropertyManager.APPLICATION_SAVE_NEEDED, Boolean.TRUE);

	// Repaint the storyboard
	storyboard.repaint();
    }


    
    public void handleSourceEventDrop(BehavioralInkStroke bStroke, int code, int x, int y) {
	GraphicalObject sourceObj = bStroke.getSourceEventObject();
	if (sourceObj == null) {
	    Point sourcePoint = bStroke.getFirstPoint();
	    sourceObj = storyboard.getRectangleOrTextContaining(sourcePoint.x, sourcePoint.y, storyboard.getCurrentLayer(), true);
	}

	
	// Map the code to the event
	SourceEvent sEvent = null;
	switch (code) {
	    
	case VLButton.SINGLE_MOUSE_CLICK_EVENT:
	    if (sourceObj instanceof AudioRectangleInkStroke) {
		// Nothing
	    } else if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new SingleMouseClickEvent((RectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new SingleMouseClickEvent((TextGraphicalObject)sourceObj);
	    }
	    break;
	    
	case VLButton.DOUBLE_MOUSE_CLICK_EVENT:
	    if (sourceObj instanceof AudioRectangleInkStroke) {
		// Nothing
	    } else if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new DoubleMouseClickEvent((RectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new DoubleMouseClickEvent((TextGraphicalObject)sourceObj);
	    }     
	    break;
	    
	case VLButton.ROLLOVER_EVENT:
	    if (sourceObj instanceof AudioRectangleInkStroke) {
		// Nothing
	    } else if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new RolloverEvent((RectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new RolloverEvent((TextGraphicalObject)sourceObj);
	    }
	    break;
	    
	case VLButton.DRAG_AND_DROP_EVENT:
	    if ((sourceObj instanceof AudioRectangleInkStroke) ||
		(sourceObj instanceof VideoRectangleInkStroke)) {
		break;

	    } else {
		GraphicalObject actionObj = bStroke.getActionObject();
		if (actionObj == null) {
		    Point destPoint = bStroke.getLastPoint();
		    actionObj = storyboard.getRectangleOrTextContaining(destPoint.x, destPoint.y, storyboard.getCurrentLayer(), true);
		}
		
		if (sourceObj instanceof RectangleInkStroke) {
		    sEvent = new DragAndDropEvent((RectangleInkStroke)sourceObj, actionObj);
		} else if (sourceObj instanceof TextGraphicalObject) {
		    sEvent = new DragAndDropEvent((TextGraphicalObject)sourceObj, actionObj);
		}
	    }   
	    break;
	    
	case VLButton.BEGIN_MEDIA_EVENT:
	    if (sourceObj instanceof ContinuousMediaRectangleInkStroke) {
		sEvent = new BeginMediaEvent((ContinuousMediaRectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new BeginMediaEvent((TextGraphicalObject)sourceObj);
	    }
	    break;
	    
	case VLButton.END_MEDIA_EVENT:
	    if (sourceObj instanceof ContinuousMediaRectangleInkStroke) {
		sEvent = new EndMediaEvent((ContinuousMediaRectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new EndMediaEvent((TextGraphicalObject)sourceObj);
	    }
	    break;
	    
	case VLButton.ENTER_TEXT_EVENT:
	    if  ((sourceObj instanceof ContinuousMediaRectangleInkStroke) ||
		 (sourceObj instanceof ImageRectangleInkStroke)) {
		break;
	    } else if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new EnterTextEvent((RectangleInkStroke)sourceObj);
	    }
	    break;
	    
	case VLButton.TEMPORAL_EVENT:
	    if (sourceObj instanceof RectangleInkStroke) {
		sEvent = new TemporalEvent((RectangleInkStroke)sourceObj);
	    } else if (sourceObj instanceof TextGraphicalObject) {
		sEvent = new TemporalEvent((TextGraphicalObject)sourceObj);
	    } else if (sourceObj instanceof BeginAnchor) {
		sEvent = new TemporalEvent((BeginAnchor)sourceObj);
	    }
	    break;
	    
	case VLButton.SYNC_MARKER_EVENT:
	    if (sourceObj instanceof ContinuousMediaRectangleInkStroke) {
		sEvent = new SyncMarkerEvent((ContinuousMediaRectangleInkStroke)sourceObj);
	    } 
	    break;

	default:
	    break;
	}
	
	if (sEvent != null) {
	    bStroke.setSourceEvent(sEvent);
	} else {
	    Toolkit.getDefaultToolkit().beep();
	}
    }



    private void handleActionDrop(BehavioralInkStroke bStroke, int code, int x, int y) {
	GraphicalObject actionObj = bStroke.getActionObject();
	if (actionObj == null) {
	    Point destPoint = bStroke.getLastPoint();
	    actionObj = storyboard.getRectangleOrTextContaining(destPoint.x, destPoint.y, storyboard.getCurrentLayer(), true);
	}

	
	// Map the code to the event or action
	DestinationAction action = null;
	switch (code) {
	    
	case VLButton.NAVIGATE_ACTION:
	    if (actionObj instanceof TextGraphicalObject) {
		TextGraphicalObject tgo = (TextGraphicalObject) actionObj;
		// What type of statement is it?
		Terminal terminal = null;
		if ((terminal = tgo.isNavigationAction()) != null) {
		    action = new NavigationAction(tgo);   
		} else {
		    action = new NavigationAction((StoryboardModel)null, actionObj);
		    if (! action.hasDestModel()) {
			return;
		    }
		}

	    } else {
		action = new NavigationAction((StoryboardModel)null, actionObj);
		if (! action.hasDestModel()) {
		    return;
		}
	    }
	    break;
	    
	case VLButton.BEGIN_MEDIA_ACTION:
	    if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		action = new BeginMediaAction((ContinuousMediaRectangleInkStroke)actionObj);
	    } else if (actionObj instanceof TextGraphicalObject) {
		action = new BeginMediaAction((TextGraphicalObject)actionObj);   
	    }
	    break;
	    
	case VLButton.END_MEDIA_ACTION:
	    if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		action = new EndMediaAction((ContinuousMediaRectangleInkStroke)actionObj);
	    } else if (actionObj instanceof TextGraphicalObject) {
		action = new EndMediaAction((TextGraphicalObject)actionObj);
	    }
	    break;
	    
	case VLButton.PAUSE_MEDIA_ACTION:
	    if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		action = new PauseMediaAction((ContinuousMediaRectangleInkStroke)actionObj);
	    } else if (actionObj instanceof TextGraphicalObject) {
		action = new PauseMediaAction((TextGraphicalObject)actionObj);
	    }
	    break;
	 
	case VLButton.SET_MEDIA_TIME_ACTION:
	    if (actionObj instanceof ContinuousMediaRectangleInkStroke) {
		action = new SetMediaTimeAction((ContinuousMediaRectangleInkStroke)actionObj);
	    }
	    break;
	    
	case VLButton.TRANSITION_ACTION:
	    if (actionObj instanceof TextGraphicalObject) {
		TextGraphicalObject tgo = (TextGraphicalObject) actionObj;
		
		// What type of statement is it?
		Terminal terminal = null;
		if ((terminal = tgo.isNavigationAction()) != null) {
		    action = new TransitionAction(tgo);   
		} else {
		    action = new TransitionAction((StoryboardModel) null, actionObj);
		    if (! action.hasDestModel()) {
			return;
		    }
		}
		
	    } else {
		action = new TransitionAction((StoryboardModel) null, actionObj);
		if (! action.hasDestModel()) {
		    return;
		}
	    }
	    break;

	case VLButton.DELAY_ACTION:
	    action = new DelayAction(actionObj);
	    break;

	case VLButton.DISPLAY_ACTION:
	    if (actionObj instanceof RectangleInkStroke) {
		action = new DisplayAction((RectangleInkStroke)actionObj);
	    } else if (actionObj instanceof TextGraphicalObject) {
		action = new DisplayAction((TextGraphicalObject)actionObj);
	    }
	    break;
	  
	case VLButton.UNDISPLAY_ACTION:
	    if (actionObj instanceof RectangleInkStroke) {
		action = new UnDisplayAction((RectangleInkStroke)actionObj);
	    } else if (actionObj instanceof TextGraphicalObject) {
		action = new UnDisplayAction((TextGraphicalObject)actionObj);
	    }
	    break; 
	    
	case VLButton.HIGHLIGHT_ACTION:
	    if (actionObj instanceof RectangleInkStroke) {
		action = new HighlightAction((RectangleInkStroke)actionObj);
	    } else if (actionObj instanceof TextGraphicalObject) {
		action = new HighlightAction((TextGraphicalObject)actionObj);
	    }
	    break;
	    
	case VLButton.UNHIGHLIGHT_ACTION:
	    if (actionObj instanceof RectangleInkStroke) {
		action = new UnHighlightAction((RectangleInkStroke)actionObj);
	    } else if (actionObj instanceof TextGraphicalObject) {
		action = new UnHighlightAction((TextGraphicalObject)actionObj);
	    }
	    break;
	 
	case VLButton.PLAY_AUDIO_ACTION:
	    action = new PlayAudioAction(actionObj);
	    break;
	  
	default:
	    break;
	}
	
	if (action != null) {
	    bStroke.setAction(action);
	} else {
	    Toolkit.getDefaultToolkit().beep();
	}
    }

}




