/*
Copyright (c) 2000 Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:

      This product includes software developed by the Group for User 
      Interface Research at the University of California at Berkeley.

4. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
// Modified by Brian P. Bailey for use with DEMAIS

/*
  Classifier.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture;

// Import needed packages
import java.lang.*;
import java.io.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture.features.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;



public class Classifier {

    // Set of examples gestures
    protected GestureSet gestureSet;
    protected boolean trained;


    // Constructors
    private Classifier() {
	this(null);
    }
    
    public Classifier(GestureSet gs) {
	gestureSet = gs;
	trained = false;
	try {
	    // Train immediately
	    train();
	    trained = true;

	} catch (TrainingException xcptn) {
	    DEMAISError.showErrorMessage(xcptn.toString(), "RubineClassifier", false);
	    DEMAISError.logException(xcptn);
	}	
    }
    
    
    public Classification classify(Gesture gesture) {
	Classification map = null;

	if (trained) {
	    // Stores mahalanobis distances from the gesture to each category
	    map = new Classification();
	    
	    // Loop through each category and perform distance computation
	    Vector enabledCategories = gestureSet.getEnabledCategories();
	    Iterator it = enabledCategories.iterator();
	    while (it.hasNext()) {
		GestureCategory gestureCategory = (GestureCategory) it.next();
		double distance = distanceToCategory(gesture, gestureCategory);
		map.put(gestureCategory.getName(), new Double(distance));
	    }
	}
	return map;
    }
  
    
    private double distanceToCategory(Gesture gesture, GestureCategory category) {
	return mahalanobisDistance(gesture.getFeatureVector(),
				   category.getMeanFeatureVector(),
				   category.getInvCovarianceMatrix());
    }
    
    
    /*
      Compute standard Mahalanobis distance:
        distance = sqrt((feature - meanFeature) * C-inverse * (feature - meanFeature))
    */
    private double mahalanobisDistance(double[] feature, double[] meanFeature, double[][] cInverse) {
	// To improve performance, compute and store difference vector (feature - meanFeature)
	double difference[] = new double[feature.length];
	for (int i=0; i<feature.length; ++i) {
	    difference[i] = feature[i] - meanFeature[i];
	}
	
	// Compute C-inverse * difference
	double tmpVector[] = new double[feature.length];
	for (int i=0; i<feature.length; ++i) {
	    double sum = 0;
	    for (int j=0; j<feature.length; ++j) {
		sum += (cInverse[i][j] * difference[j]);
	    }
	    tmpVector[i] = sum;
	}
	
	double distance = 0;
	for (int i=0; i<feature.length; ++i) {
	    distance += (difference[i] * tmpVector[i]);
	}

	if (distance < 0) {
	    distance = 0;
	    System.out.println("Error: Mahalanobis distance measure < 0");
	}
	return Math.sqrt(distance);
    }
    

    private void train() throws TrainingException {
	// Compute mean feature vectors and covariance matrices for each category
	Vector enabledCategories = gestureSet.getEnabledCategories();	
	Iterator it = enabledCategories.iterator();
	while (it.hasNext()) {
	    GestureCategory gestureCategory = (GestureCategory) it.next();
	    gestureCategory.computeMeanFeatureVector();
	    gestureCategory.computeInvCovarianceMatrix();
	}
    }    
 
}
