/*
Copyright (c) 2000 Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:

      This product includes software developed by the Group for User 
      Interface Research at the University of California at Berkeley.

4. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
// Modified by Brian P. Bailey for use with DEMAIS


/*
  Gesture.java
*/


// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture;

// Import needed packages
import java.io.*;
import java.util.*;
import java.awt.Rectangle;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.stroke.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture.features.*;



public class Gesture {
    
    public final static Class[] FEATURE_CLASSES = {
	InitAngleCosine.class, InitAngleSine.class, BoundsSize.class,
	BoundsAngle.class, EndsDistance.class, EndsAngleCosine.class,
	EndsAngleSine.class, TotalLength.class, TotalAngle.class,
	TotalAbsAngle.class, Sharpness.class, Density1.class, Density2.class,
	Curviness.class, Curviness2.class, AnglePerDistance.class
    };

    
    private TimedStroke stroke;
    private Hashtable featureTable;


    
    private Gesture() {
	this((TimedStroke)null);
    }

    public Gesture(TimedStroke ts) {
	stroke = ts;
	computeFeatureTable();
    }


    public int size() {
	return stroke.getNumPoints();
    }

    public Rectangle getBounds() {
	return stroke.getBounds();
    }
 
    public TimedStroke getPoints() {
	return stroke;
    }

    public void setPoints(TimedStroke ts) {
	stroke = ts;
	computeFeatureTable();
    }

    public void setNormalized(boolean v) {
	// Has no purpose
	//System.out.println(v);
    }


    public boolean equals(Object obj) {
	return (this == obj);
    }


    protected Feature getFeature(Class featureClass) {
	return (Feature) featureTable.get(featureClass);
    }

    protected double getFeatureValue(Class featureClass) {
	double value = 0.0;
	Feature feature = getFeature(featureClass);
	if (feature != null) {
	    value = feature.getValue();
	}
	return value;
    }


    protected double[] getFeatureVector() {
	double [] featureVector = new double[FEATURE_CLASSES.length];
	for (int i=0; i<FEATURE_CLASSES.length; ++i) {
	    featureVector[i] = getFeatureValue(FEATURE_CLASSES[i]);
	}
	return featureVector;
    }


    private void computeFeatureTable() {
	featureTable = new Hashtable();
	for (int i=0; i<FEATURE_CLASSES.length; ++i) {
	    Class featureClass = FEATURE_CLASSES[i];
	    try {
		Feature feature = (Feature) featureClass.newInstance();
		feature.setGesture(this);
		featureTable.put(featureClass, feature);
	    } catch (Exception xcptn) {
		DEMAISError.showErrorMessage(xcptn.toString(), "Gesture", false);
		DEMAISError.logException(xcptn);
	    }
	}
    }


    
    public static Gesture read(Reader reader) {
	return Gesture.read(new TokenReader(reader));
    }
    
    public static Gesture read(TokenReader r) {
	Gesture result = new Gesture();
	
	boolean done = false;
	while (!done) {
	    String token;
	    try {
		token = r.readToken().intern();
		if (token == "author") {
		    // skip it
		    r.readLine();
		}
		else if (token == "normalized") {
		    result.setNormalized(r.readBoolean());
		}
		else if (token == "creationDate") {
		    // skip it
		    r.readLine();
		}
		else if (token == "points") {
		    int numPoints = r.readInt();
		    TimedStroke stroke = new TimedStroke();
		    
		    for (int i=0; i<numPoints; ++i) {
			int x, y;
			long t;
			x = r.readInt();
			y = r.readInt();
			t = r.readLong();
			stroke.addPoint(x, y);
		    }
		    result.setPoints(stroke);
		}
		else if (token == "endgesture") {
		    done = true;
		}
	    } 

	    catch (EOFException xcptn1) {
		done = true;
	    } 
	    catch (NumberFormatException xcptn2) {
		DEMAISError.showErrorMessage(xcptn2.toString(), "Gesture", false);
		DEMAISError.logException(xcptn2);
	    }
	    catch (IOException xcptn3) {
		DEMAISError.showErrorMessage(xcptn3.toString(), "Gesture", false);
		DEMAISError.logException(xcptn3);
	    }
	}
	return result;
    }
 
}

