/*
Copyright (c) 2000 Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:

      This product includes software developed by the Group for User 
      Interface Research at the University of California at Berkeley.

4. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
// Modified by Brian P. Bailey for use with DEMAIS

/*
  GestureCategory.java
*/


// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture;

// Import needed packages
import java.lang.*;
import java.io.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture.features.*;




public class GestureCategory {
    private static final double EPSILON = 1E-6; // for singular matrix check
    private static int nameCounter;

    private String name;
    private Vector gestures; 

    private double[] meanFeatureVector;
    private double[][] invCovarianceMatrix;



    static {
	nameCounter = 1;
    }

    public GestureCategory() {
	this(null);
    }
    
    public GestureCategory(String name) {
	this(new Vector(), name);
    }
    
    public GestureCategory(Vector gestureVector, String catName) {
	if (catName == null) {
	    // invent one
	    catName = "unnamed" + Integer.toString(GestureCategory.nameCounter);
	    ++GestureCategory.nameCounter;
	}
	name = catName;
	gestures = gestureVector;

	meanFeatureVector = null;
	invCovarianceMatrix = null;
    }
    

    public String getName() {
	return name;
    }

    public void setName(String n) {
	name = n;
    }

    public void addGesture(Gesture g) {
	gestures.add(g);
    }
    
    public Vector getGestures() {
	return gestures;
    }

    
    protected double[] getMeanFeatureVector() {
	if (meanFeatureVector == null) {
	    computeMeanFeatureVector();
	}
	return meanFeatureVector;
    }
    
    
    protected double[][] getInvCovarianceMatrix() {
	if (invCovarianceMatrix == null) {
	    try {
		computeInvCovarianceMatrix();
	    } catch (TrainingException xcptn) {
		DEMAISError.showErrorMessage(xcptn.toString(), "Gesture: " + getName(), false);
		DEMAISError.logException(xcptn);
	    }

	}
	return invCovarianceMatrix;
    }


    protected void computeMeanFeatureVector() {
	// compute the mean feature vector
	int numFeatures = Gesture.FEATURE_CLASSES.length;	
	meanFeatureVector = new double[numFeatures];

	for (int featureNum = 0; featureNum < numFeatures; ++featureNum) {
	    Class featureClass = Gesture.FEATURE_CLASSES[featureNum];
	    double sum = 0;
	    Iterator it = gestures.iterator();
	    while (it.hasNext()) {
		Gesture gesture = (Gesture) it.next();
		sum += gesture.getFeatureValue(featureClass);
	    }
	    meanFeatureVector[featureNum] = sum / ((double) gestures.size());
	}
    }


    protected void computeInvCovarianceMatrix() throws TrainingException {
	int numFeatures = Gesture.FEATURE_CLASSES.length;
	double matrix[][] = new double[numFeatures][numFeatures];
	invCovarianceMatrix = new double[numFeatures][numFeatures];

	// compute the covariance matrix first, then invert it
	for (int i = 0; i < numFeatures; ++i) {
	    Class iFC = Gesture.FEATURE_CLASSES[i];
	    for (int j = i; j < numFeatures; ++j) {
		Class jFC = Gesture.FEATURE_CLASSES[j];

		double sum = 0;
		Iterator it = gestures.iterator();
		while (it.hasNext()) {
		    Gesture gesture = (Gesture) it.next();
		    double fi = gesture.getFeatureValue(iFC);
		    double fj = gesture.getFeatureValue(jFC);
		    sum += ((fi - meanFeatureVector[i]) * (fj - meanFeatureVector[j]));
		}
		matrix[i][j] = sum / ((double) (gestures.size() - 1));
		if (i != j) {
		    matrix[j][i] = matrix[i][j];
		}
	    }
	}

	// Invert the covariance matrix
	invCovarianceMatrix = Matrix.inverse(matrix);
    }




    public static GestureCategory read(Reader reader) {
	return GestureCategory.read(new TokenReader(reader));
    }
        
    public static GestureCategory read(TokenReader r) {
	GestureCategory result = new GestureCategory();
	
	boolean done = false;
	while (!done) {
	    String token;
	    String line;
	    try {
		token = r.readToken().intern();
		if (token == "name") {
		    line = r.readLine();
		    line.trim();
		    result.setName(line);
		}
		else if (token == "author") {
		    // Skip it
		    r.readLine();
		}
		else if (token == "directionInvariant") {
		    r.readBoolean();
		}
		else if (token == "orientationInvariant") {
		    r.readBoolean();
		}
		else if (token == "sizeInvariant") {
		    r.readBoolean();
		}
		else if (token == "gestures") {
		    try {
			int numGestures = Integer.parseInt(r.readToken());
			for (int i=0; i<numGestures; ++i) {
			    Gesture gesture = Gesture.read(r);
			    result.addGesture(gesture);
			}
		    }
		    catch (NumberFormatException e) {
			throw new IOException("bad argument for categories");
		    }
		}
		else if (token == "endcategory") {
		    done = true;
		}
	    }
	    catch (EOFException xcptn1) {
		done = true;
	    }
	    catch (IOException xcptn2) {
		DEMAISError.showErrorMessage(xcptn2.toString(), "GestureCategory", false);
		DEMAISError.logException(xcptn2);
	    }
	}
	return result;
    }
    
 
}
