/*
Copyright (c) 2000 Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:

      This product includes software developed by the Group for User 
      Interface Research at the University of California at Berkeley.

4. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
// Modified by Brian P. Bailey for use with DEMAIS


// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.gesture.misc;

// Import needed packages
import java.util.*;



/**
 * A Map with any kind of uniquely named Keys with numerical Values, sorted 
 * by the values rather than by the keys. You can set whether the Map is sorted 
 * each value has a threshold max value or not, and whether the sum of the 
 * values has a threshold max value or not.
 *
 * <P>
 * This software is distributed under the 
 * <A HREF="http://guir.cs.berkeley.edu/projects/COPYRIGHT.txt">
 * Berkeley Software License</A>.
 *
 * <PRE>
 * Revisions:  - GUIRLib-v1.0-1.0.0, Jun 16 1997, JH
 *               Created class
 *             - GUIRLib-v1.2-1.0.0, Jun 22 2000, JH
 *               Touched for GUIRLib release
 *             - GUIRLib-v1.3-1.0.0, Aug 11 2000, JH
 *               Touched for GUIRLib release
 *             - GUIRLib-v1.4-1.0.0, Aug 31 2000, JH
 *               Touched for GUIRLib release
 * </PRE>
 *
 * @author  (<A HREF="mailto:jasonh@cs.berkeley.edu">jasonh@cs.berkeley.edu</A>)
 * @since   JDK 1.2
 * @version GUIRLib-v1.4-1.0.0, Aug 31 2000
 */
public class SortedValueMap {

   /**
    * Ascending or descending?
    */
   protected static boolean DEFAULT_FLAG_ASCENDING = true;


   /**
    * Sort in ascending order?
    */
   protected boolean flagAscending;


   //// hashtable maps name->value
   protected Hashtable table;

   //// linkedlist maps value->name, value being sorted asc or descending
   //// soft-state, call updateList() to make it consistent
   protected LinkedList list;

   //// true if list is inconsistent and needs to be updated, false otherwise.
   protected boolean flagDirty;

   //// the comparator used to sort the values in this map
   protected Comparator comp;

   //// the comparator that knows about our internal data structs,
   //// and uses comp to sort the values
   protected Comparator internalComp;





    // Constructor
    private SortedValueMap() {
	this(null);
    }

    public SortedValueMap(Comparator newComp) {
	flagAscending = SortedValueMap.DEFAULT_FLAG_ASCENDING;
	flagDirty = false;
	table = new Hashtable(30);
	list = null;
	comp  = newComp; // compares the values in the table
	internalComp = new InternalComparator();

    } // of constructor



    // Compares the wrapped values in the table (unwraps them)
    class InternalComparator implements Comparator {
	
	public int compare(Object key1, Object key2) {
	    //// 1. Compare the values.
	    ////    Each of these are Map.Entry objects.
	    Map.Entry m1 = (Map.Entry) key1;
	    Map.Entry m2 = (Map.Entry) key2;
	    int returnVal = comp.compare(m1.getValue(), m2.getValue());
	    
	    //// 2. Allow multiple instances, so arbitrarily break ties.
	    if (returnVal == 0) {
		if (m1.getKey().hashCode() > m2.getKey().hashCode()) {
		    returnVal = 1;
		}
		else {
		    returnVal = -1;
		}
	    }
	    
	    //// 3. Flip the values if we are in descending order.
	    if (flagAscending) {
		return (returnVal);
	    }
	    else {
		return (-1 * returnVal);
	    }
	} // of compare
	
    } // of inner class InternalComparator
    



 
   /**
    * Use this BEFORE you stick in items.
    *
    *  @param flag is true if ascending order, false if descending.
    */
   public void setAscending(boolean flag) {
      flagAscending = flag;
   } // of setAscending




   /**
    * Make the sorted list consistent.
    */
   private void updateList() {
      //// 1. If we are inconsistent with the table...
      if (flagDirty == true) {
         //// 1.1. Create a new list and then sort it.
         list = new LinkedList(table.entrySet());
         Collections.sort(list, internalComp);
         flagDirty = false;
      }
   } // of method

   //===   CONSISTENCY METHODS   ===============================================
   //===========================================================================



   //===========================================================================
   //===   MAP METHODS   =======================================================

   public void clear() {
      flagDirty = true;
      table.clear();
   } // of clear

   //-----------------------------------------------------------------

   public boolean containsKey(Object key) {
      return (table.containsKey(key));
   } // of containsKey

   //-----------------------------------------------------------------

   public boolean containsValue(Object value) {
      return (table.containsValue(value));
   } // of containsValue

   //-----------------------------------------------------------------

   public Set entrySet() {
      return (table.entrySet());
   } // of entrySet

   //-----------------------------------------------------------------

   public boolean equals(Object obj) {
      return (obj.hashCode() == this.hashCode());
   } // of equals

   //-----------------------------------------------------------------

   public Object get(Object obj) {
      return (table.get(obj));
   } // of get

   //-----------------------------------------------------------------

   public boolean isEmpty() {
      return (table.isEmpty());
   } // of isEmpty

   //-----------------------------------------------------------------

   public Set keySet() {
      return (table.keySet());
   } // of keySet

   //-----------------------------------------------------------------

   public Object put(Object key, Object value) {
      flagDirty = true;
      return (table.put(key, value));
   } // of put

   //-----------------------------------------------------------------

   public void putAll(Map newMap) {
      Iterator it = newMap.keySet().iterator();
      Object   key;
      
      while (it.hasNext()) {
         key = it.next();
         this.put(key, newMap.get(key));
      }
   } // of putAll

   //-----------------------------------------------------------------

   public Object remove(Object key) {
      flagDirty = true;
      return (table.remove(key));
   } // of remove

   //-----------------------------------------------------------------

   public int size() {
      return (table.size());
   } // of size

   //-----------------------------------------------------------------

   public Collection values() {
      return (table.values());
   } // of values

   //===   MAP METHODS   =======================================================
   //===========================================================================



   //===========================================================================
   //===   OTHER MAP METHODS   =================================================

   /**
    * Get the key at the beginning of the map.
    *
    * @return whatever key is at the beginning. Will have a value that is the 
    *         largest or smallest value, depending on how the map is sorted.
    */
   public Object getFirstKey() {
      updateList();
      if (list.size() > 0) {
         Map.Entry m = (Map.Entry) list.getFirst();
         return (m.getKey());
      }
      return (null);
   } // of getFirstKey

   //-----------------------------------------------------------------

   /**
    * Get the key at the end of the map.
    *
    * @return whatever key is at the end. Will have a value that is the largest
    *         or smallest value, depending on how the map is sorted.
    */
   public Object getLastKey() {
      updateList();
      if (list.size() > 0) {
         Map.Entry m = (Map.Entry) list.getLast();
         return (m.getKey());
      }
      return (null);
   } // of getLastKey

   //-----------------------------------------------------------------

   /**
    * Get the value at the beginning of the map.
    *
    * @return whatever value is at the beginning. Either the largest or smallest
    *         value, depending on how the map is sorted.
    */
   public Object getFirstValue() {
      return (get(getFirstKey()));
   } // of getFirstValue

   //-----------------------------------------------------------------

   /**
    * Get the value at the end of the map.
    *
    * @return whatever value is at the end. Either the largest or smallest
    *         value, depending on how the map is sorted.
    */
   public Object getLastValue() {
      return (get(getLastKey()));
   } // of getLastValue

   //-----------------------------------------------------------------

   /**
    * Get the key with the largest value.
    *
    * @return whatever key has the largest value.
    */
   public Object getLargestKey() {
      if (flagAscending == true) {
         return (getLastKey());
      } 
      else {
         return (getFirstKey());
      }
   } // of getLargestKey

   //-----------------------------------------------------------------

   /**
    * Get the key with the smallest value.
    *
    * @return whatever key has the smallest value.
    */
   public Object getSmallestKey() {
      if (flagAscending == true) {
         return (getFirstKey());
      } 
      else {
         return (getLastKey());
      }
   } // of getSmallestKey

   //-----------------------------------------------------------------

   /**
    * Get the largest value stored.
    *
    * @return whatever the largest value is.
    */
   public Object getLargestValue() {
      if (flagAscending == true) {
         return (getLastValue());
      } 
      else {
         return (getFirstValue());
      }
   } // of getLargestValue

   //-----------------------------------------------------------------

   /**
    * Get the smallest value stored.
    *
    * @return whatever the smallest value is.
    */
   public Object getSmallestValue() {
      if (flagAscending == true) {
         return (getFirstValue());
      } 
      else {
         return (getLastValue());
      }
   } // of getSmallestValue

   //===   OTHER MAP METHODS   =================================================
   //===========================================================================



   //===========================================================================
   //===   TOSTRING   ==========================================================

   public String toString() {
      updateList();
      return (list.toString());
   } 



} // of class

//==============================================================================

/*
Copyright (c) 2000 Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:

      This product includes software developed by the Group for User 
      Interface Research at the University of California at Berkeley.

4. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

