/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditons and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File BehavioralInkStroke.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs;
    

// Necessary imports
import java.awt.*;
import java.awt.geom.*;
import java.io.*;
import java.util.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.core.annotation.*;
import edu.uofminnesota.DEMAIS.core.parser.*;
import edu.uofminnesota.DEMAIS.usrinterface.util.*;
import edu.uofminnesota.DEMAIS.usrinterface.multiview.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.stroke.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.behavior.*;




// InkStroke class
public class BehavioralInkStroke extends InkStroke {
    
    public static final int CLASSID = 130;
 
    private static final int CROSS_LENGTH = 6;
    private static Point DEFAULT_ARROW[];
    private static Font SEQUENCE_NUMBER_FONT = new Font("Arial", Font.BOLD, 12);
    
    // Source event triggering the behavior
    private SourceEvent sourceEvent;

    // Destination action
    private DestinationAction action;

    // Screen location of end arrow
    private Polygon arrow;
    private Polygon original_arrow;
    private int sequence;
    private Point sequenceL;



    // Static initializer
    static {
	DEFAULT_ARROW = new Point[3];
	// Define arrow in Java's coordinate system (Y increases downward)
	DEFAULT_ARROW[0] = new Point(-5, -5);
	DEFAULT_ARROW[1] = new Point(4, 0);
	DEFAULT_ARROW[2] = new Point(-5, 5);
    }


    // Do not allow default constructor to be called
    private BehavioralInkStroke() {
	this(null, null, null);
    }

    public BehavioralInkStroke(InkStroke inkStroke, SourceEvent sEvent, DestinationAction dAction) {
	super(inkStroke);

	sourceEvent = sEvent;
	action = dAction;
	arrow = null;
	original_arrow = null;
	sequence = 0;
	sequenceL = null;

	// Source event object should NEVER be null
	GraphicalObject sourceEventObj = sourceEvent.getEventObject();
	if (sourceEventObj != null) {
	    sourceEventObj.addBehavior(this);
	    if ((sourceEvent.getX() == 0) && (sourceEvent.getY() == 0)) {
		Point p = getLocationForSourceEvent(sourceEventObj);
		sourceEvent.setLocation(p.x, p.y);
	    }
	    //sequenceL = getLocationForSequence(sourceEventObj);
	}

	// Possible for actionObj to be null - navigation and transitions have no event object
	GraphicalObject actionObj = action.getEventObject();
	if (actionObj != null) {
	    actionObj.addBehavior(this);
	}

	if ((action.getX() == 0) && (action.getY() == 0)) {
	    Point p = getLocationForAction(actionObj);
	    action.setLocation(p.x, p.y);
	}

	// Compute location for the end arrow
	computeArrowLocation();
    }

    
    public BehavioralInkStroke(BehavioralInkStroke copy) {
	this(copy, copy.storyboard);
    }

    public BehavioralInkStroke(BehavioralInkStroke copy, Storyboard toStoryboard) {
	this(copy, copy.storyboard, true);
    }

    public BehavioralInkStroke(BehavioralInkStroke copy, Storyboard toStoryboard, boolean copyOfModel) {
	this(copy, null, null, toStoryboard);
    }

    public BehavioralInkStroke(BehavioralInkStroke copy, GraphicalObject toSource,
			       GraphicalObject toDest, Storyboard toStoryboard) {
	super(copy, toStoryboard);

	sourceEvent = null;
	action = null;
	sequence = 0;
	sequenceL = null;

	Point p = null;
	if (toSource != null) {
	    toSource.addBehavior(this);
	    sourceEvent = storyboard.getDefaultSourceEvent(toSource, toDest);
	    p = getLocationForSourceEvent(toSource);
	    sourceEvent.setLocation(p.x, p.y);
	    //sequenceL = getLocationForSequence(toSource);
	}
	
	if (toDest != null) {
	    toDest.addBehavior(this);
	}
	
	action = storyboard.getDefaultAction(toSource, toDest);
	if (action != null) {
	    p = getLocationForAction(toDest);
	    action.setLocation(p.x, p.y);
	}

	arrow = null;
	original_arrow = null;
	computeArrowLocation();
    }


    public BehavioralInkStroke(DataInputStream stream, Storyboard pStoryboard) throws IOException {
	super(stream, pStoryboard);

	// Read the behavior information
	sourceEvent = null;
	action = null;
	sequence = 0;
	sequenceL = null;

	// Was there a source event written?
	boolean wasSourceEvent = stream.readBoolean();
	if (wasSourceEvent) {      
	    int classid = stream.readInt();

	    switch(classid) {

	    case BeginMediaEvent.CLASSID:
		sourceEvent = new BeginMediaEvent(stream);
		break;
		
	    case DoubleMouseClickEvent.CLASSID:
		sourceEvent = new DoubleMouseClickEvent(stream);
		break;
		
	    case DragAndDropEvent.CLASSID:
		sourceEvent = new DragAndDropEvent(stream);
		break;
		
	    case EndMediaEvent.CLASSID:
		sourceEvent = new EndMediaEvent(stream);
		break;
		
	    case EnterTextEvent.CLASSID:
		sourceEvent = new EnterTextEvent(stream);
		break;
		
	    case RolloverEvent.CLASSID:
		sourceEvent = new RolloverEvent(stream);
		break;
		
	    case SingleMouseClickEvent.CLASSID:
		sourceEvent = new SingleMouseClickEvent(stream);
		break;
		
	    case SyncMarkerEvent.CLASSID:
		sourceEvent = new SyncMarkerEvent(stream);
		break;

	    case NarrationSyncMarkerEvent.CLASSID:
		sourceEvent = new NarrationSyncMarkerEvent(stream);
		break;
		
	    case TemporalEvent.CLASSID:
		sourceEvent = new TemporalEvent(stream);
		break;
				
	    default:
		throw (new IOException("Unknown source event class id"));
	    }
	}

	
	boolean wasAction = stream.readBoolean();
	if (wasAction) {

	    int classid = stream.readInt();
	    switch(classid) {
		
	    case BeginMediaAction.CLASSID:
		action = new BeginMediaAction(stream);
		break;

	    case DelayAction.CLASSID:
		action = new DelayAction(stream);
		break;

	    case DisplayAction.CLASSID:
		action = new DisplayAction(stream);
		break;
		
	    case EndMediaAction.CLASSID:
		action = new EndMediaAction(stream);
		break;
		
	    case HighlightAction.CLASSID:
		action = new HighlightAction(stream);
		break;
		
	    case NavigationAction.CLASSID:
		action = new NavigationAction(stream);
		break;
		
	    case PauseMediaAction.CLASSID:
		action = new PauseMediaAction(stream);
		break;
		
	    case PlayAudioAction.CLASSID:
		action = new PlayAudioAction(stream);
		break;
		
	    case SetMediaTimeAction.CLASSID:
		action = new SetMediaTimeAction(stream);
		break;
		
	    case TransitionAction.CLASSID:
		action = new TransitionAction(stream);
		break;
	
	    case UnDisplayAction.CLASSID:
		action = new UnDisplayAction(stream);
		break;
		
	    case UnHighlightAction.CLASSID:
		action = new UnHighlightAction(stream);
		break;
		
	    default:
		throw (new IOException("Unknown action class id"));
	    }
	}

	arrow = null;
	original_arrow = null;

	boolean wasArrow = stream.readBoolean();
	if (wasArrow) {
	    arrow = new Polygon();
	    original_arrow = new Polygon();
	    int numPoints = stream.readInt();
	    for (int i=0; i<numPoints; ++i) {
		int x = stream.readInt();
		int y = stream.readInt();
		arrow.addPoint(x, y);
		original_arrow.addPoint(x, y);
	    }
	}
    }


    public void completeIO() {
	if (sourceEvent != null) {
	    sourceEvent.completeIO();
	    GraphicalObject sourceEventObj = sourceEvent.getEventObject();
	    if (sourceEventObj != null) {
		sourceEventObj.addBehavior(this);
	    }
	    //sequenceL = getLocationForSequence(sourceEventObj);

	}
	if (action != null) {
	    action.completeIO();
	    GraphicalObject actionObj = action.getEventObject();
	    if (actionObj != null) {
		actionObj.addBehavior(this);
	    }
	}
    }



    public boolean isBehaviorDefined() {
	return (sourceEvent != null);
    }


    public Rectangle getSourceEventIconTargetBounds() {
	Rectangle r = null;
	if (sourceEvent != null) {
	    r = sourceEvent.getTargetBounds();
	}
	return r;
    }

    public GraphicalObject getSourceEventObject() {
	GraphicalObject go = null;
	if (sourceEvent != null) {
	    go = sourceEvent.getEventObject();
	}
	return go;
    }

    public SourceEvent getSourceEvent() {
	return sourceEvent;
    }

    public void setSourceEvent(SourceEvent sEvent) {
	int x = -1;
	int y = -1;
	GraphicalObject sourceObj = null;
	int listIndex = -1;

	if (sourceEvent != null) {
	    sourceObj = sourceEvent.getEventObject();
	    if (sourceObj != null) {
		listIndex = sourceObj.getBehaviors().indexOf(this);
		sourceObj.removeBehavior(this);
	    }
	    x = sourceEvent.getX();
	    y = sourceEvent.getY();
	}

	sourceEvent = sEvent; 
	if (sourceEvent == null) {
	    return;
	}

	sourceObj = sourceEvent.getEventObject();
	if ((x <= 0) && (y <= 0) && (sourceObj != null)) {
	    Point p = getLocationForSourceEvent(sourceObj);
	    x = p.x;
	    y = p.y;
	}

	sourceEvent.setLocation(x, y);
	if (sourceObj != null) {
	    sourceObj.addBehavior(this, listIndex);
	}
    }



    public Rectangle getActionIconTargetBounds() {
	Rectangle r = null;
	if (action != null) {
	    r = action.getTargetBounds();
	}
	return r;
    }

    public GraphicalObject getActionObject() {
	GraphicalObject go = null;
	if (action != null) {
	    go = action.getEventObject();
	}
	return go;
    }
    
    public DestinationAction getAction() {
	return action;
    }

    public void setAction(DestinationAction a) {
	int x = -1;
	int y = -1;
	if (action != null) {
	    GraphicalObject actionObj = action.getEventObject();
	    if (actionObj != null) {
		actionObj.removeBehavior(this);
	    }
	    x = action.getX();
	    y = action.getY();
	}

	action = a; 
	if (action == null) {
	    return;
	}

	if (x == -1) {
	    Point p = getLocationForAction(action.getEventObject());
	    x = p.x;
	    y = p.y;
	}

	action.setLocation(x, y);
	GraphicalObject actionObj = action.getEventObject();
	if (actionObj != null) {
	    actionObj.addBehavior(this);
	}
    }

       
    public void scale(double sx, double sy) {
	super.scale(sx, sy);
	if (sourceEvent != null) {
	    sourceEvent.scale(sx, sy);
	}
	if (action != null) {
	    action.scale(sx, sy);
	}

	// Now perform the scaling
	for (int i=0; i<arrow.npoints; ++i) {
	    arrow.xpoints[i] = (int) Math.round(arrow.xpoints[i] * sx);
	    arrow.ypoints[i] = (int) Math.round(arrow.ypoints[i] * sy);
	}
    }
    
    
    public void resetToOriginalScale() {
	super.resetToOriginalScale();
	if (sourceEvent != null) {
	    sourceEvent.resetToOriginalScale();
	}
	if (action != null) {
	    action.resetToOriginalScale();
	}
	
	if (original_arrow != null) {
	    for (int i=0; i<original_arrow.npoints; ++i) {
		arrow.xpoints[i] = original_arrow.xpoints[i];
		arrow.ypoints[i] = original_arrow.ypoints[i];
	    }
	}
    }


    public void doTranslate(int deltaX, int deltaY) {
	super.doTranslate(deltaX, deltaY);
	if (sourceEvent != null) {
	    sourceEvent.doTranslate(deltaX, deltaY);
	}
	if (action != null) {
	    action.doTranslate(deltaX, deltaY);
	}
	if (original_arrow != null) {
	    original_arrow.translate(deltaX, deltaY);
	}
	if (arrow != null) {
	    arrow.translate(deltaX, deltaY);
	}
    }


    public Object clone() {
	return (new BehavioralInkStroke(this));
    }

    public Object clone(Storyboard toStoryboard) {
	return (new BehavioralInkStroke(this, toStoryboard));
    }

    public Object clone(Storyboard toStoryboard, boolean copyOfModel) {
	return (new BehavioralInkStroke(this, toStoryboard, copyOfModel));
    }


    public Object clone(GraphicalObject toSource, GraphicalObject toDest, Storyboard toStoryboard) {
	return (new BehavioralInkStroke(this, toSource, toDest, toStoryboard));
    }


    public void deleteGraphicalObject(boolean deleteModel) {
	super.deleteGraphicalObject(deleteModel);
	deleteBehavior();
    }


    public void deleteBehavior() {
	if (sourceEvent != null) {
	    GraphicalObject sourceObj = sourceEvent.getEventObject();
	    if (sourceObj != null) {
		sourceObj.removeBehavior(this);
	    }
	}
	if (action != null) {
	    GraphicalObject actionObj = action.getEventObject();
	    if (actionObj != null) {
		actionObj.removeBehavior(this);
	    }
	}
    }



    public Annotation toAnnotation() {
	String sourceS = null;
	String actionS = null;

	if (sourceEvent != null) {
	    sourceS = sourceEvent.toString();
	}
	if (action != null) {
	    actionS = action.toString();
	}

	if (sourceS == null) {
	    return null;
	}
	return (new Annotation(getScope(), sourceS, actionS));
    }


    public void operationalize() {
	operationalize(getScope());
    }

    public void operationalize(DataModel dModel) {
	super.operationalize();
	setVisible(false);

	Annotation annotation = toAnnotation();
	if (annotation != null) {
	    dModel.addAnnotation(annotation);
	    
	    // Check whether this object needs mouse events
	    if (dModel instanceof StoryboardModel) {
		StoryboardModel sModel = (StoryboardModel) dModel;
		GraphicalObject sourceObject = getSourceEventObject();
		if ((sourceObject != null) && 
		    (sourceEvent.requiresRuntimeMouseEvents())) {
		    sModel.getStoryboard().addSourceEventObject(sourceObject);
		    
		    if (sourceEvent instanceof DragAndDropEvent) {
			sourceObject.setIsDraggable(true);
		    }
		}
	    }
	    
	    if (action != null) {
		// Check whether this object should be displayed initially
		GraphicalObject actionObject = getActionObject();
		if ((action.requiresRuntimeInitialRemoval()) && 
		    (actionObject != null)) {
		    actionObject.setVisible(false);

		    // Also, set the graphical objects contained within it to be invisible
		    Rectangle r = actionObject.doGetBounds();
		    Storyboard sBoard = actionObject.getStoryboard();
		    if (sBoard != null) {
			sBoard.setVisibleOnObjectsContainedBy(r, false, actionObject.getLayer(), false, actionObject);
		    }
		}
	    }
	    
	    // In some cases, we also want to add the reverse of the annotation
	    if (sourceEvent instanceof RolloverEvent) {
		if ( (action instanceof DisplayAction) ||
		     (action instanceof UnDisplayAction) ||
		     (action instanceof HighlightAction) ||
		     (action instanceof UnHighlightAction)) {
		    Annotation reverse = annotation.reverse();
		    dModel.addAnnotation(reverse);
		}
	    }
	}
    }


    // Protected methods
    protected void doRender(Graphics2D g) {

	if (! (storyboard.getModel() instanceof MultiViewModel)) {
	    GraphicalObject sourceObj = getSourceEventObject();
	    GraphicalObject actionObj = getActionObject();
	    Storyboard sourceS = null;
	    Storyboard actionS = null;
	    
	    if (sourceObj != null) {
		sourceS = sourceObj.getStoryboard();
	    }
	    if (actionObj != null) {
		actionS = actionObj.getStoryboard();
	    }
	    
	    if ((sourceObj == null) || (sourceS == null) ||
		(!sourceS.isLayerVisible(sourceObj.getLayer()))) {
		return;
	    }
	    if ((actionObj == null) || (actionS == null) ||
		(!actionS.isLayerVisible(actionObj.getLayer()))) {
		return;
	    }
	}

	super.doRender(g);

	sequence = 0;
	if ((getScaleFactorX() == 1.0) &&
	    (getScaleFactorY() == 1.0)) {
	    
	    g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			       RenderingHints.VALUE_ANTIALIAS_ON);
	    
	    // Draw a circle in the source object
	    Point sourceP = getFirstPoint();
	    g.fillArc(sourceP.x - 4, sourceP.y - 4, 9, 9, 0, 360);
	    
	    if (sourceEvent != null) {
		// Draw the source event icon
		sourceEvent.render(g, inSelectedLayer);
		
		// Determine sequence number, if any
		/*
		  GraphicalObject sourceObj = sourceEvent.getEventObject();
		  Vector behaviors = sourceObj.getBehaviors();
		  int myIndex = behaviors.indexOf(this);
		  
		  int beforeMe = 0;
		  for (int i=0; i<myIndex; ++i) {
		  BehavioralInkStroke bStroke = (BehavioralInkStroke) behaviors.elementAt(i);
		  SourceEvent sEvent = bStroke.getSourceEvent();
		  if ((sourceObj == sEvent.getEventObject()) &&
		  (sEvent.getClassId() == sourceEvent.getClassId())) {
		  ++beforeMe;
		  }
		  }
		  
		  int afterMe = 0;
		  for (int i=myIndex+1; i<behaviors.size(); ++i) {
		  BehavioralInkStroke bStroke = (BehavioralInkStroke) behaviors.elementAt(i);
		  SourceEvent sEvent = bStroke.getSourceEvent();
		  if ((sourceObj == sEvent.getEventObject()) &&
		  (sEvent.getClassId() == sourceEvent.getClassId())) {
		  ++afterMe;
		  }
		  }
		  
		  if ((beforeMe == 0) && (afterMe > 0)) {
		  sequence = 1;
		  } else if (beforeMe > 0) {
		  sequence = beforeMe + 1;
		  }
		  
		  if ((sequence > 0) && (sequenceL != null)) {
		  g.setFont(SEQUENCE_NUMBER_FONT);
		  g.drawString(Integer.toString(sequence), sequenceL.x, sequenceL.y);
		  }
		*/
	    }
	    
	    if (action != null) {
		// Draw the action icon
		action.render(g, inSelectedLayer);
	    }
	    
	}
	
	if (arrow != null) {
	    setColor(g);
	    setStroke(g);
	    g.fillPolygon(arrow);
	}
	
	g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
			   RenderingHints.VALUE_ANTIALIAS_OFF);
	
    }


    public Rectangle getBasicBounds() {
	Rectangle r = super.getBasicBounds();
	Rectangle sourceRect = null;
	Rectangle actionRect = null;
	if (sourceEvent != null) {
	    r = r.union(sourceEvent.getBounds());
	}
	if (action != null) {
	    r = r.union(action.getBounds());
	}
	if (arrow != null) {
	    r = r.union(arrow.getBounds());
	}

	return r;
    }

    protected Rectangle doGetBounds() {
	return getBasicBounds();
    }
    
    
    private Point getLocationForSourceEvent(GraphicalObject sourceEventObj) {
	TimedStroke stroke = getTimedStroke();
	
	// Locate the first point along the stroke that is outside the source and dest object
	int i = 1;
	while ((i < stroke.npoints) && 
	       (sourceEventObj.contains(stroke.xpoints[i], stroke.ypoints[i]))) {
	    ++i;
	}
	if (i >= stroke.npoints) {
	    i = 1;
	}
	
	int locX = stroke.xpoints[i];
	int locY = stroke.ypoints[i];

	// Adjust for direction
	if ( (stroke.xpoints[i] - stroke.xpoints[i-1]) < 0) {
	    // Moving to the left
	    locX -= BehaviorEvent.VLI_WIDTH;
	}
	if ( (stroke.ypoints[i] - stroke.ypoints[i-1]) < 0) {
	    // Moving towards the top of the screen
	    locY -= BehaviorEvent.VLI_HEIGHT;
	}
	
	return (new Point(locX, locY));
    }

    
    private Point getLocationForSequence(GraphicalObject sourceEventObj) {
	TimedStroke stroke = getTimedStroke();
	
	// Locate the first point along the stroke that is outside the source and dest object
	int i = 1;
	while ((i < stroke.npoints) && 
	       (sourceEventObj.contains(stroke.xpoints[i], stroke.ypoints[i]))) {
	    ++i;
	}
	i = i + 2;

	if (i >= stroke.npoints) {
	    i = stroke.npoints;
	}
		
	int locX = stroke.xpoints[i];
	int locY = stroke.ypoints[i];

	// Adjust for direction
	if ( (stroke.xpoints[i] - stroke.xpoints[i-1]) < 0) {
	    // Moving to the left
	    locX -= (BehaviorEvent.VLI_WIDTH + 3);
	} else {
	    // Moving to the right
	    locX += (BehaviorEvent.VLI_WIDTH + 3);
	}

	if ( (stroke.ypoints[i] - stroke.ypoints[i-1]) < 0) {
	    // Moving towards the top of the screen
	    locY -= (BehaviorEvent.VLI_HEIGHT + 3); 
	} else {
	    locY += (BehaviorEvent.VLI_HEIGHT + 3);
	}
	
	return (new Point(locX, locY));
    }


    private Point getLocationForAction(GraphicalObject actionObj) {
	TimedStroke stroke = getTimedStroke();
	int i = -1;
	
	if (actionObj != null) {
	    i = stroke.npoints - 2;
	    while ((i >= 0) && 
		   (actionObj.contains(stroke.xpoints[i], stroke.ypoints[i]))) {
		--i;
	    }
	}
	if (i <= 0) {
	    i = stroke.npoints - 2;
	}
	
	int locX = stroke.xpoints[i];
	int locY = stroke.ypoints[i];

	// Adjust for direction
	if ( (stroke.xpoints[i] - stroke.xpoints[i-1]) > 0) {
	    // Moving to the right
	    locX -= BehaviorEvent.VLI_WIDTH;
	}
	if ( (stroke.ypoints[i] - stroke.ypoints[i-1]) > 0) {
	    // Moving towards the bottom of the screen
	    locY -= BehaviorEvent.VLI_HEIGHT;
	}

	return (new Point(locX, locY));
    }



    private void computeArrowLocation() {
	TimedStroke stroke = getTimedStroke();
	if (stroke.npoints < 3) {
	    arrow = null;
	    return;
	}	

	int last = stroke.npoints - 1;
	int prev = stroke.npoints - 3;
	boolean up = false;

	int diff = stroke.ypoints[last] - stroke.ypoints[prev];
	if (diff < 0) {
	    // Moving towards the top of the screen
	    up = true;
	} else if (diff == 0) {
	    diff = stroke.ypoints[last] - stroke.ypoints[0];
	    if (diff < 0) {
		up = true;
	    }
	}
	
	double distance = DMath.distance(stroke.xpoints[last], stroke.ypoints[last],
					 stroke.xpoints[prev], stroke.ypoints[prev]);
	double dot_product = (stroke.xpoints[last] - stroke.xpoints[prev]);
	double alpha = Math.acos(dot_product / distance); // in radians
	if (! up) {
	    alpha = -1.0 * alpha;
	}
	    
	arrow = new Polygon();
	original_arrow = new Polygon();
	double x, y;
	int x1, y1;
	for (int i=0; i<3; ++i) {
	    x = DEFAULT_ARROW[i].getX();
	    y = DEFAULT_ARROW[i].getY();

	    // Rotate in Java's coordinate system (Y increases downward)
	    x1 = ((int)Math.round(Math.cos(alpha)*x + Math.sin(alpha)*y)) + stroke.xpoints[last];
	    y1 = ((int)Math.round(Math.cos(alpha)*y - Math.sin(alpha)*x)) + stroke.ypoints[last];
	    arrow.addPoint(x1, y1);
	    original_arrow.addPoint(x1, y1);
	}
    }


    public boolean equals(Object obj) {
	return (this == obj);
    }


    public void write(DataOutputStream stream) throws IOException {
	// Write the class id
	stream.writeInt(CLASSID);

	// Write the parent information
	super.write(stream);

	// Write the behavior information
	stream.writeBoolean(sourceEvent != null);
	if (sourceEvent != null) {
	    sourceEvent.write(stream);
	}
	
	stream.writeBoolean(action != null);
	if (action != null) {
	    action.write(stream);
	} 

	stream.writeBoolean(arrow != null);
	if (arrow != null) {
	    stream.writeInt(arrow.npoints);
	    for (int i=0; i<arrow.npoints; ++i) {
		stream.writeInt(arrow.xpoints[i]);
		stream.writeInt(arrow.ypoints[i]);
	    }
	}
    }
}

