/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File GraphicalObject.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs;
    

// Necessary imports
import java.lang.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.behavior.*;




// Graphical Object abstract class
public abstract class GraphicalObject {
    
    // An id counter for the graphical objects
    private static int GO_IDENTIFIER;

    // The width and height of a selection handle
    private static final int SELECTION_HANDLE_WIDTH;
    private static final int SELECTION_HANDLE_HEIGHT;
    private static final Color SELECTION_HANDLE_COLOR;

    // The strokes used for drawing on the storyboard
    public static final BasicStroke HEAVY_STROKE;
    public static final BasicStroke MEDIUM_STROKE;
    public static final BasicStroke LIGHT_STROKE;

    // These use the line widths specified by the user
    public static final BasicStroke DEFAULT_INK_STROKE;
    public static final BasicStroke DEFAULT_GESTURE_STROKE;



    // Drawing color for the active layer
    public static final Color ACTIVE_COLOR;

    // Drawing color for the non-active layers
    public static final Color NOT_ACTIVE_COLOR;

    // Highlight color
    public static final Color HIGHLIGHT_COLOR;



    // INSTANCE DATA

    // The identifier of the graphical object
    private int identifier;

    // The storyboard to which this object belongs
    protected Storyboard storyboard;

    // The layer in which this object resides
    private int layer;

    // Whether this object is visible
    protected boolean visible;

    // Whether this object is in the currently selected layer
    protected boolean inSelectedLayer;

    // Whether this object is selected
    protected boolean selected;

    // Whether to perform analysis on this object
    protected boolean performAnalysis;

    // A list of behaviors this object is involved in
    protected Vector behaviors;

    // The X scale factor
    protected double scaleFactorX;

    // The Y scale factor
    protected double scaleFactorY;


    // OPERATIONAL data section

    // If the mouse is over this object
    protected boolean opMouseIsOver;

    // If this object is highlighted
    protected boolean opIsHighlighted;

    // Whether this object can be dragged at runtime
    protected boolean opIsDraggable;

    // Original location of the object (before translation due to DND)
    private int opOriginalX;
    private int opOriginalY;

    // If the objects is operationalized
    protected boolean operationalized;



    // Static initializer
    static {
	GO_IDENTIFIER = -1;

	// Ensure these are EVEN numbers
	SELECTION_HANDLE_WIDTH = 8;
	SELECTION_HANDLE_HEIGHT = 8;
	SELECTION_HANDLE_COLOR = Color.black;

	HEAVY_STROKE = new BasicStroke(3.0f);
	MEDIUM_STROKE = new BasicStroke(2.0f);
	LIGHT_STROKE = new BasicStroke(1.0f);
	DEFAULT_INK_STROKE = new BasicStroke(PropertyManager.INK_STROKE_LINE_WIDTH);
	DEFAULT_GESTURE_STROKE = new BasicStroke(PropertyManager.GESTURE_STROKE_LINE_WIDTH);

	ACTIVE_COLOR = Color.black;
	NOT_ACTIVE_COLOR = new Color(125, 125, 125);
	HIGHLIGHT_COLOR = Color.red;
    }

    public static int getNextGOIdentifier() {
	++GO_IDENTIFIER;
	return GO_IDENTIFIER;
    }



    // Disallow the default constructor
    private GraphicalObject() {
    }

    public GraphicalObject(int l, boolean v, boolean s, Storyboard pStoryboard) {
	identifier = GraphicalObject.getNextGOIdentifier();
	storyboard = pStoryboard;

	layer = l;
	visible = v;
	inSelectedLayer = true;
	selected = s;
	performAnalysis = false;
	behaviors = new Vector();
	scaleFactorX = 1.0;
	scaleFactorY = 1.0;

	opMouseIsOver = false;
	opIsHighlighted = false;
	opIsDraggable = false;
	opOriginalX = -1;
	opOriginalY = -1;
	operationalized = false;
    }


    public GraphicalObject(GraphicalObject copy) {
	this(copy, copy.storyboard);
    }

    public GraphicalObject(GraphicalObject copy, Storyboard pStoryboard) {
	identifier = GraphicalObject.getNextGOIdentifier();
	storyboard = pStoryboard;

	layer = copy.layer;
	visible = copy.visible;
	inSelectedLayer = copy.inSelectedLayer;
	selected = copy.selected;
	performAnalysis = copy.performAnalysis;
	behaviors = new Vector();
	scaleFactorX = copy.scaleFactorX;
	scaleFactorY = copy.scaleFactorY;

	opMouseIsOver = false;
	opIsHighlighted = false;
	opIsDraggable = copy.opIsDraggable;
	opOriginalX = -1;
	opOriginalY = -1;
	operationalized = false;
    }


    public GraphicalObject(DataInputStream stream, Storyboard pStoryboard) throws IOException {
	// Read the class identifier
	GO_IDENTIFIER = stream.readInt();

	identifier = stream.readInt();
	storyboard = pStoryboard;

	layer = stream.readInt();
	visible = pStoryboard.isLayerVisible(layer);
	inSelectedLayer = pStoryboard.isLayerSelected(layer);

	selected = false;
	performAnalysis = stream.readBoolean();
	behaviors = new Vector();
	scaleFactorX = stream.readDouble();
	scaleFactorY = stream.readDouble();

	opMouseIsOver = false;
	opIsHighlighted = false;
	opIsDraggable = stream.readBoolean();
	opOriginalX = -1;
	opOriginalY = -1;
	operationalized = false;
    }




    // Public methods
    public void completeIO() {
    }


    public int getIdentifier() {
	return identifier;
    }


    public Storyboard getStoryboard() {
	return storyboard;
    }

    public DataModel getScope() {
	return storyboard.getModel();
    }

    public void setLayer(int l) {
	layer = l;
    }

    public int getLayer() {
	return layer;
    }


    public void setVisible(boolean v) {
	visible = v;
    }

    public boolean getVisible() {
	return visible;
    }

    public boolean isVisible() {
	return visible;
    }


    public void setInSelectedLayer(boolean sel) {
	inSelectedLayer = sel;
    }


    public double getScaleFactorX() {
	return scaleFactorX;
    }

    public double getScaleFactorY() {
	return scaleFactorY;
    }

    public void setScaleFactor(double sx, double sy) {
	scaleFactorX = sx;
	scaleFactorY = sy;
    }


    public void setSelected(boolean v) {
	selected = v;
    }

    public boolean getSelected() {
	return selected;
    }

    public boolean isSelected() {
	return selected;
    }


    public boolean performAnalysis() {
	return performAnalysis;
    }

    public void setPerformAnalysis(boolean b) {
	performAnalysis = b;
    }
        
    
    public Vector getBehaviors() {
	return behaviors;
    }

    public void addBehavior(BehavioralInkStroke b) {
	if (! behaviors.contains(b)) {
	    behaviors.add(b);
	}
    }

    public void addBehavior(BehavioralInkStroke b, int index) {
	addBehavior(b);
	/*
	  if (! behaviors.contains(b)) {
	  behaviors.insertElementAt(b, index);
	  }
	*/
    }
    
    public void removeBehavior(BehavioralInkStroke b) {
	behaviors.remove(b);
    }

    public void removeBehaviors() {
	behaviors.removeAllElements();
    }


    public boolean isMouseOver() {
	return opMouseIsOver;
    }

    public void setMouseIsOver(boolean b) {
	opMouseIsOver = b;
    }


    public boolean isHighlighted() {
	return opIsHighlighted;
    }

    public void setIsHighlighted(boolean b) {
	opIsHighlighted = b;
    }

    public boolean isDraggable() {
	return opIsDraggable;
    }

    public void setIsDraggable(boolean b) {
	opIsDraggable = b;
    }


    public boolean boundsIntersects(Rectangle r) {
	Rectangle go_rect = getBasicBounds();
	return (go_rect.intersects(r));
    }

    // For a default, return boundsIntersects(r)
    public boolean strokeIntersects(Rectangle r) {
	return boundsIntersects(r);
    }

    public boolean contains(Rectangle r) {
	Rectangle go_rect = getBasicBounds();
	return (go_rect.contains(r));
    }

    public boolean contains(int x, int y) {
	Rectangle go_rect = getBasicBounds();
	return (go_rect.contains(x, y));
    }

    public boolean containsBIS(int x, int y) {
	Rectangle r = getBasicBounds();
	Rectangle rect = new Rectangle(r.x - 8, r.y - 8, r.width + 16, r.height + 16);
	return (rect.contains(x, y));
    }

    public boolean containedBy(Rectangle r) {
	return (r.contains(getBasicBounds()));
    }


    protected void setColor(Graphics2D g) {
	if (opIsHighlighted) {
	    g.setColor(HIGHLIGHT_COLOR);
	    
	} else if (operationalized) {
	    g.setColor(ACTIVE_COLOR);

	} else if (inSelectedLayer) {
	    g.setColor(ACTIVE_COLOR);
	    
	} else {
	    g.setColor(NOT_ACTIVE_COLOR);
	}
    }

    protected void setStroke(Graphics2D g) {
	if ((opMouseIsOver) || (opIsHighlighted)) {
	    g.setStroke(MEDIUM_STROKE);
	    
	} else if ((operationalized) || 
		   (scaleFactorX < 1.0) || 
		   (scaleFactorY < 1.0)) {
	    g.setStroke(LIGHT_STROKE);

	} else if (inSelectedLayer) {
	    g.setStroke(DEFAULT_INK_STROKE);
	    
	} else {
	    g.setStroke(LIGHT_STROKE);
	}
    }

    protected void setFont(Graphics2D g) {
    }
    

    public void render(Graphics2D g) {
	if (visible) {
	    setStroke(g);
	    setColor(g);
	    setFont(g);
	    doRender(g);
	    if (selected) {
		drawSelectionHandles(g);
	    }
	}
    }

    
    public void translate(int deltaX, int deltaY, int maxWidth, int maxHeight) {
	Rectangle r = getBounds();
	// Keep object within the bounds
	if ( (r.x + deltaX) < 0) {
	    deltaX = -(r.x);
	} else if ( (r.x + r.width + deltaX) > maxWidth) {
	    deltaX = maxWidth - r.x - r.width;
	}
	
	if ( (r.y + deltaY) < 0) {
	    deltaY = -(r.y);
	} else if ( (r.y + r.height + deltaY) > maxHeight) {
	    deltaY = maxHeight - r.y - r.height;
	}
	doTranslate(deltaX, deltaY);
    }


    // Determine if the object can be translated by the amount within the defined limits
    public boolean canTranslate(int deltaX, int deltaY, int maxWidth, int maxHeight) {
	Rectangle r = getBounds();
	boolean b = true;

	if (deltaX > 0) {
	    // Check right side
	    if ((r.x + r.width + deltaX) >= maxWidth) {
		b = false;
	    }
	} else {
	    // Check left side
	    if ((r.x + deltaX) < 0) {
		b = false;
	    }
	} 

	if (deltaY > 0) {
	    // Check bottom side
	     if ((r.y + r.height + deltaY) >= maxHeight) {
		 b = false;
	     }
	} else {
	    // Check top side
	    if ((r.y + deltaY) < 0) {
		b = false;
	    }
	}
	
	return b;
    }


    public Rectangle getBounds() {
	Rectangle r = doGetBounds();
	if (selected) {
	    int offset = SELECTION_HANDLE_WIDTH / 2;
	    r = (new Rectangle(r.x - offset, r.y - offset,
			       r.width + SELECTION_HANDLE_WIDTH,
			       r.height + SELECTION_HANDLE_HEIGHT));
	}
	return r;
    }
  

    // Completely remove this object from the system
    public void deleteGraphicalObject() {
	deleteGraphicalObject(true);
    }

    public void deleteGraphicalObject(boolean deleteModel) {
    }



    public boolean equals(Object obj) {
	if (obj instanceof GraphicalObject) {
	    return (equals((GraphicalObject) obj));
	}
	return (this == obj);
    }

    public boolean equals(GraphicalObject go) {
	return ((go != null) && (identifier == go.identifier));
    }



    public void mouseHasEntered() {
	setMouseIsOver(true);
	mouseEnter();
    }

    public void mouseHasLeft() {
	mouseLeave();
	setMouseIsOver(false);
    }

    public void mouseWasSingleClicked() {
	mouseSingleClicked();
    }

    public void mouseWasDoubleClicked() {
	mouseDoubleClicked();
    }

    public void wasDraggedAndDropped() {
	draggedAndDropped();
    }



    // Provided for an operational storyboard
    public void opTranslate(int deltaX, int deltaY, int maxWidth, int maxHeight) {
	Rectangle r = getBasicBounds();
	
	if (opOriginalX == -1) {
	    opOriginalX = r.x;
	}
	if (opOriginalY == -1) {
	    opOriginalY = r.y;
	}
	translate(deltaX, deltaY, maxWidth, maxHeight);
    }



    public void resetToOriginalLocation() {
	if ((opOriginalX >= 0) || (opOriginalY >= 0)) {
	    // < 0 is the default value and it is not changed unless
	    // a translation occurs at runtime; e.g., due to drag-n-drop
	    Rectangle r = getBasicBounds();
	    int deltaX = opOriginalX - r.x;
	    int deltaY = opOriginalY - r.y;
	    doTranslate(deltaX, deltaY);
	    opOriginalX = -1;
	    opOriginalY = -1;
	}
    }


    protected void drawSelectionHandles(Graphics2D g) {
	drawSelectionHandles(g, SELECTION_HANDLE_COLOR);
    }
    
    protected void drawSelectionHandles(Graphics2D g, Color color) {
	// Set the color
	g.setColor(color);
	
	// Get strict bounds
	Rectangle rect = getBasicBounds();

	// Draw a selection handle at each corner of the rectangle
	// Halfway along each edge
	int offset = SELECTION_HANDLE_WIDTH / 2;
	g.fillRect(rect.x - offset, rect.y - offset, 
		   SELECTION_HANDLE_WIDTH,
		   SELECTION_HANDLE_HEIGHT);

	g.fillRect((rect.x + (rect.width / 2)) - offset, rect.y - offset, 
		   SELECTION_HANDLE_WIDTH,
		   SELECTION_HANDLE_HEIGHT);

	g.fillRect(rect.x + rect.width - offset, rect.y - offset, 
		   SELECTION_HANDLE_WIDTH,
		   SELECTION_HANDLE_HEIGHT);
	
	g.fillRect(rect.x + rect.width - offset, (rect.y + (rect.height / 2)) - offset, 
		   SELECTION_HANDLE_WIDTH,
		   SELECTION_HANDLE_HEIGHT);
	
	g.fillRect(rect.x + rect.width - offset, rect.y + rect.height - offset, 
		   SELECTION_HANDLE_WIDTH,
		   SELECTION_HANDLE_HEIGHT);

	g.fillRect((rect.x + (rect.width / 2)) - offset, rect.y + rect.height - offset, 
		   SELECTION_HANDLE_WIDTH,
		   SELECTION_HANDLE_HEIGHT);

	g.fillRect(rect.x - offset, rect.y + rect.height - offset, 
		   SELECTION_HANDLE_WIDTH,
		   SELECTION_HANDLE_HEIGHT);
	
	g.fillRect(rect.x - offset, (rect.y + (rect.height / 2)) - offset, 
		   SELECTION_HANDLE_WIDTH,
		   SELECTION_HANDLE_HEIGHT);
    }



    private void draggedAndDropped() {
	DragAndDropEvent dndEvent = null;
	Iterator it = behaviors.iterator();
	while (it.hasNext()) {
	    BehavioralInkStroke bStroke = (BehavioralInkStroke) it.next();
	    SourceEvent sEvent = bStroke.getSourceEvent();
	    if (sEvent instanceof DragAndDropEvent) {
		DragAndDropEvent tmpDnd = (DragAndDropEvent) sEvent;
		GraphicalObject assignedTarget = tmpDnd.getTarget();    
		if ((assignedTarget != null) &&
		    (assignedTarget.isVisible()) &&
		    (boundsIntersects(assignedTarget.getBasicBounds()))) {
		    dndEvent = tmpDnd;
		    break;
		}
	    }
	}
	
	if (dndEvent != null) {
	    GraphicalObject assignedTarget = dndEvent.getTarget();	    
	    NamedModel sourceObject = getNamedModel();
	    NamedModel targetObject = assignedTarget.getNamedModel();
	    
	    if ( (sourceObject instanceof StaticMediaModel) &&
		 (targetObject instanceof MediaModel)) {
		StaticMediaModel sourceModel = (StaticMediaModel) sourceObject;
		sourceModel.setWasDraggedAndDroppedOn(true, (MediaModel) targetObject, true);
		
		// Snap the dragged object to the target location - prevent further dragging
		storyboard.opSnapDraggableToLocation(assignedTarget.getBasicBounds());
		storyboard.removeSourceEventObject(this);
		
	    } else {
		storyboard.opResetDraggableToOriginalLocation();
	    }
	    
	} else {
	    storyboard.opResetDraggableToOriginalLocation();
	}
	
       // Use paint immediately, otherwise, a transition will blink due to a delayed repaint!
       storyboard.paintImmediately(0, 0, storyboard.getWidth(), storyboard.getHeight());
   }




    // Can be overridden, but not mandatory
    public void realize(boolean showControls, boolean recreate) {
    }
    // Unrealize this object
    public void unRealize(boolean keepVisuals) {
    }


    public void operationalize() {
	// Mark that the object is operationalized
	operationalized = true;
	
	// The layer models have already been set
	// Do not call setVisible() here
	DataModel dm = storyboard.getModel();
	if (dm != null) {
	    visible = dm.isLayerVisible(layer);
	} else {
	    visible = true;
	}
	// Unselect the object
	selected = false;
    }

    public void unOperationalize() {
	DataModel model = getScope();
	if (model != null) {
	    // The layer model information has already been set!
	    setVisible(model.isLayerVisible(layer));
	}
	setMouseIsOver(false);
	setIsHighlighted(false);
	setIsDraggable(false);
	resetToOriginalLocation();
	operationalized = false;
    }


    public void resetToOriginalScale() {
	scaleFactorX = 1.0;
	scaleFactorY = 1.0;
    }

    public void scale(double sx, double sy) {
	scaleFactorX = sx;
	scaleFactorY = sy;
    }


    // Abstract methods
    public abstract String getNamedModelId();
    public abstract NamedModel getNamedModel();
    public abstract Rectangle getBasicBounds();
    public abstract void doTranslate(int deltaX, int deltaY);
    public abstract Object clone();
    public abstract Object clone(Storyboard toStoryboard);
    public abstract Object clone(Storyboard toStoryboard, boolean copyModel);

    protected abstract void mouseEnter();
    protected abstract void mouseLeave();
    protected abstract void mouseSingleClicked();
    protected abstract void mouseDoubleClicked();
    protected abstract void doRender(Graphics2D g);
    protected abstract Rectangle doGetBounds();
    protected void mediateRender(Graphics2D g) {
    }


    public void write(DataOutputStream stream) throws IOException {
	// Write the class identifier value
	stream.writeInt(GO_IDENTIFIER);

	stream.writeInt(identifier);
	stream.writeInt(layer);
	stream.writeBoolean(performAnalysis);
	stream.writeDouble(scaleFactorX);
	stream.writeDouble(scaleFactorY);
	stream.writeBoolean(opIsDraggable);

    }

}



