/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File GraphicalObjectCollection.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs;
    

// Necessary imports
import java.lang.*;
import java.util.*;
import java.io.*;
import java.awt.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.usrinterface.base.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs.behavior.*;
import edu.uofminnesota.DEMAIS.usrinterface.util.*;



public class GraphicalObjectCollection extends LinkedList {
   
    // containing storyboard
    private Storyboard parentStoryboard;

    

    public GraphicalObjectCollection() {
	this(null, 20);
    }

    public GraphicalObjectCollection(Storyboard pStoryboard, int capacity) {
	//super(capacity);
	parentStoryboard = pStoryboard;
    }
    

    public GraphicalObjectCollection(GraphicalObjectCollection copy, Storyboard pStoryboard, 
				     boolean deepCopyOfModels) {
	super();
	parentStoryboard = pStoryboard;

	// Copy over the graphical objects
	if (deepCopyOfModels) {
	    // Due to a storyboard copy
	    copyAllObjects(copy, deepCopyOfModels);
	
	} else {
	    // Due to CM and MVs making a copy
	    Iterator it = copy.iterator();
	    while (it.hasNext()) {
		GraphicalObject go = (GraphicalObject) it.next();
		GraphicalObject go_clone = (GraphicalObject) go.clone(parentStoryboard, deepCopyOfModels);
		go_clone.setSelected(false);
		add(go_clone);
	    }
	}
    }

    public GraphicalObjectCollection(DataInputStream stream, Storyboard pStoryboard) throws IOException {
	super();
	parentStoryboard = pStoryboard;

	// Read number of graphical objects that follow
	int numObjects = stream.readInt();

	// Read the objects
	for (int i=0; i<numObjects; ++i) {
	    GraphicalObject go = null;
	    int classid = stream.readInt();

	    switch (classid) {
	    case TextGraphicalObject.CLASSID:
		go = new TextGraphicalObject(stream, pStoryboard);
		break;

	    case AudioRectangleInkStroke.CLASSID:
		// Read superclass identifiers
		stream.readInt();
		stream.readInt();
		go = new AudioRectangleInkStroke(stream, pStoryboard);
		break;

	    case VideoRectangleInkStroke.CLASSID:
		// Read superclass identifiers
		stream.readInt();
		stream.readInt();
		go = new VideoRectangleInkStroke(stream, pStoryboard);
		break;

	    case ImageRectangleInkStroke.CLASSID:
		stream.readInt();
		stream.readInt();
		go = new ImageRectangleInkStroke(stream, pStoryboard);
		break;

	    case RectangleInkStroke.CLASSID:
		stream.readInt();
		go = new RectangleInkStroke(stream, pStoryboard);
		break;

	    case BehavioralInkStroke.CLASSID:
		stream.readInt();
		go = new BehavioralInkStroke(stream, pStoryboard);
		break;

	    case InkLine.CLASSID:
		go = new InkLine(stream, pStoryboard);
		break;

	    case InkStroke.CLASSID:
		go = new InkStroke(stream, pStoryboard);
		break;

	    case BeginAnchor.CLASSID:
		go = new BeginAnchor(stream, pStoryboard);
		go.setVisible(true);
		break;

	    default:
		break;
	    }

	    if (go != null) {
		add(go);
	    }
	}
    }
    

    public void completeIO() {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.completeIO();
	}
    }


    // Due to a storyboard copy - Make an exact duplicate of the collection c
    private void copyAllObjects(GraphicalObjectCollection c, boolean copyOfModels) {
	Vector bStrokes = new Vector();
	Iterator it = c.iterator();
	while (it.hasNext()) {
	    // Copy all non-behavioral ink strokes first
	    // Do not copy any InkLines
	    GraphicalObject go = (GraphicalObject) it.next();
	    if ( (go instanceof TextGraphicalObject) || 
		 (go instanceof RectangleInkStroke) ||
		 (go instanceof BeginAnchor)) {

		// Add a clone 
		GraphicalObject go_clone = (GraphicalObject) go.clone(parentStoryboard, copyOfModels);
		go_clone.setSelected(false);
		go_clone.realize(true, true);
		add(go_clone);
		
	    } else if (go instanceof BehavioralInkStroke) {
		bStrokes.add(go);

	    } else if (go instanceof InkStroke) {
		GraphicalObject go_clone = (GraphicalObject) go.clone(parentStoryboard);
		go_clone.setSelected(false);
		add(go_clone);

	    } else if (go instanceof InkLine) {
		// Do nothing
	    }

	}
	
	// Now handle the behavioral ink strokes
	it = bStrokes.iterator();
	while (it.hasNext()) {
	    BehavioralInkStroke bStroke = (BehavioralInkStroke) it.next();
	    BehavioralInkStroke bStroke_clone = (BehavioralInkStroke) bStroke.clone(null, null, parentStoryboard);
	    bStroke_clone.setSelected(false);
	    add(bStroke_clone);

	    Point sourcePoint = bStroke_clone.getFirstPoint();
	    Point destPoint = bStroke_clone.getLastPoint();
	    int layer = bStroke_clone.getLayer();
	    int sourceLayer = layer;
	    int destLayer = layer;
	    GraphicalObject tmpSourceObject = bStroke.getSourceEventObject();
	    if (tmpSourceObject != null) {
		sourceLayer = tmpSourceObject.getLayer();
	    }
	    GraphicalObject tmpDestObject = bStroke.getActionObject();
	    if (tmpDestObject != null) {
		destLayer = tmpDestObject.getLayer();
	    }
	    GraphicalObject sourceObject = getRectangleOrTextContaining(sourcePoint.x, sourcePoint.y, sourceLayer, true);
	    GraphicalObject destObject = getRectangleOrTextContaining(destPoint.x, destPoint.y, destLayer, true);

	    SourceEvent sEvent = bStroke.getSourceEvent(); // From the original
	    DestinationAction dAction = bStroke.getAction(); // From the original
	    if (sEvent != null) {
		SourceEvent event_clone = (SourceEvent) sEvent.clone(sourceObject);
		if (event_clone instanceof DragAndDropEvent) {
		    DragAndDropEvent dndEvent = (DragAndDropEvent) event_clone;
		    dndEvent.setTarget(destObject);
		} 
		bStroke_clone.setSourceEvent(event_clone);
	    }

	    if (dAction != null) {
		DestinationAction action_clone = (DestinationAction) dAction.clone(destObject);
		bStroke_clone.setAction(action_clone);
	    }
	}
    }


    // Due to a paste operation - can be in a different storyboard, layer, or both.
    public Collection pasteObjects(Collection c, int deltaX, int deltaY, int layer, boolean selected, Storyboard parentS) {
	Vector copiedObjects = new Vector();
	int pWidth = Storyboard.WIDTH;
	int pHeight = Storyboard.HEIGHT;

	Vector bStrokes = new Vector();
	Iterator it = c.iterator();
	while (it.hasNext()) {
	    // Copy all non-behavioral ink strokes first
	    // Do not copy the BeginAnchor or any InkLines
	    GraphicalObject go = (GraphicalObject) it.next();
	    // Unselected the object being copied
	    go.setSelected(false);

	    if ( (go instanceof TextGraphicalObject) || 
		 (go instanceof RectangleInkStroke) ) {
		
		// Add a clone 
		GraphicalObject go_clone = (GraphicalObject) go.clone(parentS, true);
		if (go_clone.canTranslate(deltaX, deltaY, pWidth, pHeight)) {
		    go_clone.translate(deltaX, deltaY, pWidth, pHeight);

		} else if (go_clone.canTranslate(deltaX, 0, pWidth, pHeight)) {
		    go_clone.translate(deltaX, 0, pWidth, pHeight);
		
		} else if (go_clone.canTranslate(0, deltaY, pWidth, pHeight)) {
		    go_clone.translate(0, deltaY, pWidth, pHeight);
		
		} 
		
		go_clone.setLayer(layer);
		go_clone.setSelected(selected);
		go_clone.setVisible(parentStoryboard.getModel().isLayerVisible(layer));
		go_clone.realize(true, true);

		add(go_clone);
		copiedObjects.add(go_clone);
	    
	    } else if ((go instanceof BeginAnchor) ||
		       (go instanceof InkLine)) {
		// do nothing
		
	    } else if (go instanceof BehavioralInkStroke) {
		bStrokes.add(go);
		
	    } else if (go instanceof InkStroke) {
		GraphicalObject go_clone = (GraphicalObject) go.clone(parentS);
		if (go_clone.canTranslate(deltaX, deltaY, pWidth, pHeight)) {
		    go_clone.translate(deltaX, deltaY, pWidth, pHeight);
		    
		} else if (go_clone.canTranslate(deltaX, 0, pWidth, pHeight)) {
		    go_clone.translate(deltaX, 0, pWidth, pHeight);
		
		} else if (go_clone.canTranslate(0, deltaY, pWidth, pHeight)) {
		    go_clone.translate(0, deltaY, pWidth, pHeight);
		} 

		go_clone.setLayer(layer);
		go_clone.setSelected(selected);
		go_clone.setVisible(parentStoryboard.getModel().isLayerVisible(layer));
		add(go_clone);
		copiedObjects.add(go_clone);
	    }
	}

	if (!parentS.getModel().getLayerModel(layer).isAnnotationLayer()) {
	    // Now handle the behavioral ink strokes
	    it = bStrokes.iterator();
	    while (it.hasNext()) {
		BehavioralInkStroke bStroke = (BehavioralInkStroke) it.next();
		BehavioralInkStroke bStroke_clone = (BehavioralInkStroke) bStroke.clone(null, null, parentS);
		if (bStroke_clone.canTranslate(deltaX, deltaY, pWidth, pHeight)) {
		    bStroke_clone.translate(deltaX, deltaY, pWidth, pHeight);
		    
		} else if (bStroke_clone.canTranslate(deltaX, 0, pWidth, pHeight)) {
		    bStroke_clone.translate(deltaX, 0, pWidth, pHeight);
		    
		} else if (bStroke_clone.canTranslate(0, deltaY, pWidth, pHeight)) {
		    bStroke_clone.translate(0, deltaY, pWidth, pHeight);
		} 

		bStroke_clone.setLayer(layer);
		bStroke_clone.setSelected(selected);
		bStroke_clone.setVisible(parentStoryboard.getModel().isLayerVisible(layer));
		
		Point sourcePoint = bStroke_clone.getFirstPoint();
		Point destPoint = bStroke_clone.getLastPoint();
		GraphicalObject sourceObject = getRectangleOrTextContaining(sourcePoint.x, sourcePoint.y, layer, true);
		
		GraphicalObject destObject = getRectangleOrTextContaining(destPoint.x, destPoint.y, layer, true);
		
		// Must have a source and dest object to continue
		if (((sourceObject != null) &&
		     (destObject != null)) &&
		    ((sourceObject.getLayer() <= layer) &&
		    (destObject.getLayer() <= layer))) {
		    add(bStroke_clone);
		    copiedObjects.add(bStroke_clone);
		    
		    SourceEvent sEvent = bStroke.getSourceEvent(); // From the original
		    DestinationAction dAction = bStroke.getAction(); // From the original
		    if (sEvent != null) {
			SourceEvent event_clone = (SourceEvent) sEvent.clone(sourceObject);
			if (event_clone instanceof DragAndDropEvent) {
			    DragAndDropEvent dndEvent = (DragAndDropEvent) event_clone;
			    dndEvent.setTarget(destObject);
			} 
			bStroke_clone.setSourceEvent(event_clone);
		    }
		    
		    if (dAction != null) {
			DestinationAction action_clone = (DestinationAction) dAction.clone(destObject);
			bStroke_clone.setAction(action_clone);
		    }
		}
	    }
	}

	return copiedObjects;
    }
	



    // Add and Remove methods

    // Add object to the end of its layer
    public boolean add(Object obj) {
	add((GraphicalObject) obj);
	return true;
    }

    // Add object to the end of its layer
    public void add(GraphicalObject go) {
	int index = getLayerEndIndex(go.getLayer());
	add(index, go);
    }
    
    public void add(GraphicalObject go, int layer) {
	go.setLayer(layer);
	add(go);
    }


    public void deleteGraphicalObjects(boolean deleteModel) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.deleteGraphicalObject(deleteModel);
	}
    }


    public void delete(Object obj) {
	delete((GraphicalObject) obj);
    }

    public void delete(GraphicalObject go) {
	go.deleteGraphicalObject(true);
	super.remove(go);

	// If the object has attached behaviors, then delete them too regardless of the layer
	Vector behavior_clone = (Vector) go.getBehaviors().clone();
	Iterator it = behavior_clone.iterator();
	while (it.hasNext()) {
	    BehavioralInkStroke bStroke = (BehavioralInkStroke) it.next();

	    // Remove the behavioral ink stroke from its storyboard container,
	    // which could be in a multiview
	    bStroke.getStoryboard().deleteGraphicalObject(bStroke);
	}
    }

    public void deleteAll(Collection c) {
	Iterator it = c.iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    delete(go);
	}
    }

    public void deleteAll() {
	// Create a local set of references
	LinkedList this_clone = (LinkedList) super.clone();
	
	Iterator it = this_clone.iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    delete(go);
	}
	super.clear();
    }



    public void copyLayer(int layer, int toLayer) {
	Vector toCopy = new Vector();
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() == layer) {
		toCopy.add(go);
	    }
	}
	// This function will handle any BIS issues
	pasteObjects(toCopy, 0, 0, toLayer, false, parentStoryboard);
    }


    // This will also delete BIS's not in the current layer, but that are
    // attached to objects being removed
    public void deleteLayer(int layer) {
	Vector toDelete = new Vector();
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    int go_layer = go.getLayer();
	    if (go_layer == layer) {
		toDelete.add(go);
	    } else if (go_layer > layer) {
		go.setLayer(go_layer - 1);
	    }
	}
	deleteAll(toDelete);
    }


    public void repositionLayer(int layerNum, int toPosition) {
	int maxLayer = Math.max(layerNum, toPosition);
	Vector from = new Vector();

	// Retrieve elements in layer layerNum
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() > maxLayer) {
		// Easy performance boost - no further candidates
		break;
	    } else if (go.getLayer() == layerNum) {
		from.add(go);
	    }
	}
	
	// Remove all these elements from the main list
	removeAll(from);

	// Decrement layer for layers > layerNum
	it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() > layerNum) {
		go.setLayer(go.getLayer() - 1);
	    }
	}

	if (toPosition >= layerNum) {
	    --toPosition;
	}

	// Set layer on removed elements to toPosition
	it = from.iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.setLayer(toPosition);
	}

	// Increment layer for layers >= toPosition
	it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() >= toPosition) {
		go.setLayer(go.getLayer() + 1);
	    }
	}

	// Reinsert the removed elements
	int insertIndex = getLayerEndIndex(toPosition - 1);
	addAll(insertIndex, from);
    }


    public void flattenLayers() {
	Iterator it = iterator();
	
	// Skip the Begin Anchor, if any
	if (it.hasNext()) {
	    GraphicalObject ba = (GraphicalObject) getFirst();
	    if (ba instanceof BeginAnchor) {
		it.next();
	    }
	}
	
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.setLayer(0);
	    go.setVisible(true);
	}
    }


    public void setLayerVisible(int layer, boolean visible) {
	Iterator it = iterator();
	while(it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() == layer) {
		go.setVisible(visible);
	    }
	}
    }

    public void setLayerSelected(int layer, boolean sel) {
	Iterator it = iterator();
	while(it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() == layer) {
		go.setInSelectedLayer(sel);
	    }
	}
    }



    /*
      Method invoked only during play mode - only set visible on objects that are both contained by go and in the same or lower layer (parameterized)
    */
    public void setVisibleOnObjectsContainedBy(Rectangle r, boolean visible, int layer, boolean searchLowerLayers, GraphicalObject exempt) {
	Collection c = getObjectsContainedBy(r, layer, searchLowerLayers);
	if (exempt != null) {
	    c.remove(exempt);
	}

	Iterator it = c.iterator();
	while(it.hasNext()) {
	    GraphicalObject cand = (GraphicalObject) it.next();
	    cand.setVisible(visible);
	}
    }

    // Method invoked during play mode only
    public void setMouseIsOverOnObjectsContainedBy(Rectangle r, boolean isOver, int layer, boolean searchLowerLayers, GraphicalObject exempt) {
	Collection c = getObjectsContainedBy(r, layer, searchLowerLayers);
	if (exempt != null) {
	    c.remove(exempt);
	}

	Iterator it = c.iterator();
	while (it.hasNext()) {
	    GraphicalObject cand = (GraphicalObject) it.next();
	    cand.setMouseIsOver(isOver);
	}
    }


    // Rendering methods
    public void render(Graphics2D g) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.render(g);
	}
    }

    public void render(Collection c, Graphics2D g) {
	if (c != null) {
	    Iterator it = c.iterator();
	    while (it.hasNext()) {
		GraphicalObject go = (GraphicalObject) it.next();
		go.render(g);
	    }
	}
    }




    // Intersection and containment methods that respect layering

    // Return a list of GraphicalObjects whose rectangular bounds
    // intersect r in any way, and are in the same or lower layer
    public Collection getBoundsIntersectingObjects(Rectangle r, int layer, boolean searchLowerLayers) {
	Vector intersect = new Vector();
	Iterator it = iterator();
	
	// OK to check begin anchor
	if (searchLowerLayers) {
	    while (it.hasNext()) {
		GraphicalObject cand = (GraphicalObject) it.next();
		int cLayer = cand.getLayer();
		if ((cLayer <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    (cand.boundsIntersects(r))) {
		    intersect.add(cand);
		}
	    }
	} else {
	    while (it.hasNext()) {
		GraphicalObject cand = (GraphicalObject) it.next();
		int cLayer = cand.getLayer();
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    (cand.boundsIntersects(r))) {
		    intersect.add(cand);
		}
	    }
	}
	return intersect;
    }


    public Collection getRectangleOrTextBoundsIntersectingObjects(Rectangle r, int layer, boolean searchLowerLayers) {
	Vector intersect = new Vector();
	Iterator it = iterator();

	// Skip the Begin Anchor, if any
	if (it.hasNext()) {
	    GraphicalObject ba = (GraphicalObject) getFirst();
	    if (ba instanceof BeginAnchor) {
		it.next();
	    }
	}

	if (searchLowerLayers) {
	    while (it.hasNext()) {
		GraphicalObject cand = (GraphicalObject) it.next();
		int cLayer = cand.getLayer();
		if ((cLayer <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    ((cand instanceof RectangleInkStroke) ||
		     (cand instanceof TextGraphicalObject)) &&
		    (cand.boundsIntersects(r))) {
		    intersect.add(cand);
		}
	    }
	} else {
	    while (it.hasNext()) {
		GraphicalObject cand = (GraphicalObject) it.next();
		int cLayer = cand.getLayer();
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    ((cand instanceof RectangleInkStroke) ||
		     (cand instanceof TextGraphicalObject)) &&
		    (cand.boundsIntersects(r))) {
		    intersect.add(cand);
		}
	    }
	}
	
	return intersect;
    }
    


    // Interface for cut gesture
    public Collection getStrokeIntersectingOrContainedObjects(Rectangle r, int layer, boolean searchLowerLayers) {
	Vector candidates = new Vector();
	Iterator it = iterator();

	// Skip the Begin Anchor, if any
	if (it.hasNext()) {
	    GraphicalObject ba = (GraphicalObject) getFirst();
	    if (ba instanceof BeginAnchor) {
		it.next();
	    }
	}

	if (searchLowerLayers) {
	    while (it.hasNext()) {
		GraphicalObject cand = (GraphicalObject) it.next();
		int cLayer = cand.getLayer();
		if ((cand.getLayer() <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    (cand.boundsIntersects(r))) {
		    candidates.add(cand);
		}
	    }
	} else {
	    while (it.hasNext()) {
		GraphicalObject cand = (GraphicalObject) it.next();
		int cLayer = cand.getLayer();
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    (cand.boundsIntersects(r))) {
		    candidates.add(cand);
		}
	    }
	}

	Vector intersect = new Vector();
	it = candidates.iterator();
	while (it.hasNext()) {
	    GraphicalObject cand = (GraphicalObject) it.next();
	    if ((cand.strokeIntersects(r)) ||
		(cand.containedBy(r))) {
		intersect.add(cand);
	    }
	}
	return intersect;
    }


    public BehavioralInkStroke getBehavioralInkStrokeContaining(int x, int y, boolean bySource, int layer, boolean searchLowerLayers) {

	BehavioralInkStroke bestStroke = null;
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    BehavioralInkStroke bStroke = null;
	    int cLayer = go.getLayer();
		
	    if (searchLowerLayers) {
		if ((cLayer <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    (go instanceof BehavioralInkStroke)) {
		    bStroke = (BehavioralInkStroke) go;
		}
	    } else {
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    (go instanceof BehavioralInkStroke)) {   
		    bStroke = (BehavioralInkStroke) go;
		}
	    }
	    // Have a behavioral inkstroke to check
	    if (bStroke != null) {
		Rectangle r = null;
		if (bySource) {
		    r = bStroke.getSourceEventIconTargetBounds();
		} else {
		    r = bStroke.getActionIconTargetBounds();
		}
		if ((r != null) && (r.contains(x, y))) {
		    bestStroke = bStroke;
		    break;
		}
	    }
	}
	return bestStroke;
    }


    // Containment methods
    public GraphicalObject getRectangleContaining(int x, int y, int layer, boolean searchLowerLayers) {
	ListIterator it = listIterator(size());
	
	if (searchLowerLayers) {
	    while (it.hasPrevious()) {
		GraphicalObject go = (GraphicalObject) it.previous();
		int cLayer = go.getLayer();
		if ((go.getLayer() <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    (go instanceof RectangleInkStroke) && 
		    (go.contains(x, y))) {
		    return go;
		}
	    }

	} else {
	    while (it.hasPrevious()) {
		GraphicalObject go = (GraphicalObject) it.previous();
		int cLayer = go.getLayer();
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    (go instanceof RectangleInkStroke) && 
		    (go.contains(x, y))) {   
		    return go;
		}
	    }
	}
	return null;
    }


    public GraphicalObject getRectangleOrTextContaining(int x, int y, int layer, boolean searchLowerLayers) {
	ListIterator it = listIterator(size());

	if (searchLowerLayers) {
	    while (it.hasPrevious()) {
		GraphicalObject go = (GraphicalObject) it.previous();
		int cLayer = go.getLayer();
		if ((cLayer <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    ((go instanceof RectangleInkStroke) || 
		     (go instanceof TextGraphicalObject) ||
		     (go instanceof BeginAnchor)) &&
		    (go.contains(x, y))) {
		    return go;
		}
	    }
	} else {
	    while (it.hasPrevious()) {
		GraphicalObject go = (GraphicalObject) it.previous();
		int cLayer = go.getLayer();
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    ((go instanceof RectangleInkStroke) || 
		     (go instanceof TextGraphicalObject) ||
		     (go instanceof BeginAnchor)) &&
		    (go.contains(x, y)) ) {
		    return go;
		}
	    }
	}
	return null;
    }


    public Vector getRectangleOrTextContainingBIS(int x, int y, int layer, boolean searchLowerLayers) {
	ListIterator it = listIterator(size());
	Vector cand = new Vector();

	if (searchLowerLayers) {
	    while (it.hasPrevious()) {
		GraphicalObject go = (GraphicalObject) it.previous();
		int cLayer = go.getLayer();
		if ((cLayer <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    ((go instanceof RectangleInkStroke) || 
		     (go instanceof TextGraphicalObject) ||
		     (go instanceof BeginAnchor)) &&
		    (go.containsBIS(x, y))) {
		    cand.add(go);
		}
	    }
	} else {
	    while (it.hasPrevious()) {
		GraphicalObject go = (GraphicalObject) it.previous();
		int cLayer = go.getLayer();
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    ((go instanceof RectangleInkStroke) || 
		     (go instanceof TextGraphicalObject) ||
		     (go instanceof BeginAnchor)) &&
		    (go.containsBIS(x, y)) ) {
		    cand.add(go);
		}
	    }
	}

	// Filter the candidates
	cand = DMath.findClosestAndPeripherals(cand, x, y, 24);
	return cand;
    }


    public GraphicalObject getTextContaining(int x, int y, int layer, boolean searchLowerLayers) {
	ListIterator it = listIterator(size());

	if (searchLowerLayers) {
	    while (it.hasPrevious()) {
		GraphicalObject go = (GraphicalObject) it.previous();
		int cLayer = go.getLayer();
		if ((cLayer <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    (go instanceof TextGraphicalObject) &&
		    (go.contains(x, y))) {		    
		    return go;
		}
	    }
	} else {
	    while (it.hasPrevious()) {
		GraphicalObject go = (GraphicalObject) it.previous();
		int cLayer = go.getLayer();
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    (go instanceof TextGraphicalObject) &&
		    (go.contains(x, y))) {
		    return go;
		}
	    }
	}
	return null;
    }

  
    public Vector getObjectsContainedBy(Rectangle r, int layer, boolean searchLowerLayers) {
	Vector contained = new Vector();
	Iterator it = iterator();
	
	// Skip the Begin Anchor, if any
	if (it.hasNext()) {
	    GraphicalObject ba = (GraphicalObject) getFirst();
	    if (ba instanceof BeginAnchor) {
		it.next();
	    }
	}

	if (searchLowerLayers) {
	    while (it.hasNext()) {
		GraphicalObject cand = (GraphicalObject) it.next();
		int cLayer = cand.getLayer();
		if ((cLayer <= layer) &&
		    (isLayerVisible(cLayer)) &&
		    (cand.containedBy(r))) {
		    contained.add(cand);
		}
	    }
	} else {
	    while (it.hasNext()) {
		GraphicalObject cand = (GraphicalObject) it.next();
		int cLayer = cand.getLayer();
		if ((cLayer == layer) &&
		    (isLayerVisible(cLayer)) &&
		    (cand.containedBy(r))) {
		    contained.add(cand);
		}
	    }
	}

	return contained;
    }



    public GraphicalObject getGraphicalObjectById(int id) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getIdentifier() == id) {
		return go;
	    }
	}
	return null;
    }



    public void unSelectAllObjects() {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.setSelected(false);
	}
    }
    
    public void unSelectAllObjects(int layer) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() == layer) {
		go.setSelected(false);
	    }
	}
    }


    public Collection getSelectedGraphicalObjects(int layer) {
	Vector selectedObjs = new Vector();
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if ((go.getLayer() == layer) &&
		(go.isSelected())) {
		selectedObjs.add(go);
	    }
	}
	return selectedObjs;
    }


    public boolean isGraphicalObjectSelected(int layer) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if ((go.getLayer() == layer) &&
		(go.isSelected())) {
		return true;
	    }
	}
	return false;
    }



    // Scaling methods
    public void resetToOriginalScale() {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.resetToOriginalScale();
	}
    }


    public void scaleTo(int width, int height) {
	scaleTo(width, height, 0);
    }
    
    public void scaleTo(int width, int height, int vAdjustment) {
	double sx = ( (double) width  / (double) Storyboard.WIDTH);
	double sy = ( (double) (height - vAdjustment) / (double) Storyboard.HEIGHT);
	scale(sx, sy);
    }

    public void scale(double sx, double sy) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.scale(sx, sy);
	}
    }


    public void translate(int deltaX, int deltaY, int mWidth, int mHeight) {
	Iterator it = iterator();
	while(it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.translate(deltaX, deltaY, mWidth, mHeight);
	}
    }
    
    public void translate(Collection c, int deltaX, int deltaY, int mWidth, int mHeight) {
	Iterator it = c.iterator();
	while(it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.translate(deltaX, deltaY, mWidth, mHeight);
	}
    }




    // Mainly for any continuous media that are in this collection
    public void realize(boolean showControls, boolean recreate) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.realize(showControls, recreate);
	}
    }


    public void unRealize(boolean keepVisuals) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.unRealize(keepVisuals);
	}
    }


    public void operationalize() {
	Vector bStrokes = new Vector();
	Vector dndObjects = new Vector();

	// Operationalize the non-behavioral strokes first
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (!(go instanceof BehavioralInkStroke)) {
		// Operationalize this object
		go.operationalize();
		
	    } else {
		// Make sure the ink strokes are operationalized in order!
		bStrokes.add(go);
	    }
	}

	// Operationalize the behavioral strokes second

	/* 
	   BUG - This set of code does not enforce correct semantics required by the addition of sequence numbers on the strokes. It is possible that the order that the strokes were drawn is not consistent with the sequence numbering on those strokes. However, the code is fine for demonstrating the idea
	*/

	it = bStrokes.iterator();
	while (it.hasNext()) {
	    BehavioralInkStroke bis = (BehavioralInkStroke) it.next();
	    // Operationalize this object
	    bis.operationalize();

	    // Determine if this was a DND event, and if so,
	    // save the object for later processing
	    GraphicalObject go = bis.getSourceEventObject();
	    if ((go != null) &&
		(go.isDraggable())) {
		if (! dndObjects.contains(go)) {
		    dndObjects.add(go);
		}
	    }
	}

	// Move each DND object to the end of its current layer
	it = dndObjects.iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    int layer = go.getLayer();
	    add(getLayerEndIndex(layer), go);
	    remove(go);
	}
    }


    public void unOperationalize() {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();	    
	    // unOperationalize this object
	    go.unOperationalize();
	}
    }


    public void opResetCMObjects(GraphicalObject exempt) {
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if ((go instanceof ContinuousMediaRectangleInkStroke) &&
		(go != exempt)) {
		ContinuousMediaRectangleInkStroke cmris = (ContinuousMediaRectangleInkStroke) go;
		// stop this object
		cmris.reset();
	    }
	}
    }

    
    public Object superClone() {
	return super.clone();
    }
    
    public Object clone() {
	return (clone(parentStoryboard));
    }

    public Object clone(Storyboard pStoryboard) {
	return (clone(pStoryboard, true));
    }

    public Object clone(Storyboard pStoryboard, boolean deepCopyOfModels) {
	return (new GraphicalObjectCollection(this, pStoryboard, deepCopyOfModels));
    }


    
    // Retrieve the index at which the layer begins
    private boolean isLayerVisible(int layer) {
	return (parentStoryboard.isLayerVisible(layer) ||
		(layer == -1));
    }


    private int getLayerBeginIndex(int layer) {
	int i = 0;
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() == layer) {
		return i;
	    }
	    ++i;
	}
	return 0;
    }




    // Retrieve the index at which the layer ends 
    private int getLayerEndIndex(int layer) {
	// Simple check to improve performance
	if (size() > 0) {
	    GraphicalObject go = (GraphicalObject) getLast();
	    if (go.getLayer() <= layer) {
		return size();
	    }
	}

	int i = 0;
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    if (go.getLayer() > layer) {
		return i;
	    }
	    ++i;
	}

	return size();
    }


    
    public void write(DataOutputStream stream) throws IOException {
	// Write total number of objects
	stream.writeInt(size());
	
	// Write each graphical object
	Iterator it = iterator();
	while (it.hasNext()) {
	    GraphicalObject go = (GraphicalObject) it.next();
	    go.write(stream);
	}
    }
    
}
