/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File TextGraphicalObject.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.graphicalobjs;
    

// Necessary imports
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.JFrame;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.meal.*;
import edu.uofminnesota.DEMAIS.core.annotation.*;
import edu.uofminnesota.DEMAIS.core.parser.*;
import edu.uofminnesota.DEMAIS.usrinterface.storyboard.*;
import edu.uofminnesota.DEMAIS.usrinterface.util.*;




public class TextGraphicalObject extends GraphicalObject {

    // ID for file I/O
    public static final int CLASSID = 170;


    // Initial position of the text
    private int startX, startY;

    // Width and height of the bounding box
    private int width, height;

    // Cursor position
    private int cursorX, cursorY;

    // Cursor length
    private int cursorLength;

    // Used for blinking the cursor
    private boolean cursorToggle;

    // List index for next character
    private int insertionIndex;

    // Key listener
    private Listener listener;

    // Temp variables
    private int pX, pY, pWidth, pHeight;

    // Storyboard font - basic font from which others are derived
    private Font baseFont;

    // For scaling
    private Font originalFont;

    // Selected font
    private Font selectedFont;

    // Operational font
    private Font opFont;

    // Operational rollover font
    private Font opRolloverFont;

    // Operational highlight font 
    private Font opHighlightFont;

    // The undelying MEAL text model
    private TextModel textModel;

    // Timer for the cursor
    private javax.swing.Timer timer;

    // cached value only for navigation action text
    private String destination;

    // whether the initial text extry is taking place
    private boolean initialEntry;

    // whether the text object is currently being edited
    private boolean edited;




    // Constructors
    private TextGraphicalObject() {
	this(0, 0, 0, false, false, null);
    }

    public TextGraphicalObject(int x, int y, int layer, boolean visible, boolean selected, Storyboard pStoryboard) {
	super(layer, visible, selected, pStoryboard);

	textModel = new TextModel(PropertyManager.LOCAL_SCOPE, getScope(), null, this);

	startX = x;
	startY = y;
	width = 0;
	height = 0;

	// Will set all the font values
	setBaseFont(PropertyManager.STORYBOARD_TEXT_FONT);
	originalFont = baseFont;
	cursorX = startX;
	cursorY = startY - ((int)Math.round(cursorLength * .7));

	cursorToggle = true;
	insertionIndex = 0;

	// Steal keyboard events
	storyboard.removeKeyListener(storyboard.getStoryboardKeyListener());
	listener = new Listener();
	storyboard.addKeyListener(listener);
	storyboard.requestFocus();

	pX = -1;
	pY = -1;
	pWidth = -1;
	pHeight = -1;

	timer = new javax.swing.Timer(5000, listener);
	timerStop();
	timer.setCoalesce(true);
	timer.setRepeats(true);
	timerStart(400);
	
	destination = null;
	initialEntry = true;
	edited = false;

	// Draws the bounding box
	doRender(ACTIVE_COLOR, false, ACTIVE_COLOR);
    }


    public TextGraphicalObject(TextGraphicalObject copy) {
	this(copy, copy.storyboard);
    }

    public TextGraphicalObject(TextGraphicalObject copy, Storyboard toStoryboard) {
	this(copy, toStoryboard, true);
    }

    public TextGraphicalObject(TextGraphicalObject copy, Storyboard toStoryboard, boolean copyTModel) {
	super(copy, toStoryboard);
	startX = copy.startX;
	startY = copy.startY;
	width = copy.width;
	height = copy.height;

	cursorX = copy.cursorX;
	cursorY = copy.cursorY;
	cursorLength = copy.cursorLength;

	cursorToggle = true;

	insertionIndex = 0;

	listener = new Listener();
	pX = copy.pX;
	pY = copy.pY;
	pWidth = copy.pWidth;
	pHeight = copy.pHeight;

	baseFont = copy.baseFont;
	selectedFont = copy.selectedFont;
	opFont = copy.opFont;
	opRolloverFont = copy.opRolloverFont;
	opHighlightFont = copy.opHighlightFont; 
	originalFont = copy.originalFont;

	textModel = null;
	if (copy.textModel != null) {
	    if (copyTModel) {
		textModel = (TextModel) copy.textModel.clone(this);
	    } else {
		textModel = copy.textModel;
	    }
	}
	timer = null;
	
	destination = null;
	if (copy.destination != null) {
	    destination = new String(copy.destination);
	}
	initialEntry = false;
	edited = false;
    }


    public TextGraphicalObject(DataInputStream stream, Storyboard pStoryboard) throws IOException {
	super(stream, pStoryboard);

	startX = stream.readInt();
	startY = stream.readInt();
	width = stream.readInt();
	height = stream.readInt();

	cursorX = 0;
	cursorY = 0;

	cursorToggle = true;

	insertionIndex = 0;
	listener = new Listener();

	pX = -1;
	pY = -1;
	pWidth = -1;
	pHeight = -1;

	int fontSize = stream.readInt();

	textModel = null;
	boolean hasModel = stream.readBoolean();
	if (hasModel) {
	    textModel = new TextModel(stream, this, pStoryboard.getModel());
	}

	setBaseFont(PropertyManager.STORYBOARD_TEXT_FONT);
	setBaseFontSize(fontSize);
	originalFont = baseFont;

	timer = null;
	destination = null;
	initialEntry = false;
	edited = false;
    }



    // Public methods
    public void timerStop() {
	if (timer != null) {
	    timer.stop();
	}
    }


    public void timerStart(int ms) {
	if (timer != null) {
	    timer.setInitialDelay(ms);
	    timer.setDelay(ms);
	    timer.start();
	}
    }


    public void endTextEntry() {
	timerStop();
	unDrawCursor();
	initialEntry = false;
	edited = false;

	// Replace the original key listener on the storyboard
	storyboard.removeKeyListener(listener);
	storyboard.addKeyListener(storyboard.getStoryboardKeyListener());

	// If no text was entered, remove this object from the storyboard
	String text = getText();
	if ( (text == null) || (text.length() == 0) ) {
	    storyboard.deleteGraphicalObject(this);
	
	} else {
	    computeWidthAndHeight();
	}
    }



    // Edit this text object
    public void editText(int x, int y) {
	// Text object is being edited
	edited = true;

	// Render the text in a different color
	doRender(getColor(), true, ACTIVE_COLOR);

	// Set cursor to the character closest to (x, y)
	int closest = findClosestCharTo(x, y);

	insertionIndex = closest;
	Point p = locationOfCharAt(closest - 1);
	cursorX = p.x;
	cursorY = p.y - ((int) Math.round(cursorLength * .7));

	// Steal keyboard events
	storyboard.removeKeyListener(storyboard.getStoryboardKeyListener());
	storyboard.addKeyListener(listener);
	storyboard.requestFocus();

	cursorToggle = true;
	timer = new javax.swing.Timer(5000, listener);
	timerStop();
	timer.setCoalesce(true);
	timer.setRepeats(true);
	timerStart(400);
    }


    public String getText() {
	return textModel.getText();
    }


    public void changeFontSize(int delta) {
	int currentFontSize = baseFont.getSize();
	currentFontSize += delta;
	if (currentFontSize > 120) {
	    currentFontSize = 120;
	} else if (currentFontSize < 4) {
	    currentFontSize = 4;
	}
	setBaseFontSize(currentFontSize);
	originalFont = baseFont;

	computeWidthAndHeight();
	storyboard.paintComponent();
    }


    public void scale(double sx, double sy) {
	super.scale(sx, sy);
	pX = startX;
	pY = startY;
	pWidth = width;
	pHeight = height;

	startX = (int) Math.round(startX * sx);
	startY = (int) Math.round(startY * sy);
	width = (int) Math.round(width * sx);
	height = (int) Math.round(height * sy);

	// Scale the font - as best as possible
	int fontSize = baseFont.getSize();
	fontSize = (int) Math.round(fontSize * sx);
	if (fontSize < 4) {
	    fontSize = 4;
	}
	setBaseFontSize(fontSize);
    }


    public void resetToOriginalScale() {
	super.resetToOriginalScale();
	startX = pX;
	startY = pY;
	width = pWidth;
	height = pHeight;
	setBaseFontSize(originalFont.getSize());
    }


    public void setMouseIsOver(boolean b) {
	super.setMouseIsOver(b);
	textModel.setIsOver(b);
    }


    public void doTranslate(int deltaX, int deltaY) {
	startX += deltaX;
	startY += deltaY;
    }


    public Object clone() {
	return (new TextGraphicalObject(this));
    }

    public Object clone(Storyboard toStoryboard) {
	return (new TextGraphicalObject(this, toStoryboard));
    }

    public Object clone(Storyboard toStoryboard, boolean copyTModel) {
	return (new TextGraphicalObject(this, toStoryboard, copyTModel));
    }


    public void deleteGraphicalObject(boolean deleteModel) {
	super.deleteGraphicalObject(deleteModel);
	if (deleteModel) {
	    textModel.deleteModel();
	}
    }


    public String getNavigationDestination() {
	return destination;
    }

    public Terminal isNavigationAction() {
	Tokenizer tokenizer = new Tokenizer(getText());
	Token token = tokenizer.getNextToken();
	if (token != null) {
	    Terminal terminal = token.getTerminal();
	    if ((terminal.equals(Terminal.GO)) ||
		(terminal.equals(Terminal.JUMP)) ||
		(terminal.equals(Terminal.NAVIGATE))) {
		token = tokenizer.getNextToken();
		token = tokenizer.getNextToken();
	        if (token != null) {
		    destination = token.getLexeme();
		} else {
		    destination = null;
		}
		return terminal;
		
	    } else if (terminal.equals(Terminal.GOTO)) {
		destination = token.getLexeme();
		return terminal;
	    }
	}
	return null;
    }


    public Terminal isCMAction() {
	Tokenizer tokenizer = new Tokenizer(getText());
	Token token = tokenizer.getNextToken();
	if (token != null) {
	    Terminal terminal = token.getTerminal();
	    if ( (terminal.equals(Terminal.BEGIN)) ||
		 (terminal.equals(Terminal.END)) ||
		 (terminal.equals(Terminal.PAUSE)) ||
		 (terminal.equals(Terminal.PLAY)) || 
		 (terminal.equals(Terminal.RESTART)) ||
		 (terminal.equals(Terminal.END))) {

		return terminal;
	    }
	}
	return null;
    }


    public void display() {
	setVisible(true);
	storyboard.paintComponent();
    }

    public void unDisplay() {
	setVisible(false);
	storyboard.paintComponent();
    }

    public void highlight() {
	opIsHighlighted = true;
	computeWidthAndHeight();
	storyboard.paintComponent();
    }

    public void unHighlight() {
	opIsHighlighted = false;
	computeWidthAndHeight();
	storyboard.paintComponent();
    }


    public void realize(boolean showControls, boolean recreate) {
	super.realize(showControls, recreate);
	textModel.realize();
    }

    public void unRealize(boolean keepVisuals) {
	super.unRealize(keepVisuals);
	textModel.unRealize();
    }


    public void operationalize() {
	super.operationalize();
	textModel.operationalize();
	initialEntry = false;
	edited = false;

	computeWidthAndHeight();
	DataModel dModel = getScope();
	if ((performAnalysis) && 
	    (behaviors.size() == 0) &&
	    (dModel != null)) {
	    
	    // Free standing text object
	    Annotation annotation = Annotation.stringToAnnotation(getScope(), getText());
	    if (annotation != null) {
		try {
		    Parser.parse(annotation);
		    // Parsed successfully
		    setVisible(false);
		    dModel.addAnnotation(annotation);
		    
		} catch (Exception xcptn) {
		    // Not a parseable annotation
		}
	    }
	} 
    }
    
    
    public void unOperationalize() {
	super.unOperationalize();
	textModel.unOperationalize();
	computeWidthAndHeight();
    }
				   

    public Rectangle getBasicBounds() {
	FontMetrics fm = getFontMetrics(getFont());
	int ma = fm.getMaxAscent();
	return (new Rectangle(startX - 4, startY - ma, width + 8, height + 4));
    }

    protected Rectangle doGetBounds() {
	return getBasicBounds();
    }


    protected void mouseEnter() {
	computeWidthAndHeight();
	storyboard.paintComponent();
	textModel.mouseEnter();
    }

    protected void mouseLeave() {
	setMouseIsOver(false);
	computeWidthAndHeight();
	storyboard.paintComponent();
	textModel.mouseLeave();
    }


    protected void mouseSingleClicked() {
	textModel.mouseSingleClicked();
    }

    protected void mouseDoubleClicked() {
	textModel.mouseDoubleClicked();
    }


    public String getNamedModelId() {
	return textModel.getIdentifierId();
    }

    public NamedModel getNamedModel() {
	return textModel;
    }


    protected void setColor(Graphics2D g) {
	g.setColor(getColor());
    }

    protected void setFont(Graphics2D g) {
	g.setFont(getFont());
    }


    protected void doRender(Color c) {
	doRender(c, false, null);
    }

    protected void doRender(Color c, boolean drawSelectionHandles, Color shc) {
	Graphics2D g = (Graphics2D)storyboard.getGraphics();
	if (g == null) {
	    return;
	}
	g.setColor(c);
	setFont(g);
	
	// Draw the text
	doRender(g);
	
	if (initialEntry || edited) {
	    // Draw the bounding box
	    FontMetrics fontMetrics = g.getFontMetrics();
	    g.setStroke(LIGHT_STROKE);
	    g.setColor(shc);

	    int ma = fontMetrics.getMaxAscent();	    
	    int w = width + 6;
	    int h = height + 2;
	    
	    if (initialEntry) {
		if (textModel.size() <= 0) {
		    w = fontMetrics.charWidth('w');
		    h = fontMetrics.getHeight();
		}
	    } 
	    g.drawRect(startX - 3, startY - ma + 1, w, h);
	}

	if (drawSelectionHandles && selected) {
	    drawSelectionHandles(g, shc);
	}
    }

    protected void doRender(Graphics2D g) {
	FontMetrics fontMetrics = g.getFontMetrics();
	int x = startX;
	int y = startY;
	Iterator it = textModel.getCharSequence().iterator();
	while (it.hasNext()) {
	    char ch = ((Character) it.next()).charValue();

	    if (ch == '\n') {
		x = startX;
		y += fontMetrics.getHeight();

	    } else {
		String str = "";
		str += ch;

		int charWidth = fontMetrics.charWidth(ch);
		if (getScaleFactorX() == 1.0) {
		    g.drawString(str, x, y);
		    
		} else {    
		    if (((x + charWidth) <= (startX + width)) && 
			(y <= (startY + height))) {
			g.drawString(str, x, y);
		    }
		}
		x += charWidth;
	    }
	}
    }


    public void mediateRender(Graphics2D g, int offsetX, int offsetY) {
	FontMetrics fontMetrics = g.getFontMetrics();
	int x = offsetX;
	int y = offsetY;
	Iterator it = textModel.getCharSequence().iterator();
	while (it.hasNext()) {
	    char ch = ((Character) it.next()).charValue();

	    if (ch == '\n') {
		x = offsetX;
		y += fontMetrics.getHeight();

	    } else {
		String str = "";
		str += ch;
		int charWidth = fontMetrics.charWidth(ch);
		g.drawString(str, x, y);
		x += charWidth;
	    }
	}
    }


    // Private methods
    private Point locationOfCharAt(int location) {
	int x = startX;
	int y = startY;
	int count = 0;
	FontMetrics fm = getFontMetrics(getFont());

	Iterator it = textModel.getCharSequence().iterator();
	while ((it.hasNext()) && (count <= location)) { 
	    char c = ((Character) it.next()).charValue();
	    if (c == '\n') {
		x = startX;
		y += fm.getHeight();
	    } else {
		x += fm.charWidth(c);
	    }
	    ++count;
	}
	return (new Point(x, y));
    }

    
    private void computeWidthAndHeight() {
	int mWidth = -1; // Max Width
	int lWidth = 0;  // Current line width
	
	FontMetrics fontMetrics = getFontMetrics(getFont());
	int lHeight = fontMetrics.getHeight(); // current height

	Iterator it = textModel.getCharSequence().iterator();
	while (it.hasNext()) {
	    char c = ((Character) it.next()).charValue();
	    if (c == '\n') {
		// Increment height
		lHeight += fontMetrics.getHeight();

		// Check if current line width > max line width
		if (lWidth > mWidth) {
		    mWidth = lWidth;
		}
		lWidth = 0;
		
	    } else {
		lWidth += fontMetrics.charWidth(c); 
	    }
	}

	// Check last line
	if (lWidth > mWidth) {
	    mWidth = lWidth;
	}

	// set new width and height	
	width = mWidth;
	height = lHeight;
    }



    private void handleActionPerformed(ActionEvent event) {
	Graphics2D g = (Graphics2D) storyboard.getGraphics();
	if (cursorToggle) {
	    drawCursor(g);
	} else {
	    unDrawCursor(g);
	}
    }


    private void unDrawCursor() {
	Graphics2D g = (Graphics2D) storyboard.getGraphics();
	unDrawCursor(g);
    }

    private void unDrawCursor(Graphics2D g) {
	if (g != null) {
	    g.setColor(PropertyManager.STORYBOARD_BACKGROUND_COLOR);
	    g.setStroke(MEDIUM_STROKE);
	    g.drawLine(cursorX, cursorY, cursorX, cursorY + cursorLength);
	}
	cursorToggle = true;
    }


    private void drawCursor() {
	Graphics2D g = (Graphics2D) storyboard.getGraphics();
	drawCursor(g);
    }

    private void drawCursor(Graphics2D g) {
	if (g != null) {
	    g.setColor(ACTIVE_COLOR);
	    g.setStroke(MEDIUM_STROKE);
	    g.drawLine(cursorX, cursorY, cursorX, cursorY + cursorLength);
	}
	cursorToggle = false;
    }


    private int findClosestCharTo(int x, int y) {
	FontMetrics fm = getFontMetrics(getFont());
	int charX = startX;
	int charY = startY;
	int index = 0;
	int bestIndex = 0;

	Iterator it = textModel.getCharSequence().iterator();
	while (it.hasNext()) {
	    char ch = ((Character) it.next()).charValue();

	    if (ch == '\n') {
		charX = startX;
		charY += fm.getHeight();

	    } else {
		charX += fm.charWidth(ch);
		if ((x < charX) && (y < (charY + fm.getMaxDescent()))) {
		    bestIndex = index;
		    break;
		}
	    }
	    ++index;
	}
	return bestIndex;
    }


    
    private Color getColor() {
	Color c = ACTIVE_COLOR;
	if (opIsHighlighted) {
	    c = HIGHLIGHT_COLOR;
	    
	} else if (edited) {
	    c = PropertyManager.STORYBOARD_TEXT_EDIT_COLOR;

	} else if (operationalized) {
	    c = ACTIVE_COLOR;
	    
	} else if (inSelectedLayer) {
	    c = ACTIVE_COLOR;

	} else {
	    c = NOT_ACTIVE_COLOR;
	}
	return c;
    }


    private void setBaseFontSize(int fontSize) {
	// Increase current base font size
	Font font = new Font(baseFont.getFontName(), Font.PLAIN, fontSize);
	setBaseFont(font);
    }


    // Sets the text fonts based on font
    private void setBaseFont(Font font) {
	baseFont = font;
	if (baseFont == null) {
	    baseFont = new Font("Arial", Font.PLAIN, 18);
	}
	FontMetrics fm = getFontMetrics(font);
	cursorLength = fm.getMaxAscent() + fm.getMaxDescent() - 2;

	// Other fonts are based on the base font
	selectedFont = new Font(baseFont.getFontName(), Font.PLAIN, baseFont.getSize());
	opFont = new Font(baseFont.getFontName(), Font.PLAIN, baseFont.getSize());
	opRolloverFont = new Font(opFont.getFontName(), Font.BOLD, opFont.getSize() + 1);
	opHighlightFont = new Font(opFont.getFontName(), Font.BOLD, opFont.getSize() + 1);
    }


    public Font getFont() {
	Font cFont = baseFont;

	if (opMouseIsOver) {
	    cFont = opRolloverFont;
	    
	} else if (opIsHighlighted) {
	    cFont = opHighlightFont;
	    
	} else if (operationalized) {
	    cFont = opFont;
	}

	return cFont;
    }


    private FontMetrics getFontMetrics(Font font) {
	FontMetrics fm = null;

	// Get a graphics object, ask it for a FontMetrics object
	Graphics g = storyboard.getGraphics();
	if (g == null) {
	    JFrame frame = PropertyManager.APPLICATION_FRAME_WIDGET;
	    g = frame.getContentPane().getGraphics();
	}
	if (g != null) {
	    fm = g.getFontMetrics(font);
	}
	return fm;
    }


    private void handleKeyPressed(int keyCode) {
	// Undraw the cursor
	unDrawCursor();
	FontMetrics fm = getFontMetrics(getFont());

	if ( (keyCode == KeyEvent.VK_LEFT) ||
	     (keyCode == KeyEvent.VK_KP_LEFT)) {

	    if (insertionIndex >= 1) {
		--insertionIndex;
		char lc = textModel.charAt(insertionIndex);
		if (lc == '\n') {
		    Point p = locationOfCharAt(insertionIndex - 1);
		    cursorX = p.x;
		    cursorY = p.y - ((int) Math.round(cursorLength * .7));
		} else {
		    cursorX -= fm.charWidth(lc);
		}
	    }

	} else if ( (keyCode == KeyEvent.VK_RIGHT) ||
		    (keyCode == KeyEvent.VK_KP_RIGHT)) {

	    if (insertionIndex < textModel.size()) {
		char rc = textModel.charAt(insertionIndex);
		++insertionIndex;
		if (rc == '\n') {
		    cursorX = startX;
		    cursorY += fm.getHeight();
		} else {
		    cursorX += fm.charWidth(rc);
		}
	    }

	} else if (keyCode == KeyEvent.VK_ENTER) {
	    // Undraw the text
	    doRender(PropertyManager.STORYBOARD_BACKGROUND_COLOR, true, PropertyManager.STORYBOARD_BACKGROUND_COLOR);

	    cursorX = startX; 
	    cursorY += fm.getHeight();
	    textModel.addChar(insertionIndex, '\n');
	    ++insertionIndex;
	    computeWidthAndHeight();

	    // Draw the text
	    doRender(getColor(), true, ACTIVE_COLOR);
	    
	} else if (keyCode == KeyEvent.VK_BACK_SPACE) {
	    
	    if (textModel.size() > 0) {		
		if (insertionIndex >= 1) {
		    // Undraw the text
		    doRender(PropertyManager.STORYBOARD_BACKGROUND_COLOR, true, PropertyManager.STORYBOARD_BACKGROUND_COLOR);
		    
		    --insertionIndex;
		    char removedChar = textModel.charAt(insertionIndex);
		    textModel.removeCharAt(insertionIndex);
		    if (removedChar == '\n') {
			if ((insertionIndex - 1) > 0) {
			    Point p = locationOfCharAt(insertionIndex - 1);
			    cursorX = p.x;
			    cursorY = p.y - ((int) Math.round(cursorLength * .7));
			} else {
			    cursorX = startX;
			    cursorY = startY - ((int) Math.round(cursorLength * .7));
			}

		    } else {
			cursorX -= fm.charWidth(removedChar);
		    }
		
		    computeWidthAndHeight();
		
		    // Draw the text
		    doRender(getColor(), true, ACTIVE_COLOR);
		}
	    }
	    
	} else if (keyCode == KeyEvent.VK_DELETE) {
	    
	    if (textModel.size() > 0) {
		if (insertionIndex < textModel.size()) {
		    // Undraw the text
		    doRender(PropertyManager.STORYBOARD_BACKGROUND_COLOR, true, PropertyManager.STORYBOARD_BACKGROUND_COLOR);
		    
		    // Remove the character
		    char removedChar = textModel.charAt(insertionIndex);
		    textModel.removeCharAt(insertionIndex);
		    
		    // Compute new width and height
		    computeWidthAndHeight();
		    
		    // Draw the text
		    doRender(getColor(), true, ACTIVE_COLOR);
		}
	    }
	}
    }

    // A standard character was typed
    private void handleKeyTyped(char keyChar) {
	// Undraw the cursor
	unDrawCursor();
	
	// Undraw the text
	doRender(PropertyManager.STORYBOARD_BACKGROUND_COLOR, true, PropertyManager.STORYBOARD_BACKGROUND_COLOR);

	// Insert the character into the model
	textModel.addChar(insertionIndex, keyChar);
	++insertionIndex;
		
	// Move the cursor over
	FontMetrics fm = getFontMetrics(getFont());
	cursorX += fm.charWidth(keyChar);
	
	// Compute the possibly new width and height
	computeWidthAndHeight();
	
	// Draw the new text
	doRender(getColor(), true, ACTIVE_COLOR);
    }
    
    


    // Listener class
    class Listener implements KeyListener, ActionListener {
	
	public Listener() {
	}
	
	// Key Listener interface
	public void keyReleased(KeyEvent evt) {
	}

	public void keyPressed(KeyEvent evt) {
	    int keyCode = evt.getKeyCode();
	    if ( (keyCode == KeyEvent.VK_LEFT) ||
		 (keyCode == KeyEvent.VK_KP_LEFT) ||
		 (keyCode == KeyEvent.VK_RIGHT) ||
		 (keyCode == KeyEvent.VK_KP_RIGHT) ||
		 (keyCode == KeyEvent.VK_ENTER) ||
		 (keyCode == KeyEvent.VK_BACK_SPACE) ||
		 (keyCode == KeyEvent.VK_DELETE)) {
		handleKeyPressed(keyCode);
	    }
	}

	public void keyTyped(KeyEvent evt) {
	    char c = evt.getKeyChar();
	    if ((c != KeyEvent.CHAR_UNDEFINED) &&
		(c != '\b') &&
		(c != '\n')) {
		handleKeyTyped(c);
	    }
	}
	

	// Timer interface callback
	public void actionPerformed(ActionEvent evt) {
	    handleActionPerformed(evt);
	}
    }



    public void write(DataOutputStream stream) throws IOException {
	// Write the class id
	stream.writeInt(CLASSID);

	// Write the parent information
	super.write(stream);

	stream.writeInt(startX);
	stream.writeInt(startY);

	stream.writeInt(width);
	stream.writeInt(height);

	// Write the base font size
	stream.writeInt(baseFont.getSize());

	// Write the model
	stream.writeBoolean(textModel != null);
	if (textModel != null) {
	    textModel.write(stream);
	}
    }
}
