/*
Copyright (c) 2001 Regents of the University of Minnesota.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

3. The name of the University may not be used to endorse or promote products 
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* 
   File: Polygon2D.java
*/

// Current package
package edu.uofminnesota.DEMAIS.usrinterface.storyboard.stroke;

// Necessary imports
import java.awt.*;
import java.io.*;
import edu.uofminnesota.DEMAIS.property.*;
import edu.uofminnesota.DEMAIS.property.*;




public class Polygon2D extends Polygon {

    private static final int ANGLE_THRESHHOLD = 8;
    private static final double COS_ANGLE_THRESHHOLD = Math.cos(Math.toRadians(ANGLE_THRESHHOLD));
   
    // Length of the polygon
    private double length;
    private boolean lengthInvalid;

    // Cached computations to speed the simplification algorithm
    private int vector_a_x;
    private int vector_a_y;
    private double vector_a_length;
    private int lastx, lasty;



    public Polygon2D() {
	super();
	length = 0.0;
	lengthInvalid = true;
	lastx = lasty = 0;
    }

    public Polygon2D(int[] x, int[] y, int numPoints) {
	super(x, y, numPoints);
	computeLength();
	lastx = lasty = 0;
    }

    public Polygon2D(Polygon2D copy) {
	this(copy.xpoints, copy.ypoints, copy.npoints);  
    }


    public Polygon2D(DataInputStream stream) throws IOException {
	super();	

	// Read total number of points
	npoints = stream.readInt();
	// Read and add the points
	if (npoints > 0) {
	    xpoints = new int[npoints];
	    ypoints = new int[npoints];
	    for (int i=0; i<npoints; ++i) {
	        xpoints[i] = stream.readInt();
		ypoints[i] = stream.readInt();
	    }
	}
	computeLength();
	bounds = null;
	lastx = lasty = 0;

    }


    public double getLength() {
	if (lengthInvalid) {
	    computeLength();
	}
	return length;
    }


    public void addPoint(int x, int y) {
	if (npoints <= 0) {
	    super.addPoint(x, y);
	    lastx = x;
	    lasty = y;
	} else {
	    int dx = lastx - x;
	    int dy = lasty - y;
	    int distSq = dx*dx + dy*dy;
	    if (distSq > 2) {
		super.addPoint(x, y);
		lastx = x;
		lasty = y;
	    } 
	}
	lengthInvalid = true;
    }


    // Get the number of points in the stroke.
    public int getNumPoints() {
	return npoints;
    } 

    public Point getFirstPoint() {
	return (new Point(xpoints[0], ypoints[0]));
    }

    public Point getLastPoint() {
	int x = 0;
	int y = 0;
	if (npoints > 0) {
	    x = xpoints[npoints - 1];
	    y = ypoints[npoints - 1];
	}
	return (new Point(x, y));
    }

    public Point getNextToLastPoint() {
	int x = 0;
	int y = 0;
	if (npoints > 1) {
	    x = xpoints[npoints - 2];
	    y = ypoints[npoints - 2];
	}
	return (new Point(x, y));
    }



    // Apply the transformation matrix to each point
    public void scale(double sx, double sy) {
	for (int i=0; i<npoints; ++i) {
	    xpoints[i] = (int)Math.round(xpoints[i] * sx);
	    ypoints[i] = (int)Math.round(ypoints[i] * sy);
	}

	// Modified the parent's data set, force a recalculation of the bounds
	bounds = null;
    }


    public void resetToOriginalScale() {
	// Never used
    }
    
 
    // Fixes a bug in the JDK1.3 contains method
    // If a rectangle has a width or height of 0,
    // the structure methods return inaccurate results
    public Rectangle getBounds() {
	int x = 0;
	int y = 0;
	int w = 1;
	int h = 1;
	
	if (npoints > 0) {   
	    Rectangle r = super.getBounds();
	    x = r.x;
	    y = r.y;
	    w = r.width;
	    h = r.height;
	    if (w <= 0) {
		w = 1;
	    }
	    if (h <= 0) {
		h = 1;
	    }
	}
	return (new Rectangle(x, y, w, h));
    }

    
    public String toString() {
	String retString = "[";
	for (int i=0; i<npoints; ++i) {
	    retString += "(";
	    retString += xpoints[i];
	    retString += ", ";
	    retString += ypoints[i];
	    retString += ")";
	}
	retString += "]";
	
	return retString;
    } 
    

    public void resetStrokeToLine(int x1, int y1, int x2, int y2) {
	double increment = 0.025;
	int np = (int) Math.round(1.0 / increment);
	int xp[] = new int[np];
	int yp[] = new int[np];

	int i=0;
	for (double t=0.0; t<=1.0; t+=increment, ++i) {
	    xp[i] = (int) Math.round((x1 * (1.0 - t)) + (x2 * t));
	    yp[i] = (int) Math.round((y1 * (1.0 - t)) + (y2 * t));
	}

	this.npoints = np;
	this.xpoints = xp;
	this.ypoints = yp;
	lengthInvalid = true;
	bounds = null;
    }
    
    
    public Object clone() {
	return (new Polygon2D(this));
    }


    // Simplifies the current polygon in a destructive manner
    // Once simplified, the original polygon no longer exists.
    public void simplify() {
	if (npoints < PropertyManager.APPLICATION_MAX_TAP_POINTS) {
	    // Cannot simplify polygon
	    return;
	}

	Polygon newP = new Polygon();
	// Always add the first point
	newP.addPoint(xpoints[0], ypoints[0]);

	// compute the initial cached values
	vector_a_x = xpoints[0] - xpoints[1];
	vector_a_y = ypoints[0] - ypoints[1];
	vector_a_length = vectorLength(vector_a_x, vector_a_y);

	// See which other points we need to add
	for (int i=1; i<npoints - 1; ++i) {
	    double cos_angle = computeCosOfAngleAt(i);
	    if (Math.abs(cos_angle) < COS_ANGLE_THRESHHOLD) {
		newP.addPoint(xpoints[i], ypoints[i]);
	    }
	}
	// Always add the last point
	newP.addPoint(xpoints[npoints-1], ypoints[npoints-1]);

	// Reassign the points to *this*
	npoints = newP.npoints;
	xpoints = newP.xpoints;
	ypoints = newP.ypoints;
	lengthInvalid = true;
	bounds = null;
    }



    private void computeLength() {
	length = 0.0;
	for (int i=1; i<npoints; ++i) {
	    int dx = xpoints[i] - xpoints[i-1];
	    int dy = ypoints[i] - ypoints[i-1];
	    length += Math.sqrt(dx*dx + dy*dy);
	}
	lengthInvalid = false;
    }


    // Computes the angle at polygon point index
    private double computeCosOfAngleAt(int index) {
	// We have vector A cached
	// Compute vector B
	int vector_b_x = xpoints[index + 1] - xpoints[index];
	int vector_b_y = ypoints[index + 1] - ypoints[index];

	// compute angle between the two vectors
	double vector_b_length = vectorLength(vector_b_x, vector_b_y);
	double dot_AB = dotProduct(vector_a_x, vector_a_y, vector_b_x, vector_b_y);
	double cos_angle = (dot_AB / (vector_a_length * vector_b_length));

	// Store new cached values
	vector_a_x = -(vector_b_x);
	vector_a_y = -(vector_b_y);
	vector_a_length = vector_b_length;
	
	return cos_angle;
    }
 

    // Computes the length of a vector
    private double vectorLength(int v_x, int v_y) {
	return (Math.sqrt(dotProduct(v_x, v_y, v_x, v_y)));
    }


    // Computes the dot product between two vectors (x1,y1), (x2,y2)
    private double dotProduct(int x1, int y1, int x2, int y2) {
	return ( (double)(x1*x2 + y1*y2));
    }


    public void write(DataOutputStream stream) throws IOException {
	// Write total number of points
	stream.writeInt(npoints);
	
	// Write each point
	for (int i=0; i<npoints; ++i) {
	    stream.writeInt(xpoints[i]);
	    stream.writeInt(ypoints[i]);
	}
    }
}
